/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------------\
|                                         3do.cpp                                    |
|  ce fichier contient les structures, classes et fonctions nécessaires à la lecture |
| des fichiers 3do de total annihilation qui sont les fichiers contenant les modèles |
| 3d des objets du jeu.                                                              |
|                                                                                    |
\-----------------------------------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"
#include "particles.h"
#include "taconfig.h"

using namespace TA3D::EXCEPTION;

TEXTURE_MANAGER	texture_manager;
MODEL_MANAGER	model_manager;

	int TEXTURE_MANAGER::load_gaf(byte *data)
	{
		int nb_entry=get_gaf_nb_entry(data);
		int n_nbtex=nbtex+nb_entry;
		ANIM *n_tex=(ANIM*) malloc(sizeof(ANIM)*n_nbtex);
		int i;
		for(i=0;i<nbtex;i++)
			n_tex[i]=tex[i];
		if(tex)
			free(tex);
		tex=n_tex;
		for(i=0;i<nb_entry;i++)
			tex[nbtex+i].load_gaf(data,i,false);
		nbtex+=nb_entry;
		return 0;
	}

	const void SCRIPT_DATA::move(const float dt,const float g)
	{
		if(!is_moving)	return;
		is_moving=false;
		for(uint16 e=0;e<nb_piece;e++)
			if(flag[e]&FLAG_EXPLODE && (explosion_flag[e]&EXPLODE_BITMAPONLY)!=EXPLODE_BITMAPONLY)		// This piece is exploding
				for(byte i=0;i<3;i++) {
					if(i==1 && explosion_flag[e]&EXPLODE_FALL)
						axe[i][e].move_speed-=g;
					axe[i][e].pos+=axe[i][e].move_speed*dt;
					axe[i][e].angle+=axe[i][e].rot_speed*dt;
					is_moving=true;
					}
			else
				for(byte i=0;i<3;i++) {
					if(!axe[i][e].is_moving)	continue;
					axe[i][e].is_moving=false;
					float a=axe[i][e].move_distance;
					if(a!=0.0f) {
						axe[i][e].is_moving=true;
						is_moving=true;
						float c=axe[i][e].move_speed*dt;
						axe[i][e].move_distance-=c;
						axe[i][e].pos+=c;
						if((a>0.0f && axe[i][e].move_distance<0.0f) || (a<0.0f && axe[i][e].move_distance>0.0f)) {
							axe[i][e].pos+=axe[i][e].move_distance;
							axe[i][e].move_distance=0.0f;
							}
						}
					while(axe[i][e].angle>180.0f)	axe[i][e].angle-=360.0f;		// Maintient l'angle dans les limites
					while(axe[i][e].angle<-180.0f)	axe[i][e].angle+=360.0f;
					a=axe[i][e].rot_angle;
					if((axe[i][e].rot_speed!=0.0f || axe[i][e].rot_accel!=0.0f) && ((a!=0.0f && axe[i][e].rot_limit) || !axe[i][e].rot_limit)) {
						axe[i][e].is_moving=true;
						is_moving=true;
						float b=axe[i][e].rot_speed;
						if(b<-7200.0f)
							b=axe[i][e].rot_speed=-7200.0f;
						else if(b>7200.0f)
							b=axe[i][e].rot_speed=7200.0f;
						axe[i][e].rot_speed+=axe[i][e].rot_accel*dt;
						if(axe[i][e].rot_speed_limit)
							if((b<axe[i][e].rot_target_speed && axe[i][e].rot_speed>axe[i][e].rot_target_speed)
							|| (b>axe[i][e].rot_target_speed && axe[i][e].rot_speed<axe[i][e].rot_target_speed)) {
								axe[i][e].rot_accel=0.0f;
								axe[i][e].rot_speed=axe[i][e].rot_target_speed;
								axe[i][e].rot_speed_limit=false;
								}
						float c=axe[i][e].rot_speed*dt;
						axe[i][e].angle+=c;
						if(axe[i][e].rot_limit) {
							axe[i][e].rot_angle-=c;
							if((a>=0.0f && axe[i][e].rot_angle<0.0f) || (a<=0.0f && axe[i][e].rot_angle>0.0f)) {
								axe[i][e].angle+=axe[i][e].rot_angle;
								axe[i][e].rot_angle=0.0f;
								}
							}
						}
					}
	}

	int OBJECT::load_obj(byte *data,int offset,int dec)
	{
		destroy();					// Au cas où l'objet ne serait pas vierge

		tagObject header;				// Lit l'en-tête
		header.VersionSignature=*((int*)(data+offset));
		header.NumberOfVertexes=*((int*)(data+offset+4));
		header.NumberOfPrimitives=*((int*)(data+offset+8));
		header.OffsetToselectionPrimitive=*((int*)(data+offset+12));
		header.XFromParent=*((int*)(data+offset+16));
		header.YFromParent=*((int*)(data+offset+20));
		header.ZFromParent=*((int*)(data+offset+24));
		header.OffsetToObjectName=*((int*)(data+offset+28));
		header.Always_0=*((int*)(data+offset+32));
		header.OffsetToVertexArray=*((int*)(data+offset+36));
		header.OffsetToPrimitiveArray=*((int*)(data+offset+40));
		header.OffsetToSiblingObject=*((int*)(data+offset+44));
		header.OffsetToChildObject=*((int*)(data+offset+48));

		nb_vtx=header.NumberOfVertexes;
		nb_prim=header.NumberOfPrimitives;
		name=strdup((char*)(data+header.OffsetToObjectName));
		int i;
#ifdef DEBUG_MODE
/*		for(i=0;i<dec;i++)
			printf("  ");
		printf("%s",name);
		for(i=0;i<20-2*dec-strlen(name);i++)
			printf(" ");
		printf("-> nb_vtx=%d | nb_prim=%d\n",nb_vtx,nb_prim);*/
#endif
		if(header.OffsetToChildObject) {					// Charge récursivement les différents objets du modèle
			child=(OBJECT*) malloc(sizeof(OBJECT));
			child->init();
			child->load_obj(data,header.OffsetToChildObject,dec+1);
			}
		if(header.OffsetToSiblingObject) {					// Charge récursivement les différents objets du modèle
			next=(OBJECT*) malloc(sizeof(OBJECT));
			next->init();
			next->load_obj(data,header.OffsetToSiblingObject,dec);
			}
		points=(POINTF*) malloc(sizeof(POINTF)*nb_vtx);		// Alloue la mémoire nécessaire pour stocker les points
		int f_pos;
		float div=0.5f/65536.0f;
		pos_from_parent.x=header.XFromParent*div;
		pos_from_parent.y=header.YFromParent*div;
		pos_from_parent.z=-header.ZFromParent*div;
		f_pos=header.OffsetToVertexArray;
		for(i=0;i<nb_vtx;i++) {				// Lit le tableau de points stocké dans le fichier
			tagVertex vertex;
			vertex.x=*((int*)(data+f_pos));		f_pos+=4;
			vertex.y=*((int*)(data+f_pos));		f_pos+=4;
			vertex.z=*((int*)(data+f_pos));		f_pos+=4;
			points[i].x=vertex.x*div;
			points[i].y=vertex.y*div;
			points[i].z=-vertex.z*div;
			}
		f_pos=header.OffsetToPrimitiveArray;
		int n_index=0;
		selprim=header.OffsetToselectionPrimitive;
		for(i=0;i<nb_prim;i++) {					// Compte le nombre de primitive de chaque sorte
			tagPrimitive primitive;
			primitive.ColorIndex=*((int*)(data+f_pos));						f_pos+=4;
			primitive.NumberOfVertexIndexes=*((int*)(data+f_pos));			f_pos+=4;
			primitive.Always_0=*((int*)(data+f_pos));						f_pos+=4;
			primitive.OffsetToVertexIndexArray=*((int*)(data+f_pos));		f_pos+=4;
			primitive.OffsetToTextureName=*((int*)(data+f_pos));			f_pos+=4;
			primitive.Unknown_1=*((int*)(data+f_pos));						f_pos+=4;
			primitive.Unknown_2=*((int*)(data+f_pos));						f_pos+=4;
			primitive.Unknown_3=*((int*)(data+f_pos));						f_pos+=4;

			switch(primitive.NumberOfVertexIndexes)
			{
			case 1:		nb_p_index++;	break;
			case 2:		nb_l_index+=2;	break;
			default:
				n_index+=primitive.NumberOfVertexIndexes;
				nb_t_index++;
			};
			}
#ifdef DEBUG_MODE
//		printf("(%d,%d,%d)\n",nb_p_index,nb_l_index,nb_t_index);
#endif

		if(nb_p_index>0)				// Alloue la mémoire nécessaire pour stocker les primitives
			p_index=(GLushort*) malloc(sizeof(GLushort)*nb_p_index);
		if(nb_l_index>0)
			l_index=(GLushort*) malloc(sizeof(GLushort)*nb_l_index);
		if(nb_t_index>0) {
			tex=(int*) malloc(sizeof(int)*nb_t_index);
			usetex=(byte*) malloc(sizeof(byte)*nb_t_index);
			nb_index=(short*) malloc(sizeof(short)*nb_t_index);
			t_index=(GLushort*) malloc(sizeof(GLushort)*n_index);
			}

		f_pos=header.OffsetToPrimitiveArray;
		int pos_p=0,pos_l=0,pos_t=0,cur=0;
		int nb_diff_tex=0;
		int *index_tex = new int[nb_prim];
		int t_m=0;
		for(i=0;i<nb_prim;i++) {					// Compte le nombre de primitive de chaque sorte
			tagPrimitive primitive;
			primitive.ColorIndex=*((int*)(data+f_pos));						f_pos+=4;
			primitive.NumberOfVertexIndexes=*((int*)(data+f_pos));			f_pos+=4;
			primitive.Always_0=*((int*)(data+f_pos));						f_pos+=4;
			primitive.OffsetToVertexIndexArray=*((int*)(data+f_pos));		f_pos+=4;
			primitive.OffsetToTextureName=*((int*)(data+f_pos));			f_pos+=4;
			primitive.Unknown_1=*((int*)(data+f_pos));						f_pos+=4;
			primitive.Unknown_2=*((int*)(data+f_pos));						f_pos+=4;
			primitive.Unknown_3=*((int*)(data+f_pos));						f_pos+=4;

			switch(primitive.NumberOfVertexIndexes)
			{
			case 1:
				p_index[pos_p++]=*((short*)(data+primitive.OffsetToVertexIndexArray));
				break;
			case 2:
				l_index[pos_l++]=*((short*)(data+primitive.OffsetToVertexIndexArray));
				l_index[pos_l++]=*((short*)(data+primitive.OffsetToVertexIndexArray+2));
				break;
			default:
				nb_index[cur]=primitive.NumberOfVertexIndexes;
				tex[cur]=t_m=texture_manager.get_texture_index((char*)(data+primitive.OffsetToTextureName));
				usetex[cur]=1;
				if(t_m==-1) {
					if(primitive.ColorIndex>=0 && primitive.ColorIndex<256) {
						usetex[cur]=1;
						tex[cur]=t_m=primitive.ColorIndex;
						}
					else
						usetex[cur]=0;
					}
				if(t_m>=0)	{														// Code pour la création d'une texture propre à chaque modèle
					bool al_in=false;
					int indx=t_m;
					for(int e=0;e<nb_diff_tex;e++)
						if(index_tex[e]==indx) {
							al_in=true;
							break;
							}
					if(!al_in)
						index_tex[nb_diff_tex++]=indx;
				}
				for(int e=0;e<nb_index[cur];e++)
					t_index[pos_t++]=*((short*)(data+primitive.OffsetToVertexIndexArray+(e<<1)));
				cur++;
			};
			}
/*------------------------------Création de la texture unique pour l'unité--------------*/
		int *px = new int[nb_diff_tex],*py = new int[nb_diff_tex];			// Pour placer les différentes mini-textures sur une grande texture
		int mx=0,my=0;
		for(i=0;i<nb_diff_tex;i++) {
			int dx=texture_manager.tex[index_tex[i]].bmp[0]->w,dy=texture_manager.tex[index_tex[i]].bmp[0]->h;
			px[i]=py[i]=0;
			if(i!=0)
				for(int e=0;e<i;e++) {
					int fx=texture_manager.tex[index_tex[e]].bmp[0]->w,fy=texture_manager.tex[index_tex[e]].bmp[0]->h;
					bool found[3];
					found[0]=found[1]=found[2]=true;
					int j;

					px[i]=px[e]+fx;		py[i]=py[e];
					for(j=0;j<i;j++) {
						int gx=texture_manager.tex[index_tex[j]].bmp[0]->w,gy=texture_manager.tex[index_tex[j]].bmp[0]->h;
						if(coupe(px[i],py[i],dx,dy,px[j],py[j],gx,gy)) {
							found[0]=false;
							break;
							}
						}

					px[i]=px[e];		py[i]=py[e]+fy;
					for(j=0;j<i;j++) {
						int gx=texture_manager.tex[index_tex[j]].bmp[0]->w,gy=texture_manager.tex[index_tex[j]].bmp[0]->h;
						if(coupe(px[i],py[i],dx,dy,px[j],py[j],gx,gy)) {
							found[2]=false;
							break;
							}
						}
					px[i]=px[e]+fx;		py[i]=0;
					for(j=0;j<i;j++) {
						int gx=texture_manager.tex[index_tex[j]].bmp[0]->w,gy=texture_manager.tex[index_tex[j]].bmp[0]->h;
						if(coupe(px[i],py[i],dx,dy,px[j],py[j],gx,gy)) {
							found[1]=false;
							break;
							}
						}
					bool deborde=false;
					bool found_one=false;
					int deb=0;
					if(found[1]) {
						px[i]=px[e]+fx;
						py[i]=0;
						deborde=false;
						if(px[i]+dx>mx || py[i]+dy>my) deborde=true;
						deb=max(mx,px[i]+dx)*max(py[i]+dy,my)-mx*my;
						found_one=true;
						}
					if(found[0] && (!found_one || deborde)) {
						px[i]=px[e]+fx;
						py[i]=py[e];
						deborde=false;
						if(px[i]+dx>mx || py[i]+dy>my) deborde=true;
						deb=max(mx,px[i]+dx)*max(py[i]+dy,my)-mx*my;
						found_one=true;
						}
					if(found[2] && deborde) {
						int ax=px[i],ay=py[i];
						px[i]=px[e];
						py[i]=py[e]+fy;
						deborde=false;
						if(px[i]+dx>mx || py[i]+dy>my) deborde=true;
						int deb2=max(mx,px[i]+dx)*max(py[i]+dy,my)-mx*my;
						if(found_one && deb<deb2) {
							px[i]=ax;	py[i]=ay;
							}
						else
							found_one=true;
						}
					if(found_one)			// On a trouvé une position qui convient
						break;
					}
			if(px[i]+dx>mx) mx=px[i]+dx;
			if(py[i]+dy>my) my=py[i]+dy;
			}
		BITMAP *bmp=create_bitmap(mx,my);
		if(bmp!=NULL && mx!=0 && my!=0) {
			for(int e=0;e<expected_players;e++) {
				clear(bmp);
				bool mtex_needed=false;
				for(i=0;i<nb_diff_tex;i++)
					if(texture_manager.tex[index_tex[i]].nb_bmp==10) {
						blit(texture_manager.tex[index_tex[i]].bmp[player_color_map[e]],bmp,0,0,px[i],py[i],texture_manager.tex[index_tex[i]].bmp[player_color_map[e]]->w,texture_manager.tex[index_tex[i]].bmp[player_color_map[e]]->h);
						mtex_needed=true;
						}
					else
						blit(texture_manager.tex[index_tex[i]].bmp[0],bmp,0,0,px[i],py[i],texture_manager.tex[index_tex[i]].bmp[0]->w,texture_manager.tex[index_tex[i]].bmp[0]->h);
					dtex=e+1;
				if(g_useTextureCompression)
					allegro_gl_set_texture_format(GL_COMPRESSED_RGB_ARB);
				else
					allegro_gl_set_texture_format(GL_RGB5);
				gltex[e]=allegro_gl_make_texture(bmp);
				glBindTexture(GL_TEXTURE_2D,gltex[e]);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
				if(!mtex_needed)	break;
				}
			}
		else
			dtex=0;
		if(bmp)
			destroy_bitmap(bmp);
		int nb_total_point=0;
		for(i=0;i<nb_t_index;i++)
			nb_total_point+=nb_index[i];
		nb_total_point+=nb_l_index;
		POINTF *p=(POINTF*) malloc(sizeof(POINTF)*nb_total_point<<1);			// *2 pour le volume d'ombre
		for(i=0;i<nb_total_point-nb_l_index;i++) {
			p[i+nb_total_point]=p[i]=points[t_index[i]];
			t_index[i]=i;
			}
		for(i=nb_total_point-nb_l_index;i<nb_total_point;i++) {
			int e=i-nb_total_point+nb_l_index;
			p[i+nb_total_point]=p[i]=points[l_index[e]];
			l_index[e]=i;
			}
		if(nb_l_index==2)
			if(p[l_index[0]].x<0.0f) {
				int tmp=l_index[0];
				l_index[0]=l_index[1];
				l_index[1]=tmp;
				}
		free(points);
		points=p;
		nb_vtx=nb_total_point;

		int nb_triangle=0;
		for(i=0;i<nb_t_index;i++)
			if(i!=selprim)
				nb_triangle+=nb_index[i]-2;
		GLushort *index=(GLushort*) malloc(sizeof(GLushort)*nb_triangle*3);
		tcoord=(float*) malloc(sizeof(float)*nb_vtx<<1);
		cur=0;
		int curt=0;
		int decal=0;
		pos_t=0;
		for(i=0;i<nb_t_index;i++)
			for(int e=0;e<nb_index[i];e++) {
				if(i==selprim) {
					if(e==0) {
						sel[0]=t_index[cur];
						sel[1]=t_index[cur+1];
						sel[2]=t_index[cur+2];
						sel[3]=t_index[cur+3];
						cur+=nb_index[i];
						decal=1;
						}
					continue;
					}
				if(e<3)
					index[pos_t++]=t_index[cur];
				else {
					index[pos_t]=index[pos_t-3];	pos_t++;
					index[pos_t]=index[pos_t-2];	pos_t++;
					index[pos_t++]=t_index[cur];
					}
				tcoord[curt]=0.0f;
				tcoord[curt+1]=0.0f;
				if(usetex[i-decal]) {
					switch(e&3)
					{
					case 0:						break;
					case 1:
						tcoord[curt]+=((float)texture_manager.tex[tex[i-decal]].bmp[0]->w-1)/(mx-1);
//						tcoord[curt]+=((float)texture_manager.bmptex[tex[i-decal]]->w-1)/(mx-1);
						break;
					case 2:
						tcoord[curt]+=((float)texture_manager.tex[tex[i-decal]].bmp[0]->w-1)/(mx-1);
						tcoord[curt+1]+=((float)texture_manager.tex[tex[i-decal]].bmp[0]->h-1)/(my-1);
//						tcoord[curt]+=((float)texture_manager.bmptex[tex[i-decal]]->w-1)/(mx-1);
//						tcoord[curt+1]+=((float)texture_manager.bmptex[tex[i-decal]]->h-1)/(my-1);
						break;
					case 3:
						tcoord[curt+1]+=((float)texture_manager.tex[tex[i-decal]].bmp[0]->h-1)/(my-1);
//						tcoord[curt+1]+=((float)texture_manager.bmptex[tex[i-decal]]->h-1)/(my-1);
						break;
					};
					int indx=0;
					for(int f=0;f<nb_diff_tex;f++)
						if(tex[i-decal]==index_tex[f]) {
							indx=f;
							break;
							}
					tcoord[curt]+=((float)px[indx])/(mx-1);
					tcoord[curt+1]+=((float)py[indx])/(my-1);
					}
				cur++;
				curt+=2;
				}
		for(cur=0;cur<pos_t;cur+=3) {			// Petite inversion pour avoir un affichage correct
			GLushort t=index[cur+1];
			index[cur+1]=index[cur+2];
			index[cur+2]=t;
			}
		nb_t_index=nb_triangle*3;
		free(t_index);
		t_index=index;
		free(usetex);
		usetex=NULL;
/*--------------------------------------------------------------------------------------*/

		if(nb_t_index>0) {			// Calcule les normales pour l'éclairage
			N=(VECTOR*) malloc(sizeof(VECTOR)*nb_vtx<<1);
			for(i=0;i<nb_vtx<<1;i++)
				N[i].x=N[i].z=N[i].y=0.0f;
			for(i=0;i<nb_t_index;i+=3) {
				VECTOR AB,AC,Normal;
				AB=points[t_index[i]]>>points[t_index[i+1]];
				AC=points[t_index[i]]>>points[t_index[i+2]];
				Normal=AB*AC;	Normal.Unit();
				for(int e=0;e<3;e++)
					N[t_index[i+e]]=N[t_index[i+e]]+Normal;
				}
			for(i=0;i<nb_vtx;i++)
				N[i].Unit();
			}
		delete[] px;
		delete[] py;
		delete[] index_tex;
		return 0;
	}

	void OBJECT::create_from_2d(BITMAP *bmp,float w,float h,float max_h)
	{
		destroy();					// Au cas où l'objet ne serait pas vierge

		pos_from_parent.x=0.0f;
		pos_from_parent.y=0.0f;
		pos_from_parent.z=0.0f;
		selprim=-1;
		child=NULL;
		next=NULL;
		nb_l_index=0;
		nb_p_index=0;
		l_index=NULL;
		p_index=NULL;

		nb_vtx=256;
		nb_t_index=1350;
		points=(POINTF*) malloc(sizeof(POINTF)*nb_vtx);
		tcoord=(float*) malloc(sizeof(float)*nb_vtx<<1);
		t_index=(GLushort*) malloc(sizeof(GLushort)*nb_t_index);
		if(points==NULL || tcoord==NULL || t_index==NULL) {
			printf("ERROR: not enough memory!!\n");
			exit(1);
			}

		uint16	i;
		uint8	x,y;

		float ww = w * 0.0666666666667f;
		float hh = h * 0.0666666666667f;
		for(y=0;y<16;y++) {						// Maillage (sommets)
			uint16	seg = y<<4;
			float yy = y * 0.0666666666667f;
			for(x=0;x<16;x++) {
				uint16	offset = seg+x;
				points[offset].x=(x-7.5f)*ww;
				points[offset].z=(y-7.5f)*hh;
				tcoord[offset<<1]=x*0.0666666666667f;
				tcoord[(offset<<1)+1]=yy;
				}
			}
		for(y=0;y<15;y++) {						// Maillage (triangles)
			uint16	seg = y*15;
			for(x=0;x<15;x++) {
				uint16 offset = (seg+x)*6;
				t_index[offset++]=(y<<4)+x;
				t_index[offset++]=((y+1)<<4)+x;
				t_index[offset++]=(y<<4)+x+1;
				t_index[offset++]=(y<<4)+x+1;
				t_index[offset++]=((y+1)<<4)+x;
				t_index[offset]=((y+1)<<4)+x+1;
				}
			}

		uint32 tmp[16][16];

		uint32 med=0;
		uint32 div=0;
		for(y=0;y<16;y++)						// Carte miniature en nuances de gris
			for(x=0;x<16;x++) {
				uint32 c=0;
				uint32 n=0;
				bool zero=false;
				for(uint32 py=y*bmp->h>>4;py<(y+1)*bmp->h>>4;py++)
					for(uint32 px=x*bmp->w>>4;px<(x+1)*bmp->w>>4;px++) {
						uint32 pc=getpixel(bmp,px,py);
						c+=getr(pc)+getg(pc)+getb(pc);
						if(geta(pc)<128 || (pc&0xFFFFFF)==0xFF00FF)
							zero=true;
						n+=3;
						}
				if(zero)	{	c=-0xFFFFFF;	n=1;	}
				tmp[y][x]=c/n;
				if(!zero) {
					med+=tmp[y][x];
					div++;
					}
				}
		if(div==0)	div=1;						// Il y a des trucs bizarres des fois!
		med=(med+(div>>1))/div;
		for(y=0;y<16;y++)						// Carte miniature en nuances de gris
			for(x=0;x<16;x++)
				if(tmp[y][x]==-0xFFFFFF)
					tmp[y][x]=med;

		points[0].y=0.0f;
		for(y=1;y<16;y++)			// x=0
			points[y<<4].y=0.0f;
		for(x=1;x<16;x++)			// y=0
			points[x].y=0.0f;
		for(y=1;y<16;y++)
			for(x=1;x<16;x++) {
				int d_h0=tmp[y][x-1]-med;
				int d_h1=tmp[y-1][x]-med;
				int d_h=tmp[y][x]-med;
				float l=sqrt((float)(d_h0*d_h0+d_h1*d_h1));
				float dhx,dhy;
				if(l==0.0f)
					dhx=dhy=0.0f;
				else {
					dhx=d_h*abs(d_h0)/l;
					dhy=d_h*abs(d_h1)/l;
					}
				points[(y<<4)+x].y=(points[(y<<4)+x-1].y+dhx+points[((y-1)<<4)+x].y+dhy)*0.5f;
				}
		float maxh=0.0f,minh=0.0f;
		for(i=0;i<256;i++) {
			if(minh>points[i].y)	minh=points[i].y;
			if(maxh<points[i].y)	maxh=points[i].y;
			}
		for(y=1;y<16;y++)			// x=15
			points[(y<<4)+15].y=minh;
		for(x=1;x<16;x++)			// y=15
			points[240+x].y=minh;
		if(maxh==minh)
			for(i=0;i<256;i++)
				points[i].y=0.0f;
		else
			for(i=0;i<256;i++)
				points[i].y=(points[i].y-minh)*max_h/(maxh-minh);

		if(g_useTextureCompression)
			allegro_gl_set_texture_format(GL_COMPRESSED_RGBA_ARB);
		else
			allegro_gl_set_texture_format(GL_RGB5_A1);
		gltex[0]=allegro_gl_make_texture(bmp);	dtex=1;
		glBindTexture(GL_TEXTURE_2D,gltex[0]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);

/*--------------------------------------------------------------------------------------*/

		N=(VECTOR*) malloc(sizeof(VECTOR)*nb_vtx);
		for(i=0;i<nb_vtx;i++)
			N[i].x=N[i].z=N[i].y=0.0f;
		for(i=0;i<nb_t_index;i+=3) {
			VECTOR AB,AC,Normal;
			AB=points[t_index[i]]>>points[t_index[i+1]];
			AC=points[t_index[i]]>>points[t_index[i+2]];
			Normal=AB*AC;	Normal.Unit();
			for(int e=0;e<3;e++)
				N[t_index[i+e]]=N[t_index[i+e]]+Normal;
			}
		for(i=0;i<nb_vtx;i++)
			N[i].Unit();
	}

	bool OBJECT::draw(SCRIPT_DATA *data_s,bool sel_primitive,bool alset,bool notex,int side,bool chg_col)
	{
		glPushMatrix();
		glTranslatef(pos_from_parent.x,pos_from_parent.y,pos_from_parent.z);
		bool hide=false;
		if(script_index>=0 && data_s) {
			glTranslatef(data_s->axe[0][script_index].pos, data_s->axe[1][script_index].pos, data_s->axe[2][script_index].pos);
			glRotatef(data_s->axe[0][script_index].angle, 1.0f, 0.0f, 0.0f);
			glRotatef(data_s->axe[1][script_index].angle, 0.0f, 1.0f, 0.0f);
			glRotatef(data_s->axe[2][script_index].angle, 0.0f, 0.0f, 1.0f);
			hide=data_s->flag[script_index]&FLAG_HIDE;
			}
		bool set=false;
		if(nb_t_index>0 && nb_vtx>0 && !hide && t_index!=NULL) {
			bool activated_tex=false;
			if(surface.Flag&SURFACE_ADVANCED) {
				glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
				glEnableClientState(GL_NORMAL_ARRAY);
				alset=false;
				set=false;
				if(chg_col || notex) {
					if(surface.Flag&SURFACE_PLAYER_COLOR)
						glColor4f(player_color[side*3],player_color[side*3+1],player_color[side*3+2],surface.Color[3]);		// Couleur de matière
					else
						glColor4f(surface.Color[0],surface.Color[1],surface.Color[2],surface.Color[3]);		// Couleur de matière
					}

				if(surface.Flag&SURFACE_GOURAUD)			// Type d'éclairage
					glShadeModel (GL_SMOOTH);
				else
					glShadeModel (GL_FLAT);

				if(surface.Flag&SURFACE_LIGHTED)			// Eclairage
					glEnable(GL_LIGHTING);
				else
					glDisable(GL_LIGHTING);

				if(surface.Flag&SURFACE_BLENDED) {			// La transparence
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glEnable(GL_BLEND);
					}
				else
					glDisable(GL_BLEND);

				if(surface.Flag&SURFACE_TEXTURED && !notex) {		// Les textures et effets de texture
					activated_tex=true;
					for(int j=0;j<surface.NbTex;j++) {
						glActiveTextureARB(GL_TEXTURE0_ARB + j);
						glEnable(GL_TEXTURE_2D);
						glBindTexture(GL_TEXTURE_2D,surface.gltex[j]);
						if(j==surface.NbTex-1 && surface.Flag&SURFACE_REFLEC) {
							glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
							glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
							glEnable(GL_TEXTURE_GEN_S);
							glEnable(GL_TEXTURE_GEN_T);
							glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE_EXT);
							glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT,GL_INTERPOLATE);

							glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB_EXT,GL_TEXTURE);
							glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB_EXT,GL_SRC_COLOR);
							glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB_EXT,GL_PREVIOUS_EXT);
							glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND1_RGB_EXT,GL_SRC_COLOR);
							glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB_EXT,GL_CONSTANT_EXT);
							glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND2_RGB_EXT,GL_SRC_COLOR);
							glTexEnvfv(GL_TEXTURE_ENV,GL_TEXTURE_ENV_COLOR,surface.RColor);
							}
						else {
							glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
							glDisable(GL_TEXTURE_GEN_S);
							glDisable(GL_TEXTURE_GEN_T);
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
							}
						}
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					for(int j=0;j<surface.NbTex;j++) {
						glClientActiveTextureARB(GL_TEXTURE0_ARB + j);
						glTexCoordPointer(2, GL_FLOAT, 0, tcoord);
						}
					}
				else {
					glDisable(GL_TEXTURE_2D);
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					}
				}
			else {
				if(!alset) {
					glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
					glEnableClientState(GL_NORMAL_ARRAY);
				if(notex)
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					else
						glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					glEnable(GL_LIGHTING);
					if(notex)
						glDisable(GL_TEXTURE_2D);
					else
						glEnable(GL_TEXTURE_2D);
					alset=true;
					}
				set=true;
				if(!dtex) {
					alset=false;
					glDisable(GL_TEXTURE_2D);
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					}
				if(!notex && dtex) {
					if(side<dtex && side>=0)
						glBindTexture(GL_TEXTURE_2D,gltex[side]);
					else
						glBindTexture(GL_TEXTURE_2D,gltex[0]);
					glTexCoordPointer(2, GL_FLOAT, 0, tcoord);
					}
				}
			glVertexPointer( 3, GL_FLOAT, 0, points);
			glNormalPointer( GL_FLOAT, 0, N);
			glDrawElements(GL_TRIANGLES, nb_t_index,GL_UNSIGNED_SHORT,t_index);		// dessine le tout

			if((surface.Flag&(SURFACE_ADVANCED | SURFACE_GOURAUD))==SURFACE_ADVANCED)
				glShadeModel (GL_SMOOTH);

			if(activated_tex) {
				for(int j=0;j<surface.NbTex;j++) {
					glActiveTextureARB(GL_TEXTURE0_ARB + j);
					glDisable(GL_TEXTURE_2D);
					glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
					glDisable(GL_TEXTURE_GEN_S);
					glDisable(GL_TEXTURE_GEN_T);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
					}
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				}
			}
#ifdef DEBUG_MODE_3DO
		if(nb_l_index>0 && nb_vtx>0) {
			glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
			glDisableClientState(GL_NORMAL_ARRAY);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisable(GL_LIGHTING);
			glDisable(GL_TEXTURE_2D);
			alset=false;
			if(!set)
				glVertexPointer( 3, GL_FLOAT, 0, points);
			set=true;
			glDrawElements(GL_LINES, nb_l_index,GL_UNSIGNED_SHORT,l_index);		// dessine le tout
			}
#endif
		if(sel_primitive && selprim>=0 && nb_vtx>0 && (data_s==NULL || (data_s!=NULL && !data_s->explode))) {
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_NORMAL_ARRAY);
			glDisable(GL_LIGHTING);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_CULL_FACE);
			if(!set)
				glVertexPointer( 3, GL_FLOAT, 0, points);
			glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
			glColor3f(1.0f,1.0f,0.0f);
			glDrawElements(GL_QUADS, 4,GL_UNSIGNED_SHORT,sel);		// dessine la primitive de sélection
			if(notex) {
				float var=fabs(1.0f-(Atimer%MSEC_TO_TIMER(1000))/(float)MSEC_TO_TIMER(500));
				glColor3f(0.0f,var,0.0f);
				}
			else
				glColor3f(1.0f,1.0f,1.0f);
			glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
			glEnable(GL_CULL_FACE);
			alset=false;
			}
		if(child) {
			if(data_s==NULL || !data_s->explode)
				glPushMatrix();
			else {
				glPopMatrix();
				glPushMatrix();
				}
			alset=child->draw(data_s,sel_primitive,alset,notex,side,chg_col);
			glPopMatrix();
			}
		if(next) {
			glPopMatrix();
			glPushMatrix();
			alset=next->draw(data_s,sel_primitive,alset,notex,side,chg_col);
			glPopMatrix();
			}
		else
			glPopMatrix();
		return alset;
	}

	void OBJECT::compute_coord(SCRIPT_DATA *data_s,VECTOR *pos,bool c_part,int p_tex,VECTOR *target,POINTF *upos,MATRIX_4x4 *M,float size,VECTOR *center,bool reverse)
	{
		if(!emitter && c_part)	return;
		VECTOR opos=*pos;
		MATRIX_4x4 OM;
		if(M)
			OM=*M;
		if(script_index>=0 && data_s) {
			if(M) {
				VECTOR ipos;
				ipos.x=data_s->axe[0][script_index].pos;
				ipos.y=data_s->axe[1][script_index].pos;
				ipos.z=data_s->axe[2][script_index].pos;
				*pos=*pos+(pos_from_parent+ipos)*(*M);
				*M=RotateZ(data_s->axe[2][script_index].angle*DEG2RAD)*RotateY(data_s->axe[1][script_index].angle*DEG2RAD)*RotateX(data_s->axe[0][script_index].angle*DEG2RAD)*(*M);
				if(nb_l_index==2) {
					data_s->dir[script_index]=(points[l_index[0]]>>points[l_index[1]])*(*M);
					data_s->dir[script_index].Unit();
					ipos.x=points[l_index[0]].x;
					ipos.y=points[l_index[0]].y;
					ipos.z=points[l_index[0]].z;
					data_s->pos[script_index]=*pos+ipos*(*M);
					}
				else
					data_s->pos[script_index]=*pos;
				}
			}
		if((nb_t_index==0 || nb_vtx==0) && c_part && child==NULL && next==NULL) {			// Emission d'une particule
			VECTOR Dir;
			float life=1.0f;
			byte nb=(rand_from_table()%30)+1;
			for(byte i=0;i<nb;i++) {
				VECTOR t_mod;
				t_mod.x=((rand_from_table()%2001)-1000)*0.001f;
				t_mod.y=((rand_from_table()%2001)-1000)*0.001f;
				t_mod.z=((rand_from_table()%2001)-1000)*0.001f;
				t_mod.Unit();
				t_mod=(rand_from_table()%1001)*0.001f*size*t_mod;
				if(center)
					t_mod=t_mod+(*center);
				float dist=Dir.Norm();
				float speed=1.0f;
				if(reverse) {
					POINTF O;
					O.x=O.y=O.z=0.0f;
					Dir=*pos-(t_mod+*target);
					Dir.x+=upos->x;
					Dir.y+=upos->y;
					Dir.z+=upos->z;
					particle_engine.emit_part(O+t_mod+*target,Dir,p_tex,1,speed,life,2.0f,true);
					}
				else {
					Dir=t_mod+*target-*pos;
					Dir.x-=upos->x;
					Dir.y-=upos->y;
					Dir.z-=upos->z;
					particle_engine.emit_part(*upos+*pos,Dir,p_tex,1,speed,life,2.0f,true);
					}
				}
			}
		if(child)
			child->compute_coord(data_s,pos,c_part,p_tex,target,upos,M,size,center,reverse);
		*pos=opos;
		if(M)
			*M=OM;
		if(next)
			next->compute_coord(data_s,pos,c_part,p_tex,target,upos,M,size,center,reverse);
	}

	bool OBJECT::draw_shadow(VECTOR Dir,SCRIPT_DATA *data_s,bool alset)
	{
		glPushMatrix();
		glTranslatef(pos_from_parent.x,pos_from_parent.y,pos_from_parent.z);
		bool hide=false;
		VECTOR ODir=Dir;
		if(script_index>=0 && data_s) {
			glTranslatef(data_s->axe[0][script_index].pos, data_s->axe[1][script_index].pos, data_s->axe[2][script_index].pos);
			glRotatef(data_s->axe[0][script_index].angle, 1.0f, 0.0f, 0.0f);
			glRotatef(data_s->axe[1][script_index].angle, 0.0f, 1.0f, 0.0f);
			glRotatef(data_s->axe[2][script_index].angle, 0.0f, 0.0f, 1.0f);
			Dir=((Dir*RotateX(-data_s->axe[0][script_index].angle*DEG2RAD))*RotateY(-data_s->axe[1][script_index].angle*DEG2RAD))*RotateZ(-data_s->axe[2][script_index].angle*DEG2RAD);
			hide=data_s->flag[script_index]&FLAG_HIDE;
			}
			if(nb_t_index>0 && !hide) {
				if(!alset) {
					glDisable(GL_LIGHTING);
					glDisable(GL_TEXTURE_2D);
					glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
					glDisableClientState(GL_NORMAL_ARRAY);
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					alset=true;
					}
/*-----------------Code de calcul du cone d'ombre-------------------------*/
				if(shadow_index==NULL)
					shadow_index=(GLushort*) malloc(sizeof(GLushort)*nb_t_index*12);
				register short nb_idx=0;

				if(t_line==NULL) {									// Repère les arêtes
					t_line=(short*) malloc(sizeof(short)*nb_t_index);
					line_v_idx[0]=(short*) malloc(sizeof(short)*nb_t_index);
					line_v_idx[1]=(short*) malloc(sizeof(short)*nb_t_index);
					nb_line=0;
					for(short i=0;i<nb_t_index;i+=3)
						for(byte a=0;a<3;a++) {
							short idx=-1;
							for(short e=0;e<nb_line;e++)
								if((line_v_idx[0][e]==t_index[i+a] && line_v_idx[1][e]==t_index[i+((a+1)%3)]) || (line_v_idx[0][e]==t_index[i+((a+1)%3)] && line_v_idx[1][e]==t_index[i+a])) {
									idx=e;
									break;
									}
							if(idx==-1) {
								line_v_idx[0][nb_line]=t_index[i+a];
								line_v_idx[1][nb_line]=t_index[i+((a+1)%3)];
								idx=nb_line;	nb_line++;
								}
							t_line[i+a]=idx;
							}
						line_on = new bool[nb_line];
					}

				memset((bool*)line_on,0,nb_line);

				for(register short i=0;i<nb_t_index;i+=3) {
					if((N[t_index[i]]%Dir)>=0.0f) continue;
					line_on[t_line[i]]^=true;
					line_on[t_line[i+1]]^=true;
					line_on[t_line[i+2]]^=true;
					}
				for(short i=0;i<nb_line;i++) {
					if(!line_on[i])	continue;
					points[line_v_idx[0][i]+nb_vtx]=points[line_v_idx[0][i]]+Dir;		// Projection
					points[line_v_idx[1][i]+nb_vtx]=points[line_v_idx[1][i]]+Dir;

					shadow_index[nb_idx++]=line_v_idx[0][i];			shadow_index[nb_idx++]=line_v_idx[1][i];
					shadow_index[nb_idx++]=line_v_idx[1][i]+nb_vtx;		shadow_index[nb_idx++]=line_v_idx[0][i]+nb_vtx;
					}
				if( nb_idx ) {
					glVertexPointer( 3, GL_FLOAT, 0, points);
					glDrawElements(GL_QUADS, nb_idx,GL_UNSIGNED_SHORT,shadow_index);		// dessine le tout
					}
				}
		if(child) {
			if(data_s==NULL || !data_s->explode) {
				glPushMatrix();
				alset=child->draw_shadow(Dir,data_s,alset);
				}
			else {
				glPopMatrix();
				glPushMatrix();
				alset=child->draw_shadow(ODir,data_s,alset);
				}
			glPopMatrix();
			}
		if(next) {
			glPopMatrix();
			glPushMatrix();
			alset=next->draw_shadow(ODir,data_s,alset);
			glPopMatrix();
			}
		else
			glPopMatrix();
		return alset;
	}

	bool OBJECT::draw_shadow_basic(VECTOR Dir,SCRIPT_DATA *data_s,bool alset)
	{
		glPushMatrix();
		glTranslatef(pos_from_parent.x,pos_from_parent.y,pos_from_parent.z);
		bool hide=false;
		VECTOR ODir=Dir;
		if(script_index>=0 && data_s) {
			glTranslatef(data_s->axe[0][script_index].pos, data_s->axe[1][script_index].pos, data_s->axe[2][script_index].pos);
			glRotatef(data_s->axe[0][script_index].angle, 1.0f, 0.0f, 0.0f);
			glRotatef(data_s->axe[1][script_index].angle, 0.0f, 1.0f, 0.0f);
			glRotatef(data_s->axe[2][script_index].angle, 0.0f, 0.0f, 1.0f);
			Dir=((Dir*RotateX(-data_s->axe[0][script_index].angle*DEG2RAD))*RotateY(-data_s->axe[1][script_index].angle*DEG2RAD))*RotateZ(-data_s->axe[2][script_index].angle*DEG2RAD);
			hide=data_s->flag[script_index]&FLAG_HIDE;
			}
			if(nb_t_index>0 && !hide) {
				if(!alset) {
					glDisable(GL_LIGHTING);
					glDisable(GL_TEXTURE_2D);
					glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
					glDisableClientState(GL_NORMAL_ARRAY);
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
					alset=true;
					}
/*-----------------Code de calcul du cone d'ombre-------------------------*/
				if(shadow_index==NULL)
					shadow_index=(GLushort*) malloc(sizeof(GLushort)*nb_t_index*12);
				register short nb_idx=0;

				if(t_line==NULL) {									// Repère les arêtes
					t_line=(short*) malloc(sizeof(short)*nb_t_index);
					line_v_idx[0]=(short*) malloc(sizeof(short)*nb_t_index);
					line_v_idx[1]=(short*) malloc(sizeof(short)*nb_t_index);
					nb_line=0;
					for(short i=0;i<nb_t_index;i+=3)
						for(byte a=0;a<3;a++) {
							short idx=-1;
							for(short e=0;e<nb_line;e++)
								if((line_v_idx[0][e]==t_index[i+a] && line_v_idx[1][e]==t_index[i+((a+1)%3)]) || (line_v_idx[0][e]==t_index[i+((a+1)%3)] && line_v_idx[1][e]==t_index[i+a])) {
									idx=e;
									break;
									}
							if(idx==-1) {
								line_v_idx[0][nb_line]=t_index[i+a];
								line_v_idx[1][nb_line]=t_index[i+((a+1)%3)];
								idx=nb_line;	nb_line++;
								}
							t_line[i+a]=idx;
							}
						line_on = new bool[nb_line];
					}

//				bool line_on[nb_line];
				memset((bool*)line_on,0,nb_line);
//				for(int i=0;i<nb_line;i++)	line_on[i]=false;

				for(register short i=0;i<nb_t_index;i+=3) {
					VECTOR N=this->N[t_index[i]];
					if((N%Dir)>=0.0f) continue;
					line_on[t_line[i]]^=true;
					line_on[t_line[i+1]]^=true;
					line_on[t_line[i+2]]^=true;
					}
				for(short i=0;i<nb_line;i++) {
					if(!line_on[i])	continue;
					points[line_v_idx[0][i]+nb_vtx]=points[line_v_idx[0][i]]+Dir;		// Projection calculations
					points[line_v_idx[1][i]+nb_vtx]=points[line_v_idx[1][i]]+Dir;

					shadow_index[nb_idx++]=line_v_idx[0][i];			shadow_index[nb_idx++]=line_v_idx[1][i];
					shadow_index[nb_idx++]=line_v_idx[1][i]+nb_vtx;		shadow_index[nb_idx++]=line_v_idx[0][i]+nb_vtx;
					}
				glVertexPointer( 3, GL_FLOAT, 0, points);
				glFrontFace(GL_CW);						// 1ère passe
				glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
				glDrawElements(GL_QUADS, nb_idx,GL_UNSIGNED_SHORT,shadow_index);		// dessine le tout

				glFrontFace(GL_CCW);  						// 2nd passe
				glStencilOp(GL_KEEP,GL_KEEP,GL_DECR);
				glDrawElements(GL_QUADS, nb_idx,GL_UNSIGNED_SHORT,shadow_index);		// dessine le tout
				}
		if(child) {
			if(data_s==NULL || !data_s->explode) {
				glPushMatrix();
				alset=child->draw_shadow_basic(Dir,data_s,alset);
				}
			else {
				glPopMatrix();
				glPushMatrix();
				alset=child->draw_shadow_basic(ODir,data_s,alset);
				}
			glPopMatrix();
			}
		if(next) {
			glPopMatrix();
			glPushMatrix();
			alset=next->draw_shadow_basic(ODir,data_s,alset);
			glPopMatrix();
			}
		else
			glPopMatrix();
		return alset;
	}

	bool OBJECT::hit(VECTOR Pos,VECTOR Dir,SCRIPT_DATA *data_s,VECTOR *I,MATRIX_4x4 M)
	{
		MATRIX_4x4 OM=M;
		MATRIX_4x4 AM=Scale(1.0f);
		MATRIX_4x4 M_Dir=M;
		bool hide=false;
		VECTOR ODir=Dir;
		VECTOR OPos=Pos;
		bool is_hit=false;

		VECTOR T=pos_from_parent;
		if(script_index>=0 && data_s) {
			T.x+=data_s->axe[0][script_index].pos;
			T.y+=data_s->axe[1][script_index].pos;
			T.z+=data_s->axe[2][script_index].pos;
			M_Dir=M*RotateX(-data_s->axe[0][script_index].angle*DEG2RAD)*RotateY(-data_s->axe[1][script_index].angle*DEG2RAD)*RotateZ(-data_s->axe[2][script_index].angle*DEG2RAD);
			M=RotateX(-data_s->axe[0][script_index].angle*DEG2RAD)*RotateY(-data_s->axe[1][script_index].angle*DEG2RAD)*RotateZ(-data_s->axe[2][script_index].angle*DEG2RAD);
			AM=RotateZ(data_s->axe[2][script_index].angle*DEG2RAD)*RotateY(data_s->axe[1][script_index].angle*DEG2RAD)*RotateX(data_s->axe[0][script_index].angle*DEG2RAD);
			hide=data_s->flag[script_index]&FLAG_HIDE;
			}
		VECTOR MP;
			if(nb_t_index>0 && !hide) {
				POINTF A,B,C,O;
				O.x=O.y=O.z=0.0f;
				Pos=(Pos-T)*M;
				Dir=Dir*M_Dir;
				Dir.Unit();
//-----------------Code de calcul d'intersection--------------------------
				for(register short i=0;i<nb_t_index;i+=3) {
					A=points[t_index[i]];
					B=points[t_index[i+1]];
					C=points[t_index[i+2]];
					VECTOR AB=A>>B;
					VECTOR AC=A>>C;
					VECTOR N=AB*AC;
					if(N%Dir==0.0f)	continue;
					float dist=-((Pos-(O>>A))%N)/(N%Dir);
					VECTOR P_p=Pos+dist*Dir;

					if(is_hit && (MP-Pos)%Dir<(P_p-Pos)%Dir)	continue;

					float a,b,c;		// Coefficients pour que P soit le barycentre de A,B,C
					VECTOR AP=P_p-(O>>A);
					float pre_cal = AB.x*AC.y-AB.y*AC.x;
					if(AC.y!=0.0f && pre_cal!=0.0f) {
						b=(AP.x*AC.y-AP.y*AC.x)/pre_cal;
						a=(AP.y-b*AB.y)/AC.y;
						}
					else {
					if(AB.x!=0.0f && pre_cal!=0.0f) {
						a=(AP.y*AB.x-AP.x*AB.y)/pre_cal;
						b=(AP.x-a*AC.x)/AB.x;
						}
					else {
						pre_cal = AB.x*AC.z-AB.z*AC.x;
					if(AC.z!=0.0f && pre_cal!=0.0f) {
						b=(AP.x*AC.z-AP.z*AC.x)/pre_cal;
						a=(AP.z-b*AB.z)/AC.z;
						}
					else {
						pre_cal=-pre_cal;
					if(AB.z!=0.0f && pre_cal!=0.0f) {
						a=(AP.x*AB.z-AP.z*AB.x)/pre_cal;
						b=(AP.z-a*AC.z)/AB.z;
						}
					else {
						pre_cal = AB.y*AC.x-AB.x*AC.y;
					if(AC.x!=0.0f && pre_cal!=0.0f) {
						b=(AP.y*AC.x-AP.x*AC.y)/pre_cal;
						a=(AP.x-b*AB.x)/AC.x;
						}
					else {
					if(AB.y!=0.0f && pre_cal!=0.0f) {
						a=(AP.x*AB.y-AP.y*AB.x)/pre_cal;
						b=(AP.y-a*AC.y)/AB.y;
						}
					else continue;		// Saute le point s'il n'est pas positionnable
						}	}	}	}	}
					c=1.0f-a-b;
					if(a<0.0f || b<0.0f || c<0.0f) continue;		// Le point n'appartient pas au triangle
					MP=P_p;
					is_hit=true;
					}
				}
		if(child) {
			VECTOR MP2;
			bool nhit=child->hit(Pos,ODir,data_s,&MP2,M_Dir);
			if(nhit && !is_hit)
				MP=MP2;
			else if(nhit && is_hit)
				if((MP2-Pos)%Dir<(MP-Pos)%Dir)
					MP=MP2;
			is_hit|=nhit;
			}
		if(is_hit)
			MP=(MP*AM)+T;
		if(next) {
			VECTOR MP2;
			bool nhit=next->hit(OPos,ODir,data_s,&MP2,OM);
			Dir=ODir*OM;
			if(nhit && !is_hit)
				MP=MP2;
			else if(nhit && is_hit)
				if((MP2-OPos)%Dir<(MP-OPos)%Dir)
					MP=MP2;
			is_hit|=nhit;
			}
		if(is_hit)
			*I=MP;
		return is_hit;
	}


float OBJECT::print_struct(float Y,float X,TA3D::INTERFACES::GFX_FONT fnt)
{
//	if(nb_vtx==0 || nb_prim==0 || nb_p_index>0 || nb_l_index>0)
		gfx->print(fnt,X,Y,0.0f,0xFFFFFF,format("%s",name));
		gfx->print(fnt,320.0f,Y,0.0f,0xFFFFFF,format("(v:%d",nb_vtx));
		gfx->print(fnt,368.0f,Y,0.0f,0xFFFFFF,format(",p:%d",nb_prim));
		gfx->print(fnt,416.0f,Y,0.0f,0xFFFFFF,format(",t:%d",nb_t_index));
		gfx->print(fnt,464.0f,Y,0.0f,0xFFFFFF,format(",l:%d",nb_l_index));
		gfx->print(fnt,512.0f,Y,0.0f,0xFFFFFF,format(",p:%d)",nb_p_index));
//		allegro_gl_printf(fnt,320.0f,Y,0.0f,0xFFFFFF,"(v:%d,p:%d,t:%d,l:%d,p:%d)",nb_vtx,nb_prim,nb_t_index,nb_l_index,nb_p_index);
	Y+=8.0f;
	if(child)
		Y=child->print_struct(Y,X+8.0f,fnt);
	if(next)
		Y=next->print_struct(Y,X,fnt);
	return Y;
}

int MODEL_MANAGER::load_all(void (*progress)(float percent,char *msg))
{
	List<String> file_list;
	sint32 new_nb_models=HPIManager->GetFilelist("objects3d\\*.3do",&file_list);

	if(new_nb_models<=0)
		return -1;

	MODEL *n_model=(MODEL*) malloc(sizeof(MODEL)*(nb_models+new_nb_models));
	char **n_name=(char**) malloc(sizeof(char*)*(nb_models+new_nb_models));
	if(model) {
		memcpy(n_model,model,sizeof(MODEL)*nb_models);
		free(model);
		memcpy(n_name,name,sizeof(char*)*nb_models);
		free(name);
		}
	model=n_model;
	name=n_name;
	int i=0,n=0;
	for(List<String>::iterator e=file_list.begin();e!=file_list.end();e++) {
		if(progress!=NULL)
			progress((100.0f+n*50.0f/(new_nb_models+1))/7.0f,translation_manager.get(23));
		n++;
		model[i+nb_models].init();
		name[i+nb_models]=strdup(e->c_str());
		if(get_model(name[i+nb_models])==NULL) {				// Vérifie si le modèle n'est pas déjà chargé
			byte *data=HPIManager->PullFromHPI(*e);
			model[i+nb_models].load_3do(data);
			free(data);
			i++;
			}
		}
	nb_models+=i;

	file_list.clear();
	new_nb_models=HPIManager->GetFilelist("objects3d\\*.3dm",&file_list);

	if(new_nb_models<=0)
		return -1;

	n_model=(MODEL*) malloc(sizeof(MODEL)*(nb_models+new_nb_models));
	n_name=(char**) malloc(sizeof(char*)*(nb_models+new_nb_models));
	if(model) {
		memcpy(n_model,model,sizeof(MODEL)*nb_models);
		free(model);
		memcpy(n_name,name,sizeof(char*)*nb_models);
		free(name);
		}
	model=n_model;
	name=n_name;
	i=0;
	n=0;
	for(List<String>::iterator e=file_list.begin();e!=file_list.end();e++) {
		if(progress!=NULL)
			progress((100.0f+(50.0f+n*50.0f/(new_nb_models+1)))/7.0f,translation_manager.get(23));
		n++;
		model[i+nb_models].init();
		name[i+nb_models]=strdup(e->c_str());
		if(get_model(name[i+nb_models])==NULL) {				// Vérifie si le modèle n'est pas déjà chargé
			byte *data=HPIManager->PullFromHPI(*e);
			model[i+nb_models].load_3dm(data);
			free(data);
			i++;
			}
		}
	nb_models+=i;

	return 0;
}

void MODEL::load_asc(char *filename,float size)		// Charge un fichier au format *.ASC
{

	destroy();			// Puisqu'on charge

	float *coor[3];
	coor[0]=(float *) malloc(100000*sizeof(float));
	coor[1]=(float *) malloc(100000*sizeof(float));
	coor[2]=(float *) malloc(100000*sizeof(float));
	int *face[3];
	face[0]=(int *) malloc(100000*sizeof(int));
	face[1]=(int *) malloc(100000*sizeof(int));
	face[2]=(int *) malloc(100000*sizeof(int));

	if(coor[0]==NULL || coor[1]==NULL || coor[2]==NULL
	|| face[0]==NULL || face[1]==NULL || face[2]==NULL) {
		if(coor[0]!=NULL) free(coor[0]);
		if(coor[1]!=NULL) free(coor[1]);
		if(coor[2]!=NULL) free(coor[2]);
		if(face[0]!=NULL) free(face[0]);
		if(face[1]!=NULL) free(face[1]);
		if(face[2]!=NULL) free(face[2]);
		return;
		}

	long nbp=0,nbf=0;

	FILE  *fichier;
	char  chaine[200];
	char  *fin;
	long  i,j;
	char  temp[50];
	float x,y,z;
	int   decalage=0;
	int nbpt;
	int test;
	float dx,dy,dz;
	float xmin=0xFFFFFF,ymin=0xFFFFFF,zmin=0xFFFFFF,
		xmax=-0xFFFFFF,ymax=-0xFFFFFF,zmax=-0xFFFFFF;
  
	int StructD[4096];     // Données pour la restitution de la structure
	char *StructName[4096];
	int NbStruct=0;

	nbpt=0;
	test=0;
	nbf=0;
	nbp=0;

	if ((fichier = TA3D_OpenFile(filename,"rt"))==NULL)	{
		Console->AddEntry("Impossible d'ouvrir le fichier %s en lecture",filename);
		return;
	}

	do
	{
		// On lit le fichier contenant les informations sur l'objet
		fin=fgets(chaine,200,fichier);
		strupr(chaine);
		if (!strncmp(chaine,"VERTEX",6))	{
			if (strncmp(chaine,"VERTEX LIST",11))	{
			// Lecture des coordonnées d'un point
				i=6;

				while(chaine[i]!='X') i++;
				i+=2;
				while(chaine[i]==' ') i++;
				sscanf(chaine+i,"%f",&x);

				while(chaine[i]!='Y') i++;
				i+=2;
				while(chaine[i]==' ') i++;
				sscanf(chaine+i,"%f",&y);

				while(chaine[i]!='Z') i++;
				i+=2;
				while(chaine[i]==' ') i++;
				sscanf(chaine+i,"%f",&z);

				coor[0][nbp]=x;
				coor[1][nbp]=y;
				coor[2][nbp]=z;

				if(x<xmin) xmin=x;
				if(x>xmax) xmax=x;
				if(y<ymin) ymin=y;
				if(y>ymax) ymax=y;
				if(z<zmin) zmin=z;
				if(z>zmax) zmax=z;

				nbp++;
				}
			}
		else {
			if (!strncmp(chaine,"FACE",4))	{
				if (strncmp(chaine,"FACE LIST",9))	{
					// Lecture d'une facette
					i=j=4;
					while(chaine[i]!='A') i++;
					i+=2;
					while(chaine[i]==' ') i++;
					j=i;
					while(chaine[j]!=' ') j++;
					strncpy(temp,chaine+i,j-i);
					temp[j-i]=0;
					face[0][nbf]=atoi(temp)+decalage;

					while(chaine[i]!='B') i++;
					i+=2;
					while(chaine[i]==' ') i++;
					j=i;
					while(chaine[j]!=' ') j++;
					strncpy(temp,chaine+i,j-i);
					temp[j-i]=0;
					face[1][nbf]=atoi(temp)+decalage;

					while(chaine[i]!='C') i++;
					i+=2;
					while(chaine[i]==' ') i++;
					j=i;
					while(chaine[j]!=' ') j++;
					strncpy(temp,chaine+i,j-i);
					temp[j-i]=0;
					face[2][nbf]=atoi(temp)+decalage;

					nbf++;
					}
				}
			else
				if (!strncmp(chaine,"NAMED OBJECT",12)) {
					StructName[NbStruct]=strdup(chaine+13);
					decalage=nbp;
					StructD[NbStruct++]=nbf;
					}
				}
	} while(fin!=NULL);

	fclose(fichier);

	StructD[NbStruct]=nbf;

	OBJECT *cur = &obj;

	dx=-(xmin+xmax)*0.5f;
	dy=-(ymin+ymax)*0.5f;
	dz=-(zmin+zmax)*0.5f;
	xmin=xmax-xmin;
	ymin=ymax-ymin;
	zmin=zmax-zmin;
	xmax=sqrt(xmin*xmin+ymin*ymin+zmin*zmin);
	size=size/xmax;

	for(i=0;i<NbStruct;i++) {			// Crée les différentes parties de la meshe
		if(i>0) {
			cur->next= (OBJECT*) malloc(sizeof(OBJECT));
			cur=cur->next;
			}
		cur->init();
		cur->name=StructName[i];
		cur->nb_prim=StructD[i+1]-StructD[i];
		cur->nb_t_index=cur->nb_prim*3;
		cur->nb_vtx=cur->nb_t_index;
		cur->points = (POINTF*) malloc(sizeof(POINTF)*cur->nb_vtx);
		cur->t_index = (GLushort*) malloc(sizeof(GLushort)*cur->nb_t_index);
		cur->tcoord = (float*) malloc(sizeof(float)*cur->nb_vtx<<1);

		cur->surface.Flag=SURFACE_ADVANCED|SURFACE_GOURAUD|SURFACE_LIGHTED;
		for(int k=0;k<4;k++)
			cur->surface.Color[k]=cur->surface.RColor[k]=1.0f;

		int p_nbf=0;
		int p_nbp=0;

		for(int k=StructD[i];k<StructD[i+1];k++) {		// Compte et organise les points
			cur->t_index[p_nbf++]=p_nbp;
			cur->t_index[p_nbf++]=p_nbp+1;
			cur->t_index[p_nbf++]=p_nbp+2;

			cur->points[p_nbp].x=coor[0][face[0][k]];
			cur->points[p_nbp].y=coor[1][face[0][k]];
			cur->points[p_nbp++].z=coor[2][face[0][k]];

			cur->points[p_nbp].x=coor[0][face[1][k]];
			cur->points[p_nbp].y=coor[1][face[1][k]];
			cur->points[p_nbp++].z=coor[2][face[1][k]];

			cur->points[p_nbp].x=coor[0][face[2][k]];
			cur->points[p_nbp].y=coor[1][face[2][k]];
			cur->points[p_nbp++].z=coor[2][face[2][k]];
			}			// Fin de for(k=StructD[i];k<StructD[i+1];k++)
		}				// Fin de for(i=0;i<NbStruct;i++)

	cur=&obj;
	while(cur) {
		int removed=0;
		for(i=0;i<cur->nb_t_index;i++)				// Remove duplicate points
			for(int e=0;e<i;e++)
				if(cur->points[cur->t_index[i]].x==cur->points[cur->t_index[e]].x && cur->points[cur->t_index[i]].y==cur->points[cur->t_index[e]].y && cur->points[cur->t_index[i]].z==cur->points[cur->t_index[e]].z) {
					cur->t_index[i]=cur->t_index[e];
					removed++;
					break;
					}
		cur->nb_vtx-=removed;
		POINTF *n_points = (POINTF*) malloc(sizeof(POINTF)*cur->nb_vtx);
		int cur_pt=0;
		for(i=0;i<cur->nb_t_index;i++) {
			bool ok=false;
			for(int e=0;e<i;e++)
				if(cur->points[cur->t_index[i]].x==n_points[cur->t_index[e]].x && cur->points[cur->t_index[i]].y==n_points[cur->t_index[e]].y && cur->points[cur->t_index[i]].z==n_points[cur->t_index[e]].z) {
					cur->t_index[i]=cur->t_index[e];
					ok=true;
					break;
					}
			if(ok)	continue;
			n_points[cur_pt]=cur->points[cur->t_index[i]];
			cur->t_index[i]=cur_pt++;
			}
		free(cur->points);
		cur->points=n_points;
		for(i=0;i<cur->nb_vtx;i++) {				// Remove duplicate points
			cur->points[i].x*=size;
			cur->points[i].y*=size;
			cur->points[i].z*=size;
			}

		cur->N = (VECTOR*) malloc(sizeof(VECTOR)*cur->nb_vtx);	// Calculate normals
		for(i=0;i<cur->nb_vtx;i++)
			cur->N[i].x=cur->N[i].z=cur->N[i].y=0.0f;
		for(i=0;i<cur->nb_t_index;i+=3) {
			VECTOR AB,AC,Normal;
			AB=cur->points[cur->t_index[i]]>>cur->points[cur->t_index[i+1]];
			AC=cur->points[cur->t_index[i]]>>cur->points[cur->t_index[i+2]];
			Normal=AB*AC;	Normal.Unit();
			for(int e=0;e<3;e++)
				cur->N[cur->t_index[i+e]]=cur->N[cur->t_index[i+e]]+Normal;
			}
		for(i=0;i<cur->nb_vtx;i++)
			cur->N[i].Unit();
		cur=cur->next;
		}

	free(coor[0]);
	free(coor[1]);
	free(coor[2]);
	free(face[0]);
	free(face[1]);
	free(face[2]);
}

void OBJECT::save_3dm(FILE *dst)
{
	uint8	len = strlen(name);
	fwrite(&len,sizeof(len),1,dst);		// Write the object name
	fwrite(name,len,1,dst);

	fwrite(&pos_from_parent.x,sizeof(pos_from_parent.x),1,dst);
	fwrite(&pos_from_parent.y,sizeof(pos_from_parent.y),1,dst);
	fwrite(&pos_from_parent.z,sizeof(pos_from_parent.z),1,dst);

	fwrite(&nb_vtx,sizeof(nb_vtx),1,dst);
	if(points!=NULL)
		fwrite(points,sizeof(POINTF)*nb_vtx,1,dst);

	fwrite(sel,sizeof(GLushort)*4,1,dst);				// Selection primitive

	fwrite(&nb_p_index,sizeof(nb_p_index),1,dst);		// Write point data
	if(p_index!=NULL)
		fwrite(p_index,sizeof(GLushort)*nb_p_index,1,dst);

	fwrite(&nb_l_index,sizeof(nb_l_index),1,dst);		// Write line data
	if(l_index!=NULL)
		fwrite(l_index,sizeof(GLushort)*nb_l_index,1,dst);

	fwrite(&nb_t_index,sizeof(nb_t_index),1,dst);		// Write triangle data
	if(t_index!=NULL)
		fwrite(t_index,sizeof(GLushort)*nb_t_index,1,dst);

	fwrite(tcoord,sizeof(float)*nb_vtx<<1,1,dst);		// Write texture coordinates

	fwrite(surface.Color,sizeof(float)*4,1,dst);		// Write surface data
	fwrite(surface.RColor,sizeof(float)*4,1,dst);
	fwrite(&surface.Flag,sizeof(surface.Flag),1,dst);
	int tmp=surface.NbTex;
	if(!(surface.Flag&SURFACE_TEXTURED))
		surface.NbTex=0;
	fwrite(&surface.NbTex,sizeof(surface.NbTex),1,dst);
	surface.NbTex=tmp;
	if(surface.Flag&SURFACE_TEXTURED)
	for(uint8 i=0;i<surface.NbTex;i++) {
		BITMAP *tex;
		GLint w,h;
		glBindTexture(GL_TEXTURE_2D,surface.gltex[i]);
		glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_WIDTH,&w);
		glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_HEIGHT,&h);
		tex=create_bitmap_ex(32,w,h);
		glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_UNSIGNED_BYTE,tex->line[0]);

		fwrite(&tex->w,sizeof(tex->w),1,dst);
		fwrite(&tex->h,sizeof(tex->h),1,dst);
		for(int y=0;y<tex->h;y++)
			for(int x=0;x<tex->w;x++)
				fwrite(&((int*)(tex->line[y]))[x],4,1,dst);

		destroy_bitmap(tex);
		}

	if(surface.Flag&SURFACE_GLSL) {		// Save the shader object
		fwrite(&surface.vert_shader_size,4,1,dst);
		fwrite(surface.vert_shader_src,surface.vert_shader_size,1,dst);
		fwrite(&surface.frag_shader_size,4,1,dst);
		fwrite(surface.frag_shader_src,surface.frag_shader_size,1,dst);
		}

	if(child) {
		fputc(1,dst);
		child->save_3dm(dst);
		}
	else
		fputc(0,dst);

	if(next) {
		fputc(1,dst);
		next->save_3dm(dst);
		}
	else
		fputc(0,dst);
}

inline byte *read_from_mem(void *buf,int len,byte *data)
{
	memcpy(buf,data,len);
	return data+len;
}

byte *OBJECT::load_3dm(byte *data)
{
	uint8	len = data[0];	data++;
	name = (char*) malloc(len+1);
	data=read_from_mem(name,len,data);
	name[len]=0;

	data=read_from_mem(&pos_from_parent.x,sizeof(pos_from_parent.x),data);
	data=read_from_mem(&pos_from_parent.y,sizeof(pos_from_parent.y),data);
	data=read_from_mem(&pos_from_parent.z,sizeof(pos_from_parent.z),data);

	data=read_from_mem(&nb_vtx,sizeof(nb_vtx),data);
	if(nb_vtx>0) {
		points = (POINTF*) malloc(sizeof(POINTF)*nb_vtx<<1);
		data=read_from_mem(points,sizeof(POINTF)*nb_vtx,data);
		}
	else
		points=NULL;

	data=read_from_mem(sel,sizeof(GLushort)*4,data);

	data=read_from_mem(&nb_p_index,sizeof(nb_p_index),data);	// Read point data
	if(nb_p_index>0) {
		p_index = (GLushort*) malloc(sizeof(GLushort)*nb_p_index);
		data=read_from_mem(p_index,sizeof(GLushort)*nb_p_index,data);
		}
	else
		p_index=NULL;

	data=read_from_mem(&nb_l_index,sizeof(nb_l_index),data);	// Read line data
	if(nb_l_index>0) {
		l_index = (GLushort*) malloc(sizeof(GLushort)*nb_l_index);
		data=read_from_mem(l_index,sizeof(GLushort)*nb_l_index,data);
		}
	else
		l_index=NULL;

	data=read_from_mem(&nb_t_index,sizeof(nb_t_index),data);	// Read triangle data
	if(nb_t_index>0) {
		t_index = (GLushort*) malloc(sizeof(GLushort)*nb_t_index);
		data=read_from_mem(t_index,sizeof(GLushort)*nb_t_index,data);
		}
	else
		t_index=NULL;

	tcoord = (float*) malloc(sizeof(float)*nb_vtx<<1);
	data=read_from_mem(tcoord,sizeof(float)*nb_vtx<<1,data);

	data=read_from_mem(surface.Color,sizeof(float)*4,data);	// Read surface data
	data=read_from_mem(surface.RColor,sizeof(float)*4,data);
	data=read_from_mem(&surface.Flag,sizeof(surface.Flag),data);
	surface.NbTex=0;
	data=read_from_mem(&surface.NbTex,sizeof(surface.NbTex),data);
	for(uint8 i=0;i<surface.NbTex;i++) {
		int tex_w;
		int tex_h;
		data=read_from_mem(&tex_w,sizeof(tex_w),data);
		data=read_from_mem(&tex_h,sizeof(tex_h),data);

		BITMAP *tex = create_bitmap_ex(32,tex_w,tex_h);
		for(int y=0;y<tex->h;y++)
			for(int x=0;x<tex->w;x++)
				data=read_from_mem(&((int*)(tex->line[y]))[x],4,data);

		allegro_gl_use_alpha_channel(true);
		allegro_gl_set_texture_format(GL_RGBA8);
		surface.gltex[i]=allegro_gl_make_texture(tex);
		allegro_gl_use_alpha_channel(false);
		glBindTexture(GL_TEXTURE_2D,surface.gltex[i]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

		destroy_bitmap(tex);
		}

	if(surface.Flag&SURFACE_GLSL) {					// Fragment & Vertex shaders
		data=read_from_mem(&surface.vert_shader_size,4,data);
		surface.vert_shader_src = (char*) malloc(surface.vert_shader_size+1);
		surface.vert_shader_src[surface.vert_shader_size]=0;
		data=read_from_mem(surface.vert_shader_src,surface.vert_shader_size,data);
		data=read_from_mem(&surface.frag_shader_size,4,data);
		surface.frag_shader_src = (char*) malloc(surface.frag_shader_size+1);
		surface.frag_shader_src[surface.frag_shader_size]=0;
		data=read_from_mem(surface.frag_shader_src,surface.frag_shader_size,data);
		surface.s_shader.load_memory(surface.frag_shader_src,surface.frag_shader_size,surface.vert_shader_src,surface.vert_shader_size);
		}

	N = (VECTOR*) malloc(sizeof(VECTOR)*nb_vtx<<1);	// Calculate normals
	if(nb_t_index>0 && t_index!=NULL) {
		for(int i=0;i<nb_vtx;i++)
			N[i].x=N[i].z=N[i].y=0.0f;
		for(int i=0;i<nb_t_index;i+=3) {
			VECTOR AB,AC,Normal;
			AB=points[t_index[i]]>>points[t_index[i+1]];
			AC=points[t_index[i]]>>points[t_index[i+2]];
			Normal=AB*AC;	Normal.Unit();
			for(int e=0;e<3;e++)
				N[t_index[i+e]]=N[t_index[i+e]]+Normal;
			}
		for(int i=0;i<nb_vtx;i++)
			N[i].Unit();
		}

	byte link;
	data=read_from_mem(&link,1,data);
	if(link) {
		child = (OBJECT*) malloc(sizeof(OBJECT));
		child->init();
		data=child->load_3dm(data);
		}
	else
		child=NULL;

	data=read_from_mem(&link,1,data);
	if(link) {
		next = (OBJECT*) malloc(sizeof(OBJECT));
		next->init();
		data=next->load_3dm(data);
		}
	else
		next=NULL;
	return data;
}
