/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005   Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*----------------------------------------------------------\
|                      EngineClass.cpp                      |
|    Contient toutes les classes nécessaires au moteur 3D   |
| et au moteur physique.                                    |
|                                                           |
\----------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"					// Pour la lecture des fichiers 3D
#include "cob.h"					// Pour la lecture et l'éxecution des scripts
#include "tdf.h"					// Pour la gestion des éléments du jeu
#include "EngineClass.h"

float	player_color[30]={	0.11f,	0.28f,	0.91f,
							0.83f,	0.17f,	0.0f,
							1.0f,	1.0f,	1.0f,
							0.11f,	0.62f,	0.07f,
							0.03f,	0.12f,	0.48f,
							0.5f,	0.34f,	0.62f,
							1.0f,	1.0f,	0.0f,
							0.0f,	0.0f,	0.0f,
							0.61f,	0.8f,	0.87f,
							0.67f,	0.67f,	0.51f };

byte	player_color_map[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };

PLAYERS	players;		// Objet contenant les données sur les joueurs

int NB_PLAYERS;

void MAP_OTA::load(char *data,int ota_size)
{
	destroy();
	set_uformat(U_ASCII);
	char *pos=data;
	char *ligne=NULL;
	int nb=0;
	int index=0;
	char *limit=data+ota_size;
	char *f;
	int n_pos=0;
	do {
		nb++;
		if(ligne)
			delete[] ligne;
		ligne=get_line(pos);
		strlwr(ligne);
		while(pos[0]!=0 && pos[0]!=13 && pos[0]!=10)	{	pos++;	n_pos++;	}
		while(pos[0]==13 || pos[0]==10)	{	pos++;	n_pos++;	}

		if(strstr(ligne,"//") || strstr(ligne,"/*") || strstr(ligne,"{")) { }		// Saute les commentaires
		else if(f=strstr(ligne,"missionname=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			missionname=strdup(f+12);
			}
		else if(f=strstr(ligne,"planet=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			planet=strdup(f+7);
			}
		else if(f=strstr(ligne,"missiondescription=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			missiondescription=strdup(f+19);
			char *prec=missiondescription;
			char *cur=prec;
			int nb=0;
			while(cur[0]) {
				if(cur[0]==' ')	prec=cur;
				nb++;
				if(nb>34) {
					prec[0]='\n';
					nb=cur-prec;
					}
				cur++;
				}
			}
		else if(f=strstr(ligne,"tidalstrength=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			tidalstrength=atoi(f+14);
			}
		else if(f=strstr(ligne,"solarstrength=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			solarstrength=atoi(f+14);
			}
		else if(f=strstr(ligne,"lavaworld=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			lavaworld=(f[10]=='1');
			}
		else if(f=strstr(ligne,"killmul=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			killmul=atoi(f+8);
			}
		else if(f=strstr(ligne,"minwindspeed=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			minwindspeed=atoi(f+13);
			}
		else if(f=strstr(ligne,"maxwindspeed=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			maxwindspeed=atoi(f+13);
			}
		else if(f=strstr(ligne,"gravity=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			gravity=atoi(f+8)*0.1f;
			}
		else if(f=strstr(ligne,"numplayers=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			numplayers=strdup(f+11);
			}
		else if(f=strstr(ligne,"size=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			map_size=strdup(f+5);
			}
		else if(f=strstr(ligne,"surfacemetal=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			SurfaceMetal=atoi(f+13);
			}
		else if(f=strstr(ligne,"mohometal=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			MohoMetal=atoi(f+10);
			}
		else if(f=strstr(ligne,"specialwhat=startpos")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			index=atoi(f+20)-1;
			}
		else if(f=strstr(ligne,"xpos=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			startX[index]=atoi(f+5);
			}
		else if(f=strstr(ligne,"zpos=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			startZ[index]=atoi(f+5);
			}
		else if(f=strstr(ligne,"waterdoesdamage=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			waterdoesdamage=atoi(f+16)>0;
			}
		else if(f=strstr(ligne,"waterdamage=")) {
			if((strstr(ligne,";")))	*(strstr(ligne,";"))=0;
			waterdamage=atoi(f+12);
			}
		else if(f=strstr(ligne,"missionhint=")) {}
		else if(f=strstr(ligne,"brief=")) {}
		else if(f=strstr(ligne,"narration=")) {}
		else if(f=strstr(ligne,"glamour=")) {}
		else if(f=strstr(ligne,"lineofsight=")) {}
		else if(f=strstr(ligne,"mapping=")) {}
		else if(f=strstr(ligne,"timemul=")) {}
		else if(f=strstr(ligne,"memory=")) {}
		else if(f=strstr(ligne,"useonlyunits=")) {}
		else if(f=strstr(ligne,"schemacount=")) {}
		else if(f=strstr(ligne,"type=")) {}
		else if(f=strstr(ligne,"aiprofile=")) {}
		else if(f=strstr(ligne,"humanmetal=")) {}
		else if(f=strstr(ligne,"computermetal=")) {}
		else if(f=strstr(ligne,"humanenergy=")) {}
		else if(f=strstr(ligne,"computerenergy=")) {}
		else if(f=strstr(ligne,"meteorweapon=")) {}
		else if(f=strstr(ligne,"meteorradius=")) {}
		else if(f=strstr(ligne,"meteordensity=")) {}
		else if(f=strstr(ligne,"meteorduration=")) {}
		else if(f=strstr(ligne,"meteorinterval=")) {}
		else if(f=strstr(ligne,"killenemycommander=")) {}
		else if(f=strstr(ligne,"destroyallunits=")) {}
		else if(f=strstr(ligne,"allunitskilled=")) {}
		else if(f=strstr(ligne,"featurename=")) {}
		else {
			if(strlen(ligne)>1 && strstr(ligne,"{")==NULL && strstr(ligne,"}")==NULL && strstr(ligne,"[")==NULL)
				Console->AddEntry("(OTA) inconnu: %s",ligne);
			}

	} while(nb<1000 && pos<limit);
	delete[] ligne;
	ligne=NULL;
	if(waterdamage==0)
		waterdoesdamage=false;
}

void MAP::draw_mini(int x1,int y1,int w,int h,CAMERA *cam)			// Dessine la mini-carte
{
	if(!mini) return;		// Vérifie si la mini-carte existe
	int rw=w*mini_w/252;
	int rh=h*mini_h/252;
	x1+=w-rw>>1;
	y1+=h-rh>>1;
	float lw=mini_w/252.0f;
	float lh=mini_h/252.0f;
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,glmini);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f,0.0f);	glVertex2f(x1,y1);
		glTexCoord2f(lw,0.0f);		glVertex2f(x1+rw,y1);
		glTexCoord2f(lw,lh);		glVertex2f(x1+rw,y1+rh);
		glTexCoord2f(0.0f,lh);		glVertex2f(x1,y1+rh);
	glEnd();

	if(!cam) return;

	VECTOR A,B,C,D,P;
	A=cam->Dir+1.0f*(-cam->Side-0.75f*cam->Up);
	B=cam->Dir+1.0f*(cam->Side-0.75f*cam->Up);
	C=cam->Dir+1.0f*(cam->Side+0.75f*cam->Up);
	D=cam->Dir+1.0f*(-cam->Side+0.75f*cam->Up);
	const int nmax=64;
	float cx[4*nmax+4],cy[4*nmax+4];
	if(A.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(A.y)*A;	cx[0]=P.x;	cy[0]=P.z; }
	else {
		P=cam->Pos+10000.0f*A;	cx[0]=P.x;	cy[0]=P.z; }
	if(B.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(B.y)*B;	cx[1]=P.x;	cy[1]=P.z; }
	else {
		P=cam->Pos+10000.0f*B;	cx[1]=P.x;	cy[1]=P.z; }
	if(C.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(C.y)*C;	cx[2]=P.x;	cy[2]=P.z; }
	else {
		P=cam->Pos+10000.0f*C;	cx[2]=P.x;	cy[2]=P.z; }
	if(D.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(D.y)*D;	cx[3]=P.x;	cy[3]=P.z; }
	else {
		P=cam->Pos+10000.0f*D;	cx[3]=P.x;	cy[3]=P.z; }

	int i;
	for(i=0;i<4;i++) {
		cx[i]=(cx[i]+0.5f*map_w)*rw/map_w;
		cy[i]=(cy[i]+0.5f*map_h)*rh/map_h;
		}
	for(i=0;i<4;i++) {
		for(int e=0;e<nmax;e++) {
			cx[i*nmax+e+4]=(cx[i]*(nmax-e)+cx[(i+1)%4]*(e+1))/(nmax+1);
			cy[i*nmax+e+4]=(cy[i]*(nmax-e)+cy[(i+1)%4]*(e+1))/(nmax+1);
			}
		}
	for(i=0;i<4+(nmax<<2);i++) {
		if(cx[i]<0.0f) cx[i]=0.0f;
		else if(cx[i]>rw) cx[i]=rw;
		if(cy[i]<0.0f) cy[i]=0.0f;
		else if(cy[i]>rh) cy[i]=rh;
		}

	glDisable(GL_TEXTURE_2D);
	glColor3f(1.0f,1.0f,1.0f);
	glBegin(GL_LINE_STRIP);
		for(i=0;i<4;i++) {
			glVertex2f(cx[i]+x1,cy[i]+y1);
			for(int e=0;e<nmax;e++)
				glVertex2f(x1+cx[i*nmax+e+4],y1+cy[i*nmax+e+4]);
			}
		glVertex2f(cx[0]+x1,cy[0]+y1);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void MAP::update_player_visibility(int player_id,int px,int py,int r,bool remove)
{
	byte mask=1<<player_id;
	int r2=r*r;
	if(player_id==players.local_human_id)
		for(int y=0;y<=r;y++) {
			int x=(int)(0.5f+sqrt((float)(r2-y*y)));
			int ry=py-y;
			if(ry>=0 && ry<bloc_h_db) {
				int rx=px-x;
				int lx=px+x;
				if(rx<0)	rx=0;
				if(lx>bloc_w_db)	lx=bloc_w_db;
				if(remove)
					for(;rx<lx;rx++) {
						map_data[ry][rx].view|=mask;
						map_data[ry][rx].timer--;
						}
				else
					for(;rx<lx;rx++) {
						map_data[ry][rx].view|=mask;
						map_data[ry][rx].timer++;
						}
				}
			if(y!=0) {
				ry=py+y;
				if(ry>=0 && ry<bloc_h_db) {
					int rx=px-x;
					int lx=px+x;
					if(rx<0)	rx=0;
					if(lx>bloc_w_db)	lx=bloc_w_db;
					if(remove)
						for(;rx<lx;rx++) {
							map_data[ry][rx].view|=mask;
							map_data[ry][rx].timer--;
							}
					else
						for(;rx<lx;rx++) {
							map_data[ry][rx].view|=mask;
							map_data[ry][rx].timer++;
							}
					}
				}
			}
	else
		for(int y=0;y<=r;y++) {
			int x=(int)(0.5f+sqrt((float)(r2-y*y)));
			int ry=py-y;
			if(ry>=0 && ry<bloc_h_db) {
				int rx=px-x;
				int lx=px+x;
				if(rx<0)	rx=0;
				if(lx>bloc_w_db)	lx=bloc_w_db;
				for(;rx<lx;rx++)
					map_data[ry][rx].view|=mask;
				}
			if(y!=0) {
				ry=py+y;
				if(ry>=0 && ry<bloc_h_db) {
					int rx=px-x;
					int lx=px+x;
					if(rx<0)	rx=0;
					if(lx>bloc_w_db)	lx=bloc_w_db;
					for(;rx<lx;rx++)
						map_data[ry][rx].view|=mask;
					}
				}
			}
}

void MAP::draw(CAMERA *cam,byte player_mask,bool FLAT,float niv,float t,float dt,bool depth_only,bool check_visibility,bool draw_uw)
{
	cam->SetView();
	if(FLAT && !water)	return;
	if(FLAT)
		glTranslatef(0.0f,0.0f,sea_dec);
	int i,x,y;
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	if(!FLAT) {
		if(ntex>0)
			gfx->ReInitAllTex( true );
		else
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	VECTOR A,B,C,D,P;
	A=cam->Dir+1.0f*(-cam->Side-0.75f*cam->Up);
	B=cam->Dir+1.0f*(cam->Side-0.75f*cam->Up);
	C=cam->Dir+1.0f*(cam->Side+0.75f*cam->Up);
	D=cam->Dir+1.0f*(-cam->Side+0.75f*cam->Up);
	float cx[4],cy[4];
	if(A.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(A.y)*A;	cx[0]=P.x;	cy[0]=P.z; }
	else {
		P=cam->Pos+10000.0f*A;	cx[0]=P.x;	cy[0]=P.z; }
	if(B.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(B.y)*B;	cx[1]=P.x;	cy[1]=P.z; }
	else {
		P=cam->Pos+10000.0f*B;	cx[1]=P.x;	cy[1]=P.z; }
	if(C.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(C.y)*C;	cx[2]=P.x;	cy[2]=P.z; }
	else {
		P=cam->Pos+10000.0f*C;	cx[2]=P.x;	cy[2]=P.z; }
	if(D.y<0.0f) {
		P=cam->Pos+cam->Pos.y/fabs(D.y)*D;	cx[3]=P.x;	cy[3]=P.z; }
	else {
		P=cam->Pos+10000.0f*D;	cx[3]=P.x;	cy[3]=P.z; }

	int minx=bloc_w<<1,maxx=0;
	int miny=bloc_h<<1,maxy=0;
	for(i=0;i<4;i++) {
		cx[i]=(cx[i]+0.5f*map_w)/16.0f;
		cy[i]=(cy[i]+0.5f*map_h)/16.0f;
		if(cx[i]<minx) minx=(int)cx[i];
		if(cx[i]>maxx) maxx=(int)cx[i];
		if(cy[i]<miny) miny=(int)cy[i];
		if(cy[i]>maxy) maxy=(int)cy[i];
		}
	int y1=bloc_h,y2=0,x1=bloc_w,x2=0,mx,my;
	float limit=cam->zfar*sqrt(2.0f);
	x1=(int)((cam->Pos.x+0.5f*map_w-limit)/16.0f);
	y1=(int)((cam->Pos.z+0.5f*map_h-limit)/16.0f);
	x2=(int)((cam->Pos.x+0.5f*map_w+limit)/16.0f);
	y2=(int)((cam->Pos.z+0.5f*map_h+limit)/16.0f);
	mx=(int)((cam->Pos.x+0.5f*map_w)/16.0f);
	my=(int)((cam->Pos.z+0.5f*map_h)/16.0f);
	if(x1<minx)
		x1=minx;
	if(x2>maxx)
		x2=maxx;
	if(y1<miny)
		y1=miny;
	if(y2>maxy)
		y2=maxy;
	if(x1>mx) x1=mx;
	if(y1>my) y1=my;
	if(x2<mx) x2=mx;
	if(y2<my) y2=my;

	if((abs(my-y2)<<4)>cam->zfar+64.0f)
		y2 = my>y2 ? my - (int)(cam->zfar/16.0f)-4 : my + 4 + (int)(cam->zfar/16.0f);
	if((abs(my-y1)<<4)>cam->zfar+64.0f)
		y1 = my>y1 ? my - (int)(cam->zfar/16.0f)-4 : my + 4 + (int)(cam->zfar/16.0f);
	if((abs(mx-x2)<<4)>cam->zfar+64.0f)
		x2 = mx>x2 ? mx - (int)(cam->zfar/16.0f)-4 : mx + 4 + (int)(cam->zfar/16.0f);
	if((abs(mx-x1)<<4)>cam->zfar+64.0f)
		x1 = mx>x1 ? mx - (int)(cam->zfar/16.0f)-4 : mx + 4 + (int)(cam->zfar/16.0f);

	if(y1<0) y1=0;
	if(y2<0) y2=0;
	if(y1>=bloc_h) y1=bloc_h-1;
	if(y2>=bloc_h) y2=bloc_h-1;
	if(x1<0) x1=0;
	if(x2<0) x2=0;
	if(x1>=bloc_w) x1=bloc_w-1;
	if(x2>=bloc_w) x2=bloc_w-1;

	y1-=3;
	if(y1<0) y1=0;

	A=(cam->Dir+1.0f*(0.75f*cam->Up-cam->Side));	A.Unit();
	float ref=0.95f*(A%cam->Dir);
	float dhm=0.5f*map_h;
	float dwm=0.5f*map_w;

	if(!FLAT)
		glColor4f(1.0f,1.0f,1.0f,1.0f);

	POINTF flat[9];
	if(FLAT) {
		flat[0].x=0.0f;		flat[0].y=niv+cos(t)*0.5f;			flat[0].z=0.0f;
		flat[1].x=8.0f;		flat[1].y=niv+cos(t+1.0f)*0.5f;		flat[1].z=0.0f;
		flat[2].x=16.0f;	flat[2].y=flat[0].y;				flat[2].z=0.0f;
		flat[3].x=0.0f;		flat[3].y=niv+cos(t+1.5f)*0.5f;		flat[3].z=8.0f;
		flat[4].x=8.0f;		flat[4].y=niv+cos(t+2.5f)*0.5f;		flat[4].z=8.0f;
		flat[5].x=16.0f;	flat[5].y=flat[3].y;				flat[5].z=8.0f;
		flat[6].x=0.0f;		flat[6].y=flat[0].y;				flat[6].z=16.0f;
		flat[7].x=8.0f;		flat[7].y=flat[1].y;				flat[7].z=16.0f;
		flat[8].x=16.0f;	flat[8].y=flat[0].y;				flat[8].z=16.0f;
		}

	if(ntex>0 && !depth_only) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glEnable(GL_TEXTURE_2D);
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}

	if(FLAT)
		glTranslatef(cos(t),0.0f,sin(t));
	int old_i=-1;
	GLuint old_tex=bloc[0].tex;
	if(!depth_only)
		glBindTexture(GL_TEXTURE_2D,old_tex);
	if(!FLAT && check_visibility) {
		for(y=oy1;y<=oy2;y++)
			memset(view[y]+ox1,0,ox2-ox1+1);
		features.min_idx=features.nb_features-1;
		features.max_idx=0;
		features.list_size=0;
		ox1=x1;	ox2=x2;
		oy1=y1;	oy2=y2;
		}
	else if(!check_visibility) {
		x1=ox1;	x2=ox2;
		y1=oy1;	y2=oy2;
		for(int i=0;i<features.list_size;i++)
			features.feature[features.list[i]].draw=true;
		}
	int lavaprob=(int)(1000*dt);
	POINTF	buf_p[900];				// Tampon qui accumule les blocs pour les dessiner en chaîne
	float	buf_t[1800];
	float	buf_c[3600];
	short	buf_size=0;					// in blocs
	glEnableClientState(GL_VERTEX_ARRAY);		// vertex coordinates
	glEnableClientState(GL_COLOR_ARRAY);		// Coulors(for fog of war)
	glColorPointer(4,GL_FLOAT,0,buf_c);
	glVertexPointer( 3, GL_FLOAT, 0, buf_p);
	glTexCoordPointer(2, GL_FLOAT, 0, buf_t);
	int	ox=x1;

	bool low_def_view = cam->RPos.y>gfx->low_def_limit;		// Low detail map for mega zoom
	if(low_def_view) {							// draw the low detail map
		i=0;
		for(y=0;y<=low_h;y++) {
			int Y=y*(bloc_h_db-2)/low_h;
			for(x=0;x<=low_w;x++) {
				int X=x*(bloc_w_db-2)/low_w;
				int Z;
				Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X].view&player_mask)==0)	low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=0.1f;
				else if(!map_data[Z][X].timer)				low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=0.5f;
				else										low_col[i<<2]=low_col[(i<<2)+1]=low_col[(i<<2)+2]=1.0f;
				i++;
				}
			}
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4,GL_FLOAT,0,low_col);
		if(FLAT) {
			glTranslatef(0.0f,niv,0.0f);
			glVertexPointer( 3, GL_FLOAT, 0, low_vtx_flat);
			}
		else
			glVertexPointer( 3, GL_FLOAT, 0, low_vtx);
		glTexCoordPointer(2, GL_FLOAT, 0, low_tcoord);
		glBindTexture(GL_TEXTURE_2D,low_tex);
		glDrawElements(GL_TRIANGLE_STRIP, low_nb_idx,GL_UNSIGNED_INT,low_index);		// draw this map
		}

	if(cam->RPos.y>=900.0f) {
		memset(view[0],1,bloc_w*bloc_h);
		ox1=0;
		ox2=bloc_w-1;
		oy1=0;
		oy2=bloc_h-1;
		}

	if(cam->RPos.y<900.0f)
	for(y=y1;y<=y2;y++) {				// Balaye les blocs susceptibles d'être visibles pour dessiner ceux qui le sont
		int pre_y=y<<4;
		int Y=y<<1;
		int pre_y2=y*bloc_w;
		VECTOR T;
		T.x=-dwm;	T.y=0.0f;	T.z=pre_y-dhm;
		buf_size=0;
		ox=x1;
		for(x=x1;x<=x2;x++) {
			int X=x<<1;
			if(!FLAT && check_visibility) {
				VECTOR	V;
				V.x=(x<<4)-dwm;
				V.y=h_map[Y|1][X|1];
				V.z=pre_y-dhm+get_zdec(X|1,Y|1);
				V=V-cam->Pos;
				if(fabs(V%cam->Dir)>cam->zfar) {
					view[y][x]=0;
					continue;
					}
				float d=V.Sq();
				if(d>16384.0f)
					if((V%cam->Dir)<ref*sqrt(d)) {
						view[y][x]=0;
						continue;
						}
				if((map_data[Y][X].view&player_mask)==0 && (map_data[Y][X|1].view&player_mask)==0 && (map_data[Y|1][X].view&player_mask)==0 && (map_data[Y|1][X|1].view&player_mask)==0) {
					if(water) {
						if(map_data[Y][X].underwater && map_data[Y|1][X].underwater && map_data[Y][X|1].underwater && map_data[Y|1][X|1].underwater)
							view[y][x]=2;
						else
							view[y][x]=3;
						}
					}
				else {
					if(map_data[Y][X].timer==0 && map_data[Y][X|1].timer==0 && map_data[Y|1][X].timer==0 && map_data[Y|1][X|1].timer==0) {
						if(map_data[Y][X].underwater || map_data[Y|1][X].underwater || map_data[Y][X|1].underwater || map_data[Y|1][X|1].underwater)
							view[y][x]=2;
						else
							view[y][x]=3;
						}
					else
						view[y][x]=1;
					if(map_data[Y][X].stuff>=0 && map_data[Y][X].stuff<features.nb_features) {		// Indique comme affichables les objets présents sur le bloc
						if(features.feature[map_data[Y][X].stuff].type<0)
							map_data[Y][X].stuff=-1;
						else {
							features.feature[map_data[Y][X].stuff].draw=true;
							features.feature[map_data[Y][X].stuff].grey=(view[y][x]&2)==2;
							features.list[features.list_size++]=map_data[Y][X].stuff;
							}
						}
					if(map_data[Y][X|1].stuff>=0 && map_data[Y][X|1].stuff<features.nb_features) {
						if(features.feature[map_data[Y][X|1].stuff].type<0)
							map_data[Y][X|1].stuff=-1;
						else {
							features.feature[map_data[Y][X|1].stuff].draw=true;
							features.feature[map_data[Y][X|1].stuff].grey=(view[y][x]&2)==2;
							features.list[features.list_size++]=map_data[Y][X|1].stuff;
							}
						}
					if(map_data[Y|1][X].stuff>=0 && map_data[Y|1][X].stuff<features.nb_features) {
						if(features.feature[map_data[Y|1][X].stuff].type<0)
							map_data[Y|1][X].stuff=-1;
						else {
							features.feature[map_data[Y|1][X].stuff].draw=true;
							features.feature[map_data[Y|1][X].stuff].grey=(view[y][x]&2)==2;
							features.list[features.list_size++]=map_data[Y|1][X].stuff;
							}
						}
					if(map_data[Y|1][X|1].stuff>=0 && map_data[Y|1][X|1].stuff<features.nb_features) {
						if(features.feature[map_data[Y|1][X|1].stuff].type<0)
							map_data[Y|1][X|1].stuff=-1;
						else {
							features.feature[map_data[Y|1][X|1].stuff].draw=true;
							features.feature[map_data[Y|1][X|1].stuff].grey=(view[y][x]&2)==2;
							features.list[features.list_size++]=map_data[Y|1][X|1].stuff;
							}
						}
					}
				}
			else {
				if(view[y][x]==0)
					continue;
				if(view[y][x]==2 && !draw_uw)	continue;		// Jump this if it is under water and don't have to be drawn
				if(view[y][x]==3)
					view[y][x]=2;
				if(view[y][x]==2 && FLAT)
					view[y][x]=0;
				}
			if(low_def_view)	continue;
							// Si le joueur ne peut pas voir ce morceau, on ne le dessine pas en clair
			i=bmap[y][x];
			if(FLAT) {
				bloc[i].point=lvl[pre_y2+x];
				if(bloc[i].point==NULL || bloc[i].point[0].y<niv || bloc[i].point[1].y<niv || bloc[i].point[2].y<niv ||
					bloc[i].point[3].y<niv || bloc[i].point[4].y<niv || bloc[i].point[5].y<niv ||
					bloc[i].point[6].y<niv || bloc[i].point[7].y<niv || bloc[i].point[8].y<niv)
					bloc[i].point=flat;
				else
					continue;
				}
			else {
				if(check_visibility && bloc[i].lava && (rand_from_table()%1000000)<=lavaprob) {		// Lava emiting code moved here because of lava effect using fragment program
					POINTF POS;
					VECTOR V;
					POS.x=(x<<4)-dwm+8.0f;
					POS.z=pre_y-dhm+8.0f;
					POS.y=sealvl-10.0f;
					V.x=((rand_from_table()%201)-100);
					V.y=((rand_from_table()%51)+50);
					V.z=((rand_from_table()%201)-100);
					V.Unit();
					particle_engine.emit_lava(POS,V,1,10,(rand_from_table()%1000)*0.01f+30.0f);
					}
				bloc[i].point=lvl[pre_y2+x];
				if(bloc[i].point==NULL) {
					lvl[pre_y2+x]=bloc[i].point=(POINTF*) malloc(sizeof(POINTF)*9);
					if(tnt) {
						bloc[i].point[0].x=0.0f;	bloc[i].point[0].z=get_zdec(X,Y);
						bloc[i].point[1].x=8.0f;	bloc[i].point[1].z=get_zdec(X|1,Y);
						bloc[i].point[2].x=16.0f;	bloc[i].point[2].z=get_zdec(X+2,Y);
						bloc[i].point[3].x=0.0f;	bloc[i].point[3].z=8.0f+get_zdec(X,Y|1);
						bloc[i].point[4].x=8.0f;	bloc[i].point[4].z=8.0f+get_zdec(X|1,Y|1);
						bloc[i].point[5].x=16.0f;	bloc[i].point[5].z=8.0f+get_zdec(X+2,Y|1);
						bloc[i].point[6].x=0.0f;	bloc[i].point[6].z=16.0f+get_zdec(X,Y+2);
						bloc[i].point[7].x=8.0f;	bloc[i].point[7].z=16.0f+get_zdec(X|1,Y+2);
						bloc[i].point[8].x=16.0f;	bloc[i].point[8].z=16.0f+get_zdec(X+2,Y+2);
						bloc[i].point[0].y=get_nh(X,Y);
						bloc[i].point[1].y=get_nh(X|1,Y);
						bloc[i].point[2].y=get_nh(X+2,Y);
						bloc[i].point[3].y=get_nh(X,Y|1);
						bloc[i].point[4].y=get_nh(X|1,Y|1);
						bloc[i].point[5].y=get_nh(X+2,Y|1);
						bloc[i].point[6].y=get_nh(X,Y+2);
						bloc[i].point[7].y=get_nh(X|1,Y+2);
						bloc[i].point[8].y=get_nh(X+2,Y+2);
						}
					else {
						bloc[i].point[0].x=0.0f;	bloc[i].point[0].z=0.0f;
						bloc[i].point[1].x=8.0f;	bloc[i].point[1].z=0.0f;
						bloc[i].point[2].x=16.0f;	bloc[i].point[2].z=0.0f;
						bloc[i].point[3].x=0.0f;	bloc[i].point[3].z=8.0f;
						bloc[i].point[4].x=8.0f;	bloc[i].point[4].z=8.0f;
						bloc[i].point[5].x=16.0f;	bloc[i].point[5].z=8.0f;
						bloc[i].point[6].x=0.0f;	bloc[i].point[6].z=16.0f;
						bloc[i].point[7].x=8.0f;	bloc[i].point[7].z=16.0f;
						bloc[i].point[8].x=16.0f;	bloc[i].point[8].z=16.0f;
						bloc[i].point[0].y=get_h(X,Y);
						bloc[i].point[1].y=get_h(X|1,Y);
						bloc[i].point[2].y=get_h(X+2,Y);
						bloc[i].point[3].y=get_h(X,Y|1);
						bloc[i].point[4].y=get_h(X|1,Y|1);
						bloc[i].point[5].y=get_h(X+2,Y|1);
						bloc[i].point[6].y=get_h(X,Y+2);
						bloc[i].point[7].y=get_h(X|1,Y+2);
						bloc[i].point[8].y=get_h(X+2,Y+2);
						}
					}
				}

			if(bloc[i].tex!=old_tex || buf_size>=100 || ox+1<x) {
				glDrawElements(GL_TRIANGLE_STRIP, buf_size*13,GL_UNSIGNED_SHORT,buf_i);		// dessine le tout
				buf_size=0;
				old_tex=bloc[i].tex;
				glBindTexture(GL_TEXTURE_2D,bloc[i].tex);
				}
			ox=x;

			short buf_pos=buf_size*9;
			T.x+=x<<4;
			if(!FLAT)
				for(byte e=0;e<9;e++) {					// Copie le bloc
					buf_p[buf_pos+e].x=bloc[i].point[e].x+T.x;
					buf_p[buf_pos+e].y=bloc[i].point[e].y;
					buf_p[buf_pos+e].z=bloc[i].point[e].z+T.z;
					}
			else
				for(byte e=0;e<9;e++) {					// Copie le bloc
					buf_p[buf_pos+e].x=flat[e].x+T.x;
					buf_p[buf_pos+e].y=flat[e].y;
					buf_p[buf_pos+e].z=flat[e].z+T.z;
					}
			T.x-=x<<4;
//			memcpy(buf_t+(buf_pos<<1),bloc[i].texcoord,sizeof(float)*18);		// texture
			memcpy(buf_t+(buf_pos<<1),bloc[i].texcoord,72);		// texture
			float *color=buf_c+(buf_pos<<2);
			for(i=0;i<36;i+=4) {
				color[i]=color[i|1]=color[i|2]=1.0f;
				color[i|3]=FLAT ? 0.75f : 1.0f;
				}
			int Z;
			Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
			if((map_data[Z][X].view&player_mask)==0)	color[16]=color[17]=color[18]=0.1f;
			else if(!map_data[Z][X].timer)				color[16]=color[17]=color[18]=0.5f;
			Z=Y+get_zdec_notest(X|1,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
			if((map_data[Z][X|1].view&player_mask)==0)	color[20]=color[21]=color[22]=0.1f;
			else if(!map_data[Z][X|1].timer)			color[20]=color[21]=color[22]=0.5f;
			Z=Y+1+get_zdec_notest(X,Y|1);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
			if((map_data[Z][X].view&player_mask)==0)	color[28]=color[29]=color[30]=0.1f;
			else if(!map_data[Z][X].timer)				color[28]=color[29]=color[30]=0.5f;
			Z=Y+1+get_zdec_notest(X|1,Y|1);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
			if((map_data[Z][X|1].view&player_mask)==0)	color[32]=color[33]=color[34]=0.1f;
			else if(!map_data[Z][X|1].timer)			color[32]=color[33]=color[34]=0.5f;
			if(X>0) {
				Z=Y+get_zdec_notest(X-1,Y);				if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X-1].view&player_mask)==0)		color[12]=color[13]=color[14]=0.1f;
				else if(!map_data[Z][X-1].timer)				color[12]=color[13]=color[14]=0.5f;
				Z=Y+1+get_zdec_notest(X-1,Y|1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X-1].view&player_mask)==0)		color[24]=color[25]=color[26]=0.1f;
				else if(!map_data[Z][X-1].timer)				color[24]=color[25]=color[26]=0.5f;
				if(Y>0)
					Z=Y-1+get_zdec_notest(X-1,Y-1);		if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
					if((map_data[Z][X-1].view&player_mask)==0)		color[0]=color[1]=color[2]=0.1f;
					else if(!map_data[Z][X-1].timer)				color[0]=color[1]=color[2]=0.5f;
				}
			if(Y>0) {
				Z=Y-1+get_zdec_notest(X,Y-1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X].view&player_mask)==0)	color[4]=color[5]=color[6]=0.1f;
				else if(!map_data[Z][X].timer)				color[4]=color[5]=color[6]=0.5f;
				Z=Y-1+get_zdec_notest(X|1,Y-1);			if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
				if((map_data[Z][X|1].view&player_mask)==0)		color[8]=color[9]=color[10]=0.1f;
				else if(!map_data[Z][X|1].timer)				color[8]=color[9]=color[10]=0.5f;
				}
/*			Z=Y+get_zdec_notest(X,Y);					if(Z>=bloc_h_db-1)	Z=bloc_h_db-2;
			Z&=0xFFFFFE;
			X&=0xFFFFFE;
			if(map_data[Z][X].unit_idx!=-1 || map_data[Z+1][X].unit_idx!=-1
			|| map_data[Z][X+1].unit_idx!=-1 || map_data[Z+1][X+1].unit_idx!=-1)		// Affiche les résultats du pathfinding
				for(i=0;i<9;i++) {
					color[i<<2]=color[(i<<2)+1]=color[(i<<2)+2]=color[(i<<2)+3]=0.0f;
					color[(i<<2)+1]=1.0f;
					}*/
			buf_size++;
			}
		if(buf_size>0) {
			glDrawElements(GL_TRIANGLE_STRIP, buf_size*13,GL_UNSIGNED_SHORT,buf_i);		// dessine le tout
			buf_size=0;
			}
		}
	glDisableClientState(GL_COLOR_ARRAY);		// Couleurs(pour le brouillard de guerre)
	if(depth_only) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glEnable(GL_TEXTURE_2D);
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
		}
}

VECTOR MAP::hit(VECTOR Pos,VECTOR Dir,bool water, float length)			// Calcule l'intersection d'un rayon avec la carte(le rayon partant du dessus de la carte)
{
	if(Dir.x==0.0f && Dir.z==0.0f) {			// Solution triviale
		VECTOR P=Pos;
		P.y=get_unit_h(P.x,P.z);
		return P;
		}
	if(get_unit_h(Pos.x,Pos.z)>Pos.y)		// Cas non traité
		return Pos;
	float step=1.0f;
	if(Dir.x!=0.0f && Dir.z!=0.0f) {
		if(fabs(Dir.x)<fabs(Dir.z))
			step=1.0f/fabs(Dir.x);
		else
			step=1.0f/fabs(Dir.z);
		}
	int nb=0;
	float dwm=map_w_d;
	float dhm=map_h_d;
	Dir=(4.0f*step)*Dir;
	float len_step = Dir.Norm();
	while(((sealvl<Pos.y && water) || !water) && get_max_h((int)Pos.x,(int)Pos.z)<Pos.y) {
		if(nb>=1000 || length<0.0f)
			return Pos;
		length-=len_step;
		nb++;
		Pos=Pos+Dir;
		if(fabs(Pos.x)>dwm || fabs(Pos.z)>dhm)		// Pas de résultat
			return Pos;
		}
	length+=len_step;
	Pos=Pos-Dir;
	while(((sealvl<Pos.y && water) || !water) && get_unit_h(Pos.x,Pos.z)<Pos.y) {
		if(nb>=1000 || length<0.0f)
			return Pos;
		length-=len_step;
		nb++;
		Pos=Pos+Dir;
		if(fabs(Pos.x)>dwm || fabs(Pos.z)>dhm)		// Pas de résultat
			return Pos;
		}
	for(byte i=0;i<7;i++) {
		length+=len_step;
		Pos=Pos-Dir;		// On recommence la dernière opération mais avec plus de précision
		Dir=0.5f*Dir;
		len_step *= 0.5f;
		nb=0;
		while(((sealvl<Pos.y && water) || !water) && get_unit_h(Pos.x,Pos.z)<Pos.y) {
			if(nb>=2 || length<0.0f)
				return Pos;
			length-=len_step;
			nb++;
			Pos=Pos+Dir;
			}
		}
	return Pos;		// Meilleure solution approximative trouvée
}

void WATER::draw(float t,float X,float Y,bool flat)
{
	glActiveTextureARB(GL_TEXTURE0_ARB);
	if(flat) {
		glTranslatef(cos(t),0.0f,sin(t));
		glBegin(GL_QUADS);
			glTexCoord2f(-map_w/w+0.5f,-map_h/w+0.5f);		glVertex3f(-map_w,0.0f,-map_h);
			glTexCoord2f(map_w/w+0.5f,-map_h/w+0.5f);		glVertex3f(map_w,0.0f,-map_h);
			glTexCoord2f(map_w/w+0.5f,map_h/w+0.5f);		glVertex3f(map_w,0.0f,map_h);
			glTexCoord2f(-map_w/w+0.5f,map_h/w+0.5f);		glVertex3f(-map_w,0.0f,map_h);
		glEnd();
		return;
		}
	glTranslatef(X+cos(t),0.0f,Y+sin(t));
	glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glTranslatef(X/w,Y/w,0.0f);
	glMatrixMode(GL_MODELVIEW);
	if(!flat) {
		int i=0;
		float cv=w/s;
		for(int y=0;y<=s;y++) {
			float phaseY=y*cv;
			for(int x=0;x<=s;x++) {
				float phaseX=x*cv;
				point[i++].y=(cos(t+(X+Y+phaseX+phaseY)*0.1f)+sin(1.1f*t+(X-Y+phaseX-phaseY)*0.1f))*0.25f;
				}
			}
		normalise();
		}

	glEnableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
	glVertexPointer( 3, GL_FLOAT, 0, point);
	glNormalPointer( GL_FLOAT, 0, N);
	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, texcoord);

	glDrawElements(GL_TRIANGLES, nb_idx,GL_UNSIGNED_INT,index);		// dessine le tout

	glDisableClientState(GL_NORMAL_ARRAY);
}

void SKY::draw()
{
	glDisableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);		// Les sommets
	glVertexPointer( 3, GL_FLOAT, 0, point);
	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, texcoord);

	glDrawElements(GL_TRIANGLES, nb_idx,GL_UNSIGNED_SHORT,index);		// dessine le tout
}

void PLAYERS::show_resources()
{
	int _id = local_human_id != -1 ? local_human_id : 0;
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
	char buf[100];
	gfx->print(gfx->normal_font,130.0f,8.0f,0.0f,makeacol(0xD5,0xD5,0xD5,0xFF),"METAL");
	gfx->print(gfx->normal_font,130+(gfx->SCREEN_W_HALF-65),8.0f,0.0f,makeacol(0xFF,0xFF,0x00,0xFF),"ENERGY");
	gfx->print(gfx->normal_font,65+(gfx->SCREEN_W_HALF-65),8.0f,0.0f,makeacol(0xD5,0xD5,0xD5,0xFF),format("%d",(int)metal[_id]));
	uszprintf(buf,100,"+%f",metal_t[_id]);	*(strstr(buf,".")+2)=0;		gfx->print(gfx->normal_font,65+(gfx->SCREEN_W_HALF-65),16.0f,0.0f,makeacol(0xD5,0xD5,0xD5,0xFF),buf);
	uszprintf(buf,100,"-%f",metal_u[_id]);	*(strstr(buf,".")+2)=0;		gfx->print(gfx->normal_font,65+(gfx->SCREEN_W_HALF-65),24.0f,0.0f,makeacol(0xD5,0xD5,0xD5,0xFF),buf);
	gfx->print(gfx->normal_font,SCREEN_W-65,8.0f,0.0f,makeacol(0xFF,0xFF,0x00,0xFF),format("%d",(int)energy[_id]));
	uszprintf(buf,100,"+%f",energy_t[_id]);	*(strstr(buf,".")+2)=0;		gfx->print(gfx->normal_font,SCREEN_W-65,16.0f,0.0f,makeacol(0xFF,0xFF,0x00,0xFF),buf);
	uszprintf(buf,100,"-%f",energy_u[_id]);	*(strstr(buf,".")+2)=0;		gfx->print(gfx->normal_font,SCREEN_W-65,24.0f,0.0f,makeacol(0xFF,0xFF,0x00,0xFF),buf);

	glDisable(GL_TEXTURE_2D);

	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBegin(GL_QUADS);			// Dessine les barres de metal et d'énergie
		glColor4f(0.5f,0.5f,0.5f,0.5f);

		glVertex2f(180.0f,8.0f);					// Fond de la barre de metal
		glVertex2f(60+(gfx->SCREEN_W_HALF-65),8.0f);
		glVertex2f(60+(gfx->SCREEN_W_HALF-65),16.0f);
		glVertex2f(180.0f,16.0f);

		glColor4f(0.7f,0.7f,0.0f,0.5f);
		glVertex2f(186+(gfx->SCREEN_W_HALF-65),8.0f);		// Fond de la barre d'énergie
		glVertex2f(SCREEN_W-70,8.0f);
		glVertex2f(SCREEN_W-70,16.0f);
		glVertex2f(186+(gfx->SCREEN_W_HALF-65),16.0f);
	glEnd();
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);			// Dessine les barres de metal et d'énergie
		glColor3f(0.9f,0.9f,0.9f);

		if(metal_s[0]) {
			glVertex2f(180.0f,8.0f);					// Barre de metal
			glVertex2f(180+((gfx->SCREEN_W_HALF-65)-120)*metal[_id]/metal_s[_id],8.0f);
			glVertex2f(180+((gfx->SCREEN_W_HALF-65)-120)*metal[_id]/metal_s[_id],16.0f);
			glVertex2f(180.0f,16.0f);
			}

		glColor3f(1.0f,1.0f,0.0f);
		if(energy_s[0]) {
			glVertex2f(186+(gfx->SCREEN_W_HALF-65),8.0f);		// Barre d'énergie
			glVertex2f(186+(gfx->SCREEN_W_HALF-65)+(SCREEN_W-256-(gfx->SCREEN_W_HALF-65))*energy[_id]/energy_s[_id],8.0f);
			glVertex2f(186+(gfx->SCREEN_W_HALF-65)+(SCREEN_W-256-(gfx->SCREEN_W_HALF-65))*energy[_id]/energy_s[_id],16.0f);
			glVertex2f(186+(gfx->SCREEN_W_HALF-65),16.0f);
			}
	glEnd();
	glColor4f(1.0f,1.0f,1.0f,1.0f);
}
