/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*----------------------------------------------------------\
|                      EngineClass.h                        |
|    Contient toutes les classes nécessaires au moteur 3D   |
| et au moteur physique.                                    |
|                                                           |
\----------------------------------------------------------*/


#include "particles.h"			// Pour le moteur à particules
#include "tdf.h"				// Pour le gestionnaire de sprites
#include "taconfig.h"

#ifndef ENGINE_CLASS		// Ne déclare pas 2 fois les classes du moteur
#define ENGINE_CLASS		// Définit les classes comme déclarées

#define PARTICLE_LIMIT		100000		// pas plus de 100000 particules
#define HMAP_RESIZE			0.04f

#define H_DIV		0.5f
float const tnt_transform=1.0f/tan(63.44f*DEG2RAD)/H_DIV;
float const tnt_transform_H_DIV=1.0f/tan(63.44f*DEG2RAD);

class MAP_OTA
{
public:
	char	*missionname;
	char	*planet;
	char	*missiondescription;
	int		tidalstrength;
	int		solarstrength;
	bool	lavaworld;
	short	killmul;
	int		minwindspeed;
	int		maxwindspeed;
	float	gravity;
	char	*numplayers;
	char	*map_size;
	int		SurfaceMetal;
	int		MohoMetal;
	int		startX[10];
	int		startZ[10];
	bool	waterdoesdamage;
	int		waterdamage;

	void init()
	{
		planet=NULL;
		missionname=NULL;
		missiondescription=NULL;
		numplayers=NULL;
		map_size=NULL;
		for(int i=0;i<10;i++)
			startX[i]=startZ[i]=0;
		tidalstrength=0;
		solarstrength=22;
		lavaworld=false;
		killmul=50;
		minwindspeed=0;
		maxwindspeed=0;
		gravity=9.8f;
		SurfaceMetal=0;
		MohoMetal=0;
		waterdamage=0;
		waterdoesdamage=false;
	}

	void destroy()
	{
		if(planet)				free(planet);
		if(missionname)			free(missionname);
		if(missiondescription)	free(missiondescription);
		if(numplayers)			free(numplayers);
		if(map_size)			free(map_size);
		init();
	}

	MAP_OTA()
	{
		init();
	}

	~MAP_OTA()
	{
		destroy();
	}

private:
	inline char *get_line(char *data)
	{
		int pos=0;
		while(data[pos]!=0 && data[pos]!=13 && data[pos]!=10)	pos++;
		char *d=new char[pos+1];
		memcpy(d,data,pos);
		d[pos]=0;
		return d;
	}
public:

	void load(char *data,int ota_size);
};

class BLOC				// Blocs composant la carte
{
public:
	byte		nbindex;	// Nombre d'indices
	byte		nbpoint;	// Nombre de points
	POINTF		*point;		// Points du bloc
	float		*texcoord;	// Coordonnées de texture
	GLuint		tex;		// Indice de texture OpenGl
	bool		lava;		// Indique si le bloc est de type lave

	void init()
	{
		nbindex=nbpoint=0;
		point=NULL;
		texcoord=NULL;
		tex=0;
		lava=false;
	}

	BLOC()
	{
		init();
	}

	void destroy()
	{
		if(point) delete[] point;
		if(texcoord) delete[] texcoord;
		init();
	}
};

class SECTOR			// Structure pour regrouper les informations sur le terrain (variations d'altitude, submergé, teneur en metal, ...)
{
public:
	float	dh;					// dérivée maximale en valeur absolue de l'altitude
	bool	underwater;			// indique si le bloc est sous l'eau
	byte	view;				// Indique quels sont les joueurs voyant ce morceau de terrain
	sint32	stuff;				// Indique l'élément graphique présent sur ce secteur
	sint16	unit_idx;			// Indice de l'unité qui se trouve sur ce secteur
	uint16	timer;				// Used to count how many units are viewing this sector

	inline void init()
	{
		dh=0.0f;
		underwater=false;
		view=0;				// Personne ne peut voir ce secteur au début
		stuff=-1;
		unit_idx=-1;
		timer=0;
	}
};

#include "pathfinding.h"		// Algorithme de pathfinding

class MAP				// Données concernant la carte
{
public:
	short		ntex;		// Indique si la texture est chargée et doit être détruite
	GLuint		*tex;		// Texture de surface
	int			nbbloc;		// Nombre de blocs
	BLOC		*bloc;		// Blocs composant le terrain
	unsigned short	**bmap;		// Tableau d'indice des blocs
	float		**h_map;	// Tableau de l'élévation du terrain
	float		**ph_map;	// Tableau du relief projeté pour le calcul inverse(projection) lors de l'affichage
	byte		**ph_map_2;	// Tableau du relief projeté (multiplié par un facteur flottant) pour le calcul inverse(projection) lors de l'affichage
	SECTOR		**map_data;	// Tableau d'informations sur le terrain
	byte		**view;		// Indique quels sont les parcelles de terrain visibles à l'écran
	byte		**path;		// Tableau pour le pathfinding

	int			map_w;		// Largeur de la carte en 16ème de bloc
	int			map_h;		// Hauteur de la carte en 16ème de bloc
	int			map_w_d;	// Same values as above divided by 2
	int			map_h_d;
	int			bloc_w;		// Width in blocs
	int			bloc_h;		// Height in blocs
	int			bloc_w_db;	// Same as above but multiplied by 2
	int			bloc_h_db;
	float		map2blocdb_w;
	float		map2blocdb_h;
	BITMAP		*mini;		// Minimap
	GLuint		glmini;		// Texture OpenGl pour la minimap
	int			mini_w;
	int			mini_h;
	float		sealvl;		// Niveau de la mer
	POINTF		**lvl;		// Bloc de flottants pour le relief de la carte
	bool		water;		// Indique qu'il faut dessiner la mer
	bool		tnt;		// Indique si la carte est format tnt(format de total annihilation)
	float		sea_dec;	// Décalage de la mer
	int			ox1,ox2;	// Coordonnées de la dernière fenêtre de terrain dessinée
	int			oy1,oy2;
	GLushort	buf_i[1300];	// Pour accélérer l'affichage
	GLuint		lava_map;	// texture des zones de lave

	MAP_OTA		ota_data;	// Data read from the ota file

	float		wind;		// To handle wind
	float		wind_dir;
	VECTOR		wind_vec;

/*------------- Experimental: code for new map format -----------------------*/

/*	BLOC		**macro_bloc;	// map
	uint32		macro_w;
	uint32		macro_h;*/

/*---------------------------------------------------------------------------*/

	int			low_nb_idx;
	int			low_w,low_h;
	POINTF		*low_vtx;
	POINTF		*low_vtx_flat;
	float		*low_tcoord;
	float		*low_col;
	GLuint		*low_index;
	GLuint		low_tex;

	void init()
	{
/*------------- Experimental: code for new map format -----------------------*/

/*		macro_bloc = NULL;
		macro_w = 0;
		macro_h = 0;*/

/*---------------------------------------------------------------------------*/

		low_nb_idx=0;
		low_vtx=NULL;
		low_vtx_flat=NULL;
		low_tcoord=NULL;
		low_col=NULL;
		low_index=NULL;
		low_tex=0;

		wind=0.0f;
		wind_dir=0.0f;
		wind_vec.x=wind_vec.y=wind_vec.z=0.0f;
		ota_data.init();
		lava_map=0;
		path=NULL;
		mini_w=mini_h=252;
		ntex=0;
		tex=NULL;
		nbbloc=0;
		bloc=NULL;
		mini=NULL;
		bmap=NULL;
		h_map=NULL;
		ph_map=NULL;
		ph_map_2=NULL;
		map_data=NULL;
		sealvl=0.0f;
		glmini=0;
		lvl=NULL;
		water=true;
		tnt=false;			// Laisse la possibilité de créer un autre format de cartes
		sea_dec=0.0f;
		view=NULL;
		ox1=ox2=oy1=oy2=0;
		short buf_size=0;
		for(short i=0;i<1300;i++) {
			buf_i[i++]=0+buf_size;
			buf_i[i++]=1+buf_size;
			buf_i[i++]=3+buf_size;
			buf_i[i++]=4+buf_size;
			buf_i[i++]=6+buf_size;
			buf_i[i++]=7+buf_size;
			buf_i[i++]=7+buf_size;
			buf_i[i++]=8+buf_size;
			buf_i[i++]=4+buf_size;
			buf_i[i++]=5+buf_size;
			buf_i[i++]=1+buf_size;
			buf_i[i++]=2+buf_size;
			buf_i[i]=2+buf_size;
			buf_size+=9;
			}
	}

	MAP()
	{
		init();
	}

	void destroy()
	{
/*------------- Experimental: code for new map format -----------------------*/
/*		if(macro_bloc) {
			for(uint32 y=0;y<macro_h;y++)
				for(uint32 x=0;x<macro_w;x++)
					macro_bloc[y][x].destroy();
			delete[] macro_bloc[0];
			delete[] macro_bloc;
			macro_w=0;
			macro_h=0;
			}*/
/*---------------------------------------------------------------------------*/

		if(low_vtx)			free(low_vtx);
		if(low_vtx_flat)	free(low_vtx_flat);
		if(low_tcoord)		free(low_tcoord);
		if(low_col)			free(low_col);
		if(low_index)		free(low_index);
		if(low_tex)			glDeleteTextures(1,&low_tex);

		ota_data.destroy();
		if(lava_map)
			glDeleteTextures(1,&lava_map);
		if(path && bloc_w && bloc_h) {
			free(path[0]);
			free(path);
			}
		if(view && bloc_w && bloc_h) {
			free(view[0]);
			free(view);
			}
		if(map_data && bloc_w && bloc_h) {
			free(map_data[0]);
			free(map_data);
			}
		if(ph_map && bloc_w && bloc_h) {
			free(ph_map[0]);		// la carte est allouée d'un seul bloc
			free(ph_map);
			}
		if(ph_map_2 && bloc_w && bloc_h) {
			free(ph_map_2[0]);		// la carte est allouée d'un seul bloc
			free(ph_map_2);
			}
		if(h_map && bloc_w && bloc_h) {
			free(h_map[0]);		// la carte est allouée d'un seul bloc
			free(h_map);
			}
		if(bmap && bloc_w && bloc_h) {
			free(bmap[0]);		// la carte est allouée d'un seul bloc
			free(bmap);
			}
		if(ntex>0) {
			for(int i=0;i<ntex;i++)
				glDeleteTextures(1,&(tex[i]));
			free(tex);
			}
		if(lvl) {
			for(int i=0;i<bloc_h*bloc_w;i++)
				free(lvl[i]);
			free(lvl);
			}
		if(bloc && nbbloc>0) {
			for(int i=0;i<nbbloc;i++) {
				bloc[i].point=NULL;
				bloc[i].destroy();
				}
			free(bloc);
			}
		if(mini) {
			glDeleteTextures(1,&glmini);
			destroy_bitmap(mini);
			}
		init();
	}

	~MAP()
	{
		destroy();
	}

	void update_player_visibility(int player_id,int px,int py,int r,bool remove=false);

	void draw_mini(int x1=0,int y1=0,int w=252,int h=252,CAMERA *cam=NULL);			// Dessine la mini-carte

	inline const float get_unit_h(float x,float y)
	{
		x=(x+map_w_d)*0.125f;		// Convertit les coordonnées
		y=(y+map_h_d)*0.125f;
		int lx = bloc_w_db-1;
		int ly = bloc_h_db-1;
		if(x<0.0f) x=0.0f;
		else if(x>=lx) x=bloc_w_db-2;
		if(y<0.0f) y=0.0f;
		else if(y>=ly) y=bloc_h_db-2;
		float h[4];
		int X=(int)x,Y=(int)y;
		float dx=x-X;
		float dy=y-Y;
		h[0]=h_map[Y][X];
		if(X+1<lx)
			h[1]=h_map[Y][X+1]-h[0];
		else
			h[1]=0.0f;
		if(Y+1<ly) {
			h[2]=h_map[Y+1][X];
			if(X+1<lx)
				h[3]=h_map[Y+1][X+1]-h[2];
			else
				h[3]=0.0f;
			}
		else {
			h[2]=h[0];
			h[3]=h[1];
			}
		h[0]=h[0]+h[1]*dx;
		return h[0]+(h[2]+h[3]*dx-h[0])*dy;
	}

	inline float get_h(int x,int y)
	{
		if(x<0) x=0;
		if(y<0) y=0;
		if(x>=bloc_w_db-1) x=bloc_w_db-2;
		if(y>=bloc_h_db-1) y=bloc_h_db-2;
		return h_map[y][x];
	}

	inline float get_max_h(int x,int y)
	{
		if(x<0) x=0;
		if(y<0) y=0;
		if(x>=bloc_w_db-1) x=bloc_w_db-2;
		if(y>=bloc_h_db-1) y=bloc_h_db-2;
		float h = h_map[y][x];
		if(x<bloc_w_db-2)	h = max(h, h_map[y][x+1]);
		if(y<bloc_h_db-2) {
			h = max(h, h_map[y+1][x]);
			if(x<bloc_w_db-2)	h = max(h, h_map[y+1][x+1]);
			}
		return h;
	}

	inline float get_zdec(int x,int y)
	{
		if(x<0) x=0;
		if(y<0) y=0;
		if(x>=bloc_w_db-1) x=bloc_w_db-2;
		if(y>=bloc_h_db-1) y=bloc_h_db-2;
		return ph_map[y][x]*tnt_transform_H_DIV;
	}

	inline byte get_zdec_notest(int x,int y)
	{
		return ph_map_2[y][x];
	}

	inline float get_nh(int x,int y)
	{
		if(x<0) x=0;
		if(y<0) y=0;
		if(x>=bloc_w_db-1) x=bloc_w_db-2;
		if(y>=bloc_h_db-1) y=bloc_h_db-2;
		return ph_map[y][x];
	}

	inline void rect(int x1,int y1,int w,int h,short c,char *yardmap=NULL,bool open=false)
	{
		if(yardmap==NULL) {
			int y2=y1+h;
			int x2=x1+w;
			if(y1<0)	y1=0;
			if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
			if(x1<0)	x1=0;
			if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
			if(y2<=y1 || x2<=x1)	return;
			for(int y=y1;y<y2;y++)
				for(int x=x1;x<x2;x++)
					map_data[y][x].unit_idx=c;
			}
		else {
			int i=0;
			int y2=y1+h;
			int x2=x1+w;
			if(y1<0)	{	i-=y1*w;	y1=0;	}
			if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
			if(x1<0)	x1=0;
			if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
			int dw=w-(x2-x1);
			if(y2<=y1 || x2<=x1)	return;
			for(int y=y1;y<y2;y++) {
				for(int x=x1;x<x2;x++) {
					if(yardmap[i]=='G' || yardmap[i]=='o' || yardmap[i]=='w' || yardmap[i]=='f' || (yardmap[i]=='c' && !open) || (yardmap[i]=='C' && !open) || (yardmap[i]=='O' && open))
						map_data[y][x].unit_idx=c;
					i++;
					}
				i+=dw;
				}
			}
	}

	inline bool check_rect(int x1,int y1,int w,int h,short c)
	{
		int y2=y1+h;
		int x2=x1+w;
		if(y1<0)	y1=0;
		if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
		if(x1<0)	x1=0;
		if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
		if(y2<=y1 || x2<=x1)	return false;
		for(int y=y1;y<y2;y++)
			for(int x=x1;x<x2;x++)
				if(map_data[y][x].unit_idx!=c && map_data[y][x].unit_idx!=-1)
					return false;
		return true;
	}

	inline float check_rect_dh(int x1,int y1,int w,int h)
	{
		int y2=y1+h;
		int x2=x1+w;
		if(y1<0)	y1=0;
		if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
		if(x1<0)	x1=0;
		if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
		float max_dh=0.0f;
		bool on_water=false;
		if(y2<=y1 || x2<=x1)	return max_dh;
		for(int y=y1;y<y2;y++)
			for(int x=x1;x<x2;x++) {
				if(map_data[y][x].dh>max_dh)
					max_dh=map_data[y][x].dh;
				on_water|=map_data[y][x].underwater;
				}
		if(on_water)
			max_dh=-max_dh;
		return max_dh;
	}

	inline float check_depth(int x1,int y1,int w,int h)
	{
		int y2=y1+h;
		int x2=x1+w;
		if(y1<0)	y1=0;
		if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
		if(x1<0)	x1=0;
		if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
		float depth=0.0f;
		if(y2<=y1 || x2<=x1)	return depth;
		for(int y=y1;y<y2;y++)
			for(int x=x1;x<x2;x++) {
				float d = sealvl-h_map[y][x];
				if(d>depth)
					depth=d;
				}
		return depth;
	}

	inline bool check_vents(int x1,int y1,int w,int h,char *yard_map)
	{
		if(yard_map==NULL)	return true;
		int y2=y1+h;
		int x2=x1+w;
		if(y1<0)	y1=0;
		if(y2>bloc_h_db-1)	y2=bloc_h_db-1;
		if(x1<0)	x1=0;
		if(x2>bloc_w_db-1)	x2=bloc_w_db-1;
		int dw = w - (x2-x1);
		int i = 0;
		bool ok = true;
		if(y2<=y1 || x2<=x1)	return false;
		for(int y=y1;y<y2;y++) {
			for(int x=x1;x<x2;x++) {
				if(yard_map[i]=='G') {
					ok = false;
					if(map_data[y][x].stuff>=0) {
						int feature_id = map_data[y][x].stuff;
						if(feature_manager.feature[features.feature[feature_id].type].geothermal)
							return true;
						}
					}
				i++;
				}
			i+=dw;
			}
		return ok;
	}

	inline bool check_lava(int x1,int y1,int w,int h)
	{
		int y2=y1+h;
		int x2=x1+w;
		if(y1<0)	y1=0;
		if(y2>bloc_h-1)	y2=bloc_h-1;
		if(x1<0)	x1=0;
		if(x2>bloc_w-1)	x2=bloc_w-1;
		if(y2<=y1 || x2<=x1)	return false;
		for(int y=y1;y<y2;y++)
			for(int x=x1;x<x2;x++)
				if(bloc[bmap[y][x]].lava)
					return true;
		return false;
	}

	void draw(CAMERA *cam,byte player_mask,bool FLAT=false,float niv=0.0f,float t=0.0f,float dt=1.0f,bool depth_only=false,bool check_visibility=true,bool draw_uw=true);

	VECTOR hit(VECTOR Pos,VECTOR Dir,bool water = true, float length = 200000.0f);			// Calcule l'intersection d'un rayon avec la carte(le rayon partant du dessus de la carte)
};

class WATER
{
public:
	float		map_w;
	float		map_h;
	int			nb_vtx;
	int			nb_idx;
	VECTOR		*point;
	VECTOR		*N;
	float		*texcoord;
	GLuint		*index;
	int			s;
	float		w;

	void init()
	{
		map_w=0.0f;
		map_h=0.0f;
		nb_vtx=0;
		nb_idx=0;
		point=NULL;
		N=NULL;
		texcoord=NULL;
		index=NULL;
		s=0;
		w=1.0f;
	}

	void destroy()
	{
		if(point)		free(point);
		if(N)			free(N);
		if(texcoord)	free(texcoord);
		if(index)		free(index);
		init();
	}

	WATER()
	{
		init();
	}

	~WATER()
	{
		destroy();
	}

	void build(float m_w,float m_h,int d,float size)
	{
		destroy();
		map_w=m_w;
		map_h=m_h;
		s=d;
		w=size;

		nb_vtx=(d+1)*(d+1);
		nb_idx=d*d*6;

		point=(VECTOR*) malloc(sizeof(VECTOR)*nb_vtx);
		N=(VECTOR*) malloc(sizeof(VECTOR)*nb_vtx);
		texcoord=(float*) malloc(sizeof(float)*nb_vtx*2);
		index=(GLuint*) malloc(sizeof(GLuint)*nb_idx);

		int i=0;
		for(int y=0;y<=d;y++)
			for(int x=0;x<=d;x++) {
				point[i].x=(x-0.5f*d)*size/d;
				point[i].y=0.0f;
				point[i].z=(y-0.5f*d)*size/d;
				texcoord[i<<1]=(float)x/d;
				texcoord[(i<<1)+1]=(float)y/d;
				i++;
				}
		i=0;
		for(int y=0;y<d;y++)
			for(int x=0;x<d;x++) {
				index[i++]=y*(d+1)+x;
				index[i++]=y*(d+1)+x+1;
				index[i++]=(y+1)*(d+1)+x;
				index[i++]=(y+1)*(d+1)+x;
				index[i++]=y*(d+1)+x+1;
				index[i++]=(y+1)*(d+1)+x+1;
				}
		normalise();
	}

	void normalise()
	{
		for(int i=0;i<nb_vtx;i++)
			N[i].x=N[i].y=N[i].z=0.0f;
		for(int i=0;i<nb_idx;i+=3) {
			VECTOR n = (point[index[i+1]]-point[index[i]])*(point[index[i+2]]-point[index[i]]);
			n.Unit();
			N[index[i]]=N[index[i]]+n;
			N[index[i+1]]=N[index[i+1]]+n;
			N[index[i+2]]=N[index[i+2]]+n;
			}
		for(int i=0;i<nb_vtx;i++)
			N[i].Unit();
	}

	void draw(float t,float X,float Y,bool flat=false);
};

class SKY
{
public:
	uint16		nb_vtx;
	uint16		nb_idx;
	VECTOR		*point;
	float		*texcoord;
	GLushort	*index;
	int			s;
	float		w;

	void init()
	{
		nb_vtx=0;
		nb_idx=0;
		point=NULL;
		texcoord=NULL;
		index=NULL;
		s=0;
		w=1.0f;
	}

	void destroy()
	{
		if(point)		free(point);
		if(texcoord)	free(texcoord);
		if(index)		free(index);
		init();
	}

	SKY()
	{
		init();
	}

	~SKY()
	{
		destroy();
	}

	void build(int d,float size)
	{
		destroy();
		s=d;
		w=size;

		nb_vtx=(d+1)*((d<<1)+1);
		nb_idx=d*d*12;

		point=(VECTOR*) malloc(sizeof(VECTOR)*nb_vtx);
		texcoord=(float*) malloc(sizeof(float)*nb_vtx*2);
		index=(GLushort*) malloc(sizeof(GLushort)*nb_idx);

		int i=0;
		for(int y=0;y<=d;y++)
			for(int x=0;x<=d*2;x++) {
				point[i].x=size*cos(x/(2.0f*d)*PI*2.0f)*cos(0.5f*y/d*PI);
				point[i].y=size*sin(0.5f*y/d*PI);
				point[i].z=-size*sin(x/(2.0f*d)*PI*2.0f)*cos(0.5f*y/d*PI);
				texcoord[i<<1]=(float)x/(d*2);
				texcoord[(i<<1)+1]=1.0f-(float)y/d;
				i++;
				}
		i=0;
		for(int y=0;y<d;y++)
			for(int x=0;x<d*2;x++) {
				index[i++]=y*(d*2+1)+x+1;
				index[i++]=y*(d*2+1)+x;
				index[i++]=(y+1)*(d*2+1)+x;
				index[i++]=y*(d*2+1)+x+1;
				index[i++]=(y+1)*(d*2+1)+x;
				index[i++]=(y+1)*(d*2+1)+x+1;
				}
	}

	void draw();
};

#include "ia.h"					// Pour l'intelligence artificielle

extern int NB_PLAYERS;

#define PLAYER_CONTROL_LOCAL_HUMAN	0x0
#define PLAYER_CONTROL_REMOTE_HUMAN	0x1
#define PLAYER_CONTROL_LOCAL_AI		0x2
#define PLAYER_CONTROL_REMOTE_AI	0x3
#define PLAYER_CONTROL_NONE			0x4

#define PLAYER_CONTROL_FLAG_REMOTE	0x1
#define PLAYER_CONTROL_FLAG_AI		0x2

class PLAYERS			// Classe pour gérer les joueurs et leurs statistiques de partie
{
public:
	int			nb_player;		// Nombre de joueurs (maximum 10 joueurs)
	int			local_human_id;	// Quel est le joueur qui commande depuis cette machine??
	byte		control[10];	// Qui controle ce joueur??
	char		*nom[10];		// Noms des joueurs
	char		*side[10];		// Camp des joueurs
	float		energy[10];		// Energie des joueurs
	float		metal[10];		// Metal des joueurs
	float		metal_u[10];	// Metal utilisé
	float		energy_u[10];	// Energie utilisée
	float		metal_t[10];	// Metal extrait
	float		energy_t[10];	// Energie produite
	int			kills[10];		// Victimes
	int			losses[10];		// Pertes
	int			energy_s[10];	// Capacités de stockage d'énergie
	int			metal_s[10];	// Capacités de stockage de metal
	int			com_metal[10];	// Stockage fournit par le commandeur
	int			com_energy[10];
	bool		commander[10];	// Indique s'il y a un commandeur
	bool		annihilated[10];// Le joueur a perdu la partie??
	AI_PLAYER	ai_command[10];	// Controleurs d'intelligence artificielle
	int			nb_unit[10];	// Nombre d'unités de chaque joueur

			// For statistic purpose only
	double		energy_total[10];
	double		metal_total[10];

	inline void player_control(MAP *map)
	{
		for(byte i=0;i<nb_player;i++)
			if(control[i]==PLAYER_CONTROL_LOCAL_AI)
				ai_command[i].think(map);
	}

	inline void clear()		// Remet à 0 la taille des stocks
	{
		for(byte i=0;i<nb_player;i++) {
			energy_s[i]=metal_s[i]=0;			// Stocks
			metal_t[i]=energy_t[i]=0.0f;		// Production
			metal_u[i]=energy_u[i]=0.0f;		// Consommation
			commander[i]=false;
			annihilated[i]=true;
			nb_unit[i]=0;
			}
	}

	inline int add(char *NOM,char *SIDE,byte _control,int E=10000,int M=10000)
	{
		if(nb_player>=10)	return -1;		// Trop de joueurs déjà
		metal_u[nb_player]=0;
		energy_u[nb_player]=0;
		metal_t[nb_player]=0;
		energy_t[nb_player]=0;
		kills[nb_player]=0;
		losses[nb_player]=0;
		nom[nb_player]=strdup(NOM);
		control[nb_player]=_control;
		nb_unit[nb_player]=0;
		energy_total[nb_player]=0.0f;
		metal_total[nb_player]=0.0f;

		Console->AddEntry("adding player %s (%d) of %s with E=%d, M=%d", NOM, nb_player, SIDE, E, M );

		com_metal[nb_player] = M;
		com_energy[nb_player] = E;
		energy[nb_player] = E;
		metal[nb_player] = M;
		energy_s[nb_player] = E;
		metal_s[nb_player] = M;

		side[nb_player++]=strdup(SIDE);
		if(_control==PLAYER_CONTROL_LOCAL_HUMAN)
			local_human_id=NB_PLAYERS;
		if(_control==PLAYER_CONTROL_LOCAL_AI) {
			char filename[100];
			filename[0]=0;
			strcat(filename,"ai/");
			strcat(filename,NOM);
			strcat(filename,".ai");
			if(file_exists(filename,FA_RDONLY | FA_ARCH,NULL))						// Charge un joueur s'il existe
				ai_command[NB_PLAYERS].load(filename,NB_PLAYERS);
			else													// Sinon crée un nouveau joueur
				ai_command[NB_PLAYERS].change_name(NOM);
			ai_command[NB_PLAYERS].player_id=NB_PLAYERS;
			}
		return NB_PLAYERS++;
	}

	void show_resources();

	inline void init(int E=10000,int M=10000)		// Initialise les données des joueurs
	{
		nb_player=0;
		NB_PLAYERS=0;
		local_human_id=-1;
		for(int i=0;i<10;i++) {
			com_metal[i] = M;
			com_energy[i] = E;
			control[i] = PLAYER_CONTROL_NONE;
			ai_command[i].init();
			nom[i] = NULL;
			side[i] = NULL;
			energy[i] = E;
			metal[i] = M;
			metal_u[i] = 0;
			energy_u[i] = 0;
			metal_t[i] = 0;
			energy_t[i] = 0;
			kills[i] = 0;
			losses[i] = 0;
			energy_s[i] = E;
			metal_s[i] = M;
			nb_unit[i] = 0;
			energy_total[i] = 0.0f;
			metal_total[i] = 0.0f;
			}
	}

	inline PLAYERS()
	{
		init();
	}

	inline void destroy()
	{
		for(byte i=0;i<10;i++) {
			if(control[i]==PLAYER_CONTROL_LOCAL_AI)		// Enregistre les données de l'IA
				ai_command[i].save();
			if(nom[i])	free(nom[i]);
			if(side[i])	free(side[i]);
			ai_command[i].destroy();
			}
		init();
	}

	inline ~PLAYERS()
	{
		destroy();
	}
};

extern PLAYERS	players;		// Objet contenant les données sur les joueurs

#endif						// Fin de la déclaration des classes
