/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005   Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*----------------------------------------------------------\
|                       UnitEngine.cpp                      |
|    Contains the unit engine, which simulates units and    |
| make them interact with each other.                       |
|                                                           |
\----------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"					// Pour la lecture des fichiers 3D
#include "cob.h"					// Pour la lecture et l'éxecution des scripts
#include "tdf.h"					// Pour la gestion des éléments du jeu
#include "EngineClass.h"
#include "UnitEngine.h"

using namespace TA3D::EXCEPTION;

INGAME_UNITS units;

	void UNIT::add_mission(int mission_type,VECTOR *target,bool step,int dat,void *pointer,PATH_NODE *path,byte flags)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::add_mission );
#endif

		MISSION *new_mission=(MISSION*) malloc(sizeof(MISSION));
		new_mission->next=NULL;
		new_mission->mission=mission_type;
		new_mission->step=step;
		new_mission->time=0.0f;
		new_mission->data=dat;
		new_mission->p=pointer;
		new_mission->path=path;
		new_mission->last_d=9999999.0f;
		new_mission->flags=flags;

		if(target)
			new_mission->target=*target;

		MISSION *stop=(MISSION*) malloc(sizeof(MISSION));
		stop->mission=MISSION_STOP;
		stop->step=true;
		stop->time=0.0f;
		stop->p=NULL;
		stop->data=0;
		stop->path=NULL;
		stop->last_d=9999999.0f;
		stop->flags=flags;
		if(step) {
			stop->next=NULL;
			new_mission->next=stop;
			}
		else {
			stop->next=new_mission;
			new_mission=stop;
			}

		if(step && mission) {
//			new_mission->next=mission;
			stop->next=mission;
			mission=new_mission;
			start_mission_script(mission->mission);
			}
		else {
			if(mission) {				// Ajoute l'ordre aux autres
				MISSION *cur=mission;
				while(cur->next!=NULL)	cur=cur->next;
				if( ( cur->mission == MISSION_MOVE || cur->mission == MISSION_MOVE_CALC_PATH || cur->mission == MISSION_PATROL )
				&& ( mission_type == MISSION_MOVE || mission_type == MISSION_MOVE_CALC_PATH || mission_type == MISSION_PATROL ) ) {		// Don't stop if it's not necessary
					stop = new_mission->next;
					free( new_mission );
					new_mission = stop;
					new_mission->next = NULL;
					}
				cur->next=new_mission;
				}
			else {
				mission=new_mission;
				start_mission_script(mission->mission);
				}
			}

#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	void UNIT::set_mission(int mission_type,VECTOR *target,bool step,int dat,bool stopit,void *pointer,PATH_NODE *path,byte flags)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::set_mission );
#endif

		int old_mission=-1;
		if(mission!=NULL)	old_mission=mission->mission;
		while(mission!=NULL) 	clear_mission();			// Efface les ordres précédents

		switch(old_mission)		// Commandes de fin de mission
		{
		case MISSION_REPAIR:
		case MISSION_RECLAIM:
		case MISSION_BUILD_2:
			launch_script(get_script_index(SCRIPT_stopbuilding));
			launch_script(get_script_index(SCRIPT_stop));
			break;
		case MISSION_ATTACK:
		case MISSION_AIM:
		case MISSION_SHOOT:
			deactivate();
			break;
		};

		mission=(MISSION*) malloc(sizeof(MISSION));
		mission->next=NULL;
		mission->mission=mission_type;
		mission->step=step;
		mission->time=0.0f;
		mission->data=dat;
		mission->p=pointer;
		mission->path=path;
		mission->last_d=9999999.0f;
		mission->flags=flags;
		if(target)
			mission->target=*target;

		if(stopit) {
			MISSION *stop=(MISSION*) malloc(sizeof(MISSION));
			stop->next=mission;
			stop->mission=MISSION_STOP;
			stop->step=true;
			stop->time=0.0f;
			stop->p=NULL;
			stop->data=0;
			stop->path=NULL;
			stop->last_d=9999999.0f;
			stop->flags=flags;
			mission=stop;
			}
		else
			start_mission_script(mission->mission);
		c_time=0.0f;

#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	void UNIT::draw(CAMERA *cam,MAP *map,bool cullface,bool height_line)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::draw );
#endif

		visible=false;
		if(model==NULL) {
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;		// S'il n'y a pas de modèle associé, on quitte la fonction
			}

		int px=cur_px>>1;
		int py=cur_py>>1;
		if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	{
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;	// Unité hors de la carte
			}
		if(map->view[py][px]!=1) {
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;	// Unité non visible
			}

		VECTOR D;
		D=Pos-cam->Pos;		// Vecteur "viseur unité" partant de la caméra vers l'unité

		float dist=D.Sq();
		if(dist>=16384.0f && (D%cam->Dir)<=0.0f) {
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;
			}
		if((D%cam->Dir)>cam->zfar2) {
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;		// Si l'objet est hors champ on ne le dessine pas
			}

		visible=true;

		MATRIX_4x4 M;
		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		if(cam->RPos.y>gfx->low_def_limit) {			// for mega zoom, draw only a triangle
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glColor3f(player_color[player_color_map[owner_id]*3],player_color[player_color_map[owner_id]*3+1],player_color[player_color_map[owner_id]*3+2]);
			glBegin(GL_TRIANGLES);
				glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glVertex3f(model->center.x-unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glVertex3f(model->center.x,model->center.y,model->center.z+unit_manager.unit_type[type_id].FootprintZ*4.0f);
			glEnd();
			if(owner_id==players.local_human_id && sel) {
				glColor3f(1.0f,1.0f,0.0f);
				glBegin(GL_LINE_STRIP);
					glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x-unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x,model->center.y,model->center.z+unit_manager.unit_type[type_id].FootprintZ*4.0f);
					glVertex3f(model->center.x+unit_manager.unit_type[type_id].FootprintX*4.0f,model->center.y,model->center.z-unit_manager.unit_type[type_id].FootprintZ*4.0f);
				glEnd();
				}
			}
		else {
			float scale=unit_manager.unit_type[type_id].Scale;
			glScalef(scale,scale,scale);

			M=RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale);			// Matrice pour le calcul des positions des éléments du modèle de l'unité

			VECTOR *target=NULL,*center=NULL;
			POINTF upos;
			bool c_part=false;
			bool reverse=false;
			float size=0.0f;
			if(port[BUILD_PERCENT_LEFT]==0.0f && mission!=NULL
			&& (mission->mission==MISSION_REPAIR || mission->mission==MISSION_BUILD_2 || mission->mission==MISSION_RECLAIM))
				if(c_time>=0.25f) {
					reverse=(mission->mission==MISSION_RECLAIM);
					c_time=0.0f;
					c_part=true;
					target=&(mission->target);
					upos.x=upos.y=upos.z=0.0f;
					upos=upos+Pos;
					if(mission->p!=NULL && ((UNIT*)mission->p)->flags && ((UNIT*)mission->p)->model!=NULL) {
						size=((UNIT*)mission->p)->model->size2;
						center=&((UNIT*)mission->p)->model->center;
						}
					else if(mission->mission==MISSION_RECLAIM) {		// Récupération d'objets
						D.x=D.y=D.z=0.f;
						center=&D;
						size=32.0f;
						}
					else
						c_part=false;
					}
			if(port[BUILD_PERCENT_LEFT]==0.0f) {
				model->draw(&data,owner_id==players.local_human_id && sel,false,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				if(height_line && h>1.0f && unit_manager.unit_type[type_id].canfly) {		// For flying units, draw a line that shows how high is the unit
					cam->SetView();					// Reset the camera
					glDisable(GL_TEXTURE_2D);
					glDisable(GL_LIGHTING);
					glColor3f(1.0f,1.0f,0.0f);
					glBegin(GL_LINES);
						for(float y=Pos.y;y>Pos.y-h;y-=10.0f) {
							glVertex3f(Pos.x,y,Pos.z);
							glVertex3f(Pos.x,y-5.0f,Pos.z);
							}
					glEnd();
					}
				}
			else if(port[BUILD_PERCENT_LEFT]<=33.0f) {
				float h=4.0f*model->size2;
				double eqn[4]= { 0.0f, 1.0f, 0.0f, model->center.y+h*0.5f-h*(33.0f-port[BUILD_PERCENT_LEFT])*0.033333f};

				glClipPlane(GL_CLIP_PLANE0, eqn);
				glEnable(GL_CLIP_PLANE0);
				model->draw(&data,owner_id==players.local_human_id && sel,true,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);

				eqn[1]=-eqn[1];	eqn[3]=-eqn[3];
				glClipPlane(GL_CLIP_PLANE0, eqn);
				model->draw(&data,owner_id==players.local_human_id && sel,false,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glDisable(GL_CLIP_PLANE0);

				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				}
			else if(port[BUILD_PERCENT_LEFT]<=66.0f) {
				float h=4.0f*model->size2;
				double eqn[4]= { 0.0f, 1.0f, 0.0f, model->center.y+h*0.5f-h*(66.0f-port[BUILD_PERCENT_LEFT])*0.033333f};

				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				glClipPlane(GL_CLIP_PLANE0, eqn);
				glEnable(GL_CLIP_PLANE0);
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,c_part,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

				eqn[1]=-eqn[1];	eqn[3]=-eqn[3];
				glClipPlane(GL_CLIP_PLANE0, eqn);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glDisable(GL_CLIP_PLANE0);
				}
			else {
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				model->draw(&data,owner_id==players.local_human_id && sel,true,false,build_part,target,&upos,&M,size,center,reverse,owner_id,cullface);
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				}
			}
#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	void UNIT::draw_shadow(CAMERA *cam,VECTOR Dir,MAP *map)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::draw_shadow );
#endif

		if( model == NULL )
			Console->AddEntry( "Warning : model is NULL!!" );

		if(!visible) {
			VECTOR S_Pos = Pos-(h/Dir.y)*Dir;//map->hit(Pos,Dir);
			int px=((int)(S_Pos.x)+map->map_w_d)>>4;
			int py=((int)(S_Pos.z)+map->map_h_d)>>4;
			if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	{
#ifdef	ADVANCED_DEBUG_MODE
				GuardLeave();
#endif
				return;	// Shadow out of the map
				}
			if(map->view[py][px]!=1) {
#ifdef	ADVANCED_DEBUG_MODE
				GuardLeave();
#endif
				return;	// Unvisible shadow
				}
			}

		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		float scale=unit_manager.unit_type[type_id].Scale;
		glScalef(scale,scale,scale);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		float scale_dir=model->size2;
		scale_dir+=Pos.y;
		scale_dir*=2.0f;
		model->draw_shadow(((scale_dir*Dir*RotateX(-Angle.x*DEG2RAD))*RotateZ(-Angle.z*DEG2RAD))*RotateY(-Angle.y*DEG2RAD),&data);

#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	void UNIT::draw_shadow_basic(CAMERA *cam,VECTOR Dir,MAP *map)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::draw_shadow_basic );
#endif
		if(!visible) {
			VECTOR S_Pos = Pos-(h/Dir.y)*Dir;//map->hit(Pos,Dir);
			int px=((int)(S_Pos.x+map->map_w_d))>>4;
			int py=((int)(S_Pos.z+map->map_h_d))>>4;
			if(px<0 || py<0 || px>=map->bloc_w || py>=map->bloc_h)	{
#ifdef	ADVANCED_DEBUG_MODE
				GuardLeave();
#endif
				return;	// Shadow out of the map
				}
			if(map->view[py][px]!=1) {
#ifdef	ADVANCED_DEBUG_MODE
				GuardLeave();
#endif
				return;	// Unvisible shadow
				}
			}

		cam->SetView();					// Dessine le modèle
		glTranslatef(Pos.x,Pos.y,Pos.z);
		glRotatef(Angle.x,1.0f,0.0f,0.0f);
		glRotatef(Angle.z,0.0f,0.0f,1.0f);
		glRotatef(Angle.y,0.0f,1.0f,0.0f);
		float scale=unit_manager.unit_type[type_id].Scale;
		glScalef(scale,scale,scale);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		float scale_dir=model->size2;
		scale_dir+=Pos.y+model->center.y;
		scale_dir*=2.0f;
		model->draw_shadow_basic(((scale_dir*Dir*RotateX(-Angle.x*DEG2RAD))*RotateZ(-Angle.z*DEG2RAD))*RotateY(-Angle.y*DEG2RAD),&data);

#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	const void UNIT::run_script(const float dt,const int id,MAP *map)			// Interprète les scripts liés à l'unité
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( run_script );
#endif
		if(flags==0) {
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;	}
		if(!script_env[id].running)	{
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;	}
		if(script_env[id].wait>0.0f) {
			script_env[id].wait-=dt;
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;
			}
		if(script==NULL || script_env[id].env==NULL) {
			script_env[id].running=false;
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;	// S'il n'y a pas de script associé on quitte la fonction
			}
		sint16 script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
		sint16 pos=(script_env[id].env->cur>>8);

		if(script_id<0 || script_id>=script->nb_script)	{
			script_env[id].running=false;
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return;		// Erreur, ce n'est pas un script repertorié
			}

		float divisor=i2pwr16;
		float div=0.5f*divisor;
		POINTF O;
		O.x=O.y=O.z=0.0f;
//		bool done=false;
//		int nb_code=0;

//#define PRINT_CODE			// Affiche le code éxécuté dans la console

//		do
//		{
//			uint32 code=script->script_code[script_id][pos];			// Lit un code
//			pos++;
//			nb_code++;
//			if(nb_code>=1) done=true;			// Pour éviter que le programme ne fige à cause d'un script
//			switch(code)			// Code de l'interpréteur
			switch(script->script_code[script_id][pos++])
			{
			case SCRIPT_MOVE_OBJECT:
#ifdef PRINT_CODE
				printf("MOVE_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis==0)
						v1=-v1;
					data.axe[axis][obj].move_distance=v1*div;
					data.axe[axis][obj].move_distance-=data.axe[axis][obj].pos;
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(data.axe[axis][obj].move_distance<0.0f)
						data.axe[axis][obj].move_speed=-fabs(v2*div*0.5f);
					else
						data.axe[axis][obj].move_speed=fabs(v2*div*0.5f);
				}
				break;
			case SCRIPT_WAIT_FOR_TURN:
#ifdef PRINT_CODE
				printf("WAIT_FOR_TURN\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					float a=data.axe[axis][obj].rot_angle;
					if((data.axe[axis][obj].rot_speed!=0.0f || data.axe[axis][obj].rot_accel!=0.0f) && (a!=0.0f && data.axe[axis][obj].rot_limit))
						pos-=3;
					else if(data.axe[axis][obj].rot_speed!=data.axe[axis][obj].rot_target_speed && data.axe[axis][obj].rot_speed_limit)
						pos-=3;
				}
//				done=true;
				break;
			case SCRIPT_RANDOM_NUMBER:
#ifdef PRINT_CODE
				printf("RANDOM_NUMBER\n");
#endif
				{
					int high=script_env[id].pop();
					int low=script_env[id].pop();
					script_env[id].push((rand_from_table()%(high-low+1))+low);
				}
				break;
			case SCRIPT_GREATER_EQUAL:
#ifdef PRINT_CODE
				printf("GREATER_EQUAL\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1>=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_GREATER:
#ifdef PRINT_CODE
				printf("GREATER\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1>v2 ? 1 : 0);
				}
				break;
			case SCRIPT_LESS:
#ifdef PRINT_CODE
				printf("LESS\n");
#endif
				{
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					script_env[id].push(v1<v2 ? 1 : 0);
				}
				break;
			case SCRIPT_EXPLODE:
#ifdef PRINT_CODE
				printf("EXPLODE\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int explosion_type=script_env[id].pop();
					if(visible)					// Don't draw things which could tell the player there is something there
						particle_engine.make_fire(O+Pos,1,(int)log((float)(1+explosion_type))<<2,45.0f);
					data.flag[obj]|=FLAG_EXPLODE;
					data.explosion_flag[obj]=explosion_type;
					data.axe[0][obj].move_speed=(25.0f+(rand_from_table()%2501)*0.01f)*(rand_from_table()&1 ? 1.0f : -1.0f);
					data.axe[0][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.axe[1][obj].move_speed=25.0f+(rand_from_table()%2501)*0.01f;
					data.axe[1][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.axe[2][obj].move_speed=(25.0f+(rand_from_table()%2501)*0.01f)*(rand_from_table()&1 ? 1.0f : -1.0f);
					data.axe[2][obj].rot_speed=(rand_from_table()%7201)*0.1f-360.0f;
					data.explode=true;
				}
				break;
			case SCRIPT_TURN_OBJECT:
#ifdef PRINT_CODE
				printf("TURN_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis!=2) {
						v1=-v1;
						v2=-v2;
						}
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_angle=-v1*TA2DEG;
					data.axe[axis][obj].rot_accel=0.0f;
					data.axe[axis][obj].rot_angle-=data.axe[axis][obj].angle;
					while(data.axe[axis][obj].rot_angle>180.0f)					// Fait le tour dans le sens le plus rapide
						data.axe[axis][obj].rot_angle-=360.0f;
					while(data.axe[axis][obj].rot_angle<-180.0f)					// Fait le tour dans le sens le plus rapide
						data.axe[axis][obj].rot_angle+=360.0f;
					if(data.axe[axis][obj].rot_angle>0.0f)
						data.axe[axis][obj].rot_speed=fabs(v2*TA2DEG);
					else
						data.axe[axis][obj].rot_speed=-fabs(v2*TA2DEG);
					data.axe[axis][obj].rot_limit=true;
					data.axe[axis][obj].rot_speed_limit=false;
				}
				break;
			case SCRIPT_WAIT_FOR_MOVE:
#ifdef PRINT_CODE
				printf("WAIT_FOR_MOVE\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					float a=data.axe[axis][obj].rot_angle;
					if(data.axe[axis][obj].move_distance!=0.0f)
						pos-=3;
				}
//				done=true;
				break;
			case SCRIPT_CREATE_LOCAL_VARIABLE:
#ifdef PRINT_CODE
				printf("CREATE_LOCAL_VARIABLE\n");
#endif
				break;
			case SCRIPT_SUBTRACT:
#ifdef PRINT_CODE
				printf("SUBSTRACT\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1-v2);
				}
				break;
			case SCRIPT_GET_VALUE_FROM_PORT:
#ifdef PRINT_CODE
				printf("GET_VALUE_FROM_PORT: ");
#endif
				{
					int value=script_env[id].pop();
#ifdef PRINT_CODE
					printf("%d\n",value);
#endif
					switch(value)
					{
					case ATAN:
						{
							int v1=script_env[id].pop();
							int v2=script_env[id].pop();
							port[value]=atan((float)v1/v2);
						}
						break;
					case HYPOT:
						{
							int v1=script_env[id].pop();
							int v2=script_env[id].pop();
							port[value]=sqrt((float)(v1*v1+v2*v2));
						}
						break;
					case BUGGER_OFF:
						port[value]=map->check_rect((((int)(Pos.x+map->map_w_d))>>3)-(unit_manager.unit_type[type_id].FootprintX>>1),(((int)(Pos.z+map->map_h_d))>>3)-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx) ? 0.0f : 1.0f;
						break;
					case YARD_OPEN:
						break;
					case HEALTH:
					case BUILD_PERCENT_LEFT:
						break;
					case STANDINGMOVEORDERS:			// A faire : ajouter le support des ordres de mouvement/feu
					case STANDINGFIREORDERS:
						break;
					default:
						{
							char *op[]={"INCONNU","ACTIVATION","STANDINGMOVEORDERS","STANDINGFIREORDERS","HEALTH","INBUILDSTANCE","BUSY","PIECE_XZ","PIECE_Y",
							"UNIT_XZ","UNIT_Y","UNIT_HEIGHT","XZ_ATAN","XZ_HYPOT","ATAN","HYPOT","GROUND_HEIGHT","BUILD_PERCENT_LEFT","YARD_OPEN",
							"BUGGER_OFF","ARMORED"};
							if(value>20)
								value=0;
							printf("GET_VALUE_FROM_PORT: ");
							printf("opération non gérée : %s\n",op[value]);
						}
						break;
					};
					script_env[id].push((int)port[value]);
				}
				break;
			case SCRIPT_LESS_EQUAL:
#ifdef PRINT_CODE
				printf("LESS_EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1<=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_SPIN_OBJECT:
#ifdef PRINT_CODE
				printf("SPIN_OBJECT\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					if(axis!=2) {
						v1=-v1;
						v2=-v2;
						}
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_limit=false;
					data.axe[axis][obj].rot_speed_limit=true;
					data.axe[axis][obj].rot_target_speed=v1*TA2DEG;
					if(v2) {
						if(data.axe[axis][obj].rot_target_speed>data.axe[axis][obj].rot_speed)
							data.axe[axis][obj].rot_accel=fabs(v2*TA2DEG);
						else
							data.axe[axis][obj].rot_accel=-fabs(v2*TA2DEG);
						}
					else {
						data.axe[axis][obj].rot_accel=0;
						data.axe[axis][obj].rot_speed=data.axe[axis][obj].rot_target_speed;
						}
				}
				break;
			case SCRIPT_SLEEP:
#ifdef PRINT_CODE
				printf("SLEEP\n");
#endif
				script_env[id].wait=script_env[id].pop()*0.001f;
//				done=true;
				break;
			case SCRIPT_MULTIPLY:
#ifdef PRINT_CODE
				printf("MULTIPLY\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1*v2);
				}
				break;
			case SCRIPT_CALL_SCRIPT:
#ifdef PRINT_CODE
				printf("CALL_SCRIPT\n");
#endif
				{
				int function_id=script->script_code[script_id][pos];			// Lit un code
				pos++;
				int num_param=script->script_code[script_id][pos];			// Lit un code
				pos++;
				script_env[id].env->cur=script_id+(pos<<8);
				SCRIPT_ENV_STACK *old=script_env[id].env;
				script_env[id].env=new SCRIPT_ENV_STACK;
				script_env[id].env->init();
				script_env[id].env->next=old;
				script_env[id].env->cur=function_id;
				for(int i=0;i<num_param;i++)		// Lit les paramètres
					script_env[id].env->var[i]=script_env[id].pop();
//				done=true;
				pos=0;
				script_id=function_id;
				}
				break;
			case SCRIPT_SHOW_OBJECT:
#ifdef PRINT_CODE
				printf("SHOW_OBJECT\n");
#endif
				data.flag[script->script_code[script_id][pos++]]&=(~FLAG_HIDE);
				break;
			case SCRIPT_EQUAL:
#ifdef PRINT_CODE
				printf("EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1==v2 ? 1 : 0);
				}
				break;
			case SCRIPT_NOT_EQUAL:
#ifdef PRINT_CODE
				printf("NOT_EQUAL\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1!=v2 ? 1 : 0);
				}
				break;
			case SCRIPT_IF:
#ifdef PRINT_CODE
				printf("IF\n");
#endif
				if(script_env[id].pop()!=0)
					pos++;
				else {
					int target_offset=script->script_code[script_id][pos];			// Lit un code
					pos=target_offset-script->dec_offset[script_id];								// Déplace l'éxecution
					}
				break;
			case SCRIPT_HIDE_OBJECT:
#ifdef PRINT_CODE
				printf("HIDE_OBJECT\n");
#endif
				data.flag[script->script_code[script_id][pos++]]|=FLAG_HIDE;
				break;
			case SCRIPT_SIGNAL:
#ifdef PRINT_CODE
				printf("SIGNAL\n");
#endif
				script_env[id].env->cur=script_id+(pos<<8);			// Sauvegarde la position
				raise_signal(script_env[id].pop());					// Tue tout les processus utilisant ce signal
#ifdef	ADVANCED_DEBUG_MODE
				GuardLeave();
#endif
				return;												// Quitte la fonction d'émulation de scripts
			case SCRIPT_DONT_CACHE:
#ifdef PRINT_CODE
				printf("DONT_CACHE\n");
#endif
				pos++;
				break;
			case SCRIPT_SET_SIGNAL_MASK:
#ifdef PRINT_CODE
				printf("SET_SIGNAL_MASK\n");
#endif
				script_env[id].env->signal_mask=script_env[id].pop();
				break;
			case SCRIPT_NOT:
#ifdef PRINT_CODE
				printf("NOT\n");
#endif
				script_env[id].push(!script_env[id].pop());
				break;
			case SCRIPT_DONT_SHADE:
#ifdef PRINT_CODE
				printf("DONT_SHADE\n");
#endif
				pos++;
				break;
			case SCRIPT_EMIT_SFX:
#ifdef PRINT_CODE
				printf("EMIT_SFX: ");
#endif
				{
					int smoke_type=script_env[id].pop();
					int from_piece=script->script_code[script_id][pos++];
#ifdef PRINT_CODE
					printf("smoke_type %d from %d\n",smoke_type,from_piece);
#endif
					if(visible) {
						compute_model_coord();
						if(data.dir[from_piece].x!=0.0f || data.dir[from_piece].y!=0.0f || data.dir[from_piece].z!=0.0f) {
							VECTOR dir=data.dir[from_piece];
							switch(smoke_type)
							{
							case 0:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,fire,1,10.0f,2.5f,5.0f,true);
								break;
							case 2:
							case 3:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,0,1,10.0f);
								break;
							case 257:			// Fumée
							case 258:
								particle_engine.emit_part(O+Pos+data.pos[from_piece],dir,0,1,10.0f,10.0f,10.0f,true);
								break;
							};
							}
						else
							switch(smoke_type)
							{
							case 0:
								particle_engine.make_smoke(O+Pos+data.pos[from_piece],fire,1,0.0f,0.0f);
								break;
							case 257:
							case 258:
								particle_engine.make_smoke(O+Pos+data.pos[from_piece],0,1,10.0f);
								break;
							};
						}
				}
				break;
			case SCRIPT_PUSH_CONST:
#ifdef PRINT_CODE
				printf("PUSH_CONST\n");
#endif
				script_env[id].push(script->script_code[script_id][pos]);
				pos++;
				break;
			case SCRIPT_PUSH_VAR:
#ifdef PRINT_CODE
				printf("PUSH_VAR\n");
#endif
				script_env[id].push(script_env[id].env->var[script->script_code[script_id][pos]]);
				pos++;
				break;
			case SCRIPT_SET_VAR:
#ifdef PRINT_CODE
				printf("SET_VAR\n");
#endif
				script_env[id].env->var[script->script_code[script_id][pos]]=script_env[id].pop();
				pos++;
				break;
			case SCRIPT_PUSH_STATIC_VAR:
#ifdef PRINT_CODE
				printf("PUSH_STATIC_VAR\n");
#endif
				script_env[id].push(s_var[script->script_code[script_id][pos]]);
				pos++;
				break;
			case SCRIPT_SET_STATIC_VAR:
#ifdef PRINT_CODE
				printf("SET_STATIC_VAR\n");
#endif
				s_var[script->script_code[script_id][pos]]=script_env[id].pop();
				pos++;
				break;
			case SCRIPT_OR:
#ifdef PRINT_CODE
				printf("OR\n");
#endif
				{
					int v1=script_env[id].pop(),v2=script_env[id].pop();
					script_env[id].push(v1|v2);
				}
				break;
			case SCRIPT_START_SCRIPT:				// Transfère l'éxecution à un autre script
#ifdef PRINT_CODE
				printf("START_SCRIPT\n");
#endif
				{
				int function_id=script->script_code[script_id][pos++];			// Lit un code
				int num_param=script->script_code[script_id][pos++];			// Lit un code
				int s_id=launch_script(function_id);
				if(s_id>=0) {
					for(int i=0;i<num_param;i++)		// Lit les paramètres
						script_env[s_id].env->var[i]=script_env[id].pop();
					script_env[s_id].env->signal_mask=script_env[id].env->signal_mask;
					}
				else
					for(int i=0;i<num_param;i++)		// Enlève les paramètres de la pile
						script_env[id].pop();
//				script_env[id].running=false;		// Commenté car empêche le script de l'Atlas de fonctionner correctement
//				done=true;
				}
				break;
			case SCRIPT_RETURN:		// Retourne au script précédent
#ifdef PRINT_CODE
				printf("RETURN\n");
#endif
				{
					script_val[script_id]=script_env[id].env->var[0];
					SCRIPT_ENV_STACK *old=script_env[id].env;
					script_env[id].env=script_env[id].env->next;
					delete old;
					script_env[id].pop();		// Enlève la valeur retournée
				}
				if(script_env[id].env) {
					script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
					pos=(script_env[id].env->cur>>8);
					}
//				done=true;
				break;
			case SCRIPT_JUMP:						// Commande de saut
#ifdef PRINT_CODE
				printf("JUMP\n");
#endif
				{
					int target_offset=script->script_code[script_id][pos];			// Lit un code
					pos=target_offset-script->dec_offset[script_id];								// Déplace l'éxecution
				}
				break;
			case SCRIPT_ADD:
#ifdef PRINT_CODE
				printf("ADD\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1+v2);
				}
				break;	//added
			case SCRIPT_STOP_SPIN:
#ifdef PRINT_CODE
				printf("STOP_SPIN\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v=script_env[id].pop();
					if(axis!=2)
						v=-v;
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					data.axe[axis][obj].rot_limit=false;
					data.axe[axis][obj].rot_speed_limit=true;
					data.axe[axis][obj].rot_target_speed=0.0f;
					if(v==0) {
						data.axe[axis][obj].rot_speed=0.0f;
						data.axe[axis][obj].rot_accel=0.0f;
						}
					else {
						if(data.axe[axis][obj].rot_speed>0.0f)
							data.axe[axis][obj].rot_accel=-abs(v);
						else
							data.axe[axis][obj].rot_accel=abs(v);
						}
				}
				break;	//added
			case SCRIPT_DIVIDE:
#ifdef PRINT_CODE
				printf("DIVIDE\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v2 / v1);
				}
				break;	//added
			case SCRIPT_MOVE_PIECE_NOW:
#ifdef PRINT_CODE
				printf("MOVE_PIECE_NOW\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(axis==0)
						data.axe[axis][obj].pos=-script_env[id].pop()*div;
					else
						data.axe[axis][obj].pos=script_env[id].pop()*div;
				}
				break;	//added
			case SCRIPT_TURN_PIECE_NOW:
#ifdef PRINT_CODE
				printf("TURN_PIECE_NOW\n");
#endif
				{
					int obj=script->script_code[script_id][pos++];
					int axis=script->script_code[script_id][pos++];
					int v=script_env[id].pop();
					data.axe[axis][obj].is_moving=true;
					data.is_moving=true;
					if(axis!=2)
						v=-v;
					data.axe[axis][obj].angle=-v*TA2DEG;
				}
				break;	//added
			case SCRIPT_CACHE:
#ifdef PRINT_CODE
				printf("CACHE\n");
#endif
				pos++;
				break;	//added
			case SCRIPT_COMPARE_AND:
#ifdef PRINT_CODE
				printf("COMPARE_AND\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1 && v2);
				}
				break;	//added
			case SCRIPT_COMPARE_OR:
#ifdef PRINT_CODE
				printf("COMPARE_OR\n");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
					script_env[id].push(v1 || v2);
				}
				break;	//added
			case SCRIPT_CALL_FUNCTION:
#ifdef PRINT_CODE
				printf("CALL_FUNCTION\n");
#endif
				{
				int function_id=script->script_code[script_id][pos++];			// Lit un code
				int num_param=script->script_code[script_id][pos++];			// Lit un code
				script_env[id].env->cur=script_id+(pos<<8);
				SCRIPT_ENV_STACK *old=script_env[id].env;
				script_env[id].env=new SCRIPT_ENV_STACK;
				script_env[id].env->init();
				script_env[id].env->next=old;
				script_env[id].env->cur=function_id;
				for(int i=0;i<num_param;i++)		// Lit les paramètres
					script_env[id].env->var[i]=script_env[id].pop();
//				done=true;
				pos=0;
				script_id=function_id;
				}
				break;	//added
			case SCRIPT_GET:
				printf("GET *\n");
				{
					int v4=script_env[id].pop();
					int v3=script_env[id].pop();
					int v2=script_env[id].pop();
					int v1=script_env[id].pop();
					int val=script_env[id].pop();
					switch(val)
					{
					case ACTIVATION:
					case STANDINGMOVEORDERS:
					case STANDINGFIREORDERS:
					case HEALTH:
					case INBUILDSTANCE:
					case BUSY:
					case BUILD_PERCENT_LEFT:
					case YARD_OPEN:
					case BUGGER_OFF:
					case ARMORED:
						script_env[id].push((int)port[val]);
						break;
					case PIECE_XZ:
						compute_model_coord();
						script_env[id].push(((int)data.pos[v1].x<<16) | ((int)data.pos[v1].z));
						break;
					case PIECE_Y:
						compute_model_coord();
						script_env[id].push((int)data.pos[v1].y);
						break;
					case UNIT_XZ:
						if (v1 >= 0 && v1<units.max_unit && units.unit[v1].flags)	
							script_env[id].push(((int)units.unit[v1].Pos.x<<16) | ((int)units.unit[v1].Pos.z));
						else
							script_env[id].push(((int)Pos.x<<16) | ((int)Pos.z));
/*						if (v1 == 0)	
							script_env[id].push(((int)Pos.x<<16) | ((int)Pos.z));
						else
							script_env[id].push(0);*/
						break;
					case UNIT_Y:
						if (v1 >= 0 && v1<units.max_unit && units.unit[v1].flags)	
							script_env[id].push((int)units.unit[v1].Pos.y);
						else
							script_env[id].push((int)Pos.y);
						break;
					case UNIT_HEIGHT:
						if (v1 >= 0 && v1<units.max_unit && units.unit[v1].flags)	
							script_env[id].push((int)units.unit[v1].model->size2);
						else
							script_env[id].push((int)model->size2);
						break;
					case XZ_ATAN:
						{
							int x=v1>>16;
							int z=v1&0xFFFF;
							script_env[id].push((int)atan((float)x/z));
						}
						break;
					case XZ_HYPOT:
						{
							int x=v1>>16;
							int z=v1&0xFFFF;
							script_env[id].push((int)sqrt((float)(x*x+z*z)));
						}
						break;
					case ATAN:
						script_env[id].push((int)atan((float)v1/(v2&0xFFFF)));
						break;
					case HYPOT:
						script_env[id].push((int)sqrt((float)(v1*v1+((v2&0xFFFF)*(v2&0xFFFF)))));
						break;
					case GROUND_HEIGHT:
						script_env[id].push((int)map->get_unit_h((v1>>16)&0xFFFF,v1&0xFFFF));
//						script_env[id].push((int)map->get_unit_h(Pos.x,Pos.z));
						break;
					default:
						printf("GET constante inconnue %d", val);
					};
				}
				break;	//added
			case SCRIPT_SET_VALUE:
#ifdef PRINT_CODE
				printf("SET_VALUE *: ");
#endif
				{
					int v1=script_env[id].pop();
					int v2=script_env[id].pop();
#ifdef PRINT_CODE
					printf("%d %d\n",v1,v2);
#endif
					port[v2]=v1;
					switch(v2)
					{
					case YARD_OPEN:
						if(!map->check_rect((((int)(Pos.x+map->map_w_d))>>3)-(unit_manager.unit_type[type_id].FootprintX>>1),(((int)(Pos.z+map->map_h_d))>>3)-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx))
							port[v2]= (port[v2]==0.0f) ? 1.0f : 0.0f;
						break;
					case BUGGER_OFF:
						if(port[v2])	{
							int px=((int)(Pos.x)+map->map_w_d)>>3;
							int py=((int)(Pos.z)+map->map_h_d)>>3;
							for(int y=py-(unit_manager.unit_type[type_id].FootprintZ>>1);y<=py+(unit_manager.unit_type[type_id].FootprintZ>>1);y++)
								if(y>=0 && y<(map->bloc_h<<1)-1)
									for(int x=px-(unit_manager.unit_type[type_id].FootprintX>>1);x<=px+(unit_manager.unit_type[type_id].FootprintX>>1);x++)
										if(x>=0 && x<(map->bloc_w<<1)-1)
											if(map->map_data[y][x].unit_idx!=-1 && map->map_data[y][x].unit_idx!=idx) {
												int cur_idx=map->map_data[y][x].unit_idx;
												if(units.unit[cur_idx].owner_id==owner_id && units.unit[cur_idx].port[BUILD_PERCENT_LEFT]==0.0f && (units.unit[cur_idx].mission==NULL || units.unit[cur_idx].mission->mission!=MISSION_MOVE)) {
													VECTOR target=units.unit[cur_idx].Pos;
													target.z+=100.0f;
													PATH_NODE *path=NULL;
													if(!unit_manager.unit_type[units.unit[cur_idx].type_id].canfly) {
														float dh_max=unit_manager.unit_type[units.unit[cur_idx].type_id].MaxSlope*H_DIV;
														float h_min=map->sealvl-unit_manager.unit_type[units.unit[cur_idx].type_id].MaxWaterDepth*H_DIV;
														if(unit_manager.unit_type[units.unit[cur_idx].type_id].norestrict)
															h_min=0.0f;
														path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,dh_max,h_min,units.unit[cur_idx].Pos,target,unit_manager.unit_type[units.unit[cur_idx].type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[units.unit[cur_idx].type_id].FootprintX,unit_manager.unit_type[units.unit[cur_idx].type_id].FootprintZ,cur_idx);
														}
													else
														path=direct_path(target);
													sound_manager->PlayTDFSound( unit_manager.unit_type[units.unit[cur_idx].type_id].soundcategory, "ok1" , &units.unit[cur_idx].Pos );
//													sound_manager->PlaySound( &unit_manager.unit_type[units.unit[cur_idx].type_id].ok1 , &units.unit[cur_idx].Pos );
													units.unit[cur_idx].set_mission(MISSION_MOVE,&target,false,0,true,NULL,path);
													}
												}
							}
						break;
					};
				}
				break;	//added
			case SCRIPT_ATTACH_UNIT:
				printf("ATTACH_UNIT *\n");
				{
					int v1 = script_env[id].pop();
					int v2 = script_env[id].pop();
					UNIT *target_unit = &(units.unit[v1]);
					target_unit->attached=true;
					link_list[nb_attached]=v2;
					attached_list[nb_attached++]=target_unit->idx;
					map->rect(target_unit->cur_px-(unit_manager.unit_type[target_unit->type_id].FootprintX>>1),target_unit->cur_py-(unit_manager.unit_type[target_unit->type_id].FootprintZ>>1),unit_manager.unit_type[target_unit->type_id].FootprintX,unit_manager.unit_type[target_unit->type_id].FootprintZ,-1,unit_manager.unit_type[target_unit->type_id].yardmap,target_unit->port[YARD_OPEN]!=0.0f);
				}
				pos++;
				break;	//added
			case SCRIPT_DROP_UNIT:
				printf("DROP_UNIT *\n");
				script_env[id].pop();
				pos++;
				break;	//added
			default:
				printf("UNKNOWN %d, STOPPING SCRIPT\n",script->script_code[script_id][--pos]);
				{
					script_val[script_id]=script_env[id].env->var[0];
					SCRIPT_ENV_STACK *old=script_env[id].env;
					script_env[id].env=script_env[id].env->next;
					delete old;
				}
				if(script_env[id].env) {
					script_id=(script_env[id].env->cur&0xFF);			// Récupère l'identifiant du script en cours d'éxecution et la position d'éxecution
					pos=(script_env[id].env->cur>>8);
					}
				else
					script_env[id].running=false;
//				done=true;
			};
//		}while(!done);
		if(script_env[id].env)
			script_env[id].env->cur=script_id+(pos<<8);
#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
	}

	inline float ballistic_angle(float v,float g,float d,float y_s,float y_e)			// Calculs de ballistique pour l'angle de tir
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( ballistic_angle );
#endif
		float v2=v*v;
		float gd=g*d;
		float v2gd = v2/gd;
		float a=v2gd*(4.0f*v2gd-8.0f*(y_e-y_s)/d)-4.0f;
#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
		if(a<0.0f)				// Pas de solution
			return 360.0f;
		return RAD2DEG*atan(v2gd-0.5f*sqrt(a));
 	}

	const int UNIT::move(const float dt,MAP *map,const int *path_exec,const int key_frame)
	{
#ifdef	ADVANCED_DEBUG_MODE
		GuardEnter( UNIT::move );
#endif
		float was_open = port[YARD_OPEN];
//		if(unit_manager.unit_type[type_id].canmove)
//			map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
		sint32	o_px=cur_px;
		sint32	o_py=cur_py;
		compute_coord=true;
		VECTOR	old_V = V;			// Store the speed, so we can do some calculations

		if(port[BUILD_PERCENT_LEFT]==0.0f && unit_manager.unit_type[type_id].isfeature) {		// Turn this unit into a feature
			int x=(int)((Pos.x)+map->map_w_d)>>3;
			int y=(int)((Pos.z)+map->map_h_d)>>3;
			if(x>0 && y>0 && x<(map->bloc_w<<1) && y<(map->bloc_h<<1))
				if(map->map_data[y][x].stuff==-1) {
					int type=feature_manager.get_feature_index(unit_manager.unit_type[type_id].Unitname);
					map->map_data[y][x].stuff=features.add_feature(Pos,type);
//					if(type!=-1 && (feature_manager.feature[type].height>5 || feature_manager.feature[type].m3d))
					if(type!=-1 && feature_manager.feature[type].blocking)
						map->rect(x-(feature_manager.feature[type].footprintx>>1),y-(feature_manager.feature[type].footprintz>>1),feature_manager.feature[type].footprintx,feature_manager.feature[type].footprintz,-2-map->map_data[y][x].stuff);
				}
#ifdef	ADVANCED_DEBUG_MODE
			GuardLeave();
#endif
			return -1;
			}

		if(map->ota_data.waterdoesdamage && Pos.y<map->sealvl)
			hp-=dt*map->ota_data.waterdamage;

		bool jump_commands = (((idx+key_frame)&0xF) == 0);		// Saute certaines commandes / Jump some commands so it runs faster with lots of units

		if(hp<=0.0f) {			// L'unité est détruite
			switch(flags&0xFD)
			{
			case 1:				// Début de la mort de l'unité	(Lance le script)
				flags=4;
				{
					int param[]={ severity };
					launch_script(get_script_index(SCRIPT_killed),1,param);
					int w_id=weapons.add_weapon(weapon_manager.get_weapon_index(unit_manager.unit_type[type_id].ExplodeAs),idx);
					b_delay[0]=1.0f;
					if(w_id>=0) {
						weapons.weapon[w_id].Pos=Pos;
						weapons.weapon[w_id].target_pos=Pos;
						weapons.weapon[w_id].target=-1;
						}
				}
				break;
			case 4:				// Vérifie si le script est terminé
				if(!is_running(get_script_index(SCRIPT_killed))) {
					if(b_delay[0]<=0.0f) {
#ifdef	ADVANCED_DEBUG_MODE
						GuardLeave();
#endif
						return -1;
						}
					b_delay[0]-=dt;
					for(int i=0;i<data.nb_piece;i++)
						if((data.flag[i]&FLAG_EXPLODE)!=FLAG_EXPLODE || (data.flag[i]&FLAG_EXPLODE && (data.explosion_flag[i]&EXPLODE_BITMAPONLY)))
							data.flag[i]|=FLAG_HIDE;
					}
				break;
			default:		// It doesn't explode (it has been reclaimed for example)
				flags=4;
				{
					int param[]={ severity };
					launch_script(get_script_index(SCRIPT_killed),1,param);
					b_delay[0]=0.0f;
				}
			};
			if(data.nb_piece>0 && port[BUILD_PERCENT_LEFT]==0.0f) {
				data.move(dt,map->ota_data.gravity);
				if(c_time>=0.1f) {
					c_time=0.0f;
					POINTF O;
					O.x=O.y=O.z=0.0f;
					for(int i=0;i<data.nb_piece;i++)
						if(data.flag[i]&FLAG_EXPLODE && (data.explosion_flag[i]&EXPLODE_BITMAPONLY)!=EXPLODE_BITMAPONLY) {
							if(data.explosion_flag[i]&EXPLODE_FIRE) {
								compute_model_coord();
								particle_engine.make_smoke(O+Pos+data.pos[i],fire,1,0.0f,0.0f);
								}
							if(data.explosion_flag[i]&EXPLODE_SMOKE) {
								compute_model_coord();
								particle_engine.make_smoke(O+Pos+data.pos[i],0,1,0.0f,0.0f);
								}
							}
					}
				}
			goto script_exec;
			}
		else if(!jump_commands && do_nothing())
			if(Pos.x<-map->map_w_d || Pos.x>map->map_w_d || Pos.z<-map->map_h_d || Pos.z>map->map_h_d) {
				VECTOR target;
				target=Pos;
				if(target.x<-map->map_w_d+256)
					target.x=-map->map_w_d+256;
				else if(target.x>map->map_w_d-256)
					target.x=map->map_w_d-256;
				if(target.z<-map->map_h_d+256)
					target.z=-map->map_h_d+256;
				else if(target.z>map->map_h_d-256)
					target.z=map->map_h_d-256;
				add_mission(MISSION_MOVE_CALC_PATH,&target,true,1,NULL,NULL);		// On reste sur l'écran
				}

		flags&=0xEF;		// To fix a bug

		if(port[BUILD_PERCENT_LEFT]>0.0f)	{		// L'unité n'est pas finie
			if(!built)
				hp-=dt*0.01f*unit_manager.unit_type[type_id].MaxDamage;
			goto script_exec;
			}

		if(data.nb_piece>0)
			data.move(dt,units.g_dt);

		if(attached)
			goto script_exec;

		if(unit_manager.unit_type[type_id].canload && nb_attached>0) {
			int e=0;
			compute_model_coord();
			for(int i=0;i+e<nb_attached;i++) {
				if(units.unit[attached_list[i]].flags) {
					units.unit[attached_list[i]].Pos=Pos+data.pos[link_list[i]];
					units.unit[attached_list[i]].cur_px = ((int)(units.unit[attached_list[i]].Pos.x)+map->map_w_d)>>3;
					units.unit[attached_list[i]].cur_py = ((int)(units.unit[attached_list[i]].Pos.z)+map->map_h_d)>>3;
					units.unit[attached_list[i]].Pos.y-=units.unit[attached_list[i]].model->size2;
					units.unit[attached_list[i]].Angle=Angle;
					}
				else {
					e++;
					i--;
					continue;
					}
				attached_list[i]=attached_list[i+e];
				}
			nb_attached-=e;
			}

		if(planned_weapons>0.0f) {			// Construit des armes / build weapons
			float old=planned_weapons-(int)planned_weapons;
			int idx=-1;
			if(unit_manager.unit_type[type_id].weapon[0] && unit_manager.unit_type[type_id].weapon[0]->stockpile)	idx=0;
			else if(unit_manager.unit_type[type_id].weapon[1] && unit_manager.unit_type[type_id].weapon[1]->stockpile)	idx=0;
			else if(unit_manager.unit_type[type_id].weapon[2] && unit_manager.unit_type[type_id].weapon[2]->stockpile)	idx=0;
			if(idx!=-1 && unit_manager.unit_type[type_id].weapon[idx]->reloadtime!=0.0f)	{
				float dn=dt/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				float conso_metal=((float)unit_manager.unit_type[type_id].weapon[idx]->metalpershot)/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				float conso_energy=((float)unit_manager.unit_type[type_id].weapon[idx]->energypershot)/unit_manager.unit_type[type_id].weapon[idx]->reloadtime;
				if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
					metal_cons+=conso_metal;
					energy_cons+=conso_energy;
					planned_weapons-=dn;
					float last=planned_weapons-(int)planned_weapons;
					if((last==0.0f && last!=old) || (last>old && old>0.0f) || planned_weapons<=0.0f)		// On en a fini une / one is made
						stock[idx]++;
					if(planned_weapons<0.0f)
						planned_weapons=0.0f;
					}
				}
			}

		V_Angle.x=V_Angle.y=V_Angle.z=0.0f;
		c_time+=dt;
		if(mission) {
				mission->time+=dt;
				switch(mission->mission)
				{
				case MISSION_ATTACK:
				case MISSION_AIM:
				case MISSION_SHOOT:
					b_delay[0]+=dt;
					b_delay[1]+=dt;
					b_delay[2]+=dt;
					break;
				};
				switch(mission->mission)						// Commandes générales
				{
				case MISSION_LOAD:
					if(mission->p!=NULL) {
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags==0) {
							next_mission();
							break;
							}
						VECTOR Dir=target_unit->Pos-Pos;
						Dir.y=0.0f;
						mission->target=target_unit->Pos;
						float dist=Dir.Sq();
						int maxdist=0;
						if(unit_manager.unit_type[type_id].TransMaxUnits==1)		// Code for units like the arm atlas
							maxdist=3;
						else
							maxdist=unit_manager.unit_type[type_id].SightDistance;
						if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
							VECTOR target;
							dist=sqrt(dist);
							target=mission->target;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
							}
						else {
							if(mission->last_d>=0.0f) {
								if(unit_manager.unit_type[type_id].TransMaxUnits==1) {		// Code for units like the arm atlas
									if(nb_attached==0) {
										int param[]={ (int)(target_unit->Pos.y+target_unit->model->size2) };
										launch_script(get_script_index(SCRIPT_BeginTransport),1,param);
										target_unit->attached=true;
										link_list[nb_attached]=0;
										attached_list[nb_attached++]=target_unit->idx;
										}
									next_mission();
									}
								else {
									if(nb_attached>=unit_manager.unit_type[type_id].TransMaxUnits) {
										next_mission();
										break;
										}
									int param[]= { target_unit->idx };
									launch_script(get_script_index(SCRIPT_TransportPickup),1,param);
									}
								mission->last_d=-1.0f;
								}
							else {
								if(!is_running(get_script_index(SCRIPT_TransportPickup)))
									next_mission();
								}
							}
						}
					break;
				case MISSION_RECLAIM:
					if(mission->p!=NULL)	{		// Récupère une unité
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags && target_unit->hp<=0.0f) {		// Travail terminé
							launch_script(get_script_index(SCRIPT_stopbuilding));
							launch_script(get_script_index(SCRIPT_stop));
							target_unit->flags|=0x10;			// This unit is being reclaimed it doesn't explode!
							next_mission();
							}
						else if(target_unit->flags) {
							VECTOR Dir=target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(mission->last_d>=0.0f) {
									launch_script(get_script_index(SCRIPT_startbuilding));
									launch_script(get_script_index(SCRIPT_go));
									mission->last_d=-1.0f;
									}
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else {			// Récupère l'unité
									float recup=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].HealTime;
									if(recup>target_unit->hp)	recup=target_unit->hp;
									target_unit->hp-=recup;
									target_unit->flags|=0x10;			// This unit is being reclaimed it doesn't explode!
									if(dt>0.0f)
										metal_prod+=recup*unit_manager.unit_type[target_unit->type_id].BuildCostMetal/(dt*unit_manager.unit_type[target_unit->type_id].MaxDamage);
									}
								}
							}
						}
					else if(mission->data>=0 && mission->data<features.nb_features && features.feature[mission->data].type>=0)	{	// Récupère un élément du décors/une carcasse
						if(features.feature[mission->data].hp<=0.0f) {		// Travail terminé
							int x=((int)(features.feature[mission->data].Pos.x)+map->map_w_d-4)>>3;		// Efface l'objet
							int y=((int)(features.feature[mission->data].Pos.z)+map->map_h_d-4)>>3;
							map->rect(x-(feature_manager.feature[features.feature[mission->data].type].footprintx>>1),y-(feature_manager.feature[features.feature[mission->data].type].footprintz>>1),feature_manager.feature[features.feature[mission->data].type].footprintx,feature_manager.feature[features.feature[mission->data].type].footprintz,-1);
							features.delete_feature(mission->data);			// Supprime l'objet
							launch_script(get_script_index(SCRIPT_stopbuilding));
							launch_script(get_script_index(SCRIPT_stop));
							next_mission();
							}
						else {
							VECTOR Dir=features.feature[mission->data].Pos-Pos;
							Dir.y=0.0f;
							mission->target=features.feature[mission->data].Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(mission->last_d>=0.0f) {
									launch_script(get_script_index(SCRIPT_startbuilding));
									launch_script(get_script_index(SCRIPT_go));
									mission->last_d=-1.0f;
									}
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else {			// Récupère l'objet
									float recup=dt*unit_manager.unit_type[type_id].WorkerTime;
									if(recup>features.feature[mission->data].hp)	recup=features.feature[mission->data].hp;
									features.feature[mission->data].hp-=recup;
									if(dt>0.0f) {
										metal_prod+=recup*feature_manager.feature[features.feature[mission->data].type].metal/(dt*feature_manager.feature[features.feature[mission->data].type].damage);
										energy_prod+=recup*feature_manager.feature[features.feature[mission->data].type].energy/(dt*feature_manager.feature[features.feature[mission->data].type].damage);
										}
									}
								}
							}
						}
					else
						next_mission();
					break;
				case MISSION_GUARD:
					if(jump_commands)	break;
					if(mission->p!=NULL && ((UNIT*)mission->p)->flags && ((UNIT*)mission->p)->owner_id==owner_id) {		// On ne défend pas n'importe quoi
						if(unit_manager.unit_type[type_id].Builder) {
							if(((UNIT*)mission->p)->port[BUILD_PERCENT_LEFT]>0.0f || ((UNIT*)mission->p)->hp<unit_manager.unit_type[((UNIT*)mission->p)->type_id].MaxDamage) {		// Répare l'unité
								add_mission(MISSION_REPAIR,&((UNIT*)mission->p)->Pos,true,0,((UNIT*)mission->p),NULL);
								break;
								}
							else if(((UNIT*)mission->p)->mission!=NULL && (((UNIT*)mission->p)->mission->mission==MISSION_BUILD_2 || ((UNIT*)mission->p)->mission->mission==MISSION_REPAIR)) {		// L'aide à construire
								add_mission(MISSION_REPAIR,&((UNIT*)mission->p)->mission->target,true,0,((UNIT*)mission->p)->mission->p,NULL);
								break;
								}
							}
						if(unit_manager.unit_type[type_id].canattack) {
							if(((UNIT*)mission->p)->mission!=NULL && (((UNIT*)mission->p)->mission->mission==MISSION_SHOOT || ((UNIT*)mission->p)->mission->mission==MISSION_AIM || ((UNIT*)mission->p)->mission->mission==MISSION_ATTACK)) {		// L'aide à construire
								add_mission(MISSION_ATTACK,&((UNIT*)mission->p)->mission->target,true,0,((UNIT*)mission->p)->mission->p,NULL);
								break;
								}
							}
						if(((VECTOR)(Pos-((UNIT*)mission->p)->Pos)).Sq()>=25600.0f) {			// On reste assez près
							VECTOR target=((UNIT*)mission->p)->Pos;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,10,NULL,NULL);
							break;
							}
						}
					else
						next_mission();
					break;
				case MISSION_MOVE_CALC_PATH:
					if(jump_commands)	break;
					if(*path_exec<MAX_PATH_EXEC)
					{
						*path_exec++;
						if(unit_manager.unit_type[type_id].canfly) {
							if(mission->data<=0)
								mission->path=direct_path(mission->target);
							else {
								VECTOR Dir=mission->target-Pos;
								Dir.Unit();
								mission->path=direct_path(mission->target-(mission->data<<3)*Dir);
								}
							}
						else {
							float dh_max=unit_manager.unit_type[type_id].MaxSlope*H_DIV;
							float h_min=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
							if(unit_manager.unit_type[type_id].norestrict)
								h_min=0.0f;
							if(mission->data<=0)
								mission->path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
														dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx);
							else
								mission->path=compute_path_min_dist(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
														dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,mission->data);
							}
						c_time=0.0f;
						mission->mission=MISSION_MOVE;
					}
					break;
				case MISSION_PATROL:					// Mode patrouille
					{
						add_mission(MISSION_MOVE_CALC_PATH,&Pos,false,0,NULL,NULL,1);			// Retour à la case départ après l'éxécution de tous les ordres / back to beginning
						add_mission(MISSION_PATROL,&mission->target,false,0,NULL,NULL,1);		// Retour à la case départ après l'éxécution de tous les ordres / restart patrol order

						if(unit_manager.unit_type[type_id].canfly)
							mission->path=direct_path(mission->target);
						else {
							float dh_max=unit_manager.unit_type[type_id].MaxSlope*H_DIV;
							float h_min=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
							if(unit_manager.unit_type[type_id].norestrict)
								h_min=0.0f;
								mission->path=compute_path(map->map_data,map->path,map->h_map,map->map_w,map->map_h,map->bloc_w<<1,map->bloc_h<<1,
															dh_max,h_min,Pos,mission->target,unit_manager.unit_type[type_id].TEDclass==CLASS_SHIP,unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx);
							}
						c_time=0.0f;
						mission->mission=MISSION_MOVE;
						mission->flags=1;
					}
					break;
				case MISSION_STANDBY:
				case MISSION_VTOL_STANDBY:
					if(jump_commands)	break;
					if(mission->data>5) {
						if(mission->next)		// If there is a mission after this one
							next_mission();
						}
					else
						mission->data++;
					break;
				case MISSION_ATTACK:										// Attaque une unité / attack a unit
					if(jump_commands)	break;
					{
						UNIT *target_unit=(mission->flags&4)==4 ? NULL : (UNIT*) mission->p;
						WEAPON *target_weapon=(mission->flags&4)==4 ? (WEAPON*) mission->p : NULL;
						if((target_unit!=NULL && target_unit->flags) || (target_weapon!=NULL && target_weapon->weapon_id!=-1)) {
							VECTOR Dir=target_unit==NULL ? target_weapon->Pos-Pos : target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit==NULL ? target_weapon->Pos : target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=0;
							int mindist=0;
							if(mission->data<0)	mission->data=0;
							else mission->data%=3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;
							if(unit_manager.unit_type[type_id].weapon[mission->data]==NULL)
								mission->data=++mission->data%3;

							if(unit_manager.unit_type[type_id].weapon[mission->data]) {
								if(Dir%V<0.0f && !unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight
								&& !(unit_manager.unit_type[type_id].weapon[mission->data]->rendertype==RENDER_TYPE_LASER || unit_manager.unit_type[type_id].weapon[mission->data]->rendertype==RENDER_TYPE_LIGHTNING)
								&& unit_manager.unit_type[type_id].canfly) {
									mission->data++;
									break;		// Ne tire pas si on s'éloigne de la cible avec une arme non "laser" / dont fire if moving away from the target with a non laser weapon
									}
								if(unit_manager.unit_type[type_id].attackrunlength>0) {
									float t=sqrt(2.0f/map->ota_data.gravity*fabs(Pos.y-mission->target.y));
									mindist=(int)(t*V.Norm())-(unit_manager.unit_type[type_id].attackrunlength+2>>2);
									maxdist=mindist+(unit_manager.unit_type[type_id].attackrunlength+1>>1);
									}
								else
									maxdist=unit_manager.unit_type[type_id].weapon[mission->data]->range>>1;
								if(burst[mission->data]==0 && b_delay[mission->data]<unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime
								&& !unit_manager.unit_type[type_id].weapon[mission->data]->stockpile)	 {
									mission->data++;
									break;
									}
								}
							else {
								next_mission();
								break;
								}

							if(unit_manager.unit_type[type_id].weapon[mission->data]->commandfire && (mission->flags&8)==8) {		// Don't let it fire with a weapon if it is not allowed to
								mission->data++;
								break;
								}

							if(unit_manager.unit_type[type_id].weapon[mission->data]->stockpile && stock[mission->data]<=0) {	// We have nothing to fire / rien à tirer
								mission->data=++mission->data%3;
								break;
								}

							if(dist>maxdist*maxdist || dist<mindist*mindist) {	// Si l'unité est trop loin de sa cible / if unit isn't where it should be
								if(!unit_manager.unit_type[type_id].canmove) {		// Bah là si on peut pas bouger faut changer de cible!! / need to change target
									next_mission();
									break;
									}
								else if(!unit_manager.unit_type[type_id].canfly) {
									VECTOR target;
									target=mission->target;
									c_time=0.0f;
									add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher / tell it to go closer
									}
								}
							else {
								int param[] = { (int)(unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime*1000.0f) };
								launch_script(get_script_index(SCRIPT_SetMaxReloadTime),1,param);
								switch(mission->data)
								{
								case 0:
									launch_script(get_script_index(SCRIPT_QueryPrimary));		break;
								case 1:
									launch_script(get_script_index(SCRIPT_QuerySecondary));		break;
								case 2:
									launch_script(get_script_index(SCRIPT_QueryTertiary));		break;
								};
								mission->mission=MISSION_AIM;							// sinon on lui demande de viser / or make it aim
								}
							}
						else {
							launch_script(get_script_index(SCRIPT_TargetCleared));
							next_mission();
							}
					}
					break;
				case MISSION_AIM:											// Vise une unité / aiming code
					if(jump_commands)	break;
					if(((mission->flags&4)==4 && ((WEAPON*)(mission->p))->weapon_id!=-1) || ((mission->flags&4)!=4 && ((UNIT*)(mission->p))->flags)) {
						int query_id=-1;
						switch(mission->data)
						{
						case 0:
							query_id=get_script_index(SCRIPT_QueryPrimary);		break;
						case 1:
							query_id=get_script_index(SCRIPT_QuerySecondary);	break;
						case 2:
							query_id=get_script_index(SCRIPT_QueryTertiary);	break;
						};
						if(!is_running(query_id) && !is_running(get_script_index(SCRIPT_RequestState))) {
							if(b_delay[mission->data]>=unit_manager.unit_type[type_id].weapon[mission->data]->reloadtime || unit_manager.unit_type[type_id].weapon[mission->data]->stockpile) {
								if(unit_manager.unit_type[type_id].weapon[mission->data]->turret) {			// Si l'unité doit viser, on la fait viser / if it must aim, we make it aim
									UNIT *target_unit=(mission->flags&4)==4 ? NULL : (UNIT*) mission->p;
									WEAPON *target_weapon=(mission->flags&4)==4 ? (WEAPON*) mission->p : NULL;
									int start_piece=script_val[query_id];
									if(start_piece<0 || start_piece>=data.nb_piece)
										start_piece=0;
									VECTOR target=target_unit==NULL ? target_weapon->Pos-(Pos+data.pos[start_piece]) : target_unit->Pos+target_unit->model->center-(Pos+data.pos[start_piece]);
									float dist=target.Norm();
									target=(1.0f/dist)*target;
									VECTOR I,J,IJ,RT;
									I.x=0.0f;	I.z=1.0f;	I.y=0.0f;
									J.x=1.0f;	J.z=0.0f;	J.y=0.0f;
									IJ.x=0.0f;	IJ.z=0.0f;	IJ.y=1.0f;
									RT=target;
									RT.y=0.0f;
									RT.Unit();
									float angle=acos(I%RT)*RAD2DEG;
									if(J%RT<0.0f) angle=-angle;
									angle-=Angle.y;
									if(angle<-180.0f)	angle+=360.0f;
									else if(angle>180.0f)	angle-=360.0f;
									int aiming[]={ (int)(angle*DEG2TA), -4096 };
									if(unit_manager.unit_type[type_id].weapon[mission->data]->ballistic) {		// Calculs de ballistique / ballistic calculations
										VECTOR D=target_unit==NULL ? (Pos+data.pos[start_piece]-target_weapon->Pos) : (Pos+data.pos[start_piece]-target_unit->Pos-target_unit->model->center);
										D.y=0.0f;
										float v;
										if(unit_manager.unit_type[type_id].weapon[mission->data]->startvelocity==0.0f)
											v=unit_manager.unit_type[type_id].weapon[mission->data]->weaponvelocity;
										else
											v=unit_manager.unit_type[type_id].weapon[mission->data]->startvelocity;
										if(target_unit==NULL)
											aiming[1]=(int)(ballistic_angle(v,map->ota_data.gravity,D.Norm(),(Pos+data.pos[start_piece]).y,target_weapon->Pos.y)*DEG2TA);
										else
											aiming[1]=(int)(ballistic_angle(v,map->ota_data.gravity,D.Norm(),(Pos+data.pos[start_piece]).y,target_unit->Pos.y+target_unit->model->center.y*0.5f)*DEG2TA);
										}
									else {
										VECTOR K=target;
										K.y=0.0f;
										K.Unit();
										angle=acos(K%target)*RAD2DEG;
										if(target.y>0.0f)
											angle=-angle;
										if(angle>180.0f)	angle-=360.0f;
										if(angle<-180.0f)	angle+=360.0f;
										if(fabs(angle)>180.0f) {
											next_mission();
											break;
											}
										aiming[1]=(int)(angle*DEG2TA);
										}
									if(unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight) {
										if(target_unit==NULL)
											aim_dir=((WEAPON*)(mission->p))->Pos-(Pos+data.pos[start_piece]);
										else
											aim_dir=((UNIT*)(mission->p))->Pos+((UNIT*)(mission->p))->model->center-(Pos+data.pos[start_piece]);
										aim_dir.Unit();
										}
									else
										aim_dir=cos(aiming[1]*TA2RAD)*(cos(aiming[0]*TA2RAD+Angle.y*DEG2RAD)*I+sin(aiming[0]*TA2RAD+Angle.y*DEG2RAD)*J)+sin(aiming[1]*TA2RAD)*IJ;
									switch(mission->data)
									{
									case 0:
										launch_script(get_script_index(SCRIPT_AimPrimary),2,aiming);	break;
									case 1:
										launch_script(get_script_index(SCRIPT_AimSecondary),2,aiming);	break;
									case 2:
										launch_script(get_script_index(SCRIPT_AimTertiary),2,aiming);	break;
									};
									}
								else
									switch(mission->data)
									{
									case 0:
										launch_script(get_script_index(SCRIPT_AimPrimary));	break;
									case 1:
										launch_script(get_script_index(SCRIPT_AimSecondary));	break;
									case 2:
										launch_script(get_script_index(SCRIPT_AimTertiary));	break;
									};
								mission->time=0.0f;
								mission->mission=MISSION_SHOOT;									// (puis) on lui demande de tirer / tell it to fire
								}
							}
						}
					else {
						launch_script(get_script_index(SCRIPT_TargetCleared));
						next_mission();
						}
					break;
				case MISSION_SHOOT:											// Tire sur une unité / fire!
					if(((mission->flags&4)==4 && ((WEAPON*)(mission->p))->weapon_id!=-1) || ((mission->flags&4)!=4 && ((UNIT*)(mission->p))->flags)) {
						if(burst[mission->data]>0 && b_delay[mission->data]<unit_manager.unit_type[type_id].weapon[mission->data]->burstrate)		break;
						int query_id=-1;
						int Aim_script = -1;
						int Fire_script = -1;
						switch(mission->data)
						{
						case 0:
							query_id=get_script_index(SCRIPT_QueryPrimary);
							Aim_script=SCRIPT_AimPrimary;
							Fire_script=SCRIPT_FirePrimary;
							break;
						case 1:
							query_id=get_script_index(SCRIPT_QuerySecondary);
							Aim_script=SCRIPT_AimSecondary;
							Fire_script=SCRIPT_FireSecondary;
							break;
						case 2:
							query_id=get_script_index(SCRIPT_QueryTertiary);
							Aim_script=SCRIPT_AimTertiary;
							Fire_script=SCRIPT_FireTertiary;
							break;
						};
						if(!is_running(get_script_index(Aim_script))) {
							if((players.metal[owner_id]<unit_manager.unit_type[type_id].weapon[mission->data]->metalpershot
							|| players.energy[owner_id]<unit_manager.unit_type[type_id].weapon[mission->data]->energypershot)
							&& !unit_manager.unit_type[type_id].weapon[mission->data]->stockpile) {
								mission->mission=MISSION_ATTACK;		// Pas assez d'énergie pour tirer / not enough energy to fire
								break;
								}
							if(unit_manager.unit_type[type_id].weapon[mission->data]->stockpile && stock[mission->data]<=0) {
								mission->mission=MISSION_ATTACK;		// Plus rien pour tirer / nothing to fire
								break;
								}
							else
								stock[mission->data]--;
							players.metal[owner_id]-=unit_manager.unit_type[type_id].weapon[mission->data]->metalpershot;
							players.energy[owner_id]-=unit_manager.unit_type[type_id].weapon[mission->data]->energypershot;
							launch_script(get_script_index(Fire_script));
							if(unit_manager.unit_type[type_id].weapon[mission->data]->soundstart)	sound_manager->PlaySound( unit_manager.unit_type[type_id].weapon[mission->data]->soundstart , &Pos );
							int start_piece=script_val[query_id];
							if(start_piece>=0 && start_piece<data.nb_piece) {
								compute_model_coord();
								VECTOR Dir=data.dir[start_piece];
								if(Dir.x==0.0f && Dir.y==0.0f && Dir.z==0.0f) {
									if(unit_manager.unit_type[type_id].weapon[mission->data]->vlaunch) {
										Dir.x=0.0f;
										Dir.y=1.0f;
										Dir.z=0.0f;
										}
									else {
										Dir=aim_dir;
										if(unit_manager.unit_type[type_id].weapon[mission->data]->lineofsight && !unit_manager.unit_type[type_id].weapon[mission->data]->turret) {
											if(mission->flags&4)
												Dir=((WEAPON*)(mission->p))->Pos-(Pos+data.pos[start_piece]);
											else
												Dir=((UNIT*)(mission->p))->Pos+((UNIT*)(mission->p))->model->center-(Pos+data.pos[start_piece]);
											Dir.Unit();
											}
										}
									}
								if(mission->flags&4)
									shoot(((WEAPON*)(mission->p))->idx,Pos+data.pos[start_piece],Dir,mission->data);
								else
									shoot(((UNIT*)(mission->p))->idx,Pos+data.pos[start_piece],Dir,mission->data);
								burst[mission->data]++;
								if(burst[mission->data]>=unit_manager.unit_type[type_id].weapon[mission->data]->burst)
									burst[mission->data]=0;
								b_delay[mission->data]=0.0f;
								}
							if((mission->flags&4)!=4 && ((UNIT*)(mission->p))->hp>0) {				// La cible est-elle détruite ?? / is target destroyed ??
								if(burst[mission->data]==0) {
									mission->mission=MISSION_ATTACK;
									mission->data++;
									mission->time=0.0f;
									}
								}
							else {
								launch_script(get_script_index(SCRIPT_TargetCleared));
								next_mission();
								}
							}
						else if(mission->time>0.25f && unit_manager.unit_type[type_id].weapon[mission->data]->turret) {
							mission->mission=MISSION_AIM;
//							kill_script(get_script_index(Aim_script));
							}
						}
					else {
						launch_script(get_script_index(SCRIPT_TargetCleared));
						next_mission();
						}
					break;
				case MISSION_STOP:											// Arrête tout ce qui était en cours / stop everything running
					if(jump_commands)	break;
					if(mission->data>5) {
						if(mission->next) {
							next_mission();
							if(mission!=NULL && mission->mission==MISSION_STOP)		// Mode attente / wait mode
								mission->data=6;
							}
						}
					else {
						if(mission->data==0) {
							launch_script(get_script_index(SCRIPT_StopMoving));		// Arrête tout / stop everything
							launch_script(get_script_index(SCRIPT_stopbuilding));
							}
						mission->data++;
						}
					break;
				case MISSION_REPAIR:
					{
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit!=NULL && target_unit->flags && target_unit->port[BUILD_PERCENT_LEFT]==0.0f) {
							if(target_unit->hp>=unit_manager.unit_type[target_unit->type_id].MaxDamage || unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {
								if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
									target_unit->hp=unit_manager.unit_type[target_unit->type_id].MaxDamage;
								next_mission();
								}
							else {
								VECTOR Dir=target_unit->Pos-Pos;
								Dir.y=0.0f;
								mission->target=target_unit->Pos;
								float dist=Dir.Sq();
								int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
								if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
									VECTOR target;
									target=mission->target;
									add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
									mission->data=0;
									c_time=0.0f;
									}
								else {
									if(mission->data==0) {
										mission->data=1;
										launch_script(get_script_index(SCRIPT_startbuilding));
										launch_script(get_script_index(SCRIPT_go));
										}
									if(dist!=0.0f)
										Dir=1.0f/sqrt(dist)*Dir;
									float dest_angle=acos(Dir.z)*RAD2DEG;
									if(Dir.x<0.0f) dest_angle=-dest_angle;
									while(fabs(dest_angle-Angle.y)>180.0f) {
										if(dest_angle<Angle.y)
											Angle.y-=360.0f;
										else
											Angle.y+=360.0f;
										}
									if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle) {
										float aspeed=unit_manager.unit_type[type_id].TurnRate;
										if(dest_angle<Angle.y)
											aspeed=-aspeed;
										float a=dest_angle-Angle.y;
										V_Angle.y=aspeed*0.1f;
										float b=dest_angle-(Angle.y+dt*V_Angle.y);
										if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
											V_Angle.y=0.0f;
											Angle.y=dest_angle;
											}
										}
									else {
										float conso_metal=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostMetal))/unit_manager.unit_type[target_unit->type_id].HealTime;
										float conso_energy=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostEnergy))/unit_manager.unit_type[target_unit->type_id].HealTime;
										if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
											metal_cons+=conso_metal;
											energy_cons+=conso_energy;
											target_unit->hp+=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].HealTime;
											}
										target_unit->built=true;
										}
									}
								}
							}
						else if(target_unit!=NULL && target_unit->flags) {
							VECTOR Dir=target_unit->Pos-Pos;
							Dir.y=0.0f;
							mission->target=target_unit->Pos;
							float dist=Dir.Sq();
							int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
							if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
								VECTOR target;
								dist=sqrt(dist);
								target=mission->target;
								c_time=0.0f;
								add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
								}
							else {
								if(dist!=0.0f)
									Dir=1.0f/sqrt(dist)*Dir;
								float dest_angle=acos(Dir.z)*RAD2DEG;
								if(Dir.x<0.0f) dest_angle=-dest_angle;
								while(fabs(dest_angle-Angle.y)>180.0f) {
									if(dest_angle<Angle.y)
										Angle.y-=360.0f;
									else
										Angle.y+=360.0f;
									}
								if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
									float aspeed=unit_manager.unit_type[type_id].TurnRate;
									if(dest_angle<Angle.y)
										aspeed=-aspeed;
									float a=dest_angle-Angle.y;
									V_Angle.y=aspeed*0.1f;
									float b=dest_angle-(Angle.y+dt*V_Angle.y);
									if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
										V_Angle.y=0.0f;
										Angle.y=dest_angle;
										}
									}
								else
									mission->mission=MISSION_BUILD_2;		// Change de type de mission
								}
							}
					}
					break;
				case MISSION_BUILD_2:
					{
						UNIT *target_unit=(UNIT*) mission->p;
						if(target_unit->flags) {
							if(target_unit->port[BUILD_PERCENT_LEFT]<=0.0f) {
								target_unit->port[BUILD_PERCENT_LEFT]=0.0f;
								target_unit->launch_script(target_unit->get_script_index(SCRIPT_create));		// Script post création
								if(unit_manager.unit_type[target_unit->type_id].ActivateWhenBuilt)		// Activation automatique
									target_unit->activate();
								if(unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {		// Ordre de se déplacer
									VECTOR target=Pos;
									target.z+=128.0f;
									target_unit->set_mission(MISSION_MOVE_CALC_PATH,&target,false,5,true,NULL,NULL);		// Fait sortir l'unité du bâtiment
									}
								mission->p=NULL;
								next_mission();
								deactivate();
								}
							else {
								float conso_metal=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostMetal))/unit_manager.unit_type[target_unit->type_id].BuildTime;
								float conso_energy=((float)(unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].BuildCostEnergy))/unit_manager.unit_type[target_unit->type_id].BuildTime;
								if(players.metal[owner_id]>=conso_metal*dt && players.energy[owner_id]>=conso_energy*dt) {
									metal_cons+=conso_metal;
									energy_cons+=conso_energy;
									target_unit->port[BUILD_PERCENT_LEFT]-=dt*unit_manager.unit_type[type_id].WorkerTime*100.0f/unit_manager.unit_type[target_unit->type_id].BuildTime;
									target_unit->hp+=dt*unit_manager.unit_type[type_id].WorkerTime*unit_manager.unit_type[target_unit->type_id].MaxDamage/unit_manager.unit_type[target_unit->type_id].BuildTime;
									}
								if(unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT) {
									int script_id_buildinfo = get_script_index(SCRIPT_QueryBuildInfo);
									if(script_id_buildinfo>=0) {
										compute_model_coord();
										map->rect(target_unit->cur_px-(unit_manager.unit_type[target_unit->type_id].FootprintX>>1),target_unit->cur_py-(unit_manager.unit_type[target_unit->type_id].FootprintZ>>1),unit_manager.unit_type[target_unit->type_id].FootprintX,unit_manager.unit_type[target_unit->type_id].FootprintZ,-1,unit_manager.unit_type[target_unit->type_id].yardmap,target_unit->port[YARD_OPEN]!=0.0f);
										VECTOR old_pos = target_unit->Pos;
										target_unit->Pos=Pos+data.pos[script_val[script_id_buildinfo]];
										if(((VECTOR)(old_pos-target_unit->Pos)).Sq()>100.0f)			// It must be continuous
											target_unit->Pos = old_pos;
										else {
											target_unit->cur_px = ((int)(target_unit->Pos.x)+map->map_w_d)>>3;
											target_unit->cur_py = ((int)(target_unit->Pos.z)+map->map_h_d)>>3;
											}
										target_unit->Angle=Angle;
										target_unit->Angle.y+=data.axe[1][script_val[script_id_buildinfo]].angle;
										}
									}
								target_unit->built=true;
								}
							}
						else
							next_mission();
						}
					break;
				case MISSION_BUILD:
					if(mission->p) {
						launch_script(get_script_index(SCRIPT_startbuilding));
						mission->mission=MISSION_BUILD_2;		// Change les objectifs
						((UNIT*)(mission->p))->built=true;
						}
					else {
						VECTOR Dir=mission->target-Pos;
						Dir.y=0.0f;
						float dist=Dir.Sq();
						int maxdist=(int)(unit_manager.unit_type[type_id].BuildDistance);
						if(dist>maxdist*maxdist && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {	// Si l'unité est trop loin du chantier
							VECTOR target;
							target=mission->target;
							c_time=0.0f;
							add_mission(MISSION_MOVE_CALC_PATH,&target,true,maxdist*9/80,NULL,NULL);		// On lui demande de se rapprocher
							}
						else {
							if(dist!=0.0f)
								Dir=(1.0f/sqrt(dist))*Dir;
							float dest_angle=acos(Dir.z)*RAD2DEG;
							if(Dir.x<0.0f) dest_angle=-dest_angle;
							while(fabs(dest_angle-Angle.y)>180.0f) {
								if(dest_angle<Angle.y)
									Angle.y-=360.0f;
								else
									Angle.y+=360.0f;
								}
							if(dist>0.0f && fabs(dest_angle-Angle.y)>=unit_manager.unit_type[type_id].BuildAngle && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT) {
								float aspeed=unit_manager.unit_type[type_id].TurnRate;
								if(dest_angle<Angle.y)
									aspeed=-aspeed;
								float a=dest_angle-Angle.y;
								V_Angle.y=aspeed*0.1f;
								float b=dest_angle-(Angle.y+dt*V_Angle.y);
								if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
									V_Angle.y=0.0f;
									Angle.y=dest_angle;
									}
								}
							else {
								if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT || (unit_manager.unit_type[type_id].TEDclass==CLASS_PLANT && port[YARD_OPEN])) {
									if(map->check_rect((((int)(mission->target.x)+map->map_w_d)>>3)-(unit_manager.unit_type[mission->data].FootprintX>>1),(((int)(mission->target.z)+map->map_h_d)>>3)-(unit_manager.unit_type[mission->data].FootprintZ>>1),unit_manager.unit_type[mission->data].FootprintX,unit_manager.unit_type[mission->data].FootprintZ,-1)) {				// Vérifie s'il y a la place de construire l'unité
										mission->p=create_unit(mission->data,owner_id,mission->target,map);
										if(mission->p) {
											((UNIT*)(mission->p))->hp=0.000001f;
											((UNIT*)(mission->p))->built=true;
											}
										}
									else if(unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
										next_mission();
									}
								else {
									launch_script(get_script_index(SCRIPT_go));
									launch_script(get_script_index(SCRIPT_QueryBuildInfo));
									}
								}
							}
						}
					break;
				};

			switch(unit_manager.unit_type[type_id].TEDclass)			// Commandes particulières
			{
			case CLASS_SHIP:
				switch(mission->mission)
				{
				case MISSION_STANDBY:
					if(mission->next)
						next_mission();
					V=units.exp_dt_1*V;			// Frottements
					break;
				case MISSION_MOVE:
					{
						VECTOR J,I;
						VECTOR Target=mission->target;
						if(mission->path)
							Target=mission->path->Pos;
						J=Target-Pos;
						J.y=0.0f;
						float dist=J.Norm();
						if((dist>mission->last_d && dist<25.0f) || mission->path==NULL) {		// Là on est arrivé
							bool vide=(mission->path==NULL);
							if(mission->path)
								mission->path=next_node(mission->path);
							mission->last_d=9999999.0f;
							if(mission->path==NULL) {
								if(owner_id==players.local_human_id && !vide)
									sound_manager->PlayTDFSound( unit_manager.unit_type[type_id].soundcategory, "arrived1" , &Pos );
								launch_script(get_script_index(SCRIPT_StopMoving));
								next_mission();
								break;
								}
							}
						else
							mission->last_d=dist;
						if(dist>0.0f)
							J=1.0f/dist*J;
						float dest_angle=acos(J.z)*RAD2DEG;
						if(J.x<0.0f) dest_angle=-dest_angle;
						while(fabs(dest_angle-Angle.y)>180.0f) {
							if(dest_angle<Angle.y)
								Angle.y-=360.0f;
							else
								Angle.y+=360.0f;
							}
						if(dist>0.0f && fabs(dest_angle-Angle.y)>=1.0f) {
							float aspeed=unit_manager.unit_type[type_id].TurnRate;
							if(dest_angle<Angle.y)
								aspeed=-aspeed;
							float a=dest_angle-Angle.y;
							V_Angle.y=aspeed*0.1f;
							float b=dest_angle-(Angle.y+dt*V_Angle.y);
							if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
								V_Angle.y=0.0f;
								Angle.y=dest_angle;
								}
							}

						J.z=cos(Angle.y*DEG2RAD);
						J.x=sin(Angle.y*DEG2RAD);
						J.y=0.0f;
						I.z=-J.x;
						I.x=J.z;
						I.y=0.0f;
						V=(V%J)*J+units.exp_dt_1*(V%I)*I;
						if(!(dist<128.0f && fabs(Angle.y-dest_angle)>=1)) {
							float speed=V.Norm();
							float time_to_stop=speed/unit_manager.unit_type[type_id].BrakeRate;
							float min_dist=time_to_stop*(speed-unit_manager.unit_type[type_id].BrakeRate*0.5f*time_to_stop);
							if(min_dist>=dist && ( mission->next == NULL || ( mission->next->mission != MISSION_MOVE_CALC_PATH && mission->next->mission != MISSION_MOVE && mission->next->mission != MISSION_PATROL ) ) )			// Freine si nécessaire
								V=V-(unit_manager.unit_type[type_id].BrakeRate*dt)*J;
							else if(speed<unit_manager.unit_type[type_id].MaxVelocity)
								V=V+(unit_manager.unit_type[type_id].Acceleration*dt)*J;
							}
						else
							V=units.exp_dt_1*V;			// Frottements
					}
					break;
				default:
					V=units.exp_dt_1*V;			// Frottements
				};
				break;
			case CLASS_PLANT:
				switch(mission->mission)
				{
				case MISSION_STANDBY:
				case MISSION_BUILD:
				case MISSION_BUILD_2:
				case MISSION_REPAIR:
					break;
				default:
					next_mission();
				};
				break;
			case CLASS_WATER:
			case CLASS_VTOL:
			case CLASS_KBOT:
			case CLASS_COMMANDER:
			case CLASS_TANK:
			case CLASS_CNSTR:
				{
				VECTOR K,PV;
				K.x=0.0f;
				K.y=1.0f;
				K.z=0.0f;
				if(mission->mission!=MISSION_MOVE && ((mission->mission!=MISSION_ATTACK && mission->mission!=MISSION_AIM && mission->mission!=MISSION_SHOOT && unit_manager.unit_type[type_id].canfly) || !unit_manager.unit_type[type_id].canfly)) {
					PV=(V%K)*K;
					V=units.exp_dt_2*(V-PV)+PV;		// N'altère pas la déscente des unités
					}
				switch(mission->mission)
				{
				case MISSION_ATTACK:
				case MISSION_AIM:
				case MISSION_SHOOT:
				case MISSION_PATROL:
				case MISSION_REPAIR:
				case MISSION_BUILD:
				case MISSION_BUILD_2:
					if(unit_manager.unit_type[type_id].canfly)
						activate();
					break;
				case MISSION_MOVE:
					{
						if(unit_manager.unit_type[type_id].canfly)
							activate();
						VECTOR J,I;
						VECTOR Target=mission->target;
						if(mission->path)
							Target=mission->path->Pos;
						J=Target-Pos;
						J.y=0.0f;
						float dist=J.Norm();
						if((dist>mission->last_d && dist<15.0f) || mission->path==NULL) {
							if(mission->path)
								mission->path=next_node(mission->path);
							mission->last_d=9999999.0f;
							if(mission->path==NULL) {		// Si l'on est arrivé au bout du chemin
								sound_manager->PlayTDFSound( unit_manager.unit_type[type_id].soundcategory, "arrived1" , &Pos );
//								sound_manager->PlaySound( &unit_manager.unit_type[type_id].arrived1 , &Pos );
								launch_script(get_script_index(SCRIPT_StopMoving));
								next_mission();
								break;
								}
							}
						else
							mission->last_d=dist;
						if(dist>0.0f)
							J=1.0f/dist*J;
						float dest_angle=acos(J.z)*RAD2DEG;
						if(J.x<0.0f) dest_angle=-dest_angle;
						while(fabs(dest_angle-Angle.y)>180.0f) {
							if(dest_angle<Angle.y)
								Angle.y-=360.0f;
							else
								Angle.y+=360.0f;
							}
						if(dist>0.0f && fabs(dest_angle-Angle.y)>=1.0f) {
							float aspeed=unit_manager.unit_type[type_id].TurnRate;
							if(dest_angle<Angle.y)
								aspeed=-aspeed;
							float a=dest_angle-Angle.y;
							V_Angle.y=aspeed*0.1f;
							float b=dest_angle-(Angle.y+dt*V_Angle.y);
							if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
								V_Angle.y=0.0f;
								Angle.y=dest_angle;
								}
							}

						J.z=cos(Angle.y*DEG2RAD);
						J.x=sin(Angle.y*DEG2RAD);
						J.y=0.0f;
						I.z=-J.x;
						I.x=J.z;
						I.y=0.0f;
						V=(V%K)*K+(V%J)*J+units.exp_dt_4*(V%I)*I;
						if(!(dist<128.0f && fabs(Angle.y-dest_angle)>=1)) {
							float speed=V.Norm();
							float time_to_stop=speed/unit_manager.unit_type[type_id].BrakeRate;
							float min_dist=time_to_stop*(speed-unit_manager.unit_type[type_id].BrakeRate*0.5f*time_to_stop);
							if(min_dist>=dist && ( mission->next == NULL || ( mission->next->mission != MISSION_MOVE_CALC_PATH && mission->next->mission != MISSION_MOVE && mission->next->mission != MISSION_PATROL ) ) )			// Freine si nécessaire
								V=V-unit_manager.unit_type[type_id].BrakeRate*dt*J;
							else if(speed<unit_manager.unit_type[type_id].MaxVelocity)
								V=V+unit_manager.unit_type[type_id].Acceleration*dt*J;
							}
					}
					break;
				default:
					if(unit_manager.unit_type[type_id].canfly)
						deactivate();
					};
				}
				break;
			case CLASS_UNDEF:
			case CLASS_METAL:
			case CLASS_ENERGY:
			case CLASS_SPECIAL:
			case CLASS_FORT:
				break;
			default:
				printf("type inconnu %d\n",unit_manager.unit_type[type_id].TEDclass);
			};
			switch(mission->mission)		// Pour le code post déplacement
			{
			case MISSION_BUILD:
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
				break;
			case MISSION_GUARD:
				break;
			case MISSION_STOP:
				break;
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
				if(unit_manager.unit_type[type_id].canfly) {			// Un avion??
					activate();
					VECTOR J,I,K;
					K.x=K.z=0.0f;
					K.y=1.0f;
					VECTOR Target=mission->target;
					J=Target-Pos;
					J.y=0.0f;
					float dist=J.Norm();
					mission->last_d=dist;
					if(dist>0.0f)
						J=1.0f/dist*J;
					float dest_angle=acos(J.z)*RAD2DEG;
					if(J.x<0.0f) dest_angle=-dest_angle;
					while(fabs(dest_angle-Angle.y)>180.0f) {
						if(dest_angle<Angle.y)
							Angle.y-=360.0f;
						else
							Angle.y+=360.0f;
						}
//					if(dist>2.0f*unit_manager.unit_type[type_id].SightDistance && fabs(dest_angle-Angle.y)>=1.0f) {
					if(dist>(unit_manager.unit_type[type_id].ManeuverLeashLength>>1) && fabs(dest_angle-Angle.y)>=1.0f) {
						float aspeed=unit_manager.unit_type[type_id].TurnRate;
						if(dest_angle<Angle.y)
							aspeed=-aspeed;
						float a=dest_angle-Angle.y;
						V_Angle.y=aspeed*0.1f;
						float b=dest_angle-(Angle.y+dt*V_Angle.y);
						if((a<0.0f && b>0.0f) || (a>0.0f && b<0.0f)) {
							V_Angle.y=0.0f;
							Angle.y=dest_angle;
							}
						}

					J.z=cos(Angle.y*DEG2RAD);
					J.x=sin(Angle.y*DEG2RAD);
					J.y=0.0f;
					I.z=-J.x;
					I.x=J.z;
					I.y=0.0f;
					V=(V%K)*K+(V%J)*J+units.exp_dt_4*(V%I)*I;
					float speed=V.Norm();
					if(speed<unit_manager.unit_type[type_id].MaxVelocity)
						V=V+unit_manager.unit_type[type_id].Acceleration*dt*J;
					}
				break;
			case MISSION_MOVE:			// Oriente l'unité si elle est au sol
				if(jump_commands)	break;
				if(!unit_manager.unit_type[type_id].canfly && !unit_manager.unit_type[type_id].Upright && unit_manager.unit_type[type_id].TEDclass!=CLASS_SHIP && unit_manager.unit_type[type_id].TEDclass!=CLASS_WATER) {
					VECTOR I,J,K,A,B,C;
					MATRIX_4x4 M=RotateY((Angle.y+90.0f)*DEG2RAD);
					I.x=I.y=I.z=0.0f;
					J=K=I;
					I.x=1.0f;
					J.z=1.0f;
					K.y=1.0f;
					A=Pos-unit_manager.unit_type[type_id].FootprintZ*4.0f*I*M;
					B=Pos+(unit_manager.unit_type[type_id].FootprintX*4.0f*I-unit_manager.unit_type[type_id].FootprintZ*4.0f*J)*M;
					C=Pos+(unit_manager.unit_type[type_id].FootprintX*4.0f*I+unit_manager.unit_type[type_id].FootprintZ*4.0f*J)*M;
					A.y=map->get_unit_h(A.x,A.z);	// Projete le triangle
					B.y=map->get_unit_h(B.x,B.z);
					C.y=map->get_unit_h(C.x,C.z);
					VECTOR D=(B-A)*(B-C);
					if(D.y>=0.0f) {					// On ne met pas une unité à l'envers!!
						D.Unit();
						float angle_1=(D.z!=0.0f || D.y!=0.0f) ? acos(D.y/sqrt(D.y*D.y+D.z*D.z))*RAD2DEG : 0.0f;
						if(D.z<0.0f)	angle_1=-angle_1;
						D=D*RotateX(-angle_1*DEG2RAD);
						float angle_2=VAngle(D,K)*RAD2DEG;
						if(D.x>0.0f)	angle_2=-angle_2;
						if(fabs(angle_1-Angle.x)<=10.0f && fabs(angle_2-Angle.z)<=10.0f) {
							Angle.x=angle_1;
							Angle.z=angle_2;
							}
						}
					}
				if((mission->flags&1)==0)		// Si l'on n'est pas en mode patrouille
					break;
			default:								// Si l'unité peut attaquer d'elle même les unités enemies proches, elle le fait / Attack nearby enemies
				if(jump_commands)	break;
				if(unit_manager.unit_type[type_id].AutoFire && unit_manager.unit_type[type_id].canattack
				&& (unit_manager.unit_type[type_id].weapon[0]==NULL || (unit_manager.unit_type[type_id].weapon[0]!=NULL && !unit_manager.unit_type[type_id].weapon[0]->commandfire))
				&& (unit_manager.unit_type[type_id].weapon[1]==NULL || (unit_manager.unit_type[type_id].weapon[1]!=NULL && !unit_manager.unit_type[type_id].weapon[1]->commandfire))
				&& (unit_manager.unit_type[type_id].weapon[2]==NULL || (unit_manager.unit_type[type_id].weapon[2]!=NULL && !unit_manager.unit_type[type_id].weapon[2]->commandfire))) {
					int dx=unit_manager.unit_type[type_id].SightDistance+(int)(h+0.5f)>>3;
					int enemy_idx=-1;
					if(unit_manager.unit_type[type_id].weapon[0]!=NULL && (unit_manager.unit_type[type_id].weapon[0]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[0]->range>>4;
					if(unit_manager.unit_type[type_id].weapon[1]!=NULL && (unit_manager.unit_type[type_id].weapon[1]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[1]->range>>4;
					if(unit_manager.unit_type[type_id].weapon[2]!=NULL && (unit_manager.unit_type[type_id].weapon[2]->range>>4)>dx)
						dx=unit_manager.unit_type[type_id].weapon[2]->range>>4;
					int sx=rand_from_table()&0xF;
					int sy=rand_from_table()&0xF;
					int mask=1<<owner_id;
					for(int y=cur_py-dx+sy;y<=cur_py+dx;y+=0x10) {
						if(y>=0 && y<map->bloc_h_db-1)
							for(int x=cur_px-dx+sx;x<=cur_px+dx;x+=0x10)
								if(x>=0 && x<map->bloc_w_db-1 && (map->map_data[y][x].view&mask)==mask && map->map_data[y][x].unit_idx>-1 && map->map_data[y][x].unit_idx!=idx && units.unit[map->map_data[y][x].unit_idx].owner_id!=owner_id) {
									int cur_idx = map->map_data[y][x].unit_idx;
									if(cur_idx>=0 && cur_idx<units.max_unit && units.unit[cur_idx].flags && unit_manager.unit_type[units.unit[cur_idx].type_id].ShootMe) {
										enemy_idx = map->map_data[y][x].unit_idx;
										break;
										}
									}
						if(enemy_idx>=0)	break;
						}
					if(enemy_idx>=0)			// Si on a trouvé une unité, on l'attaque
						set_mission(MISSION_ATTACK,&(units.unit[enemy_idx].Pos),false,0,true,&(units.unit[enemy_idx]),NULL,8);	// 8 = automatic fire
					}
				if(unit_manager.unit_type[type_id].antiweapons && unit_manager.unit_type[type_id].weapon[0]) {
					float coverage=unit_manager.unit_type[type_id].weapon[0]->coverage*unit_manager.unit_type[type_id].weapon[0]->coverage;
					float range=unit_manager.unit_type[type_id].weapon[0]->range*unit_manager.unit_type[type_id].weapon[0]->range>>2;
					int enemy_idx=-1;
					byte e=0;
					for(byte i=0;i+e<mem_size;i++) {
						if(memory[i+e]<0 || memory[i+e]>=weapons.nb_weapon || weapons.weapon[memory[i+e]].weapon_id==-1) {
							e++;
							i--;
							continue;
							}
						memory[i]=memory[i+e];
						}
					mem_size-=e;
					for(uint32 f=0;f<weapons.index_list_size;f+=(rand_from_table()%10)+1) {
						uint32 i = weapons.idx_list[f];
						if(weapons.weapon[i].weapon_id!=-1 && units.unit[weapons.weapon[i].shooter_idx].owner_id!=owner_id
						&& weapon_manager.weapon[weapons.weapon[i].weapon_id].targetable)
						if(((VECTOR)(weapons.weapon[i].target_pos-Pos)).Sq()<=coverage
						&& ((VECTOR)(weapons.weapon[i].Pos-Pos)).Sq()<=range) {
							int idx=-1;
							for(e=0;e<mem_size;e++)
								if(memory[e]==i) {
									idx=i;
									break;
									}
							if(idx==-1) {
								enemy_idx=i;
								if(mem_size<10) {
									memory[mem_size]=i;
									mem_size++;
									}
								break;
								}
							}
						}
					if(enemy_idx>=0)			// Si on a trouvé une unité, on l'attaque
						add_mission(MISSION_ATTACK,&(weapons.weapon[enemy_idx].Pos),false,0,&(weapons.weapon[enemy_idx]),NULL,12);	// 12 = 4 | 8, targets a weapon and automatic fire
					}
				break;
			};
			}

		if( unit_manager.unit_type[type_id].canfly) {			// Set plane orientation
			VECTOR J,K;
			K.x=K.z=0.0f;
			K.y=1.0f;
			J = V * K;

			VECTOR virtual_G;						// Compute the apparent gravity force ( seen from the plane )
			virtual_G.x = virtual_G.z = 0.0f;		// Standard gravity vector
			virtual_G.y = -units.g_dt;
			float d = J.Sq();
			if( d )
				virtual_G = virtual_G + (((old_V - V) % J) / d) * J;		// Add the opposite of the speed derivative projected on the side of the unit

			d = virtual_G.Norm();
			if( d ) {
				virtual_G = -1.0f / d * virtual_G;

				d = sqrt(virtual_G.y*virtual_G.y+virtual_G.z*virtual_G.z);
				float angle_1= (d != 0.0f) ? acos(virtual_G.y/d)*RAD2DEG : 0.0f;
				if(virtual_G.z<0.0f)	angle_1=-angle_1;
				virtual_G=virtual_G*RotateX(-angle_1*DEG2RAD);
				float angle_2=acos( virtual_G % K )*RAD2DEG;
				if(virtual_G.x>0.0f)	angle_2=-angle_2;

				if( fabs( angle_1 - Angle.x ) < 360.0f )
					Angle.x += dt*( angle_1 - Angle.x );				// We need something continuous
				if( fabs( angle_2 - Angle.z ) < 360.0f )
					Angle.z += dt*( angle_2 - Angle.z );
				}
			}

		if(port[BUILD_PERCENT_LEFT]==0.0f) {
			Angle=Angle+dt*V_Angle;
			VECTOR OPos=Pos;
			V.y-=units.g_dt;			// L'unité subit la force de gravitation
			Pos=Pos+dt*V;			// Déplace l'unité
			cur_px = ((int)(Pos.x)+map->map_w_d)>>3;
			cur_py = ((int)(Pos.z)+map->map_h_d)>>3;
			if(mission!=NULL && (mission->flags&2)!=0)	goto script_exec;
			if(unit_manager.unit_type[type_id].canmove && unit_manager.unit_type[type_id].TEDclass!=CLASS_PLANT)
			if(!map->check_rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx)) {
				if(!unit_manager.unit_type[type_id].canfly) {
					bool back=false;
					if(map->check_rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx)) {
						Pos=OPos;
						cur_px = o_px;
						cur_py = o_py;
						back=true;
						}
					if((mission==NULL || (mission->flags&2)==0) && (mission->mission!=MISSION_STOP || mission->next==NULL)) {
						if(mission && mission->mission==MISSION_MOVE && back) {		// Recherche un autre chemin
							mission->mission=MISSION_MOVE_CALC_PATH;
							destroy_path(mission->path);
							mission->path=NULL;
							}
						VECTOR target=Pos;
						target.x+=(rand_from_table()&0x1F)-16;		// cherche ailleurs de la place
						target.z+=(rand_from_table()&0x1F)-16;
						add_mission(MISSION_MOVE_CALC_PATH,&target,true,0,NULL,NULL,2);		// On lui demande de s'éloigner
						}
					}
				else if(mission->mission!=MISSION_MOVE && mission->mission!=MISSION_BUILD && mission->mission!=MISSION_BUILD_2 && mission->mission!=MISSION_REPAIR && mission->mission!=MISSION_ATTACK && mission->mission!=MISSION_AIM && mission->mission!=MISSION_SHOOT) {
					if(map->check_rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx))
						Pos=OPos;
					if((mission==NULL || (mission->flags&2)==0) && (mission->mission!=MISSION_STOP || mission->next==NULL)) {
						VECTOR target=Pos;
						target.x+=(rand_from_table()&0x1F)-16;		// cherche ailleurs de la place pour attérrir
						target.z+=(rand_from_table()&0x1F)-16;
						add_mission(MISSION_MOVE_CALC_PATH,&target,true,0,NULL,NULL,2);		// On lui demande de s'éloigner
						}
					}
				}
			}
		script_exec:
		if(map && !jump_commands && !attached) {
			float min_h=map->get_unit_h(Pos.x,Pos.z);
			h=Pos.y-min_h;
			if(!unit_manager.unit_type[type_id].norestrict && Pos.y<map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV) {
				Pos.y=map->sealvl-unit_manager.unit_type[type_id].MaxWaterDepth*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if((unit_manager.unit_type[type_id].Floater || (unit_manager.unit_type[type_id].TEDclass&CLASS_SHIP))
				 && Pos.y<map->sealvl+(unit_manager.unit_type[type_id].AltFromSeaLevel)*H_DIV) {
				Pos.y=map->sealvl+unit_manager.unit_type[type_id].AltFromSeaLevel*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			else if(unit_manager.unit_type[type_id].TEDclass&CLASS_WATER)	{
				Pos.y=map->sealvl-unit_manager.unit_type[type_id].WaterLine*H_DIV;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if(min_h>Pos.y) {
				Pos.y=min_h;
				if(V.y<0.0f)
					V.y=0.0f;
				}
			if(unit_manager.unit_type[type_id].canfly && port[BUILD_PERCENT_LEFT]==0.0f) {
				if(mission && (mission->mission==MISSION_MOVE || mission->mission==MISSION_BUILD || mission->mission==MISSION_BUILD_2 || mission->mission==MISSION_REPAIR
				|| mission->mission==MISSION_ATTACK || mission->mission==MISSION_AIM || mission->mission==MISSION_SHOOT || nb_attached>0)) {
					float ideal_h=max(min_h,map->sealvl)+unit_manager.unit_type[type_id].CruiseAlt*H_DIV;
					V.y=(ideal_h-Pos.y)*0.5f;
					}
				else {
					float ideal_h=max(min_h,map->sealvl);
					V.y=(ideal_h-Pos.y)*0.1f;
					}
				}
			port[GROUND_HEIGHT]=Pos.y-min_h;
			}
		port[HEALTH]=hp*100/unit_manager.unit_type[type_id].MaxDamage;
		if(nb_running>0) {
			for(int i=0;i<nb_running;i++)
				run_script(dt,i,map);
			int e=0;
			for(int i=0;i+e<nb_running;) {				// Efface les scripts qui se sont arrêtés
				if(script_env[i+e].running) {
					script_env[i]=script_env[i+e];
					i++;
					}
				else {
					script_env[i+e].destroy();
					e++;
					}
				}
			nb_running-=e;
			}
		if(unit_manager.unit_type[type_id].canmove || !jump_commands) {
			map->rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,was_open != 0.0f);
			if(!attached)
				map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			}
		else if( port[YARD_OPEN] != was_open ) {
			map->rect(o_px-(unit_manager.unit_type[type_id].FootprintX>>1),o_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,-1,unit_manager.unit_type[type_id].yardmap,was_open != 0.0f);
			if(!attached)
				map->rect(cur_px-(unit_manager.unit_type[type_id].FootprintX>>1),cur_py-(unit_manager.unit_type[type_id].FootprintZ>>1),unit_manager.unit_type[type_id].FootprintX,unit_manager.unit_type[type_id].FootprintZ,idx,unit_manager.unit_type[type_id].yardmap,port[YARD_OPEN]!=0.0f);
			}
		built=false;
		attacked=false;
#ifdef	ADVANCED_DEBUG_MODE
		GuardLeave();
#endif
		return 0;
	}

	bool UNIT::hit(VECTOR P,VECTOR Dir,VECTOR *hit_vec, float length)
	{
		if(flags==0)	return false;
		if(model) {
			VECTOR c_dir=model->center+Pos-P;
			if(((VECTOR)(c_dir*Dir)).Sq()<(model->size*0.5f) && c_dir.Norm()-length <=model->size2 ) {
				float scale=unit_manager.unit_type[type_id].Scale;
//				MATRIX_4x4 M=RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale);
				MATRIX_4x4 M=RotateX(-Angle.x*DEG2RAD)*RotateZ(-Angle.z*DEG2RAD)*RotateY(-Angle.y*DEG2RAD)*Scale(1.0f/scale);
				VECTOR RP=(1.0f/scale)*(P-Pos);
				bool is_hit=model->hit(RP,Dir,&data,hit_vec,M);
				if(is_hit) {
					*hit_vec=(*hit_vec)*(RotateY(Angle.y*DEG2RAD)*RotateZ(Angle.z*DEG2RAD)*RotateX(Angle.x*DEG2RAD)*Scale(scale))+Pos;
					*hit_vec=((*hit_vec-P)%Dir)*Dir+P;
					}
				return is_hit;
				}
			}
		return false;
	}

	void UNIT::show_orders(bool only_build_commands)				// Dessine les ordres reçus
	{
		MISSION *cur=mission;
		glEnable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		VECTOR p_target=Pos;
		float rab=(Atimer%1000000)*0.000001f;
		uint32	remaining_build_commands = (unit_manager.unit_type[ type_id ].TEDclass & CLASS_PLANT) ? 0 : 0xFFFFFFF;
		while(cur) {
			if(cur->step)				// S'il s'agit d'une étape on ne la montre pas
				cur=cur->next;
			if(!only_build_commands)
			{
			int curseur=anim_cursor(CURSOR_CROSS_LINK);
			float dx=0.5f*cursor.anm[CURSOR_CROSS_LINK].ofs_x[curseur];
			float dz=0.5f*cursor.anm[CURSOR_CROSS_LINK].ofs_y[curseur];
			float sx=0.5f*(cursor.anm[CURSOR_CROSS_LINK].bmp[curseur]->w-1);
			float sy=0.5f*(cursor.anm[CURSOR_CROSS_LINK].bmp[curseur]->h-1);
			float x,y,z;
			float dist=((VECTOR)(cur->target-p_target)).Norm();
			int rec=(int)(dist/30.0f);
			switch(cur->mission)
			{
			case MISSION_LOAD:
			case MISSION_UNLOAD:
			case MISSION_GUARD:
			case MISSION_PATROL:
			case MISSION_MOVE_CALC_PATH:
			case MISSION_MOVE:
			case MISSION_BUILD:
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
			case MISSION_RECLAIM:
				if(rec>0) {
					glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_CROSS_LINK].glbmp[curseur]);
					glBegin(GL_QUADS);
						for(int i=0;i<rec;i++) {
							x=p_target.x+(cur->target.x-p_target.x)*(i+rab)/rec;
							y=p_target.y+(cur->target.y-p_target.y)*(i+rab)/rec;
							z=p_target.z+(cur->target.z-p_target.z)*(i+rab)/rec;
							x-=dx;
							y+=0.75f;
							z-=dz;
							glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
							glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
							glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
							glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
							}
					glEnd();
					}
				p_target=cur->target;
			};
			}
			glDisable(GL_DEPTH_TEST);
			switch(cur->mission)
			{
			case MISSION_LOAD:
				if(!only_build_commands)
				{
				int curseur=anim_cursor(CURSOR_LOAD);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_LOAD].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_LOAD].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_LOAD].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_LOAD].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_LOAD].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_MOVE_CALC_PATH:
			case MISSION_MOVE:
				if(!only_build_commands)
				{
				int curseur=anim_cursor(CURSOR_MOVE);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_MOVE].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_MOVE].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_MOVE].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_MOVE].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_MOVE].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_BUILD:
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				if(cur->data>=0 && cur->data<unit_manager.nb_unit && remaining_build_commands > 0 ) {
					remaining_build_commands--;
					float DX = (unit_manager.unit_type[cur->data].FootprintX<<2);
					float DZ = (unit_manager.unit_type[cur->data].FootprintZ<<2);
					float blue = 0.0f, green = 1.0f;
					if(only_build_commands) {
						blue = 1.0f;
						green = 0.0f;
						}
					glPushMatrix();
					glTranslatef(cur->target.x,cur->target.y,cur->target.z);
					glDisable(GL_CULL_FACE);
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glBegin(GL_QUADS);
						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,-DZ);			// First quad
						glVertex3f(DX,0.0f,-DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);
						glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,-DZ);			// Second quad
						glVertex3f(-DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
						glVertex3f(-DX-2.0f,5.0f,-DZ-2.0f);

						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(DX,0.0f,-DZ);			// Third quad
						glVertex3f(DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
						glVertex3f(DX+2.0f,5.0f,-DZ-2.0f);
					glEnd();
					glDisable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glPopMatrix();
					if(unit_manager.unit_type[cur->data].model!=NULL) {
						glEnable(GL_LIGHTING);
						glEnable(GL_CULL_FACE);
						glPushMatrix();
						glTranslatef(cur->target.x,cur->target.y,cur->target.z);
						glColor4f(0.0f,green,blue,0.5f);
						unit_manager.unit_type[cur->data].model->obj.draw(NULL,false,false,false);
						glPopMatrix();
						glEnable(GL_BLEND);
						glEnable(GL_TEXTURE_2D);
						glDisable(GL_LIGHTING);
						glDisable(GL_CULL_FACE);
						glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
						}
					glPushMatrix();
					glTranslatef(cur->target.x,cur->target.y,cur->target.z);
					glDisable(GL_CULL_FACE);
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glBegin(GL_QUADS);
						glColor4f(0.0f,green,blue,1.0f);
						glVertex3f(-DX,0.0f,DZ);			// Fourth quad
						glVertex3f(DX,0.0f,DZ);
						glColor4f(0.0f,green,blue,0.0f);
						glVertex3f(DX+2.0f,5.0f,DZ+2.0f);
						glVertex3f(-DX-2.0f,5.0f,DZ+2.0f);
					glEnd();
					glPopMatrix();
					glDisable(GL_BLEND);
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_CULL_FACE);
					glColor4f(1.0f,1.0f,1.0f,1.0f);
					}
				break;
			case MISSION_BUILD_2:
			case MISSION_REPAIR:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_REPAIR);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_REPAIR].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_REPAIR].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_REPAIR].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_RECLAIM:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_RECLAIM);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_RECLAIM].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_RECLAIM].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_REPAIR].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_RECLAIM].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_PATROL:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_PATROL);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_PATROL].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_PATROL].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_PATROL].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_PATROL].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_PATROL].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_GUARD:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_GUARD);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_GUARD].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_GUARD].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_GUARD].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_GUARD].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_GUARD].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			case MISSION_ATTACK:
			case MISSION_AIM:
			case MISSION_SHOOT:
				if(!only_build_commands)
				{
				if(cur->p!=NULL)
					cur->target=((UNIT*)(cur->p))->Pos;
				int curseur=anim_cursor(CURSOR_ATTACK);
				float x=cur->target.x-0.5f*cursor.anm[CURSOR_ATTACK].ofs_x[curseur];
				float y=cur->target.y+1.0f;
				float z=cur->target.z-0.5f*cursor.anm[CURSOR_ATTACK].ofs_y[curseur];
				float sx=0.5f*(cursor.anm[CURSOR_ATTACK].bmp[curseur]->w-1);
				float sy=0.5f*(cursor.anm[CURSOR_ATTACK].bmp[curseur]->h-1);
				glBindTexture(GL_TEXTURE_2D,cursor.anm[CURSOR_ATTACK].glbmp[curseur]);
				glBegin(GL_QUADS);
					glTexCoord2f(0.0f,0.0f);		glVertex3f(x,y,z);
					glTexCoord2f(1.0f,0.0f);		glVertex3f(x+sx,y,z);
					glTexCoord2f(1.0f,1.0f);		glVertex3f(x+sx,y,z+sy);
					glTexCoord2f(0.0f,1.0f);		glVertex3f(x,y,z+sy);
				glEnd();
				}
				break;
			};
			glEnable(GL_DEPTH_TEST);
			cur=cur->next;
			}
		glDisable(GL_BLEND);
	}

	bool INGAME_UNITS::select(CAMERA *cam,int sel_x[],int sel_y[])
	{
		bool selected=false;
		cam->SetView();
		MATRIX_4x4 ModelView,Project,T;
		int	viewportCoords[4] = {0, 0, 0, 0};
		glGetIntegerv(GL_VIEWPORT, viewportCoords);
		glGetFloatv(GL_MODELVIEW_MATRIX, (float*)ModelView.E);
		glGetFloatv(GL_PROJECTION_MATRIX, (float*)Project.E);
		ModelView=Transpose(ModelView);
		Project=Transpose(Project);
		float VW=(viewportCoords[2]-viewportCoords[0])*0.5f;
		float VH=-(viewportCoords[3]-viewportCoords[1])*0.5f;
		T=ModelView*Project;		// Matrice de transformation

		POINTF UPos,O;
		O.x=O.y=O.z=0.0f;
		int X1,Y1,X2,Y2;
		X1=min(sel_x[0],sel_x[1]);
		Y1=min(sel_y[0],sel_y[1]);
		X2=max(sel_x[0],sel_x[1]);
		Y2=max(sel_y[0],sel_y[1]);

		for(uint16 e=0;e<index_list_size;e++) {
			uint16 i = idx_list[e];
			if(unit[i].flags!=0 && unit[i].owner_id==players.local_human_id && unit[i].port[BUILD_PERCENT_LEFT]==0.0f && unit[i].visible) {		// Ne sélectionne que les unités achevées
				if(key[KEY_LSHIFT] && unit[i].sel) {
					selected=true;
					continue;
					}
				if(!key[KEY_LSHIFT])
					unit[i].sel=false;

				VECTOR Vec=unit[i].Pos-cam->Pos;
				float d=Vec.Sq();
				if(d>16384.0f && (Vec%cam->Dir)<=0.0f) continue;

				UPos=glNMult((O+unit[i].Pos),T);		// Transforme la position de l'unité
				UPos.x=UPos.x*VW+VW;
				UPos.y=UPos.y*VH-VH;

				if(X1<=UPos.x && X2>=UPos.x && Y1<=UPos.y && Y2>=UPos.y)
					selected=unit[i].sel=true;
				}
			}
		return selected;
	}

	int INGAME_UNITS::pick(CAMERA *cam,int sensibility)
	{
		int index=-1;

		if(nb_unit<=0)
			return index;

		VECTOR Dir;
		Dir=cam->Dir+2.0f*(mouse_x-gfx->SCREEN_W_HALF)*gfx->SCREEN_W_INV*cam->Side-1.5f*(mouse_y-gfx->SCREEN_H_HALF)*gfx->SCREEN_H_INV*cam->Up;
		Dir.Unit();		// Direction pointée par le curseur

		bool detectable=false;
		int i;

		for(uint16 e=0;e<index_list_size;e++) {
			i = idx_list[e];
			if(unit[i].flags==0 || !unit[i].visible) continue;		// Si l'unité n'existe pas on la zappe
			unit[i].flags&=0xFD;	// Enlève l'indicateur de possibilité d'intersection
			VECTOR center=unit[i].model->center+unit[i].Pos-cam->Pos;
			float size=unit[i].model->size*unit_manager.unit_type[unit[i].type_id].Scale*unit_manager.unit_type[unit[i].type_id].Scale;
			center=Dir*center;
			float dist=center.Sq();
			if(dist<size) {
				detectable=true;
				unit[i].flags|=0x2;		// Unité détectable
				}
			}

		if(!detectable)			// Si aucune unité n'est à proximité du curseur, on zappe la méthode OpenGl
			return index;

		glDisable(GL_LIGHTING);
		glDisable(GL_FOG);

		int BUFSIZE=max_unit<<2;

		GLuint *selectBuf = new GLuint[BUFSIZE];
		GLint hits;

		glSelectBuffer (BUFSIZE, selectBuf);
		glRenderMode (GL_SELECT);

		glInitNames();
		glPushName(0);

		int	viewport[4] = {0, 0, 0, 0};
		glGetIntegerv(GL_VIEWPORT, viewport);

		if(cam->Ref!=NULL)
			cam->Pos=cam->RPos+(*(cam->Ref));
		else
			cam->Pos=cam->RPos;

		VECTOR	FP;
		FP=cam->Pos+cam->Dir;

		cam->zfar2=cam->zfar*cam->zfar;
		glMatrixMode (GL_PROJECTION);			// Matrice de projection
		glLoadIdentity ();

		gluPickMatrix(mouse_x, SCREEN_H-mouse_y, sensibility, sensibility, viewport);

		glFrustum (-1.0f*cam->znear, 1.0f*cam->znear, -0.75f*cam->znear, 0.75f*cam->znear, cam->znear, cam->zfar);

		glMatrixMode (GL_MODELVIEW);

		glLoadIdentity();			// Charge la matrice identité

		gluLookAt(cam->Pos.x,cam->Pos.y,cam->Pos.z,FP.x,FP.y,FP.z,cam->Up.x,cam->Up.y,cam->Up.z);

		for(uint16 e=0;e<index_list_size;e++) {
			i = idx_list[e];
			if((unit[i].flags&0x2)==0x2) {			// Si l'unité existe et est sélectionnable
				unit[i].flags&=0xFD;
				if(unit[i].model) {
					glLoadName(i);
					VECTOR D;
					D=unit[i].Pos-cam->Pos;		// Vecteur "viseur unité" partant de la caméra vers l'unité

					float dist=D.Sq();
					if(!((dist>=16384.0f && (D%cam->Dir)<=0.0f)||(D%cam->Dir)>cam->zfar2)) {
						glPushMatrix();

						glTranslatef(unit[i].Pos.x,unit[i].Pos.y,unit[i].Pos.z);
						glRotatef(unit[i].Angle.x,1.0f,0.0f,0.0f);
						glRotatef(unit[i].Angle.y,0.0f,1.0f,0.0f);
						glRotatef(unit[i].Angle.z,0.0f,0.0f,1.0f);
						float scale=unit_manager.unit_type[unit[i].type_id].Scale;
						glScalef(scale,scale,scale);
						unit[i].model->draw(&(unit[i].data),false,true);

						glPopMatrix();
						}
					}
				}
			}

		glFlush ();

		hits = glRenderMode (GL_RENDER);

		glEnable(GL_LIGHTING);
		glEnable(GL_FOG);

		int j;
		GLuint names, *ptr;

		float nearest=2.0f;
		ptr = (GLuint *) selectBuf;
		for (i = 0; i < hits; i++) { /*  for each hit  */
			names = *ptr;
			ptr++;
			float z1=(float) *ptr/0x7fffffff; ptr++;
			float z2=(float) *ptr/0x7fffffff; ptr++;
			int v=*ptr; ptr++;

			if(z1<nearest) {
				nearest=z1;
				index=v;
				}
			}

		gfx->ReInitArrays();

		delete[] selectBuf;
		return index;
	}

	void UNIT::shoot(int target,VECTOR startpos,VECTOR Dir,int w_id)
	{
		POINTF O;
		O.x=O.y=O.z=0.0f;
		if(unit_manager.unit_type[type_id].weapon[w_id]->startsmoke && visible)
			particle_engine.make_smoke(O+startpos,0,1,0.0f);
		int w_idx=weapons.add_weapon(unit_manager.unit_type[type_id].weapon[w_id]->nb_id,idx);
		weapons.weapon[w_idx].Pos=startpos;
		if(unit_manager.unit_type[type_id].weapon[w_id]->startvelocity==0.0f && !unit_manager.unit_type[type_id].weapon[w_id]->selfprop)
			weapons.weapon[w_idx].V=unit_manager.unit_type[type_id].weapon[w_id]->weaponvelocity*Dir;
		else
			weapons.weapon[w_idx].V=unit_manager.unit_type[type_id].weapon[w_id]->startvelocity*Dir;
		if(unit_manager.unit_type[type_id].weapon[w_id]->dropped)
			weapons.weapon[w_idx].V=weapons.weapon[w_idx].V+V;
		weapons.weapon[w_idx].owner=owner_id;
		weapons.weapon[w_idx].target=target;
		if(unit_manager.unit_type[type_id].weapon[w_id]->interceptor)
			weapons.weapon[w_idx].target_pos=weapons.weapon[target].Pos;
		else
			weapons.weapon[w_idx].target_pos=units.unit[target].Pos+units.unit[target].model->center;
		weapons.weapon[w_idx].stime=0.0f;
		weapons.weapon[w_idx].visible=visible;
	}

void *create_unit(int type_id,int owner,VECTOR pos,MAP *map)
{
	int id=units.create(type_id,owner);
	if(id>=0) {
		units.unit[id].Pos=pos;
		units.unit[id].port[BUILD_PERCENT_LEFT]=100.0f;
		units.unit[id].cur_px = ((int)(units.unit[id].Pos.x)+map->map_w_d)>>3;
		units.unit[id].cur_py = ((int)(units.unit[id].Pos.z)+map->map_h_d)>>3;
		return &(units.unit[id]);
		}
	return NULL;
}

const bool can_be_built(const VECTOR Pos,MAP *map,const int unit_type_id)
{
	if(unit_type_id<0 || unit_type_id>=unit_manager.nb_unit)	return false;	// test the ID

	if(map==NULL)	return false;		// to avoid an error

	int w = unit_manager.unit_type[unit_type_id].FootprintX;
	int h = unit_manager.unit_type[unit_type_id].FootprintZ;
	int x = ((int)(Pos.x+map->map_w_d)>>3)-(w>>1);
	int y = ((int)(Pos.z+map->map_h_d)>>3)-(h>>1);
	if(x<0 || y<0 || x+w>=(map->bloc_w<<1) || y+h>=(map->bloc_h<<1))	return false;	// check if it is inside the map

	if(!map->check_rect(x,y,w,h,-1))	return false;		// There already something
	float dh = fabs(map->check_rect_dh(x,y,w,h));
	float depth = map->check_depth(x,y,w,h);

	if(dh>unit_manager.unit_type[unit_type_id].MaxSlope*H_DIV)	return false;	// Check the slope

			// Check if unit can be there
	if(depth<unit_manager.unit_type[unit_type_id].MinWaterDepth*H_DIV || depth>unit_manager.unit_type[unit_type_id].MaxWaterDepth*H_DIV)	return false;
//	if(depth>0 && (unit_manager.unit_type[unit_type_id].Category&NOTSUB))	return false;

	if(!map->check_vents(x,y,w,h,unit_manager.unit_type[unit_type_id].yardmap))	return false;

	if(map->check_lava(x+1>>1,y+1>>1,w+1>>1,h+1>>1))	return false;

	return true;
}

int INGAME_UNITS::menu_order(int omb,int player_id)							// Menu permettant de donner des ordres
{
	bool onoffable=false;
	bool canstop=false;
	bool canpatrol=false;
	bool canmove=false;
	bool canguard=false;
	bool canattack=false;
	bool canreclam=false;
	bool builders=false;
	bool canload=false;

	int signal=SIGNAL_ORDER_NONE;

	uint32 i;
	for(uint16 e=0;e<index_list_size;e++) {
		i = idx_list[e];
		if(unit[i].flags && unit[i].owner_id==player_id && unit[i].sel) {
			onoffable|=unit_manager.unit_type[unit[i].type_id].onoffable;
			canstop|=unit_manager.unit_type[unit[i].type_id].canstop;
			canmove|=unit_manager.unit_type[unit[i].type_id].canmove;
			canpatrol|=unit_manager.unit_type[unit[i].type_id].canpatrol;
			canguard|=unit_manager.unit_type[unit[i].type_id].canguard;
			canattack|=unit_manager.unit_type[unit[i].type_id].canattack;
			canreclam|=unit_manager.unit_type[unit[i].type_id].CanReclamate;
			builders|=unit_manager.unit_type[unit[i].type_id].Builder;
			canload|=unit_manager.unit_type[unit[i].type_id].canload;
			}
		}

	bool chg_page=false;
	if(mouse_b==1 && mouse_y>=420 && mouse_y<=476)
		if(mouse_x>=5 && mouse_x<30)
			chg_page=true;
	if(page==0)
		glbutton("1",5.0f,420.0f,30.0f,476.0f,chg_page);
	else
		glbutton("2",5.0f,420.0f,30.0f,476.0f,chg_page);
	if(chg_page && omb!=1)
		page=++page%2;

	if(page==0) {
		if(onoffable) {					// Fonction On/Off
			bool on=false,off=false;

			if(mouse_b==1 && mouse_y>=340 && mouse_y<=356) {
				if(mouse_x>=5 && mouse_x<62)
					on=true;
				if(mouse_x>66 && mouse_x<=123)
					off=true;
				}

			if(on && omb!=1) {
				signal=SIGNAL_ORDER_ONOFF;
				for(uint16 e=0;e<index_list_size;e++) {
					i = idx_list[e];
					if(unit[i].flags!=0 && unit[i].owner_id==player_id && unit[i].sel)
						unit[i].activate();
					}
				}
			if(off && omb!=1) {
				signal=SIGNAL_ORDER_ONOFF;
				for(uint16 e=0;e<index_list_size;e++) {
					i = idx_list[e];
					if(unit[i].flags!=0 && unit[i].owner_id==player_id && unit[i].sel)
						unit[i].deactivate();
					}
				}

			glbutton("on",5.0f,340.0f,61.5f,356.0f,on);
			glbutton("off",66.5f,340.0f,123.0f,356.0f,off);
			}

		if(canstop) {					// Fonction Stop
			bool stop=false;

			if(mouse_b==1 && mouse_y>=360 && mouse_y<=376) {
				if(mouse_x>=5 && mouse_x<62)
					stop=true;
				}

			if(stop && omb!=1) {
				signal=SIGNAL_ORDER_STOP;
				for(uint16 e=0;e<index_list_size;e++) {
					i = idx_list[e];
					if(unit[i].flags!=0 && unit[i].owner_id==player_id && unit[i].sel && unit_manager.unit_type[unit[i].type_id].canstop)
						unit[i].set_mission(MISSION_STOP);
					}
				}

			glbutton("stop",5.0f,360.0f,61.5f,376.0f,stop);
			}

		if(canmove) {					// Fonction Move
			bool move=false;

			if(mouse_b==1 && mouse_y>=360 && mouse_y<=376) {
				if(mouse_x>66 && mouse_x<=123)
					move=true;
				}

			if(move && omb!=1)
				signal=SIGNAL_ORDER_MOVE;

			glbutton("move",66.5f,360.0f,123.0f,376.0f,move);
			}

		if(canpatrol) {					// Fonction Patrol
			bool patrol=false;

			if(mouse_b==1 && mouse_y>=380 && mouse_y<=396) {
				if(mouse_x>66 && mouse_x<=123)
					patrol=true;
				}

			if(patrol && omb!=1)
				signal=SIGNAL_ORDER_PATROL;

			glbutton("patrol",66.5f,380.0f,123.0f,396.0f,patrol);
			}

		if(canguard) {					// Fonction Guard
			bool guard=false;

			if(mouse_b==1 && mouse_y>=380 && mouse_y<=396) {
				if(mouse_x>=5 && mouse_x<62)
					guard=true;
				}

			if(guard && omb!=1)
				signal=SIGNAL_ORDER_GUARD;

			glbutton("guard",5.0f,380.0f,61.5f,396.0f,guard);
			}

		if(canattack) {					// Fonction Attack
			bool attack=false;

			if(mouse_b==1 && mouse_y>=400 && mouse_y<=416) {
				if(mouse_x>66 && mouse_x<=123)
					attack=true;
				}

			if(attack && omb!=1)
				signal=SIGNAL_ORDER_ATTACK;

			glbutton("attack",66.5f,400.0f,123.0f,416.0f,attack);
			}

		if(canreclam) {					// Fonction Reclamate
			bool reclam=false;

			if(mouse_b==1 && mouse_y>=400 && mouse_y<=416) {
				if(mouse_x>=5 && mouse_x<62)
					reclam=true;
				}

			if(reclam && omb!=1)
				signal=SIGNAL_ORDER_RECLAM;

			glbutton("reclam",5.0f,400.0f,61.5f,416.0f,reclam);
			}
		}
	else if(page==1) {
		if(canload) {					// Fonction load
			bool load=false;
			bool unload=false;

			if(mouse_b==1 && mouse_y>=340 && mouse_y<=356) {
				if(mouse_x>=5 && mouse_x<62)
					load=true;
				if(mouse_x>66 && mouse_x<=123)
					unload=true;
				}

			if(load && omb!=1)
				signal=SIGNAL_ORDER_LOAD;
			if(unload && omb!=1)
				signal=SIGNAL_ORDER_UNLOAD;

			glbutton("load",5.0f,340.0f,61.5f,356.0f,load);
			glbutton("unload",66.5f,340.0f,123.0f,356.0f,unload);
			}
		}
	return signal;
}

void INGAME_UNITS::complete_menu(int index)
{
	set_uformat(U_ASCII);
	if(index<0 || index>=max_unit) return;		// On n'affiche que des données sur les unités EXISTANTES
	if(unit[index].flags==0) return;		// L'unité n'existe plus

	UNIT *target=unit[index].mission!=NULL ? (UNIT*) unit[index].mission->p : NULL;
	if(target && target->flags==0)
		target=NULL;

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	gfx->print(gfx->TA_font,150.0f,SCREEN_H-32.0f,0.0f,0xFFFFFFFF,unit_manager.unit_type[unit[index].type_id].name);
	if(target && (unit[index].mission->flags&4)!=4) {
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,175.0f+(gfx->SCREEN_W_HALF-100),SCREEN_H-32.0f,0.0f,0xFFFFFFFF,unit_manager.unit_type[target->type_id].name);
		}
	else if(unit[index].planned_weapons>0.0f) {
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,175.0f+(gfx->SCREEN_W_HALF-100),SCREEN_H-32.0f,0.0f,0xFFFFFFFF,translation_manager.get(30));
		}

	int stock=0;
	if(unit_manager.unit_type[unit[index].type_id].weapon[0] && unit_manager.unit_type[unit[index].type_id].weapon[0]->stockpile)	stock=unit[index].stock[0];
	else if(unit_manager.unit_type[unit[index].type_id].weapon[1] && unit_manager.unit_type[unit[index].type_id].weapon[1]->stockpile)	stock=unit[index].stock[1];
	else if(unit_manager.unit_type[unit[index].type_id].weapon[2] && unit_manager.unit_type[unit[index].type_id].weapon[2]->stockpile)	stock=unit[index].stock[2];
	if((unit_manager.unit_type[unit[index].type_id].Builder && unit_manager.unit_type[unit[index].type_id].TEDclass==CLASS_PLANT)
	|| unit[index].planned_weapons>0.0f || stock>0) {		// Affiche la liste de construction
		int page=unit_manager.unit_type[unit[index].type_id].page;

		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
		for(int i=6*page;i<min((int)unit_manager.unit_type[unit[index].type_id].nb_unit,6*page+6);i++) {		// Affiche les différentes images d'unités constructibles
			int nb=0;
			MISSION *m=unit[index].mission;
			while(m) {
				if((m->mission==MISSION_BUILD || m->mission==MISSION_BUILD_2) && m->data==unit_manager.unit_type[unit[index].type_id].BuildList[i])
					nb++;
				m=m->next;
				}
			if(nb>0) {
				char buf[10];
				uszprintf(buf,10,"%d",nb);
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
				gfx->print(gfx->TA_font,(i&1)*64+32-0.5f*gfx->TA_font.length(buf),128+(i-6*page>>1)*64+32-0.5f*gfx->TA_font.height(),0.0f,0xFFFFFFFF,buf);
				}
			else if(unit_manager.unit_type[unit[index].type_id].BuildList[i]==-1) {		// Il s'agit d'une arme / It's a weapon
				char buf[10];
				if((int)unit[index].planned_weapons==unit[index].planned_weapons)
					uszprintf(buf,10,"%d(%d)",(int)unit[index].planned_weapons,stock);
				else
					uszprintf(buf,10,"%d(%d)",(int)unit[index].planned_weapons+1,stock);
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
				gfx->print(gfx->TA_font,(i&1)*64+32-0.5f*gfx->TA_font.length(buf),128+(i-6*page>>1)*64+32-0.5f*gfx->TA_font.height(),0.0f,0xFFFFFFFF,buf);
				}
			}
		}

	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	char buf[10];
	glColor4f(0.9f,0.9f,0.9f,1.0f);
	uszprintf(buf,10,"+%f",unit[index].metal_prod);	*(strstr(buf,".")+2)=0;
	gfx->print(gfx->normal_font,125+(gfx->SCREEN_W_HALF-170),SCREEN_H-16,0.0f,buf);
	uszprintf(buf,10,"-%f",unit[index].metal_cons);	*(strstr(buf,".")+2)=0;
	gfx->print(gfx->normal_font,125+(gfx->SCREEN_W_HALF-170),SCREEN_H-8,0.0f,buf);

	glColor4f(1.0f,1.0f,0.0f,1.0f);
	uszprintf(buf,10,"+%f",unit[index].energy_prod);	*(strstr(buf,".")+2)=0;
	gfx->print(gfx->normal_font,125+(gfx->SCREEN_W_HALF-120),SCREEN_H-16,0.0f,buf);
	uszprintf(buf,10,"-%f",unit[index].energy_cons);	*(strstr(buf,".")+2)=0;
	gfx->print(gfx->normal_font,125+(gfx->SCREEN_W_HALF-120),SCREEN_H-8,0.0f,buf);

	glColor4f(1.0f,1.0f,1.0f,1.0f);

	glDisable(GL_TEXTURE_2D);

	glDisable(GL_BLEND);

	glBegin(GL_QUADS);			// Dessine les barres de metal et d'énergie
		glColor4f(1.0f,0.0f,0.0f,1.0f);

		glVertex2f(150.0f,SCREEN_H-16.0f);					// Fond de la barre de vie de l'unité
		glVertex2f(125+(gfx->SCREEN_W_HALF-180),SCREEN_H-16.0f);
		glVertex2f(125+(gfx->SCREEN_W_HALF-180),SCREEN_H-8.0f);
		glVertex2f(150.0f,SCREEN_H-8.0f);

		if(target && (unit[index].mission->flags&4)!=4 && !unit_manager.unit_type[target->type_id].HideDamage) {			// Si l'unité a une cible
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-16.0f);		// Fond de la barre de vie
			glVertex2f(SCREEN_W-50,SCREEN_H-16.0f);
			glVertex2f(SCREEN_W-50,SCREEN_H-8.0f);
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-8.0f);
			}
		else if(unit[index].planned_weapons>0.0f) {
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-16.0f);		// Fond de la barre de vie
			glVertex2f(SCREEN_W-50,SCREEN_H-16.0f);
			glVertex2f(SCREEN_W-50,SCREEN_H-8.0f);
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-8.0f);
			}

		glColor3f(0.0f,1.0f,0.0f);

		if(unit[index].hp>0) {
			glVertex2f(150.0f,SCREEN_H-16.0f);					// Barre de vie
			glVertex2f(150+((gfx->SCREEN_W_HALF-180)-25)*unit[index].hp/unit_manager.unit_type[unit[index].type_id].MaxDamage,SCREEN_H-16.0f);
			glVertex2f(150+((gfx->SCREEN_W_HALF-180)-25)*unit[index].hp/unit_manager.unit_type[unit[index].type_id].MaxDamage,SCREEN_H-8.0f);
			glVertex2f(150.0f,SCREEN_H-8.0f);
			}

		if(target && (unit[index].mission->flags&4)!=4 && !unit_manager.unit_type[target->type_id].HideDamage) {			// Si l'unité a une cible
			if(target->hp>0) {
				glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-16.0f);		// Barre de vie
				glVertex2f(175+(gfx->SCREEN_W_HALF-100)+(gfx->SCREEN_W_HALF-125)*target->hp/unit_manager.unit_type[target->type_id].MaxDamage,SCREEN_H-16.0f);
				glVertex2f(175+(gfx->SCREEN_W_HALF-100)+(gfx->SCREEN_W_HALF-125)*target->hp/unit_manager.unit_type[target->type_id].MaxDamage,SCREEN_H-8.0f);
				glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-8.0f);
				}
			}
		else if(unit[index].planned_weapons>0.0f) {						// construit une arme / build a weapon
			float p=1.0f-(unit[index].planned_weapons-(int)unit[index].planned_weapons);
			if(p==1.0f)	p=0.0f;
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-16.0f);
			glVertex2f(175+(gfx->SCREEN_W_HALF-100)+(gfx->SCREEN_W_HALF-125)*p,SCREEN_H-16.0f);
			glVertex2f(175+(gfx->SCREEN_W_HALF-100)+(gfx->SCREEN_W_HALF-125)*p,SCREEN_H-8.0f);
			glVertex2f(175+(gfx->SCREEN_W_HALF-100),SCREEN_H-8.0f);
			}

	glEnd();
	glColor4f(1.0f,1.0f,1.0f,1.0f);

	set_uformat(U_UTF8);
}

void INGAME_UNITS::move(float dt,MAP *map,int key_frame,bool wind_change)
{
	if(nb_unit<=0 || unit==NULL)	return;		// Pas d'unités à dessiner

	players.clear();		// Réinitialise le compteur de ressources

	uint32 i;
	for(uint16 e=0;e<index_list_size;e++) {		// Compte les stocks de ressources et les productions
		i = idx_list[e];
		if(unit[i].port[BUILD_PERCENT_LEFT]==0.0f) {
			unit[i].metal_prod=0.0f;
			unit[i].metal_cons=0.0f;
			unit[i].energy_prod=0.0f;
			unit[i].energy_cons=0.0f;
			players.metal_s[unit[i].owner_id]+=unit_manager.unit_type[unit[i].type_id].MetalStorage;
			players.energy_s[unit[i].owner_id]+=unit_manager.unit_type[unit[i].type_id].EnergyStorage;
			players.commander[unit[i].owner_id]|=(unit_manager.unit_type[unit[i].type_id].TEDclass==CLASS_COMMANDER);
			unit[i].energy_prod+=unit_manager.unit_type[unit[i].type_id].EnergyMake;
			if((unit[i].port[ACTIVATION] || !unit_manager.unit_type[unit[i].type_id].onoffable)
			&& unit_manager.unit_type[unit[i].type_id].EnergyUse<=players.energy[unit[i].owner_id]) {
				unit[i].metal_prod+=unit_manager.unit_type[unit[i].type_id].MakesMetal+unit_manager.unit_type[unit[i].type_id].MetalMake;
				if(unit_manager.unit_type[unit[i].type_id].ExtractsMetal) {	// Extracteur de métal
					int metal_base;
					int px=unit[i].cur_px;
					int py=unit[i].cur_py;
					if(map->map_data[py][px].stuff>=0)
						metal_base=feature_manager.feature[features.feature[map->map_data[py][px].stuff].type].metal;
					else {
						if((unit_manager.unit_type[unit[i].type_id].Category&0xF)==LEVEL3)
							metal_base=map->ota_data.MohoMetal+map->ota_data.SurfaceMetal;
						else
							metal_base=map->ota_data.SurfaceMetal;
						}
					unit[i].metal_prod+=10.0f*metal_base*unit_manager.unit_type[unit[i].type_id].ExtractsMetal;
					if(unit[i].just_created) {
						int param[] = { metal_base<<2 };
						unit[i].launch_script(unit[i].get_script_index(SCRIPT_SetSpeed),1,param);
						unit[i].just_created=false;
						}
					}
				if(unit_manager.unit_type[unit[i].type_id].WindGenerator) {	// Wind Generator
					unit[i].energy_prod+=map->wind*unit_manager.unit_type[unit[i].type_id].WindGenerator*0.0002f;
					if(wind_change) {
						int param[] = { (int)(map->wind*50.0f) };
						unit[i].launch_script(unit[i].get_script_index(SCRIPT_SetSpeed),1,param);
						param[0]=(int)((map->wind_dir-unit[i].Angle.y)*DEG2TA);
						unit[i].launch_script(unit[i].get_script_index(SCRIPT_SetDirection),1,param);
						unit[i].launch_script(unit[i].get_script_index(SCRIPT_go));
						}
					}
				if(unit_manager.unit_type[unit[i].type_id].TidalGenerator)	// Tidal Generator
					unit[i].energy_prod+=map->ota_data.tidalstrength;
				if(unit_manager.unit_type[unit[i].type_id].EnergyUse<0)
					unit[i].energy_prod-=unit_manager.unit_type[unit[i].type_id].EnergyUse;
				else
					unit[i].energy_cons=unit_manager.unit_type[unit[i].type_id].EnergyUse;
				}
			}
		}

	exp_dt_1=exp(-dt);
	exp_dt_2=exp(-2.0f*dt);
	exp_dt_4=exp(-4.0f*dt);
	g_dt=dt*map->ota_data.gravity;
	int path_exec=0;
	for(uint16 e=0;e<index_list_size;e++) {
		i = idx_list[e];
			if(unit[i].owner_id==players.local_human_id) {
				if(unit[i].attacked
				|| (unit[i].mission!=NULL && unit[i].mission->mission==MISSION_SHOOT))	nb_attacked+=100;
				if(unit[i].built)		nb_built++;
				}
			players.nb_unit[unit[i].owner_id]++;			// Compte les unités de chaque joueur
			if(unit[i].move(dt,map,&path_exec,key_frame)==-1)			// Vérifie si l'unité a été détruite
				kill(i,map,e);
			else {
				players.metal_t[unit[i].owner_id] += unit[i].metal_prod;
				players.metal_u[unit[i].owner_id] += unit[i].metal_cons;
				players.energy_t[unit[i].owner_id] += unit[i].energy_prod;
				players.energy_u[unit[i].owner_id] += unit[i].energy_cons;
				}
			}
	float exp_r = exp(-dt*0.1f);
	nb_attacked*=exp_r;
	nb_built*=exp_r;

	for(i=0;i<players.nb_player;i++)
		if(players.commander[i]) {
			players.metal_s[i]+=players.com_metal[i];
			players.energy_s[i]+=players.com_energy[i];
			}
	for(i=0;i<players.nb_player;i++) {
		players.metal[i]+=dt*(players.metal_t[i]-players.metal_u[i]);
		players.energy[i]+=dt*(players.energy_t[i]-players.energy_u[i]);
		players.metal_total[i]+=dt*players.metal_t[i];
		players.energy_total[i]+=dt*players.energy_t[i];
		if(players.metal[i]<0.0f)
			players.metal[i]=0.0f;
		else if(players.metal[i]>players.metal_s[i])
			players.metal[i]=players.metal_s[i];
		if(players.energy[i]<0.0f)
			players.energy[i]=0.0f;
		else if(players.energy[i]>players.energy_s[i])
			players.energy[i]=players.energy_s[i];
		}
}

int INGAME_UNITS::create(int type_id,int owner)
{
	if(type_id<0 || type_id>=unit_manager.nb_unit)	return -1;
	if(owner<0 || owner>=NB_PLAYERS)	return -1;
	if(nb_unit>=MAX_UNIT_PER_PLAYER*NB_PLAYERS)		return -1;
	nb_unit++;
	if(nb_unit>=max_unit) {
		max_unit=MAX_UNIT_PER_PLAYER*NB_PLAYERS;
		UNIT *n_unit=(UNIT*) malloc(sizeof(UNIT)*max_unit);
		uint16	*n_idx = new uint16[max_unit];
		uint16	*n_new_idx = new uint16[max_unit];
		if(index_list_size>0)
			memcpy(n_idx,idx_list,index_list_size<<1);
		if(free_idx)
			memcpy(n_new_idx,free_idx,max_unit<<1);
		if(idx_list)	delete[]	idx_list;
		if(free_idx)	delete[]	free_idx;
		idx_list = n_idx;
		free_idx = n_new_idx;
		for(uint16 i = 0; i<max_unit;i++)
			free_idx[i] = i;
		for(uint16 i = 0; i<10;i++)
			free_index_size[i] = MAX_UNIT_PER_PLAYER;
		for(int i=0;i<max_unit;i++) {
			n_unit[i].init(-1,-1,i>=nb_unit-1);
			n_unit[i].flags=0;
			n_unit[i].idx=i;
			}
		if(unit) {
			memcpy(n_unit,unit,sizeof(UNIT)*(nb_unit-1));
			free(unit);
			}
		unit=n_unit;
		}
	if(unit==NULL)
		printf("error: memory alloc failed\n");
	if(free_index_size[owner]<=0) {
		printf("unit limit reached!\n");
		return -1;
		}
	int unit_index = free_idx[owner*MAX_UNIT_PER_PLAYER+free_index_size[owner]-1];
	free_index_size[owner]--;
	idx_list[index_list_size++] = unit_index;
	unit[unit_index].init(type_id,owner);
	unit[unit_index].Angle.y=((rand_from_table()%2001)-1000)*0.01f;	// Angle de 10° maximum
	players.nb_unit[owner]++;
	return unit_index;
}

void INGAME_UNITS::draw_mini(float map_w,float map_h,int mini_w,int mini_h,SECTOR **map_data)				// Repère les unités sur la mini-carte
{
	if(nb_unit<=0 || unit==NULL)	return;		// Pas d'unités à dessiner

	float rw=128.0f*mini_w/252/map_w;
	float rh=128.0f*mini_h/252/map_h;

	glDisable(GL_TEXTURE_2D);
	glPointSize(3.0f);
	glBegin(GL_POINTS);
	int cur_id=-1;
	byte mask=1<<players.local_human_id;
	int b_w=(int)map_w>>3;
	int b_h=(int)map_h>>3;
	for(uint16 e=0;e<index_list_size;e++) {
		uint16 i = idx_list[e];
		if(unit[i].flags!=0) {
			int px=unit[i].cur_px;
			int py=unit[i].cur_py;
			if(px<0 || py<0 || px>=b_w || py>=b_h)	continue;
			if(!(map_data[py][px].view&mask) || !map_data[py][px].timer) continue;	// Unité non visible / Unit is not visible
			unit[i].flags|=0x10;
			if(cur_id!=unit[i].owner_id) {
				cur_id=unit[i].owner_id;
				uint16 used_id = 3*player_color_map[cur_id];
				glColor3f(0.5f*player_color[used_id],0.5f*player_color[used_id+1],0.5f*player_color[used_id+2]);
				}
			float pos_x=unit[i].Pos.x*rw+64.0f;
			float pos_y=unit[i].Pos.z*rh+64.0f;
			glVertex2f(pos_x-1.0f,pos_y);
			glVertex2f(pos_x+1.0f,pos_y);
			glVertex2f(pos_x,pos_y-1.0f);
			glVertex2f(pos_x,pos_y+1.0f);
			}
		}
	cur_id=-1;
	for(uint16 e=0;e<index_list_size;e++) {
		uint16 i = idx_list[e];
		if(unit[i].flags!=0 && (unit[i].flags&0x10)==0x10) {
			unit[i].flags^=0x10;
			if(cur_id!=unit[i].owner_id) {
				cur_id=unit[i].owner_id;
				uint16 used_id = 3*player_color_map[cur_id];
				glColor3f(player_color[used_id],player_color[used_id+1],player_color[used_id+2]);
				}
			glVertex2f(unit[i].Pos.x*rw+64.0f,unit[i].Pos.z*rh+64.0f);
			}
		}
	cur_id=players.local_human_id;
	for(uint16 e=0;e<index_list_size;e++) {
		uint16 i = idx_list[e];
		if(unit[i].flags!=0 && unit[i].owner_id==players.local_human_id && unit[i].sel) {
			glColor3f(1.0f,1.0f,1.0f);
			float pos_x=unit[i].Pos.x*rw+64.0f;
			float pos_y=unit[i].Pos.z*rh+64.0f;
			glVertex2f(pos_x-1.0f,pos_y);
			glVertex2f(pos_x+1.0f,pos_y);
			glVertex2f(pos_x,pos_y-1.0f);
			glVertex2f(pos_x,pos_y+1.0f);

			glColor3f(player_color[3*player_color_map[cur_id]],player_color[3*player_color_map[cur_id]+1],player_color[3*player_color_map[cur_id]+2]);
			glVertex2f(pos_x,pos_y);
			}
		}
	glEnd();
	glPointSize(1.0f);
	glEnable(GL_TEXTURE_2D);
}

void INGAME_UNITS::kill(int index,MAP *map,int prev)			// Détruit une unité
{
	if(index<0 || index>max_unit || prev<0 || prev>=index_list_size)	return;		// On ne peut pas détruire une unité qui n'existe pas
	if(unit[index].flags) {
		map->rect((((int)(unit[index].Pos.x+map->map_w*0.5f))>>3)-(unit_manager.unit_type[unit[index].type_id].FootprintX>>1),(((int)(unit[index].Pos.z+map->map_h*0.5f))>>3)-(unit_manager.unit_type[unit[index].type_id].FootprintZ>>1),unit_manager.unit_type[unit[index].type_id].FootprintX,unit_manager.unit_type[unit[index].type_id].FootprintZ,-1);
		map->update_player_visibility(unit[index].owner_id,unit[index].old_px,unit[index].old_py,unit[index].old_sight,true);
		players.nb_unit[unit[index].owner_id]--;
		players.losses[unit[index].owner_id]++;		// Statistiques
		if(unit_manager.unit_type[unit[index].type_id].canload && unit[index].nb_attached>0)
			for(int i=0;i<unit[index].nb_attached;i++)
				unit[unit[index].attached_list[i]].hp=0.0f;
		unit[index].destroy();		// Détruit l'unité
		uint16 owner = index/MAX_UNIT_PER_PLAYER;
		free_idx[MAX_UNIT_PER_PLAYER*owner + free_index_size[owner]++]=index;
		idx_list[prev]=idx_list[--index_list_size];
		nb_unit--;		// Unité détruite
		}
}

void INGAME_UNITS::draw(CAMERA *cam,MAP *map,bool underwater,bool limit,bool cullface,bool height_line)					// Dessine les unités visibles
{
	if(nb_unit<=0 || unit==NULL)	return;		// Pas d'unités à dessiner

	glEnable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,1.0f);
	float sea_lvl = limit ? map->sealvl-5.0f : map->sealvl;
	for(uint16 e=0;e<index_list_size;e++) {
		uint16 i = idx_list[e];
		if(unit[i].flags && ((unit[i].Pos.y<=map->sealvl && underwater) || (unit[i].Pos.y>=sea_lvl && !underwater))) {				// Si il y a une unité
			unit[i].draw(cam,map,cullface,height_line);
			players.annihilated[unit[i].owner_id]=false;
			}
		}
}

void INGAME_UNITS::draw_shadow(CAMERA *cam,VECTOR Dir,MAP *map,float alpha)					// Dessine les ombres des unités visibles
{
	if(nb_unit<=0 || unit==NULL)	return;		// Pas d'unités à dessiner

	if(g_useStencilTwoSide) {					// Si l'extension GL_EXT_stencil_two_side est disponible
		glEnable(GL_STENCIL_TEST);
		glEnable(GL_STENCIL_TEST_TWO_SIDE_EXT);
		glDisable(GL_CULL_FACE);
		glClear(GL_STENCIL_BUFFER_BIT);
		glDepthMask(GL_FALSE);
		glColorMask(0,0,0,0);
		glStencilFunc(GL_ALWAYS,128, 0xffffffff);

		glActiveStencilFaceEXT(GL_FRONT);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR_WRAP_EXT);
		glActiveStencilFaceEXT(GL_BACK);
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR_WRAP_EXT);

		for(uint16 e=0;e<index_list_size;e++) {
			uint16 i = idx_list[e];
			if(unit[i].flags)				// Si il y a une unité
				unit[i].draw_shadow(cam,Dir,map);
			}
		}
	else {										// Si elle ne l'est pas
		glDepthMask(GL_FALSE);
		glColorMask(0,0,0,0);

		glClear(GL_STENCIL_BUFFER_BIT);
		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_ALWAYS,128, 0xffffffff);
		glEnable(GL_CULL_FACE);

		for(uint16 e=0;e<index_list_size;e++) {
			uint16 i = idx_list[e];
			if(unit[i].flags)					// Si il y a une unité
				unit[i].draw_shadow_basic(cam,Dir,map);
			}
		}
	features.draw_shadow(cam,Dir);

	glColorMask(0xFF,0xFF,0xFF,0xFF);
	allegro_gl_set_allegro_mode();
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_NOTEQUAL,0, 0xffffffff);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glColor4f(0.0f,0.0f,0.0f,alpha);
	glBegin(GL_QUADS);
		glVertex2i(0,0);
		glVertex2i(SCREEN_W,0);
		glVertex2i(SCREEN_W,SCREEN_H);
		glVertex2i(0,SCREEN_H);
	glEnd();
	allegro_gl_unset_allegro_mode();
	glDepthMask(GL_TRUE);
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,1.0f);
}
