/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2006  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include "stdafx.h"
#include "TA3D_NameSpace.h"
#include "cLogger.h"

namespace TA3D
{
	namespace INTERFACES
	{
		cLogger::cLogger( const String &FileName, bool noInterface )
		{
			m_bNoInterface = true;

			m_File.open( FileName.c_str(), std::ios::out | std::ios::trunc );

			if( !m_File.is_open() ) // todo fix this.
				throw( "cLogger::Failed to open file for output" );

			m_bNoInterface = noInterface;

			CreateCS();

			if( !m_bNoInterface )
				InitInterface();
		}

		cLogger::~cLogger( )
		{
			if( !m_File.is_open() )
				return;

			m_File.close();

			if( !m_bNoInterface )
				DeleteInterface();

			DeleteCS();
		}

		void cLogger::LogData( const char *txt )
		{
			EnterCS();
				m_File << txt;
			m_File.flush();
			LeaveCS();            
		}

		uint32 cLogger::InterfaceMsg( const lpcImsg msg )
		{
			if( msg->MsgID != TA3D_IM_DEBUG_MSG )
				return INTERFACE_RESULT_CONTINUE;

			LogData( (const char *)(msg->lpParm1) );

			return INTERFACE_RESULT_CONTINUE;
		}
	}
} 
