/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include "stdafx.h"
#include "TA3D_NameSpace.h"
#include "cTAFileParser.h"
#include "cError.h"

namespace TA3D
{
	namespace UTILS
	{
		String cTAFileParser::GetLine( const String &Data )
		{
			std::basic_string <char>::size_type NotFound = -1;
			std::basic_string <char>::size_type iFind = Data.find_first_of( "\n" );

			if( iFind == NotFound )
				return Data;

			return Data.substr( 0, iFind );
		}

		void cTAFileParser::ProcessData(  String &Data )
		{
			if( Data.length() == 0 )
				return;

			String Line = GetLine( Data );         // extract line
			Data.erase( 0, Line.length() +1 );      // erase line from data.

			int i = (int)Line.find( "//" );        // search for start of comment.
			if( i != -1 )        // if we find a comment, we will erase everything
				Line.resize( i );// from the comment to the end of the line.

 			Line = TrimString( Line ); // strip out crap from string.
			i = (int)Line.length();

			if( i > 3 ) // check for new key.
			{
				if( Line[0] == '[' && Line[ i-1 ] == ']' )
				{
					m_cKey = ( ( m_cKey == "" ) ? Line.substr( 1, i-2 ) :
					                        m_cKey + String( "." ) +
					                        Line.substr( 1, i-2 ) );
					ProcessData( Data );
					return;
				}
			}
			else if( Line == "}" )  // close the current active key.
			{
				i = (int)m_cKey.find_last_of( "." );

				if( i != -1 )
					m_cKey = m_cKey.substr( 0, i );
				else
					m_cKey = "";

				ProcessData( Data );
				return;
			}

			// if we get here its possible its a name=value;
			// so we will search for a = and a ; if we find them we have a valid name/value
			int f1, f2;
			f1 = (int)Line.find_first_of( "=" );
			f2 = (int)Line.find_last_of( ";" );

			if( f1 != -1 && f2 != -1 )
			{
				String n, v;
				n = Line.substr( 0, f1 );
				v = Line.substr( (f1+1), (f2-(f1+1)) );

				n = m_cKey + String( "." ) + n;

				if( !m_bKeysCaseSenstive )
					InsertOrUpdate( Lowercase( n ), v );
				else
					InsertOrUpdate( n, v ); 
			}

			ProcessData( Data );
		}

		void cTAFileParser::Load( const String &FileName,  bool bClearTable )
		{
			uint32 ota_size=0;
			byte *data = NULL;
			std::ifstream   m_File;
			String tFileName = GetClientPath() + FileName;

			if(  bClearTable )
				EmptyHashTable();

			m_File.open( tFileName.c_str(), std::ios::in | std::ios::binary | std::ios::ate );
			if( m_File.is_open() )
			{
				ota_size = m_File.tellg();
				data = new byte[ota_size];
				m_File.seekg( 0, std::ios::beg );
				m_File.read( (char *)data, ota_size );
				m_File.close();         
			}
			else if( TA3D::VARS::HPIManager )
				data = TA3D::VARS::HPIManager->PullFromHPI( FileName, &ota_size );

			if( !data )
				throw String( "Unable to load file " + FileName );

			std::string tmp = String( (char *)data );
			m_cKey = "";
			free(data);

			// erase all line feeds.
			std::basic_string <char>::size_type NotFound = -1;
			std::basic_string <char>::size_type iFind = tmp.find( "\r" );
			while( iFind != NotFound )
			{
				tmp.erase( iFind, 1 );
				iFind = tmp.find( "\r" );
			}

			// now process the remaining.
			ProcessData( tmp );
		}

		cTAFileParser::cTAFileParser( const String &FileName,  bool bKeysCaseSenstive )
		{
			m_bKeysCaseSenstive = bKeysCaseSenstive;
			InitTable( 512 );
			Load( FileName );
		}

		cTAFileParser::cTAFileParser()
		{
			InitTable( 512 );
		}

		cTAFileParser::~cTAFileParser()
		{
			EmptyHashTable();
		}
		sint32 cTAFileParser::PullAsInt( const std::string &KeyName , sint32 def )
		{
			String key_to_find;
			if( !m_bKeysCaseSenstive )
				key_to_find = Lowercase( KeyName );
			else
				key_to_find = KeyName;
			if( !Exists( key_to_find ) )	return def;

			String iterFind = Find( key_to_find );

			return ( (iterFind.length() == 0) ? def : ustrtol( iterFind.c_str() , NULL, 0 ) );		// Uses ustrtol to deal with hexa numbers
		}

		real32 cTAFileParser::PullAsFloat( const std::string &KeyName , real32 def )
		{
			String key_to_find;
			if( !m_bKeysCaseSenstive )
				key_to_find = Lowercase( KeyName );
			else
				key_to_find = KeyName;
			if( !Exists( key_to_find ) )	return def;

			String iterFind = Find( key_to_find );

			return ( (iterFind.length() == 0) ? def : (float)atof( iterFind.c_str() ) );
		}

		String cTAFileParser::PullAsString( const std::string &KeyName , String def )
		{
			String key_to_find;
			if( !m_bKeysCaseSenstive )
				key_to_find = Lowercase( KeyName );
			else
				key_to_find = KeyName;
			if( !Exists( key_to_find ) )	return def;

			String iterFind = Find( key_to_find );

			return iterFind;
		}

		bool cTAFileParser::PullAsBool( const std::string &KeyName , bool def )
		{
			String key_to_find;
			if( !m_bKeysCaseSenstive )
				key_to_find = Lowercase( KeyName );
			else
				key_to_find = KeyName;
			if( !Exists( key_to_find ) )
				return def;

			String iterFind = Lowercase( Find( key_to_find ) );

			if( iterFind == "false" || iterFind == "0" || iterFind == "" )
				return false;

			return true;
		}
	}
} 
