/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------------\
|                                         fbi.cpp                                    |
|  Ce fichier contient les structures, classes et fonctions nécessaires à la lecture |
| des fichiers fbi du jeu totalannihilation qui sont les fichiers de données sur les |
| unités du jeu. Cela inclus les classes pour gérer les différents types d'unités et |
| le système de gestion de liens entre unités.                                       |
|                                                                                    |
\-----------------------------------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"
//#include "fbi.h"
#include "EngineClass.h"
#include "UnitEngine.h"

UNIT_MANAGER unit_manager;

void UNIT_TYPE::show_info(float fade,GFX_FONT fnt)
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glColor4f(0.5f,0.5f,0.5f,fade);
	int x=gfx->SCREEN_W_HALF-160;
	int y=gfx->SCREEN_H_HALF-120;
	glBegin(GL_QUADS);
		glVertex2i(x,y);
		glVertex2i(x+320,y);
		glVertex2i(x+320,y+240);
		glVertex2i(x,y+240);
	glEnd();
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glEnable(GL_TEXTURE_2D);
	glColor4f(1.0f,1.0f,1.0f,fade);
	gfx->drawtexture(glpic,x+16,y+16,x+80,y+80);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	gfx->print(fnt,x+96,y+16,0.0f,format("%s: %s",translation_manager.get(42),name));
	gfx->print(fnt,x+96,y+28,0.0f,format("%s: %s",translation_manager.get(43),Unitname));
	gfx->print(fnt,x+96,y+40,0.0f,format("%s",Description));
	gfx->print(fnt,x+96,y+52,0.0f,format("%s: %d",translation_manager.get(45),MaxDamage));
	gfx->print(fnt,x+96,y+64,0.0f,format("%s: E %d M %d",translation_manager.get(44),BuildCostEnergy,BuildCostMetal));
	gfx->print(fnt,x+16,y+100,0.0f,format("%s: %d",translation_manager.get(47),BuildTime));
	gfx->print(fnt,x+16,y+124,0.0f,format("%s:",translation_manager.get(46)));
	int Y=y+136;
	if(weapon[0])	{	gfx->print(fnt,x+16,Y,0.0f,format("%s: %d",weapon[0]->name,weapon[0]->damage));	Y+=12;	}
	if(weapon[1])	{	gfx->print(fnt,x+16,Y,0.0f,format("%s: %d",weapon[1]->name,weapon[1]->damage));	Y+=12;	}
	if(weapon[2])	{	gfx->print(fnt,x+16,Y,0.0f,format("%s: %d",weapon[2]->name,weapon[2]->damage));	Y+=12;	}
	glDisable(GL_BLEND);
}

	int UNIT_TYPE::load(char *data,int size)
	{
		set_uformat(U_ASCII);
		destroy();
		char *pos=data;
		char *f;
		char *ligne=NULL;
		int nb=0;
		int nb_inconnu=0;
		char *limit=data+size;
		while(*pos!='{') pos++;
		pos=strstr(pos,"\n")+1;
		char *lang_name="name=";
		char *lang_desc="description=";
		switch(LANG)
		{
		case TA3D_LANG_FRENCH:
			lang_name="frenchname=";
			lang_desc="frenchdescription=";
			break;
		case TA3D_LANG_GERMAN:
			lang_name="germanname=";
			lang_desc="germandescription=";
			break;
		case TA3D_LANG_SPANISH:
			lang_name="spanishname=";
			lang_desc="spanishdescription=";
			break;
		case TA3D_LANG_ITALIAN:
			lang_name="italianname=";
			lang_desc="italiandescription=";
			break;
		};
		do
		{
			nb++;
			if(ligne)
				delete[] ligne;
			ligne=get_line(pos);
			char *dup_ligne=strdup(ligne);
			strlwr(ligne);
			while(pos[0]!=0 && pos[0]!=13 && pos[0]!=10)	pos++;
			while(pos[0]==13 || pos[0]==10)	pos++;

			f=NULL;
			if(f=strstr(ligne,"unitname=")) {
				Unitname=strdup(f+9);
				if(strstr(Unitname,";"))
					*(strstr(Unitname,";"))=0;
				}
			else if(f=strstr(ligne,"version="))	version=f[8]-'0';
			else if(f=strstr(ligne,"side=")) {
				side=strdup(f+5);
				if(strstr(side,";"))
					*(strstr(side,";"))=0;
				}
			else if(f=strstr(ligne,"objectname=")) {
				ObjectName=strdup(f+11);
				if(strstr(ObjectName,";"))
					*(strstr(ObjectName,";"))=0;
				}
			else if(f=strstr(ligne,"designation=")) {
				Designation_Name=strdup(f+12);
				if(strstr(Designation_Name,";"))
					*(strstr(Designation_Name,";"))=0;
				}
			else if((f=strstr(ligne,lang_desc))!=NULL && (f==ligne || (f>ligne && *(f-1)<'a'))) {
				if(Description)	free(Description);
				Description=replace_chars(strdup(f+strlen(lang_desc)));
				if(strstr(Description,";"))
					*(strstr(Description,";"))=0;
				}
			else if((f=strstr(ligne,lang_name))!=NULL && (f==ligne || (f>ligne && *(f-1)<'a'))) {
				if(name)	free(name);
				name=replace_chars(strdup(f+strlen(lang_name)));
				if(strstr(name,";"))
					*(strstr(name,";"))=0;
				}
			else if((f=strstr(ligne,"description="))!=NULL && (f==ligne || (f>ligne && *(f-1)<'a')) && Description==NULL) {
				Description=replace_chars(strdup(f+12));
				if(strstr(Description,";"))
					*(strstr(Description,";"))=0;
				}
			else if((f=strstr(ligne,"name="))!=NULL && (f==ligne || (f>ligne && *(f-1)<'a')) && name==NULL) {
				name=replace_chars(strdup(f+5));
				if(strstr(name,";"))
					*(strstr(name,";"))=0;
				}
			else if(f=strstr(ligne,"description=")) {		// Pour éviter de surcharger les logs
				}
			else if(f=strstr(ligne,"name=")) {
				}
			else if(f=strstr(ligne,"footprintx="))			FootprintX=atoi(f+11);
			else if(f=strstr(ligne,"footprintz="))			FootprintZ=atoi(f+11);
			else if(f=strstr(ligne,"buildcostenergy="))		BuildCostEnergy=atoi(f+16);
			else if(f=strstr(ligne,"buildcostmetal="))		BuildCostMetal=atoi(f+15);
			else if(f=strstr(ligne,"maxdamage="))			MaxDamage=atoi(f+10);
			else if(f=strstr(ligne,"maxwaterdepth="))		MaxWaterDepth=atoi(f+14);
			else if(f=strstr(ligne,"minwaterdepth="))	{
				MinWaterDepth=atoi(f+14);
				if(MaxWaterDepth==0)
					MaxWaterDepth=255;
				}
			else if(f=strstr(ligne,"energyuse="))			EnergyUse=atoi(f+10);
			else if(f=strstr(ligne,"buildtime=")) {
				BuildTime=atoi(f+10);
				if(HealTime<0)
					HealTime=BuildTime;
				}
			else if(f=strstr(ligne,"workertime="))			WorkerTime=atoi(f+11);
			else if(f=strstr(ligne,"builder="))				Builder=(f[8]=='1');
			else if(f=strstr(ligne,"threed="))				ThreeD=(f[7]=='1');
			else if(f=strstr(ligne,"sightdistance="))		SightDistance=atoi(f+14)>>1;
			else if(f=strstr(ligne,"radardistance="))		RadarDistance=atoi(f+14)>>1;
			else if(f=strstr(ligne,"soundcategory=")) {
				if(strstr(f,";"))
					*(strstr(f,";"))=0;
				soundcategory = strdup(f + 14);
				}
			else if(f=strstr(ligne,"category=")) {
				if(strstr(f,"level"))		Category+=atoi(strstr(f,"level")+5);
				if(strstr(f,"arm"))			Category|=ARM;
				if(strstr(f,"weapon"))		Category|=_WEAPON;
				if(strstr(f,"commander"))	Category|=COMMANDER;
				if(strstr(f,"notair"))		Category|=NOTAIR;
				if(strstr(f,"notsub"))		Category|=NOTSUB;
				if(strstr(f,"ctrl_c"))		Category|=CTRL_C;
				}
			else if(f=strstr(ligne,"unitnumber="))			UnitNumber=atoi(f+14);
			else if(f=strstr(ligne,"canmove="))				canmove=(f[8]=='1');
			else if(f=strstr(ligne,"canpatrol="))			canpatrol=(f[10]=='1');
			else if(f=strstr(ligne,"canstop="))				canstop=(f[8]=='1');
			else if(f=strstr(ligne,"canguard="))			canguard=(f[9]=='1');
			else if(f=strstr(ligne,"maxvelocity="))			MaxVelocity=atof(f+12)*16.0f;
			else if(f=strstr(ligne,"brakerate="))			BrakeRate=atof(f+10)*16.0f;
			else if(f=strstr(ligne,"acceleration="))		Acceleration=atof(f+13)*16.0f;
			else if(f=strstr(ligne,"turnrate="))			TurnRate=atof(f+9);
			else if(f=strstr(ligne,"candgun="))				candgun=(f[8]=='1');
			else if(f=strstr(ligne,"canattack="))			canattack=(f[10]=='1');
			else if(f=strstr(ligne,"canreclamate="))		CanReclamate=(f[13]=='1');
			else if(f=strstr(ligne,"energymake="))			EnergyMake=atoi(f+11);
			else if(f=strstr(ligne,"metalmake="))			MetalMake=atof(f+10);
			else if(f=strstr(ligne,"cancapture="))			CanCapture=(f[11]=='1');
			else if(f=strstr(ligne,"hidedamage="))			HideDamage=(f[11]=='1');
			else if(f=strstr(ligne,"healtime="))			HealTime=atoi(f+9);
			else if(f=strstr(ligne,"cloakcost="))			CloakCost=atoi(f+10);
			else if(f=strstr(ligne,"cloakcostmoving="))		CloakCostMoving=atoi(f+16);
			else if(f=strstr(ligne,"builddistance="))		BuildDistance=atoi(f+14);
			else if(f=strstr(ligne,"activatewhenbuilt="))	ActivateWhenBuilt=(f[18]=='1');
			else if(f=strstr(ligne,"immunetoparalyzer="))	ImmuneToParalyzer=(f[18]=='1');
			else if(f=strstr(ligne,"sonardistance="))		SonarDistance=atoi(f+14)>>1;
			else if(f=strstr(ligne,"copyright=")) {}
			else if(f=strstr(ligne,"maxslope="))			MaxSlope=atoi(f+9);
			else if(f=strstr(ligne,"steeringmode="))		SteeringMode=atoi(f+13);
			else if(f=strstr(ligne,"bmcode="))				BMcode=atoi(f+7);
			else if(f=strstr(ligne,"zbuffer=")) {}
			else if(f=strstr(ligne,"shootme="))				ShootMe=(f[8]=='1');
			else if(f=strstr(ligne,"upright="))				Upright=(f[8]=='1');
			else if(f=strstr(ligne,"norestrict="))			norestrict=(f[11]=='1');
			else if(f=strstr(ligne,"noautofire="))			AutoFire=(f[11]!='1');
			else if(f=strstr(ligne,"energystorage="))		EnergyStorage=atoi(f+14);
			else if(f=strstr(ligne,"metalstorage="))		MetalStorage=atoi(f+13);
			else if(f=strstr(ligne,"standingmoveorder="))	StandingMoveOrder=atoi(f+18);
			else if(f=strstr(ligne,"mobilestandorders="))	MobileStandOrders=atoi(f+18);
			else if(f=strstr(ligne,"standingfireorder="))	StandingFireOrder=atoi(f+18);
			else if(f=strstr(ligne,"firestandorders="))		FireStandOrders=atoi(f+16);
			else if(f=strstr(ligne,"waterline="))			WaterLine=atoi(f+10);
			else if(f=strstr(ligne,"tedclass=")) {
				if(strstr(f,"water"))			TEDclass=CLASS_WATER;
				else if(strstr(f,"ship"))		TEDclass=CLASS_SHIP;
				else if(strstr(f,"energy"))		TEDclass=CLASS_ENERGY;
				else if(strstr(f,"vtol"))		TEDclass=CLASS_VTOL;
				else if(strstr(f,"kbot"))		TEDclass=CLASS_KBOT;
				else if(strstr(f,"plant"))		TEDclass=CLASS_PLANT;
				else if(strstr(f,"tank"))		TEDclass=CLASS_TANK;
				else if(strstr(f,"special"))	TEDclass=CLASS_SPECIAL;
				else if(strstr(f,"fort"))		TEDclass=CLASS_FORT;
				else if(strstr(f,"metal"))		TEDclass=CLASS_METAL;
				else if(strstr(f,"cnstr"))		TEDclass=CLASS_CNSTR;
				else if(strstr(f,"commander"))	TEDclass=CLASS_COMMANDER;
				else {
					printf("->tedclass id inconnu : %s\n",f);
					nb_inconnu++;
					}
				}
			else if(f=strstr(ligne,"noshadow="))			NoShadow=(f[9]=='1');
			else if(f=strstr(ligne,"antiweapons="))			antiweapons=(f[12]=='1');
			else if(f=strstr(ligne,"buildangle="))			BuildAngle=atoi(f+11);
			else if(f=strstr(ligne,"canfly="))				canfly=(f[7]=='1');
			else if(f=strstr(ligne,"canload="))				canload=(f[8]=='1');
			else if(f=strstr(ligne,"floater="))				Floater=(f[8]=='1');
			else if(f=strstr(ligne,"bankscale="))			BankScale=atoi(f+10);
			else if(f=strstr(ligne,"tidalgenerator="))		TidalGenerator=(f[15]=='1');
			else if(f=strstr(ligne,"scale="))				Scale=1.0f;
			else if(f=strstr(ligne,"corpse=")) {
				char *nom=strdup(f+7);
				if(strstr(nom,";"))
					*(strstr(nom,";"))=0;
				Corpse=model_manager.get_model(nom);
				free(nom);
				}
			else if(f=strstr(ligne,"windgenerator="))
				WindGenerator=atoi(f+14);
			else if(f=strstr(ligne,"onoffable="))			onoffable=(f[10]=='1');
			else if(f=strstr(ligne,"kamikaze="))			kamikaze=(f[9]=='1');
			else if(f=strstr(ligne,"weapon1=")) {
				char *weaponname=strdup(f+8);
				if(strstr(weaponname,";"))
					*(strstr(weaponname,";"))=0;
				Weapon1=weapon_manager.get_weapon_index(weaponname);
				free(weaponname);
				}
			else if(f=strstr(ligne,"weapon2=")) {
				char *weaponname=strdup(f+8);
				if(strstr(weaponname,";"))
					*(strstr(weaponname,";"))=0;
				Weapon2=weapon_manager.get_weapon_index(weaponname);
				free(weaponname);
				}
			else if(f=strstr(ligne,"weapon3=")) {
				char *weaponname=strdup(f+8);
				if(strstr(weaponname,";"))
					*(strstr(weaponname,";"))=0;
				Weapon3=weapon_manager.get_weapon_index(weaponname);
				free(weaponname);
				}
			else if(f=strstr(ligne,"yardmap="))	{
				f=strstr(dup_ligne,"=");
				if(strstr(f+1,";"))
					*(strstr(f+1,";"))=0;
				while(strstr(f," ")) {
					char *fm=strstr(f," ");
					memmove(fm,fm+1,strlen(fm+1)+1);
					}
				yardmap=strdup(f+1);
				}
			else if(f=strstr(ligne,"cruisealt="))			CruiseAlt=atoi(f+10);
			else if(f=strstr(ligne,"explodeas=")) {
				ExplodeAs=strdup(f+10);
				if(strstr(ExplodeAs,";"))
					*(strstr(ExplodeAs,";"))=0;
				}
			else if(f=strstr(ligne,"selfdestructas=")) {
				SelfDestructAs=strdup(f+15);
				if(strstr(SelfDestructAs,";"))
					*(strstr(SelfDestructAs,";"))=0;
				}
			else if(f=strstr(ligne,"maneuverleashlength="))	ManeuverLeashLength=atoi(f+20);
			else if(f=strstr(ligne,"defaultmissiontype=")) {
				if(strstr(f,"=standby;"))				DefaultMissionType=MISSION_STANDBY;
				else if(strstr(f,"=vtol_standby;"))		DefaultMissionType=MISSION_VTOL_STANDBY;
				else if(strstr(f,"=guard_nomove;"))		DefaultMissionType=MISSION_GUARD_NOMOVE;
				else {
					printf("->constante inconnue : %s\n",f);
					nb_inconnu++;
					}
				}
			else if(f=strstr(ligne,"transmaxunits="))		TransMaxUnits=TransportMaxUnits=atoi(f+14);
			else if(f=strstr(ligne,"transportmaxunits="))	TransMaxUnits=TransportMaxUnits=atoi(f+18);
			else if(f=strstr(ligne,"transportcapacity="))	TransportCapacity=atoi(f+18);
			else if(f=strstr(ligne,"transportsize="))		TransportSize=atoi(f+14);
			else if(f=strstr(ligne,"altfromsealevel="))		AltFromSeaLevel=atoi(f+16);
			else if(f=strstr(ligne,"movementclass=")) {
				if(strstr(f,"kbotss2"))			MovementClass=MOVEMENT_KBOTSS2;
				else if(strstr(f,"boatd3"))		MovementClass=MOVEMENT_BOATD3;
				else if(strstr(f,"tanksh3"))	MovementClass=MOVEMENT_TANKSH3;
				else if(strstr(f,"boatd6"))		MovementClass=MOVEMENT_BOATD6;
				else if(strstr(f,"boats6"))		MovementClass=MOVEMENT_BOATS6;
				else if(strstr(f,"tanksh2"))	MovementClass=MOVEMENT_TANKSH2;
				else if(strstr(f,"spid3"))		MovementClass=MOVEMENT_SPID3;
				else if(strstr(f,"tankds2"))	MovementClass=MOVEMENT_TANKDS2;
				else if(strstr(f,"boats4"))		MovementClass=MOVEMENT_BOATS4;
				else if(strstr(f,"kbotsf2"))	MovementClass=MOVEMENT_KBOTSF2;
				else if(strstr(f,"boats5"))		MovementClass=MOVEMENT_BOATS5;
				else if(strstr(f,"tankbh3"))	MovementClass=MOVEMENT_TANKBH3;
				else if(strstr(f,"tankdh3"))	MovementClass=MOVEMENT_TANKDH3;
				else {
					printf("->constante inconnue : %s\n",f);
					nb_inconnu++;
					}
				}
			else if(f=strstr(ligne,"isairbase="))			IsAirBase=(f[10]=='1');
			else if(f=strstr(ligne,"damagemodifier="))		DamageModifier=atof(f+15);
			else if(f=strstr(ligne,"makesmetal="))			MakesMetal=atof(f+11);
			else if(f=strstr(ligne,"sortbias="))			SortBias=atoi(f+9);
			else if(f=strstr(ligne,"extractsmetal="))		ExtractsMetal=atof(f+14);
			else if(f=strstr(ligne,"hoverattack="))			hoverattack=(f[12]=='1');
			else if(f=strstr(ligne,"isfeature="))			isfeature=(f[10]=='1');
			else if(f=strstr(ligne,"stealth="))				Stealth=atoi(f+8);
			else if(f=strstr(ligne,"attackrunlength="))		attackrunlength=atoi(f+16);
			else if(f=strstr(ligne,"downloadable=")) { }
			else if(f=strstr(ligne,"ovradjust=")) { }
			else if(f=strstr(ligne,"ai_limit=")) { }
			else if(f=strstr(ligne,"ai_weight=")) { }
			if(f==NULL && strstr(ligne,"}")==NULL) {
				Console->AddEntry("inconnu: %s",ligne);
				nb_inconnu++;
				}
			if(dup_ligne)
				free(dup_ligne);
		}while(strstr(ligne,"}")==NULL && nb<1000 && pos<limit);
		delete[] ligne;
		if(Weapon1>-1)	weapon[0]=&(weapon_manager.weapon[Weapon1]);
		if(Weapon2>-1)	weapon[1]=&(weapon_manager.weapon[Weapon2]);
		if(Weapon3>-1)	weapon[2]=&(weapon_manager.weapon[Weapon3]);
		if(Unitname) {
			model=model_manager.get_model(Unitname);
			if(model==NULL)
				Console->AddEntry("%s sans modèle 3D!",Unitname);
			}
		else
			Console->AddEntry("attention: unité sans nom!");
		return nb_inconnu;
	}

	int UNIT_MANAGER::unit_build_menu(int index,int omb)				// Affiche et gère le menu des unités
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		glDisable(GL_TEXTURE_2D);
		glColor4f(0.5f,0.5f,0.5f,0.5f);
		glBegin(GL_QUADS);
			glVertex2f(0.0f,0.0f);			// Barre latérale gauche
			glVertex2f(128.0f,0.0f);
			glVertex2f(128.0f,SCREEN_H);
			glVertex2f(0.0f,SCREEN_H);

			glVertex2f(128.0f,SCREEN_H-32.0f);		// Barre inférieure
			glVertex2f(SCREEN_W,SCREEN_H-32.0f);
			glVertex2f(SCREEN_W,SCREEN_H);
			glVertex2f(128.0f,SCREEN_H);

			glVertex2f(128.0f,0.0f);				// Barre supérieure
			glVertex2f(SCREEN_W,0.0f);
			glVertex2f(SCREEN_W,32.0f);
			glVertex2f(128.0f,32.0f);
		glEnd();
		glColor4f(1.0f,1.0f,1.0f,0.75f);

		glEnable(GL_TEXTURE_2D);

		if(index<0 || index>=nb_unit) return -1;		// L'indice est incorrect

		int page=unit_type[index].page;

		int sel=-1;

		for(int i=6*page;i<min((int)unit_type[index].nb_unit,6*page+6);i++) {		// Affiche les différentes images d'unités constructibles
			if(mouse_x>=(i&1)*64 && mouse_x<(i&1)*64+64 && mouse_y>=128+(i-6*page>>1)*64 && mouse_y<192+(i-6*page>>1)*64) {
				glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
				sel=unit_type[index].BuildList[i];
				if(sel==-1)
					sel=-2;
				}
			else
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			if(unit_type[index].BuildList[i]==-1)					// Construction d'armes
				gfx->drawtexture(unit_type[index].glpic,(i&1)*64,128+(i-6*page>>1)*64,(i&1)*64+64,192+(i-6*page>>1)*64);
			else
				gfx->drawtexture(unit_type[unit_type[index].BuildList[i]].glpic,(i&1)*64,128+(i-6*page>>1)*64,(i&1)*64+64,192+(i-6*page>>1)*64);
			}
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glDisable(GL_BLEND);

		if(unit_type[index].nb_unit>0) {
			bool prec=false,suiv=false;

			if(mouse_b==1 && mouse_y>=320 && mouse_y<=336) {
				if(mouse_x>=5 && mouse_x<62)
					prec=true;
				if(mouse_x>66 && mouse_x<=123)
					suiv=true;
				}

			if(prec && omb!=1)
				unit_type[index].page=(unit_type[index].page+((unit_type[index].nb_unit+5)/6)-1)%((unit_type[index].nb_unit+5)/6);
			if(suiv && omb!=1)
				unit_type[index].page=(unit_type[index].page+1)%((unit_type[index].nb_unit+5)/6);

			glbutton("prec.",5.0f,320.0f,61.5f,336.0f,prec);
			glbutton("suiv.",66.5f,320.0f,123.0f,336.0f,suiv);
			}

		if(sel>-1) {
			set_uformat(U_ASCII);
			glEnable(GL_BLEND);
			glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
			gfx->print(gfx->TA_font,127.0f,SCREEN_H-48,0.0f,0x0,format("%s M:%d E:%d HP:%d",unit_type[sel].name,unit_type[sel].BuildCostMetal,unit_type[sel].BuildCostEnergy,unit_type[sel].MaxDamage));
			gfx->print(gfx->TA_font,129.0f,SCREEN_H-48,0.0f,0x0,format("%s M:%d E:%d HP:%d",unit_type[sel].name,unit_type[sel].BuildCostMetal,unit_type[sel].BuildCostEnergy,unit_type[sel].MaxDamage));
			gfx->print(gfx->TA_font,128.0f,SCREEN_H-49,0.0f,0x0,format("%s M:%d E:%d HP:%d",unit_type[sel].name,unit_type[sel].BuildCostMetal,unit_type[sel].BuildCostEnergy,unit_type[sel].MaxDamage));
			gfx->print(gfx->TA_font,128.0f,SCREEN_H-47,0.0f,0x0,format("%s M:%d E:%d HP:%d",unit_type[sel].name,unit_type[sel].BuildCostMetal,unit_type[sel].BuildCostEnergy,unit_type[sel].MaxDamage));
			glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
			gfx->print(gfx->TA_font,128.0f,SCREEN_H-48,0.0f,0xFFFFFF,format("%s M:%d E:%d HP:%d",unit_type[sel].name,unit_type[sel].BuildCostMetal,unit_type[sel].BuildCostEnergy,unit_type[sel].MaxDamage));
			if(unit_type[sel].Description) {
				glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
				gfx->print(gfx->TA_font,127.0f,SCREEN_H-40,0.0f,0x0,format("%s",unit_type[sel].Description));
				gfx->print(gfx->TA_font,129.0f,SCREEN_H-40,0.0f,0x0,format("%s",unit_type[sel].Description));
				gfx->print(gfx->TA_font,128.0f,SCREEN_H-41,0.0f,0x0,format("%s",unit_type[sel].Description));
				gfx->print(gfx->TA_font,128.0f,SCREEN_H-39,0.0f,0x0,format("%s",unit_type[sel].Description));
				glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
				gfx->print(gfx->TA_font,128.0f,SCREEN_H-40,0.0f,0xFFFFFF,format("%s",unit_type[sel].Description));
				}
			glDisable(GL_BLEND);
			set_uformat(U_UTF8);
			}

		return sel;
	}

int load_all_units()
{
	unit_manager.init();
	int nb_inconnu=0;
	List<String> file_list;
	uint32 nb_file = HPIManager->GetFilelist("units\\*.fbi",&file_list);

	for(List<String>::iterator i=file_list.begin();i!=file_list.end();i++) {
		char *nom=strdup(strstr(i->c_str(),"\\")+1);			// Vérifie si l'unité n'est pas déjà chargée
		*(strstr(nom,"."))=0;
		strupr(nom);

		if(unit_manager.get_unit_index(nom)==-1) {
			uint32 file_size=0;
			byte *data=HPIManager->PullFromHPI(*i,&file_size);
			nb_inconnu+=unit_manager.load_unit(data,file_size);
			if(unit_manager.unit_type[unit_manager.nb_unit-1].Unitname) {
				String nom_pcx = String("unitpics\\")+String(unit_manager.unit_type[unit_manager.nb_unit-1].Unitname)+String(".pcx");
				byte *dat=HPIManager->PullFromHPI(nom_pcx);
				if(dat) {
					unit_manager.unit_type[unit_manager.nb_unit-1].unitpic=load_memory_pcx(dat,pal);
					if(unit_manager.unit_type[unit_manager.nb_unit-1].unitpic) {
						allegro_gl_use_alpha_channel(false);
						if(g_useTextureCompression)
							allegro_gl_set_texture_format(GL_COMPRESSED_RGB_ARB);
						else
							allegro_gl_set_texture_format(-1);
						unit_manager.unit_type[unit_manager.nb_unit-1].glpic=allegro_gl_make_texture(unit_manager.unit_type[unit_manager.nb_unit-1].unitpic);
						glBindTexture(GL_TEXTURE_2D,unit_manager.unit_type[unit_manager.nb_unit-1].glpic);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
						}
					free(dat);
					}
				else
					unit_manager.unit_type[unit_manager.nb_unit-1].unitpic=NULL;
				}
			free(data);
			}
		free(nom);
		allegro_gl_flip();
		}

	for(uint32 i=0;i<unit_manager.nb_unit;i++)
		unit_manager.load_script_file(unit_manager.unit_type[i].Unitname);

	unit_manager.Identify();

//	unit_manager.load_sound_file();

	unit_manager.gather_all_build_data();

	return nb_inconnu;
}
