/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------\
|                                     ia.cpp                                   |
|       Ce module est responsable de l'intelligence artificielle               |
|                                                                              |
\-----------------------------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"					// Pour la lecture des fichiers 3D
#include "cob.h"					// Pour la lecture et l'éxecution des scripts
#include "tdf.h"					// Pour la gestion des éléments du jeu
#include "EngineClass.h"			// Inclus le moteur
#include "UnitEngine.h"				// Inclus le moteur
//#include "ia.h"

void BRAIN::save(FILE *file)		// Enregistre le réseau de neurones
{
	fwrite("BRAIN",5,1,file);			// Identifiant du format du fichier
	fwrite(&n,sizeof(int),1,file);		// Nombre d'entrées
	fwrite(&p,sizeof(int),1,file);		// Nombre de sorties
	fwrite(&q,sizeof(int),1,file);		// Taille de la couche intermédiaire

	for(int i=n;i<nb_neuron;i++)		// Enregistre les poids
		if(i<n+q)
			fwrite(neuron[i].weight,sizeof(float)*n,1,file);
		else
			fwrite(neuron[i].weight,sizeof(float)*q,1,file);
}

int BRAIN::load(FILE *file)		// Charge le réseau de neurones
{
	char tmp[6];

	fread(tmp,5,1,file);				// Identifiant du format du fichier
	tmp[5]=0;
	if(strcmp(tmp,"BRAIN")!=0)	{		// Vérifie si le fichier est bien du type attendu
		return 1;
		}

	destroy();		// Au cas où

	fread(&n,sizeof(int),1,file);		// Nombre d'entrées
	fread(&p,sizeof(int),1,file);		// Nombre de sorties
	fread(&q,sizeof(int),1,file);		// Taille de la couche intermédiaire
	nb_neuron=p+q+n;

	neuron=(NEURON*) malloc(sizeof(NEURON)*nb_neuron);
	n_out=(float*) malloc(sizeof(float)*p);

	for(int i=0;i<p;i++)
		n_out[i]=0.0f;

	for(int i=0;i<n;i++)
		neuron[i].weight=NULL;

	for(int i=n;i<nb_neuron;i++)		// Lit les poids
		if(i<n+q) {
			neuron[i].weight=(float*) malloc(sizeof(float)*n);
			fread(neuron[i].weight,sizeof(float)*n,1,file);
			}
		else {
			neuron[i].weight=(float*) malloc(sizeof(float)*q);
			fread(neuron[i].weight,sizeof(float)*q,1,file);
			}

	return 0;
}

BRAIN *copy_brain(BRAIN *brain,BRAIN *dst)			// Copie un réseau de neurones
{
	BRAIN *copy=dst;
	if(copy==NULL)
		copy=(BRAIN*) malloc(sizeof(BRAIN));
	copy->init();
	copy->nb_neuron=brain->nb_neuron;
	copy->n=brain->n;
	copy->p=brain->p;
	copy->q=brain->q;
	copy->neuron=(NEURON*) malloc(sizeof(NEURON)*copy->nb_neuron);
	copy->n_out=(float*) malloc(sizeof(float)*copy->p);
	for(int i=0;i<brain->nb_neuron;i++) {
		if(i<brain->p)
			copy->n_out[i]=0.0f;
		if(brain->neuron[i].weight==NULL)
			copy->neuron[i].weight=NULL;
		else {
			if(i>=copy->n && i<copy->nb_neuron-copy->p) {
				copy->neuron[i].weight=(float*) malloc(sizeof(float)*brain->q);
				for(int e=0;e<brain->q;e++)
					copy->neuron[i].weight[e]=brain->neuron[i].weight[e];
				}
			else if(i>=copy->n) {
				copy->neuron[i].weight=(float*) malloc(sizeof(float)*brain->n);
				for(int e=0;e<brain->n;e++)
					copy->neuron[i].weight[e]=brain->neuron[i].weight[e];
				}
			else
				copy->neuron[i].weight=NULL;
			}
		}
	return copy;
}

inline byte int2brain_value(int a)			// Evaluateur de quantité pour réseau de neurones
{
	if(a==0)	return BRAIN_VALUE_NULL;
	if(a<=5)	return BRAIN_VALUE_LOW;
	if(a<=15)	return BRAIN_VALUE_MEDIUM;
	if(a<=50)	return BRAIN_VALUE_HIGH;
	return BRAIN_VALUE_MAX;
}

inline int get_bits(float bits[],byte v,int pos)	// Remplit le tableau de bits (entrée des réseaux de neurones)
{
	for(int i=0;i<BRAIN_VALUE_BITS;i++)
		bits[pos++]=((v>>i)&1) ? 1.0f : 0.0f;
	return pos;
}

void AI_PLAYER::scan_unit()
{
}

void AI_PLAYER::refresh_unit_weights()
{
}

void AI_PLAYER::think(MAP *map)				// La vrai fonction qui simule l'Intelligence Artificielle
{
	float Conv = 1.0f / SECS_TO_TIMER(1);

	if( first_refresh ) {	first_refresh = false;	last_refresh = Atimer;	}

	if( (Atimer - last_refresh) * Conv >= 10.0f ) {
		last_refresh = Atimer;
		}
	return;							// Shortcut to prevent execution of this function because AI will be finished later
}
