/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------\
|                                      ia.h                                    |
|       Ce module est responsable de l'intelligence artificielle               |
|                                                                              |
\-----------------------------------------------------------------------------*/

#ifndef ARTIFICIAL_INTELLIGENCE
#define ARTIFICIAL_INTELLIGENCE

struct NEURON
{
	float	var;			// Variable pour les opérations
	float	*weight;		// Poids des différents neurones sources
};

class BRAIN		// Réseau de NEURON à n entrées et p sorties
{
public:
	int		nb_neuron;		// Nombre de NEURON
	NEURON	*neuron;		// Tableau de NEURON
	int		n;				// Nombre d'entrées dans le réseau
	int		p;				// Nombre de sorties du réseau
	int		q;				// Taille des rangs
	float	*n_out;			// Tableau de sortie

	inline void init()
	{
		nb_neuron=0;
		neuron=NULL;
		n=p=q=0;
		n_out=NULL;
	}

	inline void destroy()
	{
		if(nb_neuron>0 && neuron) {
			for(int i=0;i<nb_neuron;i++)
				if(neuron[i].weight!=NULL)
					free(neuron[i].weight);
			free(neuron);
			}
		if(n_out)	free(n_out);
		init();
	}

	BRAIN()
	{
		init();
	}

	~BRAIN()
	{
		destroy();
	}

	void build(int nb_in,int nb_out,int rg)				// Crée le réseau de neurones
	{
		destroy();
		q=rg;
		nb_neuron=q+nb_in+nb_out;		// Nombre de couches x nombre d'entrées + nombre de sorties
		n=nb_in;
		p=nb_out;
		neuron=(NEURON*) malloc(sizeof(NEURON)*nb_neuron);
		n_out=(float*) malloc(sizeof(float)*nb_out);
		for(int i=0;i<nb_neuron;i++) {
			neuron[i].var=0.0f;
			neuron[i].weight=NULL;
			if(i<nb_out)
				n_out[i]=0.0f;
			if(i>=n && i<nb_neuron-p) {
				neuron[i].weight=(float*) malloc(sizeof(float)*n);
				for(int e=0;e<n;e++)
					neuron[i].weight[e]=(rand()%2001)*0.001f-1.0f;
				}
			else if(i>=n) {
				neuron[i].weight=(float*) malloc(sizeof(float)*q);
				for(int e=0;e<q;e++)
					neuron[i].weight[e]=(rand()%2001)*0.001f-1.0f;
				}
			}
	}

	inline void active_neuron(int i)
	{
		if(neuron[i].weight==NULL)	return;
		if(i<n)
			return;
		neuron[i].var=0.0f;
		if(i<nb_neuron-p)
			for(int e=0;e<n;e++)
				neuron[i].var+=neuron[e].var*neuron[i].weight[e];
		else
			for(int e=0;e<q;e++)
				neuron[i].var+=neuron[n+e].var*neuron[i].weight[e];
		neuron[i].var=1.0f/(1.0f+exp(-neuron[i].var));
	}

	inline float *work(float entry[],bool seuil=false)			// Fait bosser un peu le réseau de NEURON et renvoie les valeurs calculées par un NEURON
	{
		if(nb_neuron<0)	return NULL;		// Pas de NEURON à faire bosser
		int i;
		for(i=0;i<n;i++)		// Prépare le réseau au calcul
			neuron[i].var=entry[i];
		for(i=n;i<nb_neuron;i++)
			active_neuron(i);
		if(!seuil)
			for(i=0;i<p;i++)		// Récupère le résultat du calcul
				n_out[i]=neuron[n+q+i].var;
		else
			for(i=0;i<p;i++)		// Récupère le résultat du calcul
				n_out[i]=neuron[n+q+i].var>=0.5f ? 1.0f : 0.0f;
		return n_out;
	}

	inline void mutation()			// Déclenche une mutation dans le réseau de neurones
	{
		int index=(rand()%(nb_neuron-n))+n;
		int mod_w=0;
		if(index<nb_neuron-p)	mod_w=rand()%n;
		else mod_w=rand()%q;
		neuron[index].weight[mod_w]+=((rand()%200001)-100000)*0.00001f;
	}

	inline void learn(float *result,float coef=1.0f)		// Corrige les défauts
	{
		for(int i=0;i<p;i++)
			n_out[i]=(result[i]-n_out[i])*(n_out[i]+0.01f)*(1.01f-n_out[i]);

		float *diff = new float[q];
		for(int i=0;i<q;i++)
			diff[i]=0.0f;

		for(int i=0;i<p;i++)		// Neurones de sortie
			for(int e=0;e<q;e++) {
				diff[e]+=(n_out[i]+0.01f)*(1.01f-n_out[i])*neuron[n+q+i].weight[e]*neuron[n+e].var;
				neuron[n+q+i].weight[e]+=coef*n_out[i]*neuron[n+e].var;
				}

									// Neurones des couches intermédiaires
		for(int i=0;i<q;i++)
			for(int e=0;e<n;e++)
				neuron[n+i].weight[e]+=coef*diff[i]*neuron[e].var;
		delete[] diff;
	}

	void save(FILE *file);		// Enregistre le réseau de neurones

	int load(FILE *file);		// Charge le réseau de neurones
};

BRAIN *copy_brain(BRAIN *brain,BRAIN *dst=NULL);		// Copie un réseau de neurones

#define	ORDER_ATTACK		0x0			// Order units to attack
#define	ORDER_ARMY			0x1			// Order to build an army
#define	ORDER_METAL			0x2			// Order to gather metal
#define	ORDER_ENERGY		0x3			// Order to gather energy
#define	ORDER_DEFENSE		0x4			// Order to build defensive units ( AA towers, but also atomic weapons ...)
#define	ORDER_FACTORY		0x5			// Order to build factories
#define	ORDER_BUILDER		0x6			// Order to build construction units

#define NB_ORDERS			0x5

#define BRAIN_VALUE_NULL	0x0
#define BRAIN_VALUE_LOW		0x1
#define BRAIN_VALUE_MEDIUM	0x2
#define BRAIN_VALUE_HIGH	0x4
#define BRAIN_VALUE_MAX		0x8

#define BRAIN_VALUE_BITS	0x4			// Nombre de bits nécessaires pour coder une valeur pour un réseau de neurones

struct AI_WEIGHT
{
	uint32	idx;					// ID of the unit
	float	w;						// Weight given to this unit
};

#define	AI_TYPE_EASY		0x0
#define AI_TYPE_MEDIUM		0x1
#define AI_TYPE_HARD		0x2
#define AI_TYPE_BLOODY		0x3

class AI_PLAYER					// Classe gérant les joueurs controlés par l'IA
{
public:
	char		*name;			// Attention faudrait pas qu'il se prenne pour quelqu'un!! -> indique aussi le fichier correspondant à l'IA (faut sauvegarder les cervelles)
	BRAIN		decider;		// Réseau de neurones d'analyse de la partie et de décision
	BRAIN		anticiper;		// Réseau de neurones voué à l'analyse et à l'anticipation des mouvements ennemis
	byte		player_id;		// Identifiant du joueur
	int			unit_id;		// Identifiant d'unité pour parcourir les unités

	byte		AI_type;		// Which AI do we have to use?

	std::vector< AI_WEIGHT >	weights;	// Vector of weights used to decide what to build

	float		order_weight[NB_ORDERS];	// weights of orders

	uint32		last_refresh;
	bool		first_refresh;

	inline void init()
	{
		name=strdup("default ai");
		decider.init();
		anticiper.init();
		player_id = 0;
		unit_id = 0;
		AI_type = AI_TYPE_EASY;

		weights.clear();

		order_weight[ORDER_ATTACK] = 0.0f;		// That's not a commander rush!
		order_weight[ORDER_ARMY] = 0.5f;
		order_weight[ORDER_METAL] = 2.0f;
		order_weight[ORDER_ENERGY] = 3.0f;
		order_weight[ORDER_DEFENSE] = 0.0f;
		order_weight[ORDER_FACTORY] = 1.0f;
		order_weight[ORDER_BUILDER] = 5.0f;

		last_refresh = Atimer;
		first_refresh = true;
	}

	inline void destroy()
	{
		if(name)	free(name);
		name = NULL;
		decider.destroy();
		anticiper.destroy();
		player_id=0;
		unit_id=0;
		weights.clear();
	}

	AI_PLAYER()
	{
		init();
	}

	~AI_PLAYER()
	{
		destroy();
	}

	inline void change_name(char *new_name)		// Change le nom de l'IA (conduit à la création d'un nouveau fichier)
	{
		if(name)
			free(name);
		name=strdup(new_name);
	}

	void save()
	{
		String filename = format( "ai/%s.ai", name );
		FILE *file=TA3D_OpenFile(filename,"wb");

		byte l=(byte)strlen(name);
		fwrite(&l,1,1,file);		// Nom de l'IA
		fwrite(name,l,1,file);

		decider.save(file);			// Réseau de décision

		anticiper.save(file);		// Réseau d'analyse

		fclose(file);
	}

	void load(char *filename,int id=0)
	{
		FILE *file=TA3D_OpenFile(filename,"rb");

		byte l;
		fread(&l,1,1,file);
		if(name)
			free(name);
		name=(char*) malloc(l+1);
		name[l]=0;
		fread(name,l,1,file);

		decider.load(file);

		anticiper.load(file);

		fclose(file);

		player_id=id;
	}

	void scan_unit();

	void refresh_unit_weights();

	void think(MAP *map);
};

#endif
