/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*
**  File: main.cpp
** Notes: The applications main entry point. 
*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif

#include "stdafx.h"					// standard pch inheritance.
#include "TA3D_NameSpace.h"			// our namespace, a MUST have.
#include "intro.h"					// intro prototypes,   TODO: phase out.
#include "menu.h"					// menu prototypes      TODO: phase out.
#include "cTA3D_Engine.h"			// The engine class.
#include "ta3dbase.h"				// Just for the LANG var

// below defination defines accuracy of timer i guess.
#define precision   MSEC_TO_TIMER(1)

// timer variable, incremented by preccission, set by main.
volatile int Atimer;
extern int start;     // for now.

using namespace TA3D::EXCEPTION;
// intrupt timer, driven by allgegro.
void Timer()
{
	Atimer+=precision;
}
END_OF_FUNCTION(Timer); /* I guess allegro needs this. */


/*
** Function: LoadConfigFile
**    Notes: This function will eventually load our config file if it exists.
**             config files will be stored as 'tdf' format and thus loaded as text,
**             using the cTAFileParser class.
**           If something goes wrong you can safely throw a string for an error.
**             The call to this function is tried, but it only catches exceptions
**             and strings, ie throw( "LoadConfigFile: some error occured" );
*/
void LoadConfigFile( void )
{
	GuardEnter( LoadConfigFile );

	cTAFileParser *cfgFile;

	try { // we need to try catch this cause the config file may not exists
		 // and if it don't exists it will throw an error on reading it, which
		 // will be caught in our main function and the application will exit.
		cfgFile = new TA3D::UTILS::cTAFileParser(  "ta3d.cfg" );
	}
	catch( ... )
	{
		GuardLeave();
		return;
	}

	TA3D::VARS::lp_CONFIG->fps_limit = cfgFile->PullAsFloat( "TA3D.FPS Limit" );
	TA3D::VARS::lp_CONFIG->shadow_r  = cfgFile->PullAsFloat( "TA3D.Shadow R" );
	TA3D::VARS::lp_CONFIG->timefactor = cfgFile->PullAsFloat( "TA3D.Time Factor" );

	TA3D::VARS::lp_CONFIG->shadow_quality = cfgFile->PullAsInt( "TA3D.Shadow Quality" );
	TA3D::VARS::lp_CONFIG->priority_level = cfgFile->PullAsInt( "TA3D.Priority Level" );
	TA3D::VARS::lp_CONFIG->fsaa = cfgFile->PullAsInt( "TA3D.FSAA" );
	TA3D::VARS::lp_CONFIG->Lang = cfgFile->PullAsInt( "TA3D.Language" );
	TA3D::VARS::lp_CONFIG->water_quality = cfgFile->PullAsInt( "TA3D.Water Quality" );
	TA3D::VARS::lp_CONFIG->screen_width = cfgFile->PullAsInt( "TA3D.Screen Width" );
	TA3D::VARS::lp_CONFIG->screen_height = cfgFile->PullAsInt( "TA3D.Screen Height" );

	TA3D::VARS::lp_CONFIG->showfps = cfgFile->PullAsBool( "TA3D.Show FPS" );
	TA3D::VARS::lp_CONFIG->wireframe = cfgFile->PullAsBool( "TA3D.Show Wireframe" );
	TA3D::VARS::lp_CONFIG->particle = cfgFile->PullAsBool( "TA3D.Show particles" );
	TA3D::VARS::lp_CONFIG->trees = cfgFile->PullAsBool( "TA3D.Show Trees" );
	TA3D::VARS::lp_CONFIG->shadow = cfgFile->PullAsBool( "TA3D.Show Shadows" );
	TA3D::VARS::lp_CONFIG->height_line = cfgFile->PullAsBool( "TA3D.Show Height Lines" );
	TA3D::VARS::lp_CONFIG->fullscreen = cfgFile->PullAsBool( "TA3D.Show FullScreen" );

	delete cfgFile; 

	LANG = lp_CONFIG->Lang;

	GuardLeave();
}

/*
** Function: SaveConfigFile
**    Notes: Upon application exit this will write out our config file.
**             note that if the application fails to startup then a config file
**             will never be generated. See LoadConfigFile for notes on format.
**           If something goes wrong you can safely throw a string for an error.
**             The call to this function is tried, but it only catches exceptions
**             and strings, ie throw( "LoadConfigFile: some error occured" );
*/
void SaveConfigFile( void )
{
	GuardEnter( SaveConfigFile );
	lp_CONFIG->Lang = LANG;

	if( !TA3D::VARS::lp_CONFIG )
	{
		GuardLeave();
		return;
	}
	String FileName = GetClientPath() + "ta3d.cfg";
	std::ofstream   m_File;

	m_File.open( FileName.c_str(), std::ios::out | std::ios::trunc );

	if( !m_File.is_open() )
	{
		GuardLeave();
		return;
	}

	m_File << "[TA3D]\n{\n";
	m_File << "        FPS Limit=" << TA3D::VARS::lp_CONFIG->fps_limit << ";\n";
	m_File << "         Shadow R=" << TA3D::VARS::lp_CONFIG->shadow_r << ";\n";
	m_File << "      Time Factor=" << TA3D::VARS::lp_CONFIG->timefactor << ";\n";
	m_File << "   Shadow Quality=" << TA3D::VARS::lp_CONFIG->shadow_quality << ";// 0->100\n";
	m_File << "   Priority Level=" << TA3D::VARS::lp_CONFIG->priority_level << ";// 0, 1, 2\n";
	m_File << "             FSAA=" << TA3D::VARS::lp_CONFIG->fsaa << ";\n";
	m_File << "         Language=" << TA3D::VARS::lp_CONFIG->Lang << ";\n";
	m_File << "    Water Quality=" << TA3D::VARS::lp_CONFIG->water_quality << ";//0->3\n";
	m_File << "     Screen Width=" << TA3D::VARS::lp_CONFIG->screen_width << ";\n";
	m_File << "    Screen Height=" << TA3D::VARS::lp_CONFIG->screen_height << ";\n";
	m_File << "         Show FPS=" << TA3D::VARS::lp_CONFIG->showfps << ";\n";
	m_File << "   Show Wireframe=" << TA3D::VARS::lp_CONFIG->wireframe << ";\n";
	m_File << "   Show particles=" << TA3D::VARS::lp_CONFIG->particle << ";\n";
	m_File << "       Show Trees=" << TA3D::VARS::lp_CONFIG->trees << ";\n";
	m_File << "     Show Shadows=" << TA3D::VARS::lp_CONFIG->shadow << ";\n";
	m_File << " Show Height Lines=" << TA3D::VARS::lp_CONFIG->height_line << ";\n";
	m_File << "   Show FullScreen=" << TA3D::VARS::lp_CONFIG->fullscreen << ";\n";
	m_File << "}\n";

	m_File.flush();
	m_File.close();

	GuardLeave();
}

/*
** Function: ParseCommandLine
**    Notes: this will eventually break down any command line arguments passed to
**              the application at run time.  It don't do anything yet, but eventually
**              we will be adding lots of command parms.
**           If something goes wrong you can safely throw a string for an error.
**             The call to this function is tried, but it only catches exceptions
**             and strings, ie throw( "LoadConfigFile: some error occured" );
**           Remember if you throw an error, or generate one, you are responsible for
**             cleaning up what you initialized!
*/
void ParseCommandLine( int argc, char *argv[] )
{
	GuardEnter( ParseCommandLine );
	GuardLeave();
}

/*
** Function: main
**    Notes: Whats this for anyhow? :)  Just kidding, this is where it all begin baby!
*/
int main(int argc,char *argv[])
{
	GuardStart( main ); // start guard.
		GuardInfo( "Constructing config.\n" ); // what we are doing.
		TA3D::VARS::lp_CONFIG = new TA3D::TA3DCONFIG;
	GuardCatch();      // close guard
	if( IsExceptionInProgress() ) // if guard threw an error this will be true.
	{
		GuardDisplayAndLogError();   // record and display the error.
		exit(1);                      // we outa here.
	}

	GuardStart( main );
		LoadConfigFile(); /* Load Config File */
	GuardCatch();
	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();

		delete TA3D::VARS::lp_CONFIG;
		exit(1);
	}

	GuardStart( main );
		ParseCommandLine( argc, argv );   /* process any command line args passed */
	GuardCatch();
	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();

		delete TA3D::VARS::lp_CONFIG;
		exit(1);
	}


	/*
	** Allright we have now created our config structure, loaded our config file
	**   and processed our command line arguments, its time to create the main engine
	**   which in turn will create nearly everything it will need.
	*/
	TA3D::cTA3D_Engine *Engine;    // class pointer to our engine.
	GuardStart( main );
		Engine = new TA3D::cTA3D_Engine;  // Create and intialize engine.
	GuardCatch();
	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();

		delete TA3D::VARS::lp_CONFIG;
		exit(1);
	}

	/* we can use guard here cause the entine will start in a 'paused' state
	**     it will continue to init code and set itself as ready when all data it needs
	**     to continue is loaded.  at which point we can poll the engine.
	**     Once the engine is unpaused and in a go state we can no longer use GuardStart
	**     as the engine will use it, until the engine exits.
	*/
	GuardStart( main );
		Engine->Start();
	GuardCatch();
	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();
		// We need to guard deleting the engine in case something bad goes wrong during
		//   the cleanup process.
		GuardStart( main );
			delete Engine;
			delete TA3D::VARS::lp_CONFIG;
		GuardCatch();

		exit(1);
	}

	// ok, if we are here, our thread in engine class is running
	//   and doing some work loading up alot of crap so while its doing that
	//   we are going to show our intro, but first we need to start our timer.
	install_int_ex( Timer, precision);
	start=Atimer;      // Initalize timer.

	// while our engine does some loading and intlaizing, lets show our intro.
	GuardStart( main );
		play_intro();
	GuardCatch();
	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();
		// We need to guard deleting the engine in case something bad goes wrong during
		//   the cleanup process, note that this might not exit right away because it
		//   might take a few seconds to kill the thread.
		GuardStart( main );
			delete Engine;
			delete TA3D::VARS::lp_CONFIG;
		GuardCatch();

		exit(1);
	}

	// The main menu call will eventually not be here, instead
	//   we will turn control over to our engine, but for now we'll call the
	//   menu this way.
	// Engine->UnPause(); // unpause the engine, the engine will finish initing
	// While( !Engine->Ready() ) { rest( 100 ); } // while its not ready rest a bit.
	//  engines thread should now be running and managing everything for us.
	//  we will continue to keep checking if its running and if so call a helper function
	//  in the engine to do some non-cricticle stuff, every 100 ms or so.
	// While( Engine-> IsRunning() ) { Engine->DoSomeNonImportantShit(); rest( 100 ); }
	main_menu();

	// if we get here its time to exit, so delete the engine, the engine should clean itself
	//   up and we are outa here, but first lets save our config file, we
	//   need to try/catch this but no worries for now since its not doing anything.
	GuardStart( main );
		SaveConfigFile();
	GuardCatch();

	if( IsExceptionInProgress() )
	{
		GuardDisplayAndLogError();
		GuardStart( main );
			delete Engine;
			delete TA3D::VARS::lp_CONFIG;
		GuardCatch();

		exit(1);
	}

	GuardStart( main )
		delete Engine;
	GuardCatch();

	// if something bad happens log and show the error but don't exit
	//    so that we can delete the config var next.
	if( IsExceptionInProgress() )
		GuardDisplayAndLogError();

	GuardStart( main )
		delete TA3D::VARS::lp_CONFIG;
	GuardCatch();

	return 0; // thats it folks.
}
END_OF_MAIN();
/* Allegro needs END_OF_MAIN(); I guess. */ 
