/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------\
|                                    menu.cpp                                  |
|       Ce module contient les routines du menu de TA3D                        |
|                                                                              |
\-----------------------------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "3do.h"					// Pour la lecture des fichiers 3D
#include "cob.h"					// Pour la lecture et l'éxecution des scripts
//#include "fbi.h"					// Pour la gestion des unités
#include "EngineClass.h"			// Inclus le moteur
#include "tnt.h"					// Inclus le chargeur de cartes
#include "menu.h"
#include "gui.h"
#include "taconfig.h"

using namespace TA3D::EXCEPTION;

#define size	20.0f

void main_menu(void)
{
	GuardEnter( main_menu );

	GLuint tex_formatA,tex_format;
	if(g_useTextureCompression) {
		tex_formatA=GL_COMPRESSED_RGBA_ARB;
		tex_format=GL_COMPRESSED_RGB_ARB;
		}
	else {
		tex_formatA=GL_RGBA8;
		tex_format=GL_RGB8;
		}

	bool done=false;

	gfx->set_2D_mode();

	gfx->ReInitTexSys();

	glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);

	int i;

	float dt=0.0f;
	int time=Atimer;
	float Conv=1.0f/SECS_TO_TIMER(1);

	float h=gfx->TA_font.height();

	BITMAP *tst[3];
	GLuint mnu[3];
	GLuint flame = gfx->load_masked_texture( "gfx/fire.tga","gfx/firea.tga", FILTER_LINEAR );
	mnu[0] = gfx->load_texture( "gfx/exit.tga", FILTER_LINEAR );
	mnu[1] = gfx->load_texture( "gfx/options.tga", FILTER_LINEAR );
	mnu[2] = gfx->load_texture( "gfx/play.tga", FILTER_LINEAR );
	tst[0] = load_bitmap("gfx/exit.tga",NULL);
	tst[1] = load_bitmap("gfx/options.tga",NULL);
	tst[2] = load_bitmap("gfx/play.tga",NULL);

	int px[3]={	320,	160,	480 };
	int py[3]={	400,	300,	300 };

	VECTOR part[100];
	for(i=0;i<100;i++) {
		part[i].x=-size;	part[i].y=-size;
		part[i].z=0.0f;
		}

	int n=0;

	cursor_type=CURSOR_DEFAULT;

	do
	{
		do
		{
			dt=(Atimer-time)*Conv;
			rest(1);
		}while(dt<0.01f);
		time=Atimer;

		if(key[KEY_ESC])	done=true;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		gfx->drawtexture(gfx->glfond,0.0f,0.0f,640.0f,480.0);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,320.0f-gfx->TA_font.length(TA3D_ENGINE_VERSION)*0.5f,448.0f,0.0f,0xFFFFFFFF,TA3D_ENGINE_VERSION);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		int index=-1;
		for(i=0;i<3;i++)
			for(int e=0;e<10;e+=2)
				if(px[i]-120-e<mouse_x*gfx->SCREEN_W_TO_640 && px[i]+120+e>mouse_x*gfx->SCREEN_W_TO_640 && py[i]-40-e<mouse_y*gfx->SCREEN_H_TO_480 && py[i]+40+e>mouse_y*gfx->SCREEN_H_TO_480) {
					index=i;
					i=3;
					e=10;
					}

		for(i=0;i<3;i++) {
			if(i==index)
				glBlendFunc(GL_ONE,GL_ONE);
			else
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			for(int e=0;e<10;e+=2)
				gfx->drawtexture(mnu[i],px[i]-120-e,py[i]-40-e,px[i]+120+e,py[i]+40+e);
			}
		glColor4f(1.0f,1.0f,1.0f,1.0f);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		switch(index)
		{
		case -1:
			part[n%100].x=-size;
			part[n%100].y=-size;
			break;
		case 0:
		case 1:
		case 2:
			part[n%100].x=px[index]+120.0f*cos(n*0.15f)+((rand()%2001)-1000)*0.01f;
			part[n%100].y=py[index]+40.0f*cos(n*0.255f)+((rand()%2001)-1000)*0.01f;
			break;
		};

		for(i=0;i<100;i++) {
			int d=100-((n-i)%100);
			glColor4f(d*0.01f,d*0.01f,d*0.01f,d*0.01f);
			glPushMatrix();
			glTranslatef(part[i].x,part[i].y,0.0f);
			glRotatef((i&1) ? n : -n,0.0f,0.0f,1.0f);
			glTranslatef(-part[i].x,-part[i].y,0.0f);
			gfx->drawtexture(flame,part[i].x-size,part[i].y-size,part[i].x+size,part[i].y+size);
			glPopMatrix();
			}
		n++;
		glColor4f(1.0f,1.0f,1.0f,1.0f);

		glDisable(GL_BLEND);

		draw_cursor();

		if(mouse_b) {
			gfx->SCREEN_W_TO_640 = 1.0f;				// To have mouse sensibility undependent from the resolution
			gfx->SCREEN_H_TO_480 = 1.0f;
			switch(index)
			{
			case 0:
				done=true;
				break;
			case 1:
				config_menu();
				break;
			case 2:
				setup_game();
				break;
			};
			gfx->SCREEN_W_TO_640 = 640.0f / SCREEN_W;				// To have mouse sensibility undependent from the resolution
			gfx->SCREEN_H_TO_480 = 480.0f / SCREEN_H;
			cursor_type=CURSOR_DEFAULT;		// Curseur par standard
			}

		gfx->flip();
	}while(!done);

	destroy_bitmap(tst[0]);
	destroy_bitmap(tst[1]);
	destroy_bitmap(tst[2]);

	gfx->set_2D_mode();

	glDeleteTextures(1,&flame);
	glDeleteTextures(1,&mnu[0]);
	glDeleteTextures(1,&mnu[1]);
	glDeleteTextures(1,&mnu[2]);

	GuardLeave();
}

uint32 GetMultiPlayerMapList( std::list<std::string> *li )
{
	std::list< String > map_list;
	uint32 n = HPIManager->GetFilelist("maps\\*.tnt",&map_list);
	std::list< String >::iterator i_map;
	uint32 count;

	if( n < 1 )
		return 0;

	cTAFileParser *MapData;
	bool isNetworkGame;
	count = 0;

	for( i_map=map_list.begin(); i_map!=map_list.end(); i_map++ )
	{
		*i_map=i_map->substr(5,i_map->length()-9);
		MapData = new cTAFileParser( String( "maps\\" ) + *i_map + String( ".ota" ) );

		sint32 SCount = MapData->PullAsInt( "GlobalHeader.SCHEMACOUNT" );
		isNetworkGame = false;
		for( sint32 i = 0; i < SCount; i++ )
		{
			String KeyName = format( "GlobalHeader.Schema %d.Type", i );
			String KeyVal = MapData->PullAsString( KeyName );
			if( StartsWith( KeyVal, "Network" ) )
			{
				isNetworkGame = true;
				break;
			}
		}

		delete MapData;
		MapData = NULL;

		if( isNetworkGame )
		{
			li->push_back( *i_map );
			count++;
		}
	}

	li->sort();
	return count;
} 

char *select_map(String *def_choice)		// Cette fonction affiche un menu permettant à l'utilisateur de choisir une carte dans une liste et de la prévisualiser à l'écran
{
	gfx->SCREEN_W_TO_640 = 640.0f / SCREEN_W;				// To have mouse sensibility undependent from the resolution
	gfx->SCREEN_H_TO_480 = 480.0f / SCREEN_H;

	bool done=false;

	gfx->set_2D_mode();

	gfx->ReInitTexSys();

	set_uformat(U_ASCII);		// Juste histoire d'avoir un affichage correct des textes

	glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);

	int i;

	float dt=0.0f;
	int time=Atimer;
	float Conv=1.0f/SECS_TO_TIMER(1);

	float h=gfx->TA_font.height();
	float data_h = 24.0f;

	List< String > map_list;
	uint32 n = GetMultiPlayerMapList( &map_list );
	List< String >::iterator i_map;

	if(n==0)	{
		Popup(translation_manager.get(48),translation_manager.get(49));
		Console->AddEntry("no maps found!!");
		reset_mouse();
		return NULL;
		}

	char *choice=NULL;

	int dec=0;			// Pour la gestion de la liste déroulante
	int amz=mouse_z;
	int o_sel=-1;
	int sel_index=0;
	int index=-1;
	int m_timer=Atimer;

	bool dmini=false;		// Pour la minimap
	GLuint mini;
	int dx=0;
	int dy=0;

	bool ok_status=false;		// Pour les boutons
	bool cancel_status=false;
	bool o_ok_status=false;
	bool o_cancel_status=false;

	MAP_OTA	map_data;

	if( def_choice ) {
		*def_choice = def_choice->substr(5,def_choice->length()-9);
		i = 0;
		for( i_map = map_list.begin() ; i_map != map_list.end() ; i_map++, i++ )
			if( *i_map == *def_choice ) {
				sel_index = i;
				break;
				}
		}

	do
	{
		do
		{
			dt=(Atimer-time)*Conv;
			rest(1);
		}while(dt<0.02f);
		time=Atimer;

		o_ok_status=ok_status;
		o_cancel_status=cancel_status;
		ok_status=false;
		cancel_status=false;
		if(mouse_b==1 && mouse_y*gfx->SCREEN_H_TO_480>=440 && mouse_y*gfx->SCREEN_H_TO_480<=460) {
			cancel_status=(mouse_x*gfx->SCREEN_W_TO_640>=360 && mouse_x*gfx->SCREEN_W_TO_640<=440);
			ok_status=(mouse_x*gfx->SCREEN_W_TO_640>=520 && mouse_x*gfx->SCREEN_W_TO_640<=600);
			}

		if(mouse_b==0 && !ok_status && o_ok_status) {		// Click sur ok
			if(sel_index!=-1) {
				done=true;
				i_map=map_list.begin();
				for(i=0;i<sel_index && i_map!=map_list.end();i++)	i_map++;
				String tmp = String("maps\\") + *i_map + String(".tnt");
				choice=strdup(tmp.c_str());		// Copie le nom de la carte sélectionnée
				}
			}
		if(mouse_b==0 && !cancel_status && o_cancel_status)			// Click sur retour
			done=true;

		if(key[KEY_ESC]) {
			while(key[KEY_ESC])
				rest(1);
			done=true;
			}
		if(key[KEY_ENTER]) {
			if(sel_index!=-1) {
				done=true;
				i_map=map_list.begin();
				for(i=0;i<sel_index && i_map!=map_list.end();i++)	i_map++;
				String tmp = String("maps\\") + *i_map + String(".tnt");
				choice=strdup(tmp.c_str());		// Copie le nom de la carte sélectionnée
				}
			}
		bool check=false;
		if((Atimer-m_timer)*Conv>=0.1f) {
			if(key[KEY_DOWN])	{	sel_index++;	m_timer=Atimer;	check=true;	}
			if(key[KEY_UP])		{	sel_index--;	m_timer=Atimer;	check=true;	}
			if(o_sel!=sel_index && sel_index==-1)	sel_index=0;
			if(sel_index<-1)	sel_index=-1;
			if(sel_index>=n)	sel_index=n-1;
			}

		if(check) {
			if(sel_index>=dec+(int)(440.0f/h)-1)
				dec=sel_index-(int)(440.0f/h)+1;
			if(sel_index>=0 && sel_index<dec)
				dec=sel_index;
			}

		dec+=amz-mouse_z;
		if((Atimer-m_timer)*Conv>=0.2f) {
			if(mouse_x*gfx->SCREEN_W_TO_640>=15.0f && mouse_x*gfx->SCREEN_W_TO_640<320.0f && mouse_y*gfx->SCREEN_H_TO_480>=30.0f && mouse_y*gfx->SCREEN_H_TO_480<30.0f+h) {
				m_timer=Atimer;
				dec--;
				}
			if(mouse_x*gfx->SCREEN_W_TO_640>=15.0f && mouse_x*gfx->SCREEN_W_TO_640<320.0f && mouse_y*gfx->SCREEN_H_TO_480>=470.0f-h && mouse_y*gfx->SCREEN_H_TO_480<470.0f) {
				m_timer=Atimer;
				dec++;
				}
			}
		if(dec<0) dec=0;
		if(dec>n+1-440.0f/h) dec=(int)(n+1-440.0f/h);
		amz=mouse_z;

		if(mouse_x*gfx->SCREEN_W_TO_640>=15.0f && mouse_x*gfx->SCREEN_W_TO_640<320.0f && mouse_y*gfx->SCREEN_H_TO_480>=30.0f && mouse_y*gfx->SCREEN_H_TO_480<470.0f) {
			index=(int)((mouse_y*gfx->SCREEN_H_TO_480-30.0f)/h+dec);
			if(index>=n) index=-1;		// On ne sort pas de l'intervalle fixé
			if(mouse_b==1)
				sel_index=index;
			}
		else
			index=-1;

		if(sel_index!=o_sel && sel_index>=0) {
			if(dmini)
				glDeleteTextures(1,&mini);
			i_map=map_list.begin();
			for(i=0;i<sel_index && i_map!=map_list.end();i++)	i_map++;
			String tmp=String("maps\\") + *i_map + String(".tnt");
			mini=load_tnt_minimap_fast((char*)tmp.c_str(),&dx,&dy);
			dmini=(mini!=0);
			tmp=String("maps\\") + *i_map + String(".ota");								// Read the ota file
			uint32 ota_size=0;
			byte *data = HPIManager->PullFromHPI(tmp,&ota_size);
			if(data) {
				map_data.load((char*)data,ota_size);
				free(data);
				}
			else
				map_data.destroy();
			}

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		gfx->drawtexture(gfx->glfond,0.0f,0.0f,640.0f,480.0);

		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.0f,0.0f,0.0f,0.5f);
		gfx->rectfill(10.0,25.0f,630.0f,470.0f);
		glColor4f(1.0f,1.0f,1.0f,0.5f);
		gfx->rect(10.0f,25.0f,630.0f,470.0f);
		gfx->rect(347.0f,50.0f,603.0f,306.0f);
		gfx->line(320.0f,26.0f,320.0f,469.0f);
		glEnable(GL_TEXTURE_2D);

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,320.0f-gfx->TA_font.length(translation_manager.get(15))*0.5f,6.0f,0.0f,0xFFFFFFFF,translation_manager.get(15));

		i=0;
		for(i_map=map_list.begin();i_map!=map_list.end();i_map++) {
			if(30.0f+h*(i-dec)<470.0f-h && h*(i-dec)>=0) {
				if(i==index || i==sel_index) {
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					glBegin(GL_QUADS);
						if(i==sel_index)
							glColor4f(0.0f,0.0f,0.5f,0.5f);
						else
							glColor4f(0.25f,0.25f,0.5f,0.5f);
						glVertex2f(15.0f,30.0f+h*(i-dec));
						glVertex2f(319.0f,30.0f+h*(i-dec));
						if(i==sel_index)
							glColor4f(0.0f,0.0f,1.0f,0.5f);
						else
							glColor4f(0.5f,0.5f,1.0f,0.5f);
						glVertex2f(319.0f,30.0f+h*(i-dec+0.5f));
						glVertex2f(15.0f,30.0f+h*(i-dec+0.5f));

						glVertex2f(15.0f,30.0f+h*(i-dec+0.5f));
						glVertex2f(319.0f,30.0f+h*(i-dec+0.5f));
						if(i==sel_index)
							glColor4f(0.0f,0.0f,0.5f,0.5f);
						else
							glColor4f(0.25f,0.25f,0.5f,0.5f);
						glVertex2f(319.0f,30.0f+h*(i-dec+1));
						glVertex2f(15.0f,30.0f+h*(i-dec+1));
					glEnd();
					glColor4f(1.0f,1.0f,1.0f,0.5f);
					glEnable(GL_TEXTURE_2D);
					glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
					}
				gfx->print(gfx->TA_font,20.0f,30.0f+h*(i-dec),0.0f,0xFFFFFFFF,i_map->c_str());
				}
			i++;
			}

		glColor4f(1.0f,1.0f,1.0f,1.0f);

		glDisable(GL_BLEND);

		if(dmini) {
			glBindTexture(GL_TEXTURE_2D,mini);
			float ldx=dx*256.0f/252.0f*0.5f;
			float ldy=dy*256.0f/252.0f*0.5f;
			glBegin(GL_QUADS);

				glTexCoord2f(0.0f,0.0f);
				glVertex2f(475.0f-ldx,178.0f-ldy);

				glTexCoord2f(dx/252.0f,0.0f);
				glVertex2f(475.0f+ldx,178.0f-ldy);
					
				glTexCoord2f(dx/252.0f,dy/252.0f);
				glVertex2f(475.0f+ldx,178.0f+ldy);
			
				glTexCoord2f(0.0f,dy/252.0f);
				glVertex2f(475.0f-ldx,178.0f+ldy);

			glEnd();
			}
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		if(map_data.missionname)
			gfx->print(gfx->TA_font,480.0f-gfx->TA_font.length(map_data.missionname)*0.5f,320.0f,0.0f,0xFFFFFFFF,map_data.missionname);
		if(map_data.missiondescription) {
			glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
			if(strstr(map_data.missiondescription,"\n")) {
				char *rem=strstr(map_data.missiondescription,"\n");
				rem[0]=0;
				gfx->print(gfx->normal_font,340.0f,320.0f+data_h+3.0f,0.0f,0xFFFFFFFF,map_data.missiondescription);
				rem[0]='\n';
				if(strstr(rem+1,"\n")) {
					rem=strstr(rem+1,"\n");
					rem[0]=0;
					gfx->print(gfx->normal_font,340.0f,320.0f+data_h+13.0f,0.0f,0xFFFFFFFF,rem+1);
					rem[0]='\n';
					gfx->print(gfx->normal_font,340.0f,320.0f+data_h+23.0f,0.0f,0xFFFFFFFF,rem+1);
					}
				else
					gfx->print(gfx->normal_font,340.0f,320.0f+data_h+13.0f,0.0f,0xFFFFFFFF,rem+1);
				}
			else
				gfx->print(gfx->normal_font,340.0f,320.0f+data_h,0.0f,0xFFFFFFFF,map_data.missiondescription);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			}
		if(map_data.numplayers) {
			float len=(gfx->TA_font.length(translation_manager.get(31))+gfx->TA_font.length(map_data.numplayers))*0.5f;
			gfx->print(gfx->TA_font,480.0f-len,320.0f+data_h*3.0f,0.0f,0xFFFFFFFF,translation_manager.get(31));
			gfx->print(gfx->TA_font,480.0f-len+gfx->TA_font.length(translation_manager.get(31)),320.0f+data_h*3.0f,0.0f,0xFFFFFFFF,map_data.numplayers);
			}

		o_sel=sel_index;

		glDisable(GL_BLEND);

		glbutton(translation_manager.get(17),360.0f,440.0f,440.0f,460.0f,cancel_status);
		glbutton(translation_manager.get(16),520.0f,440.0f,600.0f,460.0f,ok_status);

		draw_cursor();

		gfx->flip();
	}while(!done);

	if(dmini)
		glDeleteTextures(1,&mini);

	gfx->unset_2D_mode();

	set_uformat(U_UTF8);

	return choice;
}

void config_menu(void)
{
	gfx->SCREEN_W_TO_640 = 640.0f / SCREEN_W;				// To have mouse sensibility undependent from the resolution
	gfx->SCREEN_H_TO_480 = 480.0f / SCREEN_H;

	lp_CONFIG->Lang = LANG;

	TA3DCONFIG	saved_config = *lp_CONFIG;

	TA3D_CONFIG	config_manager;
	config_manager.add_option("showfps",true,false,&lp_CONFIG->showfps);
	config_manager.add_option("fps_limit",false,true,&lp_CONFIG->fps_limit);
	config_manager.add_option("wireframe",true,false,&lp_CONFIG->wireframe);
	config_manager.add_option("particle",true,false,&lp_CONFIG->particle);
	config_manager.add_option("trees",true,false,&lp_CONFIG->trees);
	config_manager.add_option("shadow",true,false,&lp_CONFIG->shadow);
	config_manager.add_option("shadow_quality",false,false,&lp_CONFIG->shadow_quality);
	config_manager.add_option("shadow_r",false,true,&lp_CONFIG->shadow_r);
	config_manager.add_option("priority_level",false,false,&lp_CONFIG->priority_level);
	config_manager.add_option("timefactor",false,true,&lp_CONFIG->timefactor);
	config_manager.add_option("screen_width",false,false,&lp_CONFIG->screen_width);
	config_manager.add_option("screen_height",false,false,&lp_CONFIG->screen_height);
	config_manager.add_option("fullscreen",true,false,&lp_CONFIG->fullscreen);
	config_manager.add_option("fsaa",false,false,&lp_CONFIG->fsaa);
	config_manager.add_option("water_quality",false,false,&lp_CONFIG->water_quality);
	config_manager.add_option("LANG",false,false,&lp_CONFIG->Lang);
	config_manager.add_option("height_line",true,false,&lp_CONFIG->height_line);

	bool done=false;

	gfx->set_2D_mode();

	gfx->ReInitTexSys();

	glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);

	int i;

	float dt=0.0f;
	int time=Atimer;
	float Conv=1.0f/SECS_TO_TIMER(1);

	float h=gfx->TA_font.height();
	int	amb=mouse_b;

	bool save=false;

	bool ok_status=false;		// Pour les boutons
	bool cancel_status=false;
	bool o_ok_status=false;
	bool o_cancel_status=false;
	int dec=0;
	int max_opts=(int)(400/h)+1;
	int amz=mouse_z;

	do
	{
		do
		{
			dt=(Atimer-time)*Conv;
			rest(1);
		}while(dt<0.02f);
		time=Atimer;

		if(key[KEY_ESC]) {
			while(key[KEY_ESC])
				rest(1);
			done=true;
			}

		o_ok_status=ok_status;
		o_cancel_status=cancel_status;
		ok_status=false;
		cancel_status=false;
		if(mouse_b==1 && mouse_y*gfx->SCREEN_H_TO_480>=440 && mouse_y*gfx->SCREEN_H_TO_480<=460) {
			ok_status=(mouse_x*gfx->SCREEN_W_TO_640>=440 && mouse_x*gfx->SCREEN_W_TO_640<=520);
			cancel_status=(mouse_x*gfx->SCREEN_W_TO_640>=120 && mouse_x*gfx->SCREEN_W_TO_640<=200);
			}

		if(mouse_b==0 && !ok_status && o_ok_status) {		// Click sur ok
			done=true;
			save=true;
			}
		if(mouse_b==0 && !cancel_status && o_cancel_status)			// Click sur retour
			done=true;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		gfx->drawtexture(gfx->glfond,0.0f,0.0f,640.0f,480.0);

		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.0f,0.0f,0.0f,0.5f);
		gfx->rectfill(10.0f,25.0f,630.0f,470.0f);
		glColor4f(1.0f,1.0f,1.0f,0.5f);
		gfx->rect(10.0f,25.0f,630.0f,470.0f);
		glEnable(GL_TEXTURE_2D);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		gfx->print(gfx->TA_font,320.0f-0.5f*gfx->TA_font.length(translation_manager.get(18)),6.0f,0.0f,0xFFFFFFFF,translation_manager.get(18));
		for(int i=0;i<min(config_manager.nb_options,max_opts);i++) {		// affiche les options
			int e=i+dec;
			gfx->print(gfx->TA_font,32.0f,32.0f+i*h,0.0f,0xFFFFFFFF,config_manager.options[e].name);
			char buf[100];
			if(config_manager.options[e].boolean)
				uszprintf(buf,100,"%s",config_manager.get_option(e) ? "on" : "off");
			else if(config_manager.options[e].floatting)
				uszprintf(buf,100,"%f",config_manager.get_option(e));
			else
				uszprintf(buf,100,"%d",(int)config_manager.get_option(e));
			gfx->print(gfx->TA_font,600.0f-gfx->TA_font.length(buf),32.0f+i*h,0.0f,0xFFFFFFFF,buf);
			}

		glDisable(GL_BLEND);
		glbutton(translation_manager.get(16),440.0f,440.0f,520.0f,460.0f,ok_status);
		glbutton(translation_manager.get(21),120.0f,440.0f,200.0f,460.0f,cancel_status);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		if(mouse_x*gfx->SCREEN_W_TO_640>=32 && mouse_x*gfx->SCREEN_W_TO_640<=600 && mouse_y*gfx->SCREEN_H_TO_480>=32 && mouse_y*gfx->SCREEN_H_TO_480<32+h*min(config_manager.nb_options,max_opts)) {
			int idx=(int)((mouse_y*gfx->SCREEN_H_TO_480-32)/h);
			glDisable(GL_TEXTURE_2D);
			glBegin(GL_QUADS);
				glColor4f(0.5f,0.5f,1.0f,0.0f);
				glVertex2f(32.0f,32.0f+idx*h);
				glVertex2f(600.0f,32.0f+idx*h);
				glColor4f(0.5f,0.5f,1.0f,0.5f);
				glVertex2f(600.0f,32.0f+(idx+0.5f)*h);
				glVertex2f(32.0f,32.0f+(idx+0.5f)*h);

				glVertex2f(32.0f,32.0f+(idx+0.5f)*h);
				glVertex2f(600.0f,32.0f+(idx+0.5f)*h);
				glColor4f(0.5f,0.5f,1.0f,0.0f);
				glVertex2f(600.0f,32.0f+(idx+1)*h);
				glVertex2f(32.0f,32.0f+(idx+1)*h);
			glEnd();
			glEnable(GL_TEXTURE_2D);
			glColor4f(1.0f,1.0f,1.0f,1.0f);
			idx+=dec;
			if(mouse_b==1 && amb==0) {
				if(config_manager.options[idx].boolean)
					config_manager.set_option(config_manager.options[idx].name,config_manager.get_option(idx)==1.0f ? 0.0f : 1.0f);
				else {
					char titre[100],msg[100];
					uszprintf(titre,100,translation_manager.get(19),config_manager.options[idx].name);
					uszprintf(msg,100,translation_manager.get(20),config_manager.options[idx].name);
					String ans = msg_box( gfx->TA_font , titre,msg,true);
					config_manager.set_option(config_manager.options[idx].name,atof(ans.c_str()));
					ans.clear();
					switch(idx)
					{
					case 1:			// fps_limit
						if(config_manager.get_option(idx)==0.0f)
							config_manager.set_option(config_manager.options[idx].name,-1.0f);
						break;
					case 6:			// shadow_quality
						if(config_manager.get_option(idx)<=0.0f)
							config_manager.set_option(config_manager.options[idx].name,1.0f);
						else if(config_manager.get_option(idx)>12.0f)
							config_manager.set_option(config_manager.options[idx].name,12.0f);
						break;
					case 7:			// shadow_r
						if(config_manager.get_option(idx)<0.0f)
							config_manager.set_option(config_manager.options[idx].name,0.0f);
						else if(config_manager.get_option(idx)>100.0f)
							config_manager.set_option(config_manager.options[idx].name,100.0f);
						break;
					case 8:			// priority_level
						if(config_manager.get_option(idx)<0.0f)
							config_manager.set_option(config_manager.options[idx].name,0.0f);
						else if(config_manager.get_option(idx)>2.0f)
							config_manager.set_option(config_manager.options[idx].name,2.0f);
						break;
					case 9:			// timefactor
						if(config_manager.get_option(idx)<=0.0f)
							config_manager.set_option(config_manager.options[idx].name,1.0f);
						else if(config_manager.get_option(idx)>10.0f)
							config_manager.set_option(config_manager.options[idx].name,10.0f);
						break;
					case 10:		// screen_width
						if(config_manager.get_option(idx)<=0.0f)
							config_manager.set_option(config_manager.options[idx].name,1.0f);
						break;
					case 11:		// screen_height
						if(config_manager.get_option(idx)<=0.0f)
							config_manager.set_option(config_manager.options[idx].name,1.0f);
						break;
					case 14:		// water_quality
						if(config_manager.get_option(idx)<0.0f)
							config_manager.set_option(config_manager.options[idx].name,0.0f);
						else if(config_manager.get_option(idx)>3.0f)
							config_manager.set_option(config_manager.options[idx].name,3.0f);
						break;
					};
					}
				}
			}
		amb=mouse_b;

		glDisable(GL_BLEND);

		if(max_opts<config_manager.nb_options) {
			glDisable(GL_TEXTURE_2D);
			glBegin(GL_QUADS);
				glColor3f(0.4f,0.4f,0.8f);
				glVertex2i(4,32);
				glVertex2i(28,32);

				glColor3f(0.2f,0.2f,0.4f);
				glVertex2i(28,440);
				glVertex2i(4,440);
			glEnd();
			int pos = 32 + dec*376/(config_manager.nb_options-max_opts);
			glBegin(GL_QUADS);
				glColor3f(0.8f,0.4f,0.4f);
				glVertex2i(4,pos);
				glVertex2i(28,pos);
				glColor3f(0.4f,0.2f,0.2f);
				glVertex2i(28,32+pos);
				glVertex2i(4,32+pos);
			glEnd();
			if(mouse_x*gfx->SCREEN_W_TO_640>=4 && mouse_x*gfx->SCREEN_W_TO_640<=28 && mouse_y*gfx->SCREEN_H_TO_480>=32 && mouse_y*gfx->SCREEN_H_TO_480<440 && mouse_b==1)
				dec=(((int)(mouse_y*gfx->SCREEN_W_TO_640)-48)*(config_manager.nb_options-max_opts)+188)/376;
			else
				dec+=amz-mouse_z;
			amz=mouse_z;
			if(dec<0)	dec=0;
			if(dec+max_opts>config_manager.nb_options)	dec=config_manager.nb_options-max_opts;
			}

		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f,1.0f,1.0f);

		draw_cursor();

		gfx->flip();
	}while(!done);

	if(!save)
		*lp_CONFIG = saved_config;

	LANG = lp_CONFIG->Lang;

	gfx->unset_2D_mode();
}

void stats_menu(void)
{
	gfx->SCREEN_W_TO_640 = 640.0f / SCREEN_W;				// To have mouse sensibility undependent from the resolution
	gfx->SCREEN_H_TO_480 = 480.0f / SCREEN_H;

	cursor_type=CURSOR_DEFAULT;

	bool done=false;

	gfx->set_2D_mode();

	gfx->ReInitTexSys();

	glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);

	int i;

	float dt=0.0f;
	int time=Atimer;
	float Conv=1.0f/SECS_TO_TIMER(1);

	float h=gfx->TA_font.height();
	int	amb=mouse_b;

	bool ok_status=false;		// Pour les boutons
	bool o_ok_status=false;
	int amz=mouse_z;

	while(key[KEY_ESC])	rest(1);

	do
	{
		do
		{
			dt=(Atimer-time)*Conv;
			rest(1);
		}while(dt<0.02f);
		time=Atimer;

		if(key[KEY_ESC]) {
			while(key[KEY_ESC])
				rest(1);
			done=true;
			}

		o_ok_status=ok_status;
		ok_status=false;
		if(mouse_b==1 && mouse_y*gfx->SCREEN_H_TO_480>=440 && mouse_y*gfx->SCREEN_H_TO_480<=460)
			ok_status=(mouse_x*gfx->SCREEN_W_TO_640>=240 && mouse_x*gfx->SCREEN_W_TO_640<=400);

		if(mouse_b==0 && !ok_status && o_ok_status)		// Click sur ok
			done=true;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		gfx->drawtexture(gfx->glfond,0.0f,0.0f,640.0f,480.0f);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		gfx->print(gfx->TA_font,320.0f-0.5f*gfx->TA_font.length(translation_manager.get(32)),8.0f,0.0f,0xFFFFFFFF,translation_manager.get(32));

		glDisable(GL_TEXTURE_2D);

		glColor4f(0.0f,0.0f,0.0f,0.5f);
		gfx->rectfill(8,32,632,432);

		glDisable(GL_BLEND);
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		gfx->rect(8,32,632,432);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.5f,0.5f,0.5f,0.5f);
		glBegin(GL_LINES);
		for(int i=0;i<10;i++) {
			glVertex2i(32,72+36*i);
			glVertex2i(608,72+36*i);
			}
		glColor4f(1.0f,1.0f,1.0f,0.5f);
			glVertex2i(240,48);		glVertex2i(240,424);
			glVertex2i(318,48);		glVertex2i(318,424);
			glVertex2i(396,48);		glVertex2i(396,424);
			glVertex2i(474,48);		glVertex2i(474,424);
			glVertex2i(552,48);		glVertex2i(552,424);
		glEnd();
		glEnable(GL_TEXTURE_2D);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,136.0f-0.5f*gfx->TA_font.length(translation_manager.get(33)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(33));
		gfx->print(gfx->TA_font,276.0f-0.5f*gfx->TA_font.length(translation_manager.get(34)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(34));
		gfx->print(gfx->TA_font,354.0f-0.5f*gfx->TA_font.length(translation_manager.get(35)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(35));
		gfx->print(gfx->TA_font,432.0f-0.5f*gfx->TA_font.length(translation_manager.get(36)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(36));
		gfx->print(gfx->TA_font,510.0f-0.5f*gfx->TA_font.length(translation_manager.get(37)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(37));
		gfx->print(gfx->TA_font,588.0f-0.5f*gfx->TA_font.length(translation_manager.get(38)),48.0f,0.0f,0xFFFFFFFF,translation_manager.get(38));

		for(int i=0;i<players.nb_player;i++) {
			char txt[100];
			float y=72.0f+18.0f-h*0.5f+36.0f*i;
			gfx->print(gfx->TA_font,32.0f,y,0.0f,0xFFFFFFFF,players.nom[i]);
			gfx->print(gfx->TA_font,276.0f-0.5f*gfx->TA_font.length(players.side[i]),y,0.0f,0xFFFFFFFF,players.side[i]);
			uszprintf(txt,100,"%d",players.losses[i]);					gfx->print(gfx->TA_font,354.0f-0.5f*gfx->TA_font.length(txt),y,0.0f,0xFFFFFFFF,txt);
			uszprintf(txt,100,"%d",players.kills[i]);					gfx->print(gfx->TA_font,432.0f-0.5f*gfx->TA_font.length(txt),y,0.0f,0xFFFFFFFF,txt);
			uszprintf(txt,100,"%d",(int)players.energy_total[i]);		gfx->print(gfx->TA_font,510.0f-0.5f*gfx->TA_font.length(txt),y,0.0f,0xFFFFFFFF,txt);
			uszprintf(txt,100,"%d",(int)players.metal_total[i]);		gfx->print(gfx->TA_font,588.0f-0.5f*gfx->TA_font.length(txt),y,0.0f,0xFFFFFFFF,txt);
			}

		glDisable(GL_BLEND);
		glbutton(translation_manager.get(16),240.0f,440.0f,400.0f,460.0f,ok_status);

		amb=mouse_b;

		glEnable(GL_TEXTURE_2D);
		glColor3f(1.0f,1.0f,1.0f);

		draw_cursor();

		gfx->flip();
	}while(!done);

	gfx->unset_2D_mode();
}

void setup_game(void)
{
	List<String> map_list;
	uint32 n = HPIManager->GetFilelist("maps\\*.tnt",&map_list);

	if( n == 0 ) {
		Popup(translation_manager.get(48),translation_manager.get(49));
		Console->AddEntry("no maps found!!");
		reset_mouse();
		return;
		}

	cursor_type=CURSOR_DEFAULT;

	float resize_w = SCREEN_W / 640.0f;
	float resize_h = SCREEN_H / 480.0f;

	uint16	player_str_n = 3;
	uint16	side_str_n = 2;
	uint16	ai_level_str_n = 4;
	String	player_str[3] = { translation_manager.get(51), translation_manager.get(52), translation_manager.get(53) };
	byte	player_control[3] = { PLAYER_CONTROL_LOCAL_HUMAN, PLAYER_CONTROL_LOCAL_AI, PLAYER_CONTROL_NONE };
	String	ai_level_str[4] = { translation_manager.get(54), translation_manager.get(55), translation_manager.get(56), translation_manager.get(57) };
	String	side_str[2] = { "arm","core" };

	GAME_DATA game_data;

	game_data.map_filename = strdup( map_list.begin()->c_str() );
	game_data.nb_players = 2;
	game_data.game_script = strdup("scripts/default.c");

	map_list.clear();

	game_data.player_names[0] = player_str[0];
	game_data.player_sides[0] = side_str[0];
	game_data.player_control[0] = player_control[0];
	game_data.ai_level[0] = AI_TYPE_EASY;

	game_data.player_names[1] = player_str[1];
	game_data.player_sides[1] = side_str[1];
	game_data.player_control[1] = player_control[1];
	game_data.ai_level[1] = AI_TYPE_EASY;

	for(uint16 i = 2; i < 10 ; i++) {
		game_data.player_names[i] = player_str[2];
		game_data.player_sides[i] = side_str[0];
		game_data.player_control[i] = player_control[2];
		game_data.ai_level[i] = AI_TYPE_EASY;
		}

	gfx->set_2D_mode();

	WND w_game;				// Window used to manage the interface
	w_game.Title=translation_manager.get(50);
	w_game.x=0;					w_game.y=0;
	w_game.width=SCREEN_W+1;	w_game.height=SCREEN_H+1;
	w_game.Lock=true;
	w_game.NbObj=137;
	w_game.Objets=new GUIOBJ[w_game.NbObj];
	w_game.draw_borders = false;
	w_game.show_title = false;
	w_game.background = gfx->glfond;

	int dx, dy;
	GLuint glimg = load_tnt_minimap_fast(game_data.map_filename,&dx,&dy);
	float ldx = dx*70.0f/252.0f;
	float ldy = dy*70.0f/252.0f;

			// OK
	w_game.Objets[0].create_button(530-40,432,530+40,456,translation_manager.get(16),NULL);
			// Cancel
	w_game.Objets[1].create_button(640/3-40,432,640/3+40,456,translation_manager.get(17),NULL);

	w_game.Objets[2].create_box(20,20,620,460,0xFFFFFFFF);

			// Title
	w_game.Objets[3].create_text(320-(int)(gfx->TA_font.length(translation_manager.get(50))/2),4,translation_manager.get(50),0xFFFFFFFF,1.0f);

	w_game.Objets[4].create_line(440,22,440,459,0xFFFFFFFF);
	w_game.Objets[5].create_img(460,40,600,180,glimg);
	w_game.Objets[5].x1 = (int)(530-ldx);
	w_game.Objets[5].y1 = (int)(110-ldy);
	w_game.Objets[5].x2 = (int)(530+ldx);
	w_game.Objets[5].y2 = (int)(110+ldy);
	w_game.Objets[5].u2 = dx/252.0f;
	w_game.Objets[5].v2 = dy/252.0f;
	w_game.Objets[6].create_box(460,40,600,180,0xFFFFFFFF);

	w_game.Objets[7].create_text(40,40,translation_manager.get(33),0xFFFFFFFF,1.0f);
	w_game.Objets[8].create_text(160,40,translation_manager.get(34),0xFFFFFFFF,1.0f);
	const uint8	_N = 12;
	for(uint16 i = 0;i< 10; i++) {
		w_game.Objets[9+i*_N].create_box(30,60+20*i,150,79+20*i, 0xFF3F3F3F );
		w_game.Objets[9+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[10+i*_N].create_text(40,62+20*i, game_data.player_names[i].c_str(), 0xFFFFFFFF, 1.0f );

		w_game.Objets[11+i*_N].create_box(150,60+20*i,200,79+20*i, 0xFF3F3F3F );
		w_game.Objets[11+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[12+i*_N].create_text(160,62+20*i,game_data.player_sides[i].c_str(), 0xFFFFFFFF, 1.0f );

		w_game.Objets[13+i*_N].create_box(200,60+20*i,280,79+20*i, 0xFF3F3F3F );
		w_game.Objets[13+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[14+i*_N].create_text(210,62+20*i,game_data.player_control[i] & PLAYER_CONTROL_FLAG_AI ? ai_level_str[game_data.ai_level[i]].c_str() : "", 0xFFFFFFFF, 1.0f );

		w_game.Objets[15+i*_N].create_box(280,60+20*i,300,79+20*i, 0xFF3F3F3F );
		w_game.Objets[15+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[16+i*_N].create_box(282,62+20*i,298,77+20*i, gfx->makeintcol(player_color[player_color_map[i]*3], player_color[player_color_map[i]*3+1], player_color[player_color_map[i]*3+2]) );
		w_game.Objets[16+i*_N].Flag |= FLAG_FILL | (game_data.player_control[i] == PLAYER_CONTROL_NONE ? FLAG_HIDDEN : 0 );

		w_game.Objets[17+i*_N].create_box(300,60+20*i,365,79+20*i, 0xFF3F3F3F );
		w_game.Objets[17+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[18+i*_N].create_text(310,62+20*i,strdup(format("%d",game_data.energy[i]).c_str()), 0xFFFFFFFF, 1.0f );

		w_game.Objets[19+i*_N].create_box(365,60+20*i,430,79+20*i, 0xFF3F3F3F );
		w_game.Objets[19+i*_N].Flag |= FLAG_HIGHLIGHT;
		w_game.Objets[20+i*_N].create_text(375,62+20*i,strdup(format("%d",game_data.metal[i]).c_str()), 0xFFFFFFFF, 1.0f );
		}
	w_game.Objets[129].create_text(210,40,translation_manager.get(58),0xFFFFFFFF,1.0f);
	w_game.Objets[130].create_line(150,40,150,260,0xFFFFFFFF);
	w_game.Objets[131].create_line(200,40,200,260,0xFFFFFFFF);
	w_game.Objets[132].create_line(280,40,280,260,0xFFFFFFFF);
	w_game.Objets[133].create_line(300,40,300,260,0xFFFFFFFF);
	w_game.Objets[134].create_line(365,40,365,260,0xFFFFFFFF);
	w_game.Objets[135].create_text(310,40,translation_manager.get(37),0xFFFFFFFF,1.0f);
	w_game.Objets[136].create_text(375,40,translation_manager.get(38),0xFFFFFFFF,1.0f);

	for(uint16 i = 0; i < w_game.NbObj; i++) {
		w_game.Objets[i].x1 = (int)(w_game.Objets[i].x1*resize_w);
		w_game.Objets[i].y1 = (int)(w_game.Objets[i].y1*resize_h);
		w_game.Objets[i].x2 = (int)(w_game.Objets[i].x2*resize_w);
		w_game.Objets[i].y2 = (int)(w_game.Objets[i].y2*resize_h);
		w_game.Objets[i].s *= resize_w;
		}

	bool done=false;

	bool start_game = false;

	int amx,amy,amb,amz;

	do
	{
		amx=mouse_x;
		amy=mouse_y;
		amz=mouse_z;
		amb=mouse_b;

		rest(10);

		poll_mouse();		// obtient les coordonnées de la souris
		
		w_game.check(amx,amy,amb);		// gestion de l'interface utilisateur graphique

		for(uint16 i = 0; i<10; i++) {		// Monitor events on the table used to display data
			if(w_game.Objets[9+i*_N].Etat) {	// Change player type
				uint16 e = 0;
				for( uint16 f = 0 ; f<player_str_n ; f++ )
					if( w_game.Objets[10+i*_N].Text[0] == player_str[f].c_str() ) {	e = f;	break;	}
				e = ++e % player_str_n;

				if( player_control[e] == PLAYER_CONTROL_LOCAL_HUMAN )					// We can only have one local human player ( or it crashes )
					for( uint16 f = 0 ; f<10 ; f++ )
						if( f!= i && game_data.player_control[f] == PLAYER_CONTROL_LOCAL_HUMAN ) {		// If we already have a local human player pass this player type value
							e = ++e % player_str_n;
							break;
							}

				game_data.player_names[i] = player_str[e];								// Update game data
				game_data.player_control[i] = player_control[e];

				w_game.Objets[10+i*_N].Text[0] = player_str[e];			// Update gui
				w_game.Objets[14+i*_N].Text[0] = (game_data.player_control[i] & PLAYER_CONTROL_FLAG_AI) ? ai_level_str[game_data.ai_level[i]] : String("");
				if( player_control[e] == PLAYER_CONTROL_NONE )
					w_game.Objets[16+i*_N].Flag |= FLAG_HIDDEN;
				else
					w_game.Objets[16+i*_N].Flag &= ~FLAG_HIDDEN;
				}
			if(w_game.Objets[11+i*_N].Etat) {	// Change player side
				uint16 e = 0;
				for( uint16 f = 0 ; f<side_str_n ; f++ )
					if( w_game.Objets[12+i*_N].Text[0] == side_str[f].c_str() ) {	e = f;	break;	}
				e = ++e % side_str_n;
				w_game.Objets[12+i*_N].Text[0] = side_str[e];			// Update gui

				game_data.player_sides[i] = side_str[e];								// update game data
				}
			if(w_game.Objets[13+i*_N].Etat) {	// Change player level (for AI)
				uint16 e = 0;
				for( uint16 f = 0 ; f<ai_level_str_n ; f++ )
					if( w_game.Objets[14+i*_N].Text[0] == ai_level_str[f].c_str() ) {	e = f;	break;	}
				e = ++e % ai_level_str_n;
				w_game.Objets[14+i*_N].Text[0] = game_data.player_control[i] & PLAYER_CONTROL_FLAG_AI ? ai_level_str[e] : String("");			// Update gui

				game_data.ai_level[i] = e;								// update game data
				}
			if(w_game.Objets[15+i*_N].Etat) {	// Change player color
				sint16 e = player_color_map[i];
				sint16 f = -1;
				for( sint16 g = 0; g<10 ; g++ )						// Look for the next color
					if( game_data.player_control[g] == PLAYER_CONTROL_NONE && player_color_map[g] > e && (f == -1 || player_color_map[g] < player_color_map[f]) )
						f = g;
				if( f == -1 )
					for( uint16 g = 0; g<10 ; g++ )
						if( game_data.player_control[g] == PLAYER_CONTROL_NONE && (f == -1 || player_color_map[g] < player_color_map[f]) )
							f = g;
				if( f != -1 ) {
					sint16 g = player_color_map[f];
					player_color_map[i] = g;								// update game data
					player_color_map[f] = e;

					w_game.Objets[16+i*_N].Data = gfx->makeintcol(player_color[player_color_map[i]*3],player_color[player_color_map[i]*3+1],player_color[player_color_map[i]*3+2]);			// Update gui
					w_game.Objets[16+f*_N].Data = gfx->makeintcol(player_color[player_color_map[f]*3],player_color[player_color_map[f]*3+1],player_color[player_color_map[f]*3+2]);			// Update gui
					}
				}
			if(w_game.Objets[17+i*_N].Etat) {	// Change player color
				game_data.energy[i] = (game_data.energy[i] + 500) % 10500;
				if( game_data.energy[i] == 0 ) game_data.energy[i] = 500;

				w_game.Objets[18+i*_N].Text[0] = format("%d",game_data.energy[i]);			// Update gui
				}
			if(w_game.Objets[19+i*_N].Etat) {	// Change player color
				game_data.metal[i] = (game_data.metal[i] + 500) % 10500;
				if( game_data.metal[i] == 0 ) game_data.metal[i] = 500;

				w_game.Objets[20+i*_N].Text[0] = format("%d",game_data.metal[i]);			// Update gui
				}
			}

		if(w_game.Objets[5].Etat) {		// Clic on the mini-map
			gfx->unset_2D_mode();
			reset_mouse();
			String map_filename = game_data.map_filename;
			char *new_map = select_map( &map_filename );

			gfx->SCREEN_W_TO_640 = 1.0f;				// To have mouse sensibility undependent from the resolution
			gfx->SCREEN_H_TO_480 = 1.0f;
			cursor_type=CURSOR_DEFAULT;
			gfx->set_2D_mode();

			if(new_map) {
				if(game_data.map_filename)	free(game_data.map_filename);
				game_data.map_filename = new_map;
				glimg=load_tnt_minimap_fast(game_data.map_filename,&dx,&dy);
				ldx = dx*70.0f/252.0f;
				ldy = dy*70.0f/252.0f;
				w_game.Objets[5].x1 = (int)((530-ldx)*resize_w);
				w_game.Objets[5].y1 = (int)((110-ldy)*resize_h);
				w_game.Objets[5].x2 = (int)((530+ldx)*resize_w);
				w_game.Objets[5].y2 = (int)((110+ldy)*resize_h);
				w_game.Objets[5].u2 = dx/252.0f;
				w_game.Objets[5].v2 = dy/252.0f;
				}

			w_game.Objets[5].Data = glimg;		// Synchronize the picture on GUI
			}

		if(w_game.Objets[0].Etat) {
			done=true;		// En cas de click sur "OK", on quitte la fenêtre
			start_game = true;
			}
		if(w_game.Objets[1].Etat) done=true;		// En cas de click sur "retour", on quitte la fenêtre

		if(key[KEY_ESC]) done=true;			// Quitte si on appuie sur echap
					// Efface tout
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		w_game.draw();				// Dessine la fenêtre d'édition

		glEnable(GL_TEXTURE_2D);
		gfx->set_color(0xFFFFFFFF);
		draw_cursor(resize_w,resize_h);
		
					// Affiche
		gfx->flip();
	}while(!done);

	w_game.destroy();		// Free memory used by the window

	gfx->unset_2D_mode();	// Quitte le mode de dessin d'allegro

	gfx->destroy_texture(glimg);

	reset_mouse();
	while(key[KEY_ESC])	rest(1);

	if(start_game) {
		if(game_data.map_filename && game_data.game_script) {
			game_data.nb_players = 0;
			for(uint16 i = 0 ; i<10 ; i++)		// Move players to the top of the vector, so it's easier to access data
				if( game_data.player_control[i] != PLAYER_CONTROL_NONE ) {
					game_data.player_control[game_data.nb_players] = game_data.player_control[i];
					game_data.player_names[game_data.nb_players] = game_data.player_names[i];
					game_data.player_sides[game_data.nb_players] = game_data.player_sides[i];
					game_data.ai_level[game_data.nb_players] = game_data.ai_level[i];
					game_data.energy[game_data.nb_players] = game_data.energy[i];
					game_data.metal[game_data.nb_players] = game_data.metal[i];
					uint16 e = player_color_map[game_data.nb_players];
					player_color_map[game_data.nb_players] = player_color_map[i];
					player_color_map[i] = e;
					game_data.nb_players++;
					}

			gfx->unset_2D_mode();
			GuardStart( play );
				play(&game_data);
			GuardCatch();
			if( IsExceptionInProgress() ) // if guard threw an error this will be true.
			{
				GuardDisplayAndLogError();   // record and display the error.
				exit(1);                      // we outa here.
			}
			gfx->set_2D_mode();
			gfx->ReInitTexSys();
			glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);
			}

		while(key[KEY_ESC])	rest(1);
		}
}

void battle_room(void)				// Let players create/join a game
{
	gfx->SCREEN_W_TO_640 = 640.0f / SCREEN_W;				// To have mouse sensibility undependent from the resolution
	gfx->SCREEN_H_TO_480 = 480.0f / SCREEN_H;

	bool done=false;

	gfx->set_2D_mode();

	gfx->ReInitTexSys();

	set_uformat(U_ASCII);		// Juste histoire d'avoir un affichage correct des textes

	glScalef(SCREEN_W/640.0f,SCREEN_H/480.0f,1.0f);

//	Network	TA3D_network;
//	TA3D_network.HostGame("my_game","1234",2,0);

	std::list<String>	servers;					// the server list
	std::list<String>::iterator	l_s_servers;		// iterator to read the server list
	servers.push_front(translation_manager.get(41));

	int i;

	float dt=0.0f;
	int time=Atimer;
	float Conv=1.0f/SECS_TO_TIMER(1);

	float h=gfx->TA_font.height();

	bool ok_status=false;		// Pour les boutons
	bool cancel_status=false;
	bool o_ok_status=false;
	bool o_cancel_status=false;

	do
	{
		do
		{
			dt=(Atimer-time)*Conv;
			rest(1);
		}while(dt<0.02f);
		time=Atimer;

		o_ok_status=ok_status;
		o_cancel_status=cancel_status;
		ok_status=false;
		cancel_status=false;
		if(mouse_b==1 && mouse_y*gfx->SCREEN_H_TO_480>=440 && mouse_y*gfx->SCREEN_H_TO_480<=460) {
			ok_status=(mouse_x*gfx->SCREEN_W_TO_640>=360 && mouse_x*gfx->SCREEN_W_TO_640<=440);
			cancel_status=(mouse_x*gfx->SCREEN_W_TO_640>=520 && mouse_x*gfx->SCREEN_W_TO_640<=600);
			}

		if(mouse_b==0 && !ok_status && o_ok_status) {		// Click sur ok
			done=true;
			}
		if(mouse_b==0 && !cancel_status && o_cancel_status)			// Click sur retour
			done=true;

		if(key[KEY_ESC]) {
			while(key[KEY_ESC])
				rest(1);
			done=true;
			}
		if(key[KEY_ENTER])
			done=true;
		bool check=false;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Efface l'écran

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		gfx->drawtexture(gfx->glfond,0.0f,0.0f,640.0f,480.0);

		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.0f,0.0f,0.0f,0.5f);
		glBegin(GL_QUADS);
			glVertex2f(10.0f,25.0f);
			glVertex2f(630.0f,25.0f);
			glVertex2f(630.0f,470.0f);
			glVertex2f(10.0f,470.0f);
		glEnd();
		glColor4f(1.0f,1.0f,1.0f,0.5f);
		glBegin(GL_LINE_LOOP);
			glVertex2f(10.0f,25.0f);
			glVertex2f(630.0f,25.0f);
			glVertex2f(630.0f,470.0f);
			glVertex2f(10.0f,470.0f);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glVertex2f(347.0f,50.0f);
			glVertex2f(603.0f,50.0f);
			glVertex2f(603.0f,306.0f);
			glVertex2f(347.0f,306.0f);
		glEnd();
		glBegin(GL_LINES);
			glVertex2f(320.0f,26.0f);
			glVertex2f(320.0f,469.0f);
			glVertex2f(10.0f,32.0f+h);
			glVertex2f(320.0f,32.0f+h);
		glEnd();
		glEnable(GL_TEXTURE_2D);

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gfx->print(gfx->TA_font,320.0f-gfx->TA_font.length(translation_manager.get(39))*0.5f,0.0f,0.0f,0xFFFFFFFF,translation_manager.get(39));
		gfx->print(gfx->TA_font,165.0f-gfx->TA_font.length(translation_manager.get(40))*0.5f,32.0f,0.0f,0xFFFFFFFF,translation_manager.get(40));

		glColor4f(1.0f,1.0f,1.0f,1.0f);

		int i=0;
		for(l_s_servers=servers.begin();l_s_servers!=servers.end();l_s_servers++) {
			gfx->print(gfx->TA_font,20.0f,h*i+64.0f,0.0f,0xFFFFFFFF,l_s_servers->c_str());
			i++;
			}

		glDisable(GL_BLEND);

		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

		glDisable(GL_BLEND);

		glbutton(translation_manager.get(16),360.0f,440.0f,440.0f,460.0f,ok_status);
		glbutton(translation_manager.get(17),520.0f,440.0f,600.0f,460.0f,cancel_status);

		draw_cursor();

		gfx->flip();
	}while(!done);

	gfx->unset_2D_mode();

	set_uformat(U_UTF8);
}
