/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

#include "ta3dbase.h"

#ifndef TA3D_BASIC_ENGINE
int expected_players;

int LANG=TA3D_LANG_ENGLISH;

int CURSOR_MOVE;
int CURSOR_GREEN;
int CURSOR_CROSS;
int CURSOR_RED;
int CURSOR_LOAD;
int CURSOR_GUARD;
int CURSOR_PATROL;
int CURSOR_REPAIR;
int CURSOR_ATTACK;
int CURSOR_BLUE;
int CURSOR_AIR_LOAD;
int CURSOR_BOMB_ATTACK;
int CURSOR_BALANCE;
int CURSOR_RECLAIM;
int CURSOR_WAIT;
int CURSOR_CANT_ATTACK;
int CURSOR_CROSS_LINK;
CAMERA *game_cam=NULL;

ANIMS cursor;

int fire;
int build_part;

int start=0;
int cursor_type=CURSOR_DEFAULT;
#endif

/*---------------------------------------------------------------------------------\
|                              void init_rand_table()                              |
|        Build the random table used in critical code                              |
\---------------------------------------------------------------------------------*/

#define RAND_TABLE_SIZE		0x100000
#define RAND_TABLE_MASK		0xFFFFF

int table_pos=0;
uint32 rand_table[RAND_TABLE_SIZE];

void init_rand_table()
{
	srand( (unsigned)time( NULL ) );
	for(int i=0;i<RAND_TABLE_SIZE;i++)
		rand_table[i]=rand();
}

/*---------------------------------------------------------------------------------\
|                              int rand_from_table()                               |
|        Get a "random" number from the table                                      |
\---------------------------------------------------------------------------------*/

const int rand_from_table()
{
	table_pos=++table_pos&RAND_TABLE_MASK;
	return rand_table[table_pos];
}

#ifndef TA3D_BASIC_ENGINE

/*---------------------------------------------------------------------------------\
|                              BITMAP *load_memory_pcx(...)                        |
|        Charge le fichier pcx dont data pointe les données                        |
\---------------------------------------------------------------------------------*/
BITMAP *load_memory_pcx(byte *data,RGB *cpal)
{
	if(cpal==NULL) return NULL;		// On ne peut pas charger de fichier 8bits sans palette graphique
	short width,height;
	width=*((short*)(data+8));		// Dimensions
	height=*((short*)(data+10));
	width-=*((short*)(data+4));		// Dimensions
	height-=*((short*)(data+6));

	width++;
	height++;

	BITMAP *pcx=create_bitmap(width,height);
	clear(pcx);

	int pos=128;
	for(int y=0;y<pcx->h;y++) {
		int x=0;
		do
		{
			int c=data[pos++];
			if(c>191) {
				int l=c-192;
				c=data[pos++];
				int col=makecol(cpal[c].r<<2,cpal[c].g<<2,cpal[c].b<<2);
				for(;l>0 && x<pcx->w;l--)
					((int*)(pcx->line[y]))[x++]=col;
				}
			else
				((int*)(pcx->line[y]))[x++]=makecol(cpal[c].r<<2,cpal[c].g<<2,cpal[c].b<<2);
		}while(x<pcx->w);
		}
	return pcx;
}
#endif

/*---------------------------------------------------------------------------------\
|                                   void PutTex(...)                               |
|        Procédure qui affiche la texture Tex dans le rectangle (x1,y1)-(x2,y2)    |
\---------------------------------------------------------------------------------*/

void PutTex(GLuint Tex,float x1,float y1,float x2,float y2)
{
	glBindTexture(GL_TEXTURE_2D,Tex);
	glBegin(GL_QUADS);

		glTexCoord2f(0.0f,0.0f);
		glVertex2f(x1,y1);

		glTexCoord2f(1.0f,0.0f);
		glVertex2f(x2,y1);
					
		glTexCoord2f(1.0f,1.0f);
		glVertex2f(x2,y2);
			
		glTexCoord2f(0.0f,1.0f);
		glVertex2f(x1,y2);

	glEnd();
}

GLuint LoadTex(char *file)
{
	set_color_depth(32);
	allegro_gl_use_alpha_channel(true);
	BITMAP *bmp=load_bitmap(file,NULL);
	GLuint gl_bmp;
	if(g_useTextureCompression)
		allegro_gl_set_texture_format(GL_COMPRESSED_RGBA_ARB);
	else
		allegro_gl_set_texture_format(GL_RGBA8);
	gl_bmp=allegro_gl_make_texture(bmp);
	allegro_gl_use_alpha_channel(false);
	glBindTexture(GL_TEXTURE_2D, gl_bmp);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	allegro_gl_set_texture_format(-1);
	destroy_bitmap(bmp);
	return gl_bmp;
}

GLuint LoadMaskedTex(char *file,char *filealpha)
{
	set_color_depth(32);
	BITMAP *bmp=load_bitmap(file,NULL);
	BITMAP *alpha;
	set_color_depth(8);
	alpha=load_bitmap(filealpha,NULL);
	set_color_depth(32);
	for(int y=0;y<bmp->h;y++)
		for(int x=0;x<bmp->w;x++)
			bmp->line[y][(x<<2)+3]=alpha->line[y][x];
	GLuint gl_bmp;
	allegro_gl_use_alpha_channel(true);
	if(g_useTextureCompression)
		allegro_gl_set_texture_format(GL_COMPRESSED_RGBA_ARB);
	else
		allegro_gl_set_texture_format(GL_RGBA8);
	gl_bmp=allegro_gl_make_texture(bmp);
	allegro_gl_use_alpha_channel(false);
	glBindTexture(GL_TEXTURE_2D, gl_bmp);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	destroy_bitmap(bmp);
	destroy_bitmap(alpha);
	return gl_bmp;
}

BITMAP *LoadMaskedTexBmp(char *file,char *filealpha)
{
	set_color_depth(32);
	BITMAP *bmp=load_bitmap(file,NULL);
	set_color_depth(8);
	BITMAP *alpha=load_bitmap(filealpha,NULL);
	set_color_depth(32);
	for(int y=0;y<bmp->h;y++)
		for(int x=0;x<bmp->w;x++)
			bmp->line[y][(x<<2)+3]=alpha->line[y][x];
	destroy_bitmap(alpha);
	return bmp;
}

#ifndef TA3D_BASIC_ENGINE

#include "3do.h"					// For 3DO/3DM management
#include "cob.h"					// For unit scripts management
#include "tdf.h"					// For 2D features
#include "EngineClass.h"			// The Core Engine
#include "UnitEngine.h"				// The Unit Engine
#include "tnt.h"					// The TNT loaded
#include "script.h"					// The game script manager
#include "ia.h"						// AI Engine

void SIDEDATA::load_data()
{
	destroy();

	uint32 file_size=0;
	byte *data=HPIManager->PullFromHPI("gamedata\\sidedata.tdf",&file_size);		// Charge les données sur les sons
	if(data==NULL)
		data=HPIManager->PullFromHPI("sidedata.tdf",&file_size);
	if(data) {
		char *ligne=NULL;
		char *pos=(char*)data;
		int n_pos=0;
		int side_id=-1;
		do {
			if(ligne)
				delete[] ligne;
			ligne=get_line(pos);
			while(pos[0]!=0 && pos[0]!=13 && pos[0]!=10)	{	pos++;	n_pos++; }
			while(pos[0]==13 || pos[0]==10)	{ pos++;	n_pos++; }

			if(n_pos>=file_size)	break;

			if(strstr(ligne,";"))
				*(strstr(ligne,";"))=0;
			if(strstr(ligne,"[SIDE")) {
				if(strstr(ligne,"]"))	*(strstr(ligne,"]"))=0;
				side_id=atoi(strstr(ligne,"[SIDE")+5);
				if(nb_side<=side_id)	nb_side=side_id+1;
				}
			else if(side_id>=0) {
				if(strstr(ligne,"name="))
					side_name[side_id]=strdup(strstr(ligne,"name=")+5);
				else if(strstr(ligne,"nameprefix="))
					side_pref[side_id]=strdup(strstr(ligne,"nameprefix=")+11);
				else if(strstr(ligne,"commander="))
					side_com[side_id]=strdup(strstr(ligne,"commander=")+10);
				}
		} while(n_pos<file_size);
		if(ligne)
			delete[] ligne;
		free(data);
		}
	FILE *mod_data = TA3D_OpenFile("ta3d.mod","rb");
	if(mod_data) {
		char buf[100];
		while(fgets(buf,100,mod_data)) {		// Lit tout le fichier / Read the file
			strlwr(buf);
			char *f=buf;
			while(f[0]!=0)
				if(f[0]==10 || f[0]==13)	f[0]=0;
				else	f++;
			if(strstr(buf,"unit_ext="))
				unit_ext=strdup(strstr(buf,"unit_ext=")+9);
			}
		fclose(mod_data);
		}
	else {
		unit_ext=strdup(".fbi");
		}
}

int	SIDEDATA::get_side_id(char *side)
{
	for(int i=0;i<nb_side;i++)
		if(strcasecmp(side,side_name[i])==0)
			return i;
	return -1;
}

SIDEDATA ta3d_sidedata;

#endif

void reset_mouse()
{
	int amx = mouse_x;
	int amy = mouse_y;

	remove_mouse();
	install_mouse();

	position_mouse(amx, amy);
}
