/*  TA3D, a remake of Total Annihilation
    Copyright (C) 2005  Roland BROCHARD

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA*/

/*-----------------------------------------------------------------------------------\
|                                     taconfig.cpp                                   |
|  ce fichier contient les structures, classes et fonctions nécessaires à la gestion |
| de la configuration de TA3D: sauvegarde, chargement, modification des options du   |
| jeu.                                                                               |
|                                                                                    |
\-----------------------------------------------------------------------------------*/

#ifdef CWDEBUG
#include <libcwd/sys.h>
#include <libcwd/debug.h>
#endif
#include "stdafx.h"
#include "matrix.h"
#include "TA3D_NameSpace.h"
#include "ta3dbase.h"
#include "taconfig.h"

TRANSLATION	translation_manager;

void TA3D_CONFIG::add_option(char *name,bool b,bool f,void *p)
{
	CONFIG_OPTION	*n_options = new CONFIG_OPTION[nb_options+1];
	if(options!=NULL && nb_options>0)
		for(int i=0;i<nb_options;i++)
			n_options[i]=options[i];
	if(options)	delete[] options;
	options=n_options;
	options[nb_options].data=p;
	char *abc = strdup( name );
	strlwr( abc );
	options[nb_options].name=abc;
	options[nb_options].boolean=b;
	options[nb_options].floatting=f;
	if(p==NULL) {
		options[nb_options].d=0;
		options[nb_options].f=0.0f;
		}
	nb_options++;
}

float TA3D_CONFIG::get_option(char *name)
{
	if(options==NULL || nb_options<=0)	return 0.0f;
	for(int i=0;i<nb_options;i++)
		if(strcasecmp(name,options[i].name)==0) {
			if(options[i].data) {
				if(options[i].boolean)
					return *((bool*)options[i].data);
				else if(options[i].floatting)
					return *((float*)options[i].data);
				else
					return *((short*)options[i].data);
				}
			else {
				if(options[i].floatting)
					return options[i].f;
				else
					return options[i].d;
				}
			}
	return 0.0f;
}

float TA3D_CONFIG::get_option(int idx)
{
	if(options==NULL || nb_options<=0 || idx<0 || idx>=nb_options)	return 0.0f;
	if(options[idx].data) {
		if(options[idx].boolean)
			return *((bool*)options[idx].data);
		else if(options[idx].floatting)
			return *((float*)options[idx].data);
		else
			return *((short*)options[idx].data);
		}
	else {
		if(options[idx].floatting)
			return options[idx].f;
		else
			return options[idx].d;
		}
}

void TA3D_CONFIG::set_option(char *name,float v)
{
	if(options==NULL || nb_options<=0)	return;
	for(int i=0;i<nb_options;i++)
		if(strcasecmp(name,options[i].name)==0) {
			if(options[i].data) {
				if(options[i].boolean)
					*((bool*)options[i].data)=( (v>0.0f) ? true : false);
				else if(options[i].floatting)
					*((float*)options[i].data)=v;
				else
					*((short*)options[i].data)=(short)v;
				}
			else {
				if(options[i].floatting)
					options[i].f=v;
				else
					options[i].d=(short)v;
				}
			return;
			}
}

void TRANSLATION::load(char *filename)				// Load data from a ressource file
{
	destroy();
	FILE *src = TA3D_OpenFile(filename,"rb");
	if(src) {
		char buf[200];
		int cur=0;
		while(fgets(buf,200,src)) {
			char *f=buf;
			while(f[0]!=0) {
				f++;
				if(f[0]==10 || f[0]==13)	f[0]=0;
				}
			f=buf;
			while(f[0]==' ' || f[0]==8 || f[0]==9)	f++;
			if(StartsWith(f,"nb_string=")) {
				destroy();
				nb_strings=atoi(f+10);
				for(int i=0;i<5;i++) {
					string[i]=(char**)malloc(sizeof(char*)*nb_strings);
					for(int e=0;e<nb_strings;e++)
						string[i][e]=NULL;
					}
				}
			else {
				if(StartsWith(f,"-----"))	cur++;
				else if(StartsWith(f,"english="))
					string[TA3D_LANG_ENGLISH][cur]=strdup(f+8);
				else if(StartsWith(f,"french="))
					string[TA3D_LANG_FRENCH][cur]=strdup(f+7);
				else if(StartsWith(f,"spanish="))
					string[TA3D_LANG_SPANISH][cur]=strdup(f+8);
				else if(StartsWith(f,"german="))
					string[TA3D_LANG_GERMAN][cur]=strdup(f+7);
				else if(StartsWith(f,"italian="))
					string[TA3D_LANG_ITALIAN][cur]=strdup(f+8);
				}
			}
		fclose(src);
		}
	else
		Console->AddEntry("error: resource unavailable");
}

char *TRANSLATION::get(int str_idx)
{
	int lang=LANG;
	if(string[lang][str_idx]==NULL) {
		lang=0;
		while(string[lang][str_idx]==NULL) {
			lang++;
			if(lang==5)				// string not found!
				return NULL;
			}
		}
	return string[lang][str_idx];
}
